!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_analyse.F90
!! This file contains the main analysis driver
!!
!!
!! ##############################################################################################


#include "qrm_common.h"

!> @brief This is the driver routine for the analysis phase.

!> This routine performa a number of symbolic operations in
!! preparation for the numerical factorization:
!! -# it computes the graph of the matrix removing duplicates
!! -# it detects the presence of singletons in the matrix
!! -# it computes a column permutation in order to reduce the fill-in
!! -# it computes the elimination tree
!! -# it postorders the elimination tree
!! -# it computes the rowcount
!! -# it performs an amalgamation of the elimination tree (with fill)
!! -# it computes a row permutation of the matrix in order to have
!!    a global staircase structure
!! -# merges the singletons into the results of the above operations
!! -# it compresses the results of the above operations to get a size
!!    which is proportional to the number of nodes in the amalgamated tree
!! -# it doea a symbolic factorization which completely characterizes the
!!    structure of fronts etc.
!! -# it computes a tree traversal order which reduces the search space
!!    for scheduling tasks in the numerical factorization
!!
!! [in] qrm_mat       a qrm_spmat_type data which contains the input matrix.
!!                          On output qrm_mat%adata will contain the results of the
!!                          analysis phase
!!
!! [in,out] qrm_spfct a qrm_spfct_type data. On output qrm_spfct%adata will
!!                          contain the results of the analysis phase
!!
!! [in] transp        a character saying whether to do analysis on A or A'
!!

subroutine cqrm_analyse(qrm_mat, qrm_spfct, transp, info)
  use cqrm_spmat_mod
  use cqrm_spfct_mod
  use qrm_common_mod
  use qrm_error_mod
  use cqrm_utils_mod
  use qrm_string_mod
  use qrm_dscr_mod
  use cqrm_analysis_mod, savesym => cqrm_analyse

  implicit none

  type(cqrm_spmat_type), target   :: qrm_mat
  type(cqrm_spfct_type), target   :: qrm_spfct
  character, optional             :: transp
  integer, optional               :: info

  type(qrm_dscr_type)             :: qrm_dscr

  character(len=*), parameter     :: name='qrm_analyse'

  ! init the descriptor
  call qrm_dscr_init(qrm_dscr)

  ! submit the task(s)
  call cqrm_analyse_async(qrm_dscr, qrm_mat, qrm_spfct, transp)

  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, info)

  ! destroy the descriptor
  call qrm_dscr_destroy(qrm_dscr)

  return
end subroutine cqrm_analyse



subroutine cqrm_analyse_async(qrm_dscr, qrm_mat, qrm_spfct, transp)
  use cqrm_spmat_mod
  use cqrm_spfct_mod
  use qrm_common_mod
  use qrm_error_mod
  use cqrm_utils_mod
  use qrm_string_mod
  use iso_c_binding
  use qrm_dscr_mod
  use cqrm_analysis_mod, savesym => cqrm_analyse_async
#if defined (have_starpu)
  use cqrm_starpu_mod
#endif
  implicit none

  type(qrm_dscr_type), target     :: qrm_dscr
  type(cqrm_spmat_type), target   :: qrm_mat
  type(cqrm_spfct_type), target   :: qrm_spfct
  character, optional             :: transp

#if defined (have_starpu)
  type(c_ptr), target             :: qrm_mat_c, qrm_spfct_c, dscrpt
  character(kind=c_char), target  :: transp_c
#endif

  ! error management
  integer                         :: info
  character(len=*), parameter     :: name='qrm_analyse_async'

  if(qrm_dscr%info.ne.0) return

  info = 0

  call cqrm_spmat_check(qrm_mat, info)
  __QRM_INFO_CHECK(info, name,'qrm_check_spmat',9999)
  
  ! make sure all the tasks on this matrix (e.g., the analysis) have completed
  call cqrm_spfct_sync(qrm_spfct)

  call cqrm_spfct_check(qrm_spfct, qrm_analyse_, info)
  __QRM_INFO_CHECK(info, name,'qrm_check_spfct',9999)

  if(allocated(qrm_spfct%adata)) then
     call qrm_adata_cleanup(qrm_spfct%adata, info)
     __QRM_INFO_CHECK(info, name,'qrm_adata_cleanup',9999)
  else
     call qrm_adata_init(qrm_spfct%adata, info)
     __QRM_INFO_CHECK(info, name,'qrm_adata_init',9999)
  end if

  qrm_spfct%sym = qrm_mat%sym
  
#if defined(have_starpu)

  if(present(transp)) then
     transp_c = qrm_str_tolower(transp)
  else
     transp_c = qrm_no_transp
  end if

  qrm_mat_c   = c_loc(qrm_mat)
  qrm_spfct_c = c_loc(qrm_spfct)
  dscrpt      = c_loc(qrm_dscr)

  call fstarpu_insert_task((/ cqrm_analysis_cl,                      &
       FSTARPU_VALUE    , c_loc(dscrpt)      , FSTARPU_SZ_C_PTR,     &
       FSTARPU_VALUE    , c_loc(qrm_mat_c)   , FSTARPU_SZ_C_PTR,     &
       FSTARPU_VALUE    , c_loc(qrm_spfct_c) , FSTARPU_SZ_C_PTR,     &
       FSTARPU_VALUE    , c_loc(transp_c)    , FSTARPU_SZ_CHARACTER, &
       FSTARPU_RW       , qrm_spfct%adata%hdl  ,                       &
       FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),                       &
       C_NULL_PTR /))

#else
  call cqrm_analysis_core(qrm_mat, qrm_spfct, qrm_dscr, transp)
  __QRM_INFO_CHECK(info, name,'qrm_analysis_core',9999)
#endif

  qrm_spfct%adata%ok = .true.
  return

9999 continue
  if(allocated(qrm_spfct%adata)) qrm_spfct%adata%ok = .false.
  call qrm_error_set(qrm_dscr%info, info)
  return
end subroutine cqrm_analyse_async


#if defined (have_starpu)
recursive subroutine cqrm_analysis_cpu_func(buffers, cl_arg) bind(C)
    use qrm_common_mod
    use cqrm_spmat_mod
    use cqrm_spfct_mod
    use fstarpu_mod
    use iso_c_binding
    use qrm_dscr_mod
    use cqrm_analysis_mod
    implicit none

    type(c_ptr), value             :: cl_arg
    type(c_ptr), value             :: buffers

    type(c_ptr), target            :: qrm_mat_c, qrm_spfct_c, qrm_dscr_c

    type(qrm_dscr_type), pointer   :: qrm_dscr
    type(cqrm_spmat_type), pointer :: qrm_mat
    type(cqrm_spfct_type), pointer :: qrm_spfct
    character(kind=c_char), target :: transp
    integer                        :: info

    call fstarpu_unpack_arg(cl_arg,(/ &
         c_loc(qrm_dscr_c),           &
         c_loc(qrm_mat_c),            &
         c_loc(qrm_spfct_c),          &
         c_loc(transp)                &
         /))

    ! until StarPU has a proper error handling method we have to check
    ! manually
    call c_f_pointer(qrm_dscr_c, qrm_dscr)
    if(qrm_dscr%info.ne.0) return

    call c_f_pointer(qrm_mat_c, qrm_mat)
    call c_f_pointer(qrm_spfct_c, qrm_spfct)

    call cqrm_analysis_core(qrm_mat, qrm_spfct, qrm_dscr, transp)

    return

  end subroutine cqrm_analysis_cpu_func
#endif




subroutine cqrm_analysis_core(qrm_mat, qrm_spfct, qrm_dscr, transp)

  use cqrm_spmat_mod
  use cqrm_spfct_mod
  use qrm_common_mod
  use qrm_error_mod
  use cqrm_utils_mod
  use qrm_string_mod
  use qrm_dscr_mod
  use cqrm_analysis_mod, savesym => cqrm_analysis_core

  implicit none

  type(cqrm_spmat_type)           :: qrm_mat
  type(cqrm_spfct_type), target   :: qrm_spfct

  type(qrm_dscr_type)             :: qrm_dscr
  character, optional             :: transp

  type(cqrm_spmat_type), target   :: graph
  integer, allocatable            :: parent(:), rc(:)
  integer, allocatable            :: stair(:), nvar(:)
  integer                         :: ncsing, nrsing
  integer, pointer                :: cperm_p(:), rperm_p(:)
  integer, pointer                :: cperm(:), rperm(:), tmp(:)
  integer                         :: i, nth, bignodes, ilrg, iwrd
  integer                         :: ts, to, te, tp, trc, ta, trp, tc, tsy, tpr, tre, tcm, tr, tg
  character                       :: itransp
  real(kind(1.d0))                :: lrg, wrd, rtmp
  type(cqrm_spmat_type)           :: tmp_mat

  ! error management
  integer                         :: err
  character(len=*), parameter     :: name='qrm_analysis_core'

  err = 0

  __QRM_PRNT_DBG('("Entering the analysis driver")')

  qrm_spfct%nz = qrm_mat%nz

  nullify(cperm_p,rperm_p,cperm,rperm,tmp)

  if(present(transp)) then
     itransp = qrm_str_tolower(transp)
  else
     itransp = qrm_no_transp
  end if
  call system_clock(ts)
  ! call cqrm_compute_graph(qrm_mat, graph, itransp, err)
  if(itransp.eq.qrm_conj_transp) then
     ! call qrm_spmat_transpose(qrm_mat)
     qrm_spfct%m  =   qrm_mat%n
     qrm_spfct%n  =   qrm_mat%m
     tmp_mat%m    =   qrm_mat%n
     tmp_mat%n    =   qrm_mat%m
     tmp_mat%nz   =   qrm_mat%nz
     tmp_mat%irn  =>  qrm_mat%jcn
     tmp_mat%jcn  =>  qrm_mat%irn
     tmp_mat%sym  =   qrm_mat%sym
     tmp_mat%fmt  =   qrm_mat%fmt
  else
     qrm_spfct%m  =   qrm_mat%m
     qrm_spfct%n  =   qrm_mat%n
     tmp_mat%m    =   qrm_mat%m
     tmp_mat%n    =   qrm_mat%n
     tmp_mat%nz   =   qrm_mat%nz
     tmp_mat%irn  =>  qrm_mat%irn
     tmp_mat%jcn  =>  qrm_mat%jcn
     tmp_mat%sym  =   qrm_mat%sym
     tmp_mat%fmt  =   qrm_mat%fmt
  end if

  
  if(qrm_mat%sym.gt.0) then
     call qrm_spmat_convert(tmp_mat, graph, fmt='csc', values=.false., dups=.true., uplo='b', info=err)
  else
     call qrm_spmat_convert(tmp_mat, graph, fmt='csc', values=.false., dups=.true., info=err)
  end if
  __QRM_INFO_CHECK(err, name,'qrm_compute_graph',9999)
  
  ! the cperm array will contain
  call qrm_alloc(qrm_spfct%adata%cperm, graph%n, err)
  __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)
  call qrm_alloc(qrm_spfct%adata%rperm, graph%m, err)
  __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)

  cperm => qrm_spfct%adata%cperm
  rperm => qrm_spfct%adata%rperm

  call system_clock(tg)
  ! Time to go for the ordering
  call cqrm_do_ordering(graph, qrm_spfct, cperm, qrm_spfct%cperm_in, err)
  __QRM_INFO_CHECK(err, name,'qrm_do_ordering',9999)
  call system_clock(to)

  ! build the elimination tree
  call qrm_alloc(parent, graph%n, err)
  __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)

  call cqrm_elim_tree(graph, cperm, parent, err)
  __QRM_INFO_CHECK(err, name,'qrm_elim_tree',9999)
  call system_clock(te)

  ! compute a postorder traversal of the tree
  call qrm_postorder(parent, graph%n, cperm, info=err)
  __QRM_INFO_CHECK(err, name,'qrm_postorder',9999)
  call system_clock(tp)

  call qrm_alloc(rc, graph%n, err)
  __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)

  ! do the symbolic facto
  call cqrm_rowcount(graph, parent, cperm, rc, err)
  __QRM_INFO_CHECK(err, name,'qrm_rowcount',9999)
  call system_clock(trc)

  call qrm_alloc(nvar, graph%n, err)
  __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)

  ! amalgamate the tree
  call qrm_amalg_tree(graph%n, parent, rc, cperm, nvar, &
       qrm_spfct%icntl(qrm_minamalg_), &
       qrm_spfct%rcntl(1), err)
  __QRM_INFO_CHECK(err, name,'qrm_amalg_tree',9999)
  call system_clock(ta)

  call qrm_alloc(stair, graph%n, err)
  __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)

  ! Compute the row permutation to put the matrix in staircase form
  if(qrm_mat%sym.eq.0) then
     call cqrm_rowperm(graph, cperm, rperm, nvar, stair, err)
     __QRM_INFO_CHECK(err, name,'qrm_rowperm',9999)
     call system_clock(trp)
  end if
     
  ! Rework all the data and compress it to nnodes size instead of n
  call qrm_compress_data(qrm_spfct%adata, cperm, parent, rc, stair, graph%n, err)
  __QRM_INFO_CHECK(err, name,'qrm_compress_data',9999)
  call system_clock(tc)

  ! Do the symbolic facto: estimate fronts size, flops, parallelism etc.
  call cqrm_symbolic(graph, qrm_spfct, err)
  __QRM_INFO_CHECK(err, name,'qrm_symbolic',9999)
  call system_clock(tsy)

  nth=qrm_get_num_threads(qrm_dscr)
  
  ! find a layer such that all the subtrees rooted there are treated sequentially
  call qrm_prune_tree(qrm_spfct%adata, nth, err)
  __QRM_INFO_CHECK(err, name,'qrm_prune_tree',9999)
  call system_clock(tpr)

  ! reorder the tree in order to minimize something
  call qrm_reorder_tree(qrm_spfct%adata, err)
  __QRM_INFO_CHECK(err, name,'qrm_reorder_tree',9999)
  call system_clock(tre)

  ! estimate the memory consumption
  call cqrm_compute_memory(qrm_spfct, itransp)
  call system_clock(tcm, tr)

  ! compute the number of rows in each front and some global stats
  ! call qrm_prnt_array(qrm_spfct%adata%cperm, 'cp')
  ! call qrm_prnt_array(qrm_spfct%adata%rc, 'rc')
  ! call qrm_prnt_array(qrm_spfct%adata%cp_ptr(1:qrm_spfct%adata%nnodes+1), 'pt')
  ! call qrm_prnt_array(qrm_spfct%adata%rperm, 'rp')
  qrm_spfct%adata%ok = .true.
  bignodes = 0
  lrg      = 0.d0
  wrd      = 0.d0

  do i=1, qrm_spfct%adata%nnodes
     if(qrm_spfct%adata%small(i).eq.0) bignodes = bignodes+1
     if(graph%sym.eq.0) then
        if(min(qrm_spfct%adata%rc(i),qrm_spfct%adata%nfrows(i)).le.0) cycle
        rtmp = real(qrm_spfct%adata%rc(i),kind(1.d0))*real(qrm_spfct%adata%nfrows(i),kind(1.d0))
        if(rtmp .gt. lrg) then
           lrg  = rtmp
           ilrg = i
        end if
        rtmp = real(qrm_spfct%adata%rc(i),kind(1.d0))/real(qrm_spfct%adata%nfrows(i),kind(1.d0))
        rtmp = max(rtmp,1.d0/rtmp)
        if(rtmp.gt.wrd) then
           wrd  = rtmp
           iwrd = i
        end if
     else if(graph%sym.gt.0) then
        if(qrm_spfct%adata%rc(i).le.0) cycle
        rtmp = real(qrm_spfct%adata%rc(i),kind(1.d0))
        if(rtmp .gt. lrg) then
           lrg  = rtmp
           ilrg = i
        end if
     end if
  end do

  __QRM_PRNT_DBG('("  Total ana time                      : ",f7.4)')real(tcm-ts)/real(tr)
  __QRM_PRNT_DBG('("     Graph    time                    : ",f7.4)')real(tg-ts)/real(tr)
  __QRM_PRNT_DBG('("     Ordering time                    : ",f7.4)')real(to-tg)/real(tr)
  __QRM_PRNT_DBG('("     Elimtree time                    : ",f7.4)')real(te-to)/real(tr)
  __QRM_PRNT_DBG('("     Postorde time                    : ",f7.4)')real(tp-te)/real(tr)
  __QRM_PRNT_DBG('("     Rowcount time                    : ",f7.4)')real(trc-tp)/real(tr)
  __QRM_PRNT_DBG('("     Amalgama time                    : ",f7.4)')real(ta-trc)/real(tr)
  __QRM_PRNT_DBG('("     Rowperm  time                    : ",f7.4)')real(trp-ta)/real(tr)
  __QRM_PRNT_DBG('("     Compress time                    : ",f7.4)')real(tc-trp)/real(tr)
  __QRM_PRNT_DBG('("     Symbolic time                    : ",f7.4)')real(tsy-tc)/real(tr)
  __QRM_PRNT_DBG('("     Prune    time                    : ",f7.4)')real(tpr-tsy)/real(tr)
  __QRM_PRNT_DBG('("     Reorder  time                    : ",f7.4)')real(tre-tpr)/real(tr)
  __QRM_PRNT_DBG('("     Compmemo time                    : ",f7.4)')real(tcm-tre)/real(tr)
  __QRM_PRNT_DBG('("  # nodes in the tree                 : ",i10)')qrm_spfct%adata%nnodes
  __QRM_PRNT_DBG('("  # big nodes                         : ",i10)')bignodes
  __QRM_PRNT_DBG('("  # nleaves                           : ",i20)')qrm_spfct%adata%nleaves
  if(graph%sym.eq.0) then
     __QRM_PRNT_DBG('("  Largest  node (m,n)                 : ",i6,2x,i6)')qrm_spfct%adata%nfrows(ilrg), &
          & qrm_spfct%adata%rc(ilrg)
     __QRM_PRNT_DBG('("  Weirdest node (m,n)                 : ",i6,2x,i6)')qrm_spfct%adata%nfrows(iwrd), &
          & qrm_spfct%adata%rc(iwrd)
     __QRM_PRNT_DBG('("  Estimated nonzeroes in H            : ",i20)')qrm_spfct%gstats(qrm_e_nnz_h_)
  else if(graph%sym.gt.0) then
     __QRM_PRNT_DBG('("  Largest  node (m)                   : ",i6,2x,i6)')qrm_spfct%adata%rc(ilrg)
  end if
  __QRM_PRNT_DBG('("  Estimated nonzeroes in R            : ",i20)')qrm_spfct%gstats(qrm_e_nnz_r_)
  __QRM_PRNT_DBG('("  Estimated total flops at facto      : ",i20)')qrm_spfct%gstats(qrm_e_facto_flops_)
  __QRM_PRNT_DBG('("  Estimated memory peak at facto (MB) : ",f20.3)')&
       & real(qrm_spfct%gstats(qrm_e_facto_mempeak_))/1e6
  if(qrm_print_etree.ge.0) &
       call qrm_print_tree('atree.dot',qrm_spfct%adata,small=qrm_print_etree)

9999 continue
  ! cleanup and return
  if(err.eq.0) call qrm_dealloc(parent, err)
  if(err.eq.0) call qrm_dealloc(rc,     err)
  if(err.eq.0) call qrm_dealloc(nvar,   err)
  if(err.eq.0) call qrm_dealloc(stair,  err)
  if(err.eq.0) call qrm_spmat_dealloc(graph, info=err)
  if(err.eq.0) call qrm_spmat_destroy(graph, info=err)

  call qrm_error_set(qrm_dscr%info, err)

  return

end subroutine cqrm_analysis_core
