!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_do_scotch.F90
!! This file contains the routine that computes a SCOTCH permutation of the input matrix
!!
!!
!! ##############################################################################################


#include "qrm_common.h"

!> Please refer to:
!!
!! <em>"SCOTCH 5.1 User's guide</em>.
!! Technical report, LaBRI, September 2008.
!! F. Pellegrini.
!!
!! for the details of the reordering method.
!!
!! @param[in] graph  the graph associated to the matrix to be ordered.
!!
!! @param[out] cperm an integer array containing the new column order
!!
subroutine dqrm_do_scotch(graph, cperm, info)

  use dqrm_spmat_mod
  use qrm_error_mod
  use dqrm_analysis_mod, savesym => dqrm_do_scotch
  use qrm_mem_mod
  use iso_c_binding
  implicit none

#if defined(have_scotch)
  include "scotchf.h"
#endif

  type(dqrm_spmat_type), target :: graph
  integer                       :: cperm(:)
  integer, optional             :: info

  ! TODO: fix this with conditional compilation
#if defined(have_scotch)
  integer                        :: i, cblknbr, ts, te, tr
  type(dqrm_spmat_type), target  :: ata_graph
  type(dqrm_spmat_type), pointer :: scotch_graph
  integer, allocatable           :: iperm(:)
  real(kind(1.d0))               :: grafdat(scotch_graphdim), stradat(scotch_stratdim), orderdat(scotch_orderdim)

  ! error management
  integer                     :: err
  character(len=*), parameter :: name='qrm_do_scotch'

  err = 0

  if(graph%sym.gt.0) then
     scotch_graph => graph
  else
     call dqrm_ata_graph(graph, ata_graph, err)
     __QRM_INFO_CHECK(err, name,'qrm_ata_graph',9999)
     ! TODO: make ata_graph return the ata graph in csc format instead
     ! of playing with pointers
     ata_graph%jptr => ata_graph%iptr
     ata_graph%irn  => ata_graph%jcn 
     scotch_graph => ata_graph
  end if
  
  info = 0
  call scotchfgraphinit(grafdat, err)
  call scotchfstratinit(stradat, err)
  if(err .ne. 0) then
     err = 19
     call qrm_error_print(err, name)
     goto 9999
  end if

  call scotchfgraphbuild(grafdat, 1, scotch_graph%n, scotch_graph%jptr(1), &
       & scotch_graph%jptr(2), scotch_graph%jptr, scotch_graph%jptr, scotch_graph%nz, &
       & scotch_graph%irn, scotch_graph%irn, err)
  if(err .ne. 0) then
     err = 19
     call qrm_error_print(err, name)
     goto 9999
  end if

  call scotchfgraphorder(grafdat, stradat, grafdat, cperm, cblknbr, &
       & grafdat, grafdat, info)
  if(err .ne. 0) then
     err = 19
     call qrm_error_print(err, name)
     goto 9999
  end if

  call scotchfgraphexit(grafdat)
  call scotchfstratexit(stradat)

9999 continue
  nullify(scotch_graph)

  ! cleanup and return
  if(graph%sym.eq.0) then
     nullify(ata_graph%jptr)
     nullify(ata_graph%irn)
     call dqrm_spmat_destroy(ata_graph, info=err)
  end if

  if(present(info)) info=err
  return

#endif

  return
end subroutine dqrm_do_scotch
