!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_do_colamd.F90
!! This file contains the routine that computes a COLAMD permutation of the input matrix
!!
!!
!! ##############################################################################################


#include "qrm_common.h"


!> @brief This subroutine computes the fill reducing ordering using COLAMD

!> Please refer to:
!!
!! <em>"A column approximate minimum degree ordering algorithm"</em>,
!! T. A. Davis, J. R. Gilbert, S. Larimore, E. Ng, ACM Transactions on
!! Mathematical Software, vol 30, no. 3, Sept. 2004, pp. 353-376.
!! 
!! for the details of the reordering method.
!!
!! @param[in] graph  the graph associated to the matrix to be ordered.
!!
!! @param[out] cperm an integer array containing the new column order
!!
subroutine sqrm_do_colamd(graph, cperm, info)

  use sqrm_spmat_mod
  use qrm_error_mod
  use sqrm_analysis_mod, savesym => sqrm_do_colamd
  use qrm_mem_mod
  use iso_c_binding

  implicit none

  type(sqrm_spmat_type)       :: graph
  integer, target             :: cperm(:)
  integer, optional           :: info

  interface
     subroutine qrm_colamd(n_row, n_col, Alen, A, p, err) bind(c, name='qrm_colamd')
       use iso_c_binding
       integer(c_int), value :: n_row, n_col, Alen
       integer(c_int)        :: A(*), p(*), err
     end subroutine qrm_colamd
  end interface

  interface
     subroutine qrm_colamd_recommended(Alen, nnz, n_row, n_col) bind(c, name='qrm_colamd_recommended')
       use iso_c_binding
       integer(c_int), value :: nnz, n_row, n_col
       integer(c_int)        :: Alen
     end subroutine qrm_colamd_recommended
  end interface

#if defined(have_colamd)
  interface
     function amd_order(n, ap, ai, p, control, info) bind(c)
       use iso_c_binding
       integer               :: amd_order
       integer(c_int), value :: n
       integer               :: ap(*), ai(*), p(*)
       type(c_ptr), value    :: control, info
     end function amd_order
  end interface
#endif
  
  type(sqrm_spmat_type)           :: gcopy
  integer                         :: i, idx, cnt, tmp, alen
  ! error management
  integer                         :: err, err2
  character(len=*), parameter     :: name='qrm_do_colamd'

  err = 0; err2 = 0

  if(graph%sym.eq.0) then
     
     ! at this point we have to make a copy of the graph.
     ! this is a huge waste of mem but we don't have a choice
     ! since ccolamd destroys the graph which, instead, we want to
     ! save for successive computations
     
     ! compute the memory required by colamd (a lot) and allocate
     call qrm_colamd_recommended(alen, graph%nz, graph%m, graph%n)
     call qrm_alloc(gcopy%irn, alen, err)
     __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)
     call qrm_alloc(gcopy%jptr, graph%n+1)
     
     ! make the copy
     call sqrm_spmat_copy(graph, gcopy, values=.false., info=err)
     __QRM_INFO_CHECK(err, name,'qrm_spmat_copy',9999)
     
     ! colamd wants 0 based indices (argh!)
     gcopy%irn(1:gcopy%nz) = gcopy%irn(1:gcopy%nz)-1
     gcopy%jptr(1:gcopy%n+1) = gcopy%jptr(1:gcopy%n+1)-1
     
     ! call colamd
     call qrm_colamd(gcopy%m, gcopy%n, alen, gcopy%irn, gcopy%jptr, err)
     __QRM_INFO_CHECK(err, name,'qrm_colamd',9999)
     
     ! colamd return a 0-based permutation (re-argh!)
     cperm = gcopy%jptr(1:graph%n)+1
     
     call qrm_dealloc(gcopy%jptr)

  else
     ! symmetric matrices. call AMD

     ! amd wants 0 based indices (argh!)
#if defined(have_colamd)
     graph%irn(1:graph%nz)   = graph%irn(1:graph%nz)-1
     graph%jptr(1:graph%n+1) = graph%jptr(1:graph%n+1)-1

     err = amd_order(graph%n, graph%jptr, graph%irn, cperm, c_null_ptr, c_null_ptr)
     ! amd return a 0-based permutation (re-argh!)
     cperm                   = cperm+1
     graph%irn(1:graph%nz)   = graph%irn(1:graph%nz)+1
     graph%jptr(1:graph%n+1) = graph%jptr(1:graph%n+1)+1
#endif
     
  end if
9999 continue
     ! cleanup and return
     call sqrm_spmat_destroy(gcopy, info=err2)

  if(present(info)) info = merge(err, err2, err.ne.0)
  return

end subroutine sqrm_do_colamd
