!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


! This routine is a variant of the LAPACK _GEMQRT routine which allows
! for handling matrices that have a staircase structure

subroutine sqrm_gemqrt(side, trans,   &
     & m, n, k, nb, stair, ofs,       &
     & v, ldv,                        &
     & t, ldt,                        &
     & c, ldc,                        &
     & work, info)

  use qrm_parameters_mod
  implicit none

  character    :: side, trans
  integer      :: info, k, ldv, ldc, m, n, nb, ldt, ofs
  integer      :: stair(*)
  real(r32) ::  v( ldv, * ), c( ldc, * ), t( ldt, * ), work( * )

  logical      :: left, right, tran, notran
  integer      :: i, j, ib, ldwork, kf, q, e, ef, im
  logical      :: lsame

  info   = 0
  left   = lsame( side,  qrm_left )
  right  = lsame( side,  qrm_right )
  tran   = lsame( trans, qrm_transp )
  notran = lsame( trans, qrm_no_transp )

  if( left ) then
     ldwork = max( 1, n )
     q = m
  else if ( right ) then
     ldwork = max( 1, m )
     q = n
  end if
  if( .not.left .and. .not.right ) then
     info = -1
  else if( .not.tran .and. .not.notran ) then
     info = -2
  else if( m.lt.0 ) then
     info = -3
  else if( n.lt.0 ) then
     info = -4
  else if( ldv.lt.max( 1, q ) ) then
     info = -8
  else if( ldt.lt.nb ) then
     info = -10
  else if( ldc.lt.max( 1, m ) ) then
     info = -12
  end if

  if( info.ne.0 ) then
     call xerbla( 'dgemqrt', -info )
     return
  end if

  if( m.eq.0 .or. n.eq.0 .or. k.eq.0 ) return

  e = min( m-ofs+1, k )

  if( left .and. tran ) then

     i = ofs
     do j = 1, e, nb
        ib = min( nb, e-j+1 )
        if(stair(1).lt.0) then
           im = max(m - i + 1,ib)
        else
           im = max(min(stair(j+ib-1)-i+1,m),ib)
        end if

        if(im.gt.0) call slarfb( qrm_left, qrm_transp, &
             qrm_forward, qrm_colwise,                      &  
             im, n, ib,                                     &  
             v( i, j ), ldv,                                &  
             t( 1, j ), ldt,                                &  
             c( i, 1 ), ldc,                                &  
             work, ldwork )
        i = i+ib
     end do

  else if( right .and. notran ) then

     write(*,'("sqrm_gemqrt: not implemented")')
     info = -13

  else if( left .and. notran ) then

     ef = ((e-1)/nb)*nb+ofs

     i  = ef
     do j = ef, 1, -nb
        ib = min( nb, e-j+1 )
        if(stair(1).lt.0) then
           im = max(m - i + 1,ib)
        else
           im = max(min(stair(j+ib-1)-i+1,m),ib)
        end if

        call slarfb( qrm_left, qrm_no_transp, &
             qrm_forward, qrm_colwise,        &
             im, n, ib,                       &
             v( i, j ), ldv,                  &
             t( 1, j ), ldt,                  &
             c( i, 1 ), ldc,                  &
             work, ldwork )
        i = i-nb
     end do

  else if( right .and. tran ) then

     write(*,'("sqrm_gemqrt: not implemented")')
     info = -13

  end if

  return

end subroutine sqrm_gemqrt
