!! ##############################################################################################
!!
!! Copyright 2012-2018 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains synchronous and asynchronous routines for
!! applying Householder transformations to a matrix
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine cqrm_dsmat_gemqr(trans, a, t, x, ib, bh, work, prio, info)

  use cqrm_dsmat_mod, alias=>cqrm_dsmat_gemqr
  use qrm_dscr_mod

  implicit none

  character                   :: trans
  type(cqrm_dsmat_type)       :: a, t, x
  type(cqrm_ws_type)          :: work

  integer                     :: ib, bh
  integer, optional           :: prio
  integer, optional           :: info

  type(qrm_dscr_type)         :: qrm_dscr
  
  ! error handling
  integer                        :: err
  character(len=*), parameter    :: name='qrm_dsmat_gemqr'

  call qrm_dscr_init(qrm_dscr)

  call cqrm_dsmat_gemqr_async(qrm_dscr, trans, a, t, x, ib, bh, work, prio)

  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, info)

  call qrm_dscr_destroy(qrm_dscr)

  return
end subroutine cqrm_dsmat_gemqr

subroutine cqrm_dsmat_gemqr_async(qrm_dscr, trans, a, t, x, ib, bh, work, prio)

  use cqrm_dsmat_mod, alias=>cqrm_dsmat_gemqr_async
  use qrm_dscr_mod
  use cqrm_dsmat_task_mod

  implicit none

  type(qrm_dscr_type)         :: qrm_dscr
  character                   :: trans
  type(cqrm_dsmat_type)       :: a, t, x
  type(cqrm_ws_type)          :: work
  integer                     :: ib, bh
  integer, optional           :: prio

  integer                     :: i, j, k, s, l, nbr, nbc, nbcx, nbe, first, iprio, ibh

  ! error handling
  integer                     :: info
  character(len=*), parameter :: name='qrm_dsmat_gemqr_async'

  if(qrm_dscr%info.ne.0) return
  info = 0

  __QRM_IFPRESENT3(iprio, prio, 0)

  nbr  = size(a%blocks,1)
  nbc  = size(a%blocks,2)
  nbe  = min(nbr,nbc)
  nbcx = size(x%blocks,2)

  if(bh.le.0) then
     ibh = nbr
  else
     ibh = bh
  end if
  
  if(trans .eq. qrm_conj_transp) then
     do k=1, nbe
        do first=k+1, nbr
           if(qrm_allocated(a%blocks(first,k)%stair)) exit
        end do
        do i = k, first-1, ibh
           do j=1, nbcx
              if(qrm_allocated(t%blocks(i,k)%c))       &
              call cqrm_higemqrt_task(qrm_dscr,        &
                   qrm_conj_transp,                    &
                   a%blocks(i,k),                      &
                   t%blocks(i,k),                      &
                   x%blocks(i,j),                      &
                   1, 1,                               &
                   a%nb, ib, work, iprio)               
           end do                                       
           do l=i+1, min(i+ibh-1,first-1)                
              do j=1, nbcx                              
                 if(qrm_allocated(a%blocks(l,k)%c))    &
                 call cqrm_hitpmqrt_task(qrm_dscr,     &
                      qrm_conj_transp,                 &
                      a%blocks(l,k),                   &
                      t%blocks(l,k),                   &
                      x%blocks(i,j),                   &
                      x%blocks(l,j),                   &
                      1, 1,                            &
                      a%nb, ib, qrm_square,            &
                      work, iprio)
              end do
           end do
        end do
        s = ibh
        do while (s.le.first-1-k+1)
           do i = k, first-1-s, 2*s
              l = i+s
              if(l.le.first-1) then
                 do j=1, nbcx
                    if(qrm_allocated(a%blocks(l,k)%c)) &
                    call cqrm_hitpmqrt_task(qrm_dscr,  &
                         qrm_conj_transp,              &
                         a%blocks(l,k),                &
                         t%blocks(l,nbe+k),            &
                         x%blocks(i,j),                &
                         x%blocks(l,j),                &
                         1, 1,                         &
                         a%nb, ib, qrm_triang,         &
                         work, iprio)
                 end do
              end if
           end do
           s = s*2
        end do

        do l=first, nbr
           do j=1, nbcx
              if(qrm_allocated(a%blocks(l,k)%c))       &  
                   call cqrm_hitpmqrt_task(qrm_dscr,   &
                   qrm_conj_transp,                    &
                   a%blocks(l,k),                      &
                   t%blocks(l,k),                      &
                   x%blocks(k,j),                      &
                   x%blocks(l,j),                      &
                   1, 1,                               &
                   a%nb, ib, qrm_square,               &
                   work, iprio)
           end do
        end do
     end do
  else
     do k=nbe, 1, -1
        do first=k+1, nbr
           if(qrm_allocated(a%blocks(first,k)%stair)) exit
        end do

        do l=nbr, first, -1
           do j=1, nbcx
              if(qrm_allocated(a%blocks(l,k)%c))       &
                   call cqrm_hitpmqrt_task(qrm_dscr,   &
                   qrm_no_transp,                      &
                   a%blocks(l,k),                      &
                   t%blocks(l,k),                      &
                   x%blocks(k,j),                      &
                   x%blocks(l,j),                      &
                   1, 1,                               &
                   a%nb, ib, qrm_square,               &
                   work, iprio)
           end do
        end do

        s = ibh
        do while (s.le.first-1-k+1)
           s = s*2
        end do
        do while (s.ge.ibh)
           do i = k, first-1-s, 2*s
              l = i+s
              if(l.le.first-1) then
                 do j=1, nbcx
                    if(qrm_allocated(a%blocks(l,k)%c))      &
                         call cqrm_hitpmqrt_task(qrm_dscr,  &
                         qrm_no_transp,                     &
                         a%blocks(l,k),                     &
                         t%blocks(l,nbe+k),                 &
                         x%blocks(i,j),                     &
                         x%blocks(l,j),                     &
                         1, 1,                              &
                         a%nb, ib, qrm_triang,              &
                         work, iprio)
                 end do
              end if
           end do
           s = s/2
        end do

        do i = k, first-1, ibh
           do l=min(i+ibh-1,first-1), i+1, -1
              do j=1, nbcx
                 if(qrm_allocated(a%blocks(l,k)%c))         &
                 call cqrm_hitpmqrt_task(qrm_dscr,          &
                      qrm_no_transp,                        &
                      a%blocks(l,k),                        &
                      t%blocks(l,k),                        &
                      x%blocks(i,j),                        &
                      x%blocks(l,j),                        &
                      1, 1,                                 &
                      a%nb, ib, qrm_square,                 &
                      work, iprio)
              end do

           end do
           do j=1, nbcx
              if(qrm_allocated(t%blocks(i,k)%c))            &
              call cqrm_higemqrt_task(qrm_dscr,             &
                   qrm_no_transp,                           &
                   a%blocks(i,k),                           &
                   t%blocks(i,k),                           &
                   x%blocks(i,j),                           &
                   1, 1,                                    &
                   a%nb, ib, work, iprio)
           end do
        end do
     end do
  end if

9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return

end subroutine cqrm_dsmat_gemqr_async
