!! ##############################################################################################
!!
!! Copyright 2012-2018 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains synchronous and asynchronous routines for
!! computing the QR factorization of a matrix
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine dqrm_dsmat_geqr(a, t, ib, bh, work, prio, prepart, info)

  use dqrm_dsmat_mod, alias=>dqrm_dsmat_geqr
  use qrm_dscr_mod
  implicit none

  type(dqrm_dsmat_type)       :: a, t
  type(dqrm_ws_type)          :: work
  integer                     :: ib, bh
  integer, optional           :: prio
  integer, optional           :: prepart  ! number of rightmost block-columns to pre-partition
  integer, optional           :: info

  type(qrm_dscr_type)         :: qrm_dscr

  ! error handling
  integer                     :: err
  character(len=*), parameter :: name='qrm_dsmat_geqr'

  err = 0
  call qrm_dscr_init(qrm_dscr)

  call dqrm_dsmat_geqr_async(qrm_dscr, a, t, ib, bh, work, prio, prepart)

  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, info)

  call qrm_dscr_destroy(qrm_dscr)

9999 continue
  return
end subroutine dqrm_dsmat_geqr

subroutine dqrm_dsmat_geqr_async(qrm_dscr, a, t, ib, bh, work, prio, prepart)

  use dqrm_dsmat_mod, alias=>dqrm_dsmat_geqr_async
  use qrm_dscr_mod
  implicit none

  type(qrm_dscr_type)         :: qrm_dscr
  type(dqrm_dsmat_type)       :: a, t
  type(dqrm_ws_type)          :: work
  integer                     :: ib, bh
  integer, optional           :: prio
  integer, optional           :: prepart  ! number of rightmost block-columns to pre-partition

  integer                     :: i, j, k, s, l, nbr, nbc, nbe, first, iprio, iprepart, ibh

  ! error handling
  integer                     :: info
  character(len=*), parameter :: name='qrm_dsmat_geqr_async'

  if(qrm_dscr%info.ne.0) return
  info = 0

  if(min(a%m,a%n).eq.0) return

  __QRM_IFPRESENT3(iprio, prio, 0)
  
  nbr = size(a%blocks,1)
  nbc = size(a%blocks,2)
  nbe = min(nbr,nbc)

  if(present(prepart)) then
     iprepart = min(nbc,max(prepart,0))
  else
     iprepart = 0
  end if

  if(bh.le.0) then
     ibh = nbr
  else
     ibh = bh
  end if
  
  ! prepartition
  do j= nbc, nbc-iprepart+1, -1
     do i = 1, nbr
        call dqrm_block_partition(a%blocks(i,j))
     end do
  end do

  do k=1, min(nbr,nbc)
     ! find first sub-diagonal tile not being traversed by the staircase
     do first=k+1, nbr
        if(qrm_allocated(a%blocks(first,k)%stair)) exit
     end do

     do i = k, first-1, ibh
        call dqrm_block_partition(a%blocks(i,k), t%blocks(i,k))
        call dqrm_higeqrt(qrm_dscr,              &
             a%blocks(i,k),                      &
             t%blocks(i,k),                      &
             a%nb, ib, work, iprio)
        do j=k+1, nbc
           call dqrm_higemqrt(qrm_dscr,          &
                a%blocks(i,k),                   &
                t%blocks(i,k),                   &
                a%blocks(i,j),                   &
                a%nb, ib,                        &
                work, iprio)
        end do
        do l=i+1, min(i+ibh-1,first-1)
           call dqrm_block_partition(a%blocks(l,k), t%blocks(l,k))
           call dqrm_hitpqrt(qrm_dscr,           &
                a%blocks(i,k),                   &
                a%blocks(l,k),                   &
                t%blocks(l,k),                   &
                a%nb, ib, qrm_square,            &
                work, iprio)
           do j=k+1, nbc
              call dqrm_hitpmqrt(qrm_dscr,       &
                   a%blocks(l,k),                &
                   t%blocks(l,k),                &
                   a%blocks(i,j),                &
                   a%blocks(l,j),                &
                   a%nb, ib, qrm_square,         &
                   work, iprio)
           end do
           call dqrm_block_unpartition(a%blocks(l,k), t%blocks(l,k))
        end do
     end do
     s = ibh
     do while (s.le.first-k)
        do i = k, first-1-s, 2*s
           l = i+s
           if(l.le.first-1) then
              call dqrm_block_partition(t%blocks(l,nbe+k))
              call dqrm_hitpqrt(qrm_dscr,        &
                   a%blocks(i,k),                &
                   a%blocks(l,k),                &
                   t%blocks(l,nbe+k),            &
                   a%nb, ib, qrm_triang,         &
                   work, iprio)
              do j=k+1, nbc
                 call dqrm_hitpmqrt(qrm_dscr,    &
                      a%blocks(l,k),             &
                      t%blocks(l,nbe+k),         &
                      a%blocks(i,j),             &
                      a%blocks(l,j),             &
                      a%nb, ib, qrm_triang,      &
                      work, iprio)
              end do
              call dqrm_block_unpartition(a%blocks(l,k), t%blocks(l,k))
              call dqrm_block_unpartition(t%blocks(l,nbe+k))
           end if
        end do
        s = s*2
     end do

     do l=first, nbr
        call dqrm_block_partition(a%blocks(l,k), t%blocks(l,k))
        call dqrm_hitpqrt(qrm_dscr,              &
             a%blocks(k,k),                      &
             a%blocks(l,k),                      &
             t%blocks(l,k),                      &
             a%nb, ib, qrm_square,               &
             work, iprio)                         
        do j=k+1, nbc                             
           call dqrm_hitpmqrt(qrm_dscr,          &
                a%blocks(l,k),                   &
                t%blocks(l,k),                   &
                a%blocks(k,j),                   &
                a%blocks(l,j),                   &
                a%nb, ib, qrm_square,            &
                work, iprio)
        end do
        call dqrm_block_unpartition(a%blocks(l,k), t%blocks(l,k))
     end do
          
     call dqrm_block_unpartition(a%blocks(k,k), t%blocks(k,k))
  end do

  ! prepartition
  do j= nbc, nbc-iprepart+1, -1
     do i = 1, nbr
        call dqrm_block_unpartition(a%blocks(i,j))
     end do
  end do

9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return
end subroutine dqrm_dsmat_geqr_async


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!
!!!!  GEQRT
!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine dqrm_higeqrt(qrm_dscr, a, t, nb, ib, work, prio)

  use dqrm_dsmat_mod, alias=>dqrm_higeqrt
  use qrm_dscr_mod
  use dqrm_dsmat_task_mod

  implicit none

  type(qrm_dscr_type)   :: qrm_dscr
  type(dqrm_block_type) :: a, t
  integer               :: nb, ib
  type(dqrm_ws_type)    :: work
  integer               :: prio

  integer               :: k, j, parts, i

  if(qrm_dscr%info.ne.0)      return
  if(.not.qrm_allocated(a%c)) return
  
  if(a%partitioned) then
     parts = (size(a%c,2)-1)/nb+1
     do k=1, min(parts,(size(a%c,1)-1)/nb+1)
        call dqrm_higeqrt_task(qrm_dscr, a, t, k, nb, ib, work, prio)
        do j=k+1, parts
           call dqrm_higemqrt_task(qrm_dscr, qrm_transp, a, t, a, k, j, nb, ib, work, prio)
        end do
     end do
  else
     call dqrm_higeqrt_task(qrm_dscr, a, t, 1, nb, ib, work, prio)
  end if

  return
end subroutine dqrm_higeqrt


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!
!!!!  GEMQRT
!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine dqrm_higemqrt(qrm_dscr, a, t, c, nb, ib, work, prio)

  use dqrm_dsmat_mod, alias=>dqrm_higemqrt
  use qrm_dscr_mod
  use dqrm_dsmat_task_mod

  implicit none

  type(qrm_dscr_type)   :: qrm_dscr
  type(dqrm_block_type) :: a, t, c
  type(dqrm_ws_type)    :: work
  integer               :: nb, ib
  integer               :: prio

  integer               :: k, j, partst, partsc

  if(qrm_dscr%info.ne.0)      return
  if(.not.qrm_allocated(t%c)) return
  
  if(t%partitioned) then
     partst = min((size(t%c,2)-1)/nb+1,(size(t%c,1)-1)/nb+1)
  else
     partst = 1
  end if

  if(c%partitioned) then
     partsc = (size(c%c,2)-1)/nb+1
  else
     partsc = 1
  end if

  ! FIXME: not the cleanest. find a better way to compute the number
  ! of panels
  do k=1,  min(partst,(size(c%c,1)-1)/nb+1)
     do j=1, partsc
        call dqrm_higemqrt_task(qrm_dscr, qrm_transp, a, t, c, k, j, nb, ib, work, prio)
     end do
  end do

  return
end subroutine dqrm_higemqrt


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!
!!!!  TPQRT
!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine dqrm_hitpqrt(qrm_dscr, a1, a2, t, nb, ib, ts, work, prio)

  use dqrm_dsmat_mod, alias=>dqrm_hitpqrt
  use qrm_dscr_mod
  use dqrm_dsmat_task_mod

  implicit none

  type(qrm_dscr_type)   :: qrm_dscr
  type(dqrm_block_type) :: a1, a2, t
  type(dqrm_ws_type)    :: work
  integer               :: nb, ib
  character             :: ts
  integer               :: prio

  integer               :: k, j, parts

  if(qrm_dscr%info.ne.0)       return
  if(.not.qrm_allocated(a2%c)) return
  
  if(a1%partitioned) then
     parts = (size(a1%c,2)-1)/nb+1
     do k=1, parts
        call dqrm_hitpqrt_task(qrm_dscr, a1, a2, t, k, nb, ib, ts, work, prio)
        do j=k+1, parts
           call dqrm_hitpmqrt_task(qrm_dscr, qrm_transp, a2, t, a1, a2, k, j, nb, ib, ts, work, prio)
        end do
     end do
  else
     call dqrm_hitpqrt_task(qrm_dscr, a1, a2, t, 1, nb, ib, ts, work, prio)
  end if

  return

end subroutine dqrm_hitpqrt


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!!!
!!!!  TPMQRT
!!!!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

subroutine dqrm_hitpmqrt(qrm_dscr, v, t, c1, c2, nb, ib, ts, work, prio)

  use dqrm_dsmat_mod, alias=>dqrm_hitpmqrt
  use qrm_dscr_mod
  use dqrm_dsmat_task_mod

  implicit none

  type(qrm_dscr_type)   :: qrm_dscr
  type(dqrm_block_type) :: v, t, c1, c2
  type(dqrm_ws_type)    :: work
  integer               :: nb, ib
  character             :: ts
  integer               :: prio

  integer               :: k, j, partsv, partsc

  if(qrm_dscr%info.ne.0)      return
  if(.not.qrm_allocated(v%c)) return
  
  if(v%partitioned) then
     partsv = (size(v%c,2)-1)/nb+1
  else
     partsv = 1
  end if

  if(c1%partitioned) then
     ! TODO: if .not. c2%partitioned raise error
     partsc = (size(c1%c,2)-1)/nb+1
  else
     partsc = 1
  end if

  do k=1, partsv
     do j=1, partsc
        call dqrm_hitpmqrt_task(qrm_dscr, qrm_transp, v, t, c1, c2, k, j, nb, ib, ts, work, prio)
     end do
  end do

  return
end subroutine dqrm_hitpmqrt





subroutine dqrm_dsmat_init_t(qrm_dsmat_a, qrm_dsmat_t, ib, bh, pin, seq, info)

#if defined (have_starpu)
  use fstarpu_mod
  use qrm_starpu_common_mod
#endif
  use qrm_error_mod
  use qrm_mem_mod
  use qrm_parameters_mod
  use dqrm_dsmat_mod, alias=>dqrm_dsmat_init_t
  use qrm_dscr_mod

  implicit none

  type(dqrm_dsmat_type), intent(in)            :: qrm_dsmat_a
  type(dqrm_dsmat_type), intent(inout), target :: qrm_dsmat_t
  integer, intent(in)                          :: ib, bh
  integer, optional                            :: info
  logical, optional                            :: pin, seq

  integer                                      :: i, j, is, js, parts, nbr, nbc, nb, ibh
  type(c_ptr)                                  :: bc_filter
  logical                                      :: iseq
  
  ! error handling
  integer                                      :: err
  character(len=*), parameter                  :: name='qrm_dsmat_init_t'

  err = 0

  if(present(seq)) then
     iseq = seq
  else
     iseq = .false.
  end if

  if(.not.qrm_dsmat_a%inited) return
  
  nbr = size(qrm_dsmat_a%blocks,1)
  nbc = size(qrm_dsmat_a%blocks,2)
  nb  = qrm_dsmat_a%nb

  if(bh.le.0) then
     ibh = nbr
  else
     ibh = bh
  end if
  
  if(min(nbr,nbc).le.0) return
  
  ! FIXME: add error checks

#if defined (have_starpu)
  if(.not.iseq) then
     bc_filter = qrm_get_bc_filter()
     call fstarpu_data_filter_set_filter_arg (bc_filter, nb)
  end if
#endif

  if((ibh.ge.nbr) .or. (ibh.le.0)) then
     ! facto has flat reduction tree (no need for T2)
     allocate(qrm_dsmat_t%blocks(nbr,min(nbr,nbc)))
  else
     ! facto has hybrid reduction tree (T2 is needed)
     allocate(qrm_dsmat_t%blocks(nbr,2*min(nbr,nbc)))
  end if

  do j=1, nbc
     do i=j, nbr
        if(.not.qrm_allocated(qrm_dsmat_a%blocks(i,j)%c)) cycle

        is = size(qrm_dsmat_a%blocks(i,j)%c,1)
        js = size(qrm_dsmat_a%blocks(i,j)%c,2)

        if(mod(i-j,ibh).eq.0) then
           ! on this block a geqrt is done and thus we need to
           ! allocate extra space for copying the householder
           ! vectors. This copy allows for preventing useless
           ! dependencies in StarPU:
           !   |TTTT|
           !   |1000|
           !   |V100|
           !   |VV10|
           !   |VVV1|
           call qrm_alloc(qrm_dsmat_t%blocks(i,j)%c, ib+is, js, err, pin)
           __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)
  
           if (i.ne.j ) then
              ! the block is below the diagonal. this block is first
              ! reduced into a triangle with a geqrf and then
              ! annihilated with a tpqrt. Must allocate T2 for this
              ! second operation
              call qrm_alloc(qrm_dsmat_t%blocks(i,j+min(nbr,nbc))%c, ib, js, err, pin)
              __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)
              qrm_dsmat_t%blocks(i,j+min(nbr,nbc))%c = qrm_dzero
           end if
        else
           call qrm_alloc(qrm_dsmat_t%blocks(i,j)%c, ib, js, err, pin)
           __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)
        end if

        qrm_dsmat_t%blocks(i,j)%c = qrm_dzero

#if defined (have_starpu)
        if(.not.iseq) then
           ! register the block to starpu
           call fstarpu_matrix_data_register(qrm_dsmat_t%blocks(i,j)%hdl, 0, &
                c_loc(qrm_dsmat_t%blocks(i,j)%c),                            &
                size(qrm_dsmat_t%blocks(i,j)%c,1),                           &
                size(qrm_dsmat_t%blocks(i,j)%c,1),                           &
                size(qrm_dsmat_t%blocks(i,j)%c,2),                           &
                c_sizeof(qrm_dsmat_t%blocks(i,j)%c(1,1)))
           if(nb.lt.js) then
              ! plan a dynamic partitioning
              parts = (js-1)/nb+1
              allocate(qrm_dsmat_t%blocks(i,j)%shdls(parts))
              call fstarpu_data_partition_plan(qrm_dsmat_t%blocks(i,j)%hdl, bc_filter, qrm_dsmat_t%blocks(i,j)%shdls)
           end if

           if((mod(i-j,ibh).eq.0) .and. (i.ne.j)) then
              ! register the block to starpu
              call fstarpu_matrix_data_register(qrm_dsmat_t%blocks(i,j+min(nbr,nbc))%hdl, 0, &
                   c_loc(qrm_dsmat_t%blocks(i,j+min(nbr,nbc))%c),                            &
                   size(qrm_dsmat_t%blocks(i,j+min(nbr,nbc))%c,1),                           &
                   size(qrm_dsmat_t%blocks(i,j+min(nbr,nbc))%c,1),                           &
                   size(qrm_dsmat_t%blocks(i,j+min(nbr,nbc))%c,2),                           &
                   c_sizeof(qrm_dsmat_t%blocks(i,j+min(nbr,nbc))%c(1,1)))
              if(nb.lt.js) then
                 ! plan a dynamic partitioning
                 parts = (js-1)/nb+1
                 allocate(qrm_dsmat_t%blocks(i,j+min(nbr,nbc))%shdls(parts))
                 call fstarpu_data_partition_plan(qrm_dsmat_t%blocks(i,j+min(nbr,nbc))%hdl,  &
                      bc_filter, qrm_dsmat_t%blocks(i,j+min(nbr,nbc))%shdls)
              end if
           end if
        end if
#endif
        qrm_dsmat_t%blocks(i,j)%partitioned = .false.
        if((mod(i-j,ibh).eq.0) .and. (i.ne.j)) then
           qrm_dsmat_t%blocks(i,j+min(nbr,nbc))%partitioned = .false.
        end if
     end do
  end do

  qrm_dsmat_t%inited = .true.

9999 continue 
  ! should do cleanup
  if(present(info)) info = err
  return
end subroutine dqrm_dsmat_init_t
