!! ##############################################################################################
!!
!! Copyright 2012-2018 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains synchronous and asynchronous routines for
!! computing the frobenius norm of a matrix
!!
!! ##############################################################################################

#include "qrm_common.h"


subroutine sqrm_dsmat_nrm(qrm_dsmat, nrm, m, n, info)

  use sqrm_dsmat_mod, alias=>sqrm_dsmat_nrm
  use qrm_dscr_mod

  implicit none

  type(sqrm_dsmat_type), target  :: qrm_dsmat
  real(r32)                      :: nrm
  integer, optional              :: m, n, info
  
  type(qrm_dscr_type)            :: qrm_dscr

  ! error handling
  integer                        :: err
  character(len=*), parameter    :: name='qrm_dsmat_nrm'

  err = 0

  if(.not.qrm_dsmat%inited) then
     err = 1000
     call qrm_error_print(err, name)
     goto 9999
  end if

  call qrm_dscr_init(qrm_dscr, nocuda=.true.)

  call sqrm_dsmat_nrm_async(qrm_dscr, qrm_dsmat, nrm, m, n)

  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, err)

  call qrm_dscr_destroy(qrm_dscr)

9999 continue
  if(present(info)) info = err
  return

end subroutine sqrm_dsmat_nrm


subroutine sqrm_dsmat_nrm_async(qrm_dscr, qrm_dsmat, nrm, m, n)
  use sqrm_dsmat_mod, alias=>sqrm_dsmat_nrm_async
  use qrm_dscr_mod
  use qrm_mem_mod
  use iso_c_binding
  use sqrm_dsmat_task_mod
#if defined (have_starpu)
  use sqrm_starpu_mod
  use qrm_starpu_common_mod
#endif

  implicit none

  type(qrm_dscr_type)               :: qrm_dscr
  type(sqrm_dsmat_type)             :: qrm_dsmat
  real(r32)                         :: nrm
  integer, optional                 :: m, n

  integer                           :: i, j, im, in, mb, nb, mm, nn
  type(sqrm_ws_type)                :: ssq
#if defined (have_mpi)
  real(r32)                         :: err
#endif

  ! error handling
  integer                        :: info
  character(len=*), parameter    :: name='qrm_dsmat_nrm_async'

  if(qrm_dscr%info.ne.0) return
  info = 0

  if(.not.qrm_dsmat%inited) then
     info = 1000
     call qrm_error_print(info, name)
     goto 9999
  end if

  __QRM_IFPRESENT3(im, m, qrm_dsmat%m)
  __QRM_IFPRESENT3(in, n, qrm_dsmat%n)

  mb  = (im-1)/qrm_dsmat%mb+1
  nb  = (in-1)/qrm_dsmat%mb+1

  call sqrm_nrm_ws_init(ssq)

  do i=1, mb
     if(i.eq.mb) then
        mm=im-(i-1)*qrm_dsmat%mb
     else
        mm=qrm_dsmat%mb
     end if
     do j=1, nb
        if(j.eq.nb) then
           nn=in-(j-1)*qrm_dsmat%mb
        else
           nn=qrm_dsmat%mb
        end if
        call sqrm_block_nrm_task(qrm_dscr, qrm_dsmat%blocks(i,j), mm, nn, ssq)
     end do 
  end do

  call sqrm_nrm_ws_destroy(ssq, nrm)
  
9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return

contains


  
  subroutine sqrm_nrm_ws_destroy(ssq, nrm)
    use qrm_mem_mod
    implicit none
    type(sqrm_ws_type)                :: ssq
    real(r32)                         :: nrm
    
#if defined (have_starpu)
    ! maybe these two are not needed
    call fstarpu_data_acquire(ssq%hdl, FSTARPU_R)
    call fstarpu_data_release(ssq%hdl)
    call fstarpu_data_unregister(ssq%hdl)
#endif
    
    nrm = ssq%c(1,1)*sqrt(ssq%c(2,1))
    call qrm_dealloc(ssq%c)
    
    return
  end subroutine sqrm_nrm_ws_destroy


  
  subroutine sqrm_nrm_ws_init(ssq)
    use qrm_parameters_mod
    use qrm_mem_mod
    implicit none
    type(sqrm_ws_type)                :: ssq

    call qrm_alloc(ssq%c,2,1)

    ssq%c(:,1) = (/ qrm_szero, qrm_sone /)

#if defined (have_starpu)
    call fstarpu_vector_data_register(ssq%hdl, 0, c_loc(ssq%c), 2, c_sizeof(ssq%c(1,1)))
    call fstarpu_data_set_reduction_methods(ssq%hdl, sqrm_lassq_red_cl, sqrm_lassq_ini_cl)
#endif
    
    return
  end subroutine sqrm_nrm_ws_init

    
end subroutine sqrm_dsmat_nrm_async
