!! ##############################################################################################
!!
!! Copyright 2012-2018 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains synchronous and asynchronous routines for
!! making the sum of two matrices
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine zqrm_dsmat_axpy(a, b, ia, ja, ib, jb, m, n, l, alpha, info)
  use qrm_parameters_mod
  use zqrm_dsmat_mod, alias=>zqrm_dsmat_axpy
  use qrm_dscr_mod

  implicit none


  type(zqrm_dsmat_type)          :: a                    ! This is the source
                                                         ! matrix.

  type(zqrm_dsmat_type)          :: b                    ! This is the destination
                                                         ! matrix.

  integer, optional              :: ia, ja, ib, jb       ! ia and ja are the staring row and col
                                                         ! in the A matrix matrix. ib and jb
                                                         ! the same for B
  integer, optional              :: m, n, l              ! m and n are the
                                                         ! number of rows and cols to
                                                         ! be copied. l is the number of
                                                         ! rows in the upper trapezoidal
                                                         ! part of the block (same meaning
                                                         ! as in tpqrt
  complex(r64), optional         :: alpha                ! alpha scalar
  integer, optional              :: info

  type(qrm_dscr_type)            :: qrm_dscr

  ! error handling
  integer                        :: err
  character(len=*), parameter    :: name='qrm_dsmat_axpy'

  err = 0

  if(.not.b%inited) then
     err = 1000
     call qrm_error_print(err, name)
     goto 9999
  end if

  call qrm_dscr_init(qrm_dscr, nocuda=.true.)

  call zqrm_dsmat_axpy_async(qrm_dscr, a, b, ia=ia, ja=ja, ib=ib, jb=jb, m=m, n=n, l=l, alpha=alpha)
  
  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, err)

  call qrm_dscr_destroy(qrm_dscr)

9999 continue
  if(present(info)) info = err
  return

end subroutine zqrm_dsmat_axpy


subroutine zqrm_dsmat_axpy_async(qrm_dscr, a, b, ia, ja, ib, jb, m, n, l, alpha)
  use zqrm_dsmat_mod, alias=>zqrm_dsmat_axpy_async
  use qrm_dscr_mod
  use qrm_mem_mod
  use iso_c_binding
  use zqrm_dsmat_task_mod
  use qrm_parameters_mod

  implicit none

  type(qrm_dscr_type), target    :: qrm_dscr
  type(zqrm_dsmat_type)          :: a                    ! This is the source
                                                         ! matrix.

  type(zqrm_dsmat_type)          :: b                    ! This is the destination
                                                         ! matrix.

  integer, optional              :: ia, ja, ib, jb       ! ia and ja are the staring row and col
                                                         ! in the A matrix matrix. ib and jb
                                                         ! the same for B
  integer, optional              :: m, n, l              ! m and n are the
                                                         ! number of rows and cols to
                                                         ! be copied. l is the number of
                                                         ! rows in the upper trapezoidal
                                                         ! part of the block (same meaning
                                                         ! as in tpqrt
  complex(r64), optional            :: alpha                ! alpha scalar

  integer   :: a_m, a_n, a_mb
  integer   :: b_m, b_n, b_mb
  integer   :: a_i, a_j
  integer   :: b_i, b_j
  integer   :: im, in, il
  integer   :: a_r, a_c
  integer   :: b_r, b_c
  integer   :: rb, cb
  integer   :: a_bi, a_bj, a_bm, a_bn, a_bl
  integer   :: b_bi, b_bj, b_bm, b_bn, b_bl
  integer   :: bc_i, bc_j, bc_m, bc_n, bc_l
  integer   :: off, r, c, tmp
  complex(r64) :: ialpha
  
  ! error handling
  integer                        :: info
  character(len=*), parameter    :: name='qrm_dsmat_axpy_async'

  if(qrm_dscr%info.ne.0) return
  info = 0

  __QRM_IFPRESENT3(im,m,a%m)
  __QRM_IFPRESENT3(in,n,a%n)
  __QRM_IFPRESENT3(il,l,0)
  __QRM_IFPRESENT3(a_i,ia,1)
  __QRM_IFPRESENT3(a_j,ja,1)
  __QRM_IFPRESENT3(b_i,ib,1)
  __QRM_IFPRESENT3(b_j,jb,1)
  __QRM_IFPRESENT3(ialpha,alpha,qrm_zone)

  a_m = a%m; a_n=a%n; a_mb=a%mb
  b_m = b%m; b_n=b%n; b_mb=b%mb
  
  if(min(im,in).le.0) return

  if(.not.b%inited) then
     info = 1000
     call qrm_error_print(info, name)
     goto 9999
  end if

  bc_j = a_j
  bc_i = a_i
  c    = 1

  do
     if(bc_j.ge.a_j+in) exit
     a_c  = (bc_j-1)/a_mb+1
     b_c  = (bc_j-a_j+b_j-1)/b_mb+1
     bc_n = min(min(a_c*a_mb, b_c*b_mb - b_j+a_j),a_j+in-1) - bc_j + 1
     bc_m = min( im-il+ bc_j +bc_n - a_j, im )

     ! if(min(bc_m,bc_n).le.0) cycle
     
     if( bc_j.ge.a_j+il ) then
        bc_l = 0
     else
        bc_l = bc_m-( im-il+ bc_j - a_j)
     end if

     ! write(*,'("Bcol: ",i2,"   -- a_c:",i2,"  b_c:",i2)')c,a_c,b_c
     ! write(*,'(" -- bc_i: ",i3)', advance='no')bc_i
     ! write(*,'(" -- bc_j: ",i3)', advance='no')bc_j
     ! write(*,'(" -- bc_m: ",i3)', advance='no')bc_m
     ! write(*,'(" -- bc_n: ",i3)', advance='no')bc_n
     ! write(*,'(" -- bc_l: ",i3)')bc_l

     r    = 1
     bc_i = a_i
     
     do 
        if(bc_i.ge.a_i+bc_m) exit
        a_r  = (bc_i-1)/a_mb+1
        b_r  = (bc_i-a_i+b_i-1)/b_mb+1
        a_bm = min(min(a_r*a_mb, b_r*b_mb - b_i+a_i),a_i+bc_m-1) - bc_i + 1
        a_bl = max(a_bm - max(a_i + (bc_m-bc_l) - bc_i, 0),0)
        off  = max(0, bc_i - (a_i+bc_m-bc_l-1)-1)
        a_bi = bc_i
        a_bj = bc_j+off
        b_bi = a_bi-a_i+b_i
        b_bj = a_bj-a_j+b_j
        a_bn = bc_n-off
        a_bi = a_bi - (a_r-1)*a_mb
        a_bj = a_bj - (a_c-1)*a_mb
        b_bi = b_bi - (b_r-1)*b_mb
        b_bj = b_bj - (b_c-1)*b_mb

        if(min(a_bm,a_bn).gt.0) then
           ! write(*,'(" -- a_r:",i2," a_c:",i2,"  b_r:",i2," b_c:",i2)', advance='no')a_r,a_c,b_r,b_c
           ! write(*,'(", a_bi: ",i3)', advance='no')a_bi
           ! write(*,'(", a_bj: ",i3)', advance='no')a_bj
           ! write(*,'(", b_bi: ",i3)', advance='no')b_bi
           ! write(*,'(", b_bj: ",i3)', advance='no')b_bj
           ! write(*,'(", a_bm: ",i3)', advance='no')a_bm
           ! write(*,'(", a_bn: ",i3)', advance='no')a_bn
           ! write(*,'(", a_bl: ",i3)')a_bl
           ! write(*,'(" ")')

           call zqrm_block_axpy_task(qrm_dscr, ialpha, &
                a%blocks(a_r,a_c), b%blocks(b_r,b_c),   &
                ia=a_bi, ja=a_bj, ib=b_bi, jb=b_bj,     &
                m=a_bm, n=a_bn, l=a_bl)
        end if
        bc_i = bc_i+a_bm
     end do

     bc_j = bc_j + bc_n
     c = c+1
     ! write(*,'(" ")')
     ! write(*,'(" ")')
     
  end do


9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return
end subroutine zqrm_dsmat_axpy_async
