!! ##############################################################################################
!!
!! Copyright 2012-2018 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains synchronous and asynchronous routines for
!! copying a matrix
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine zqrm_dsmat_extadd(a, b, i, j, m, n, l, inout, op, rowmap, colmap, info)

  use zqrm_dsmat_mod, alias=>zqrm_dsmat_extadd
  use qrm_dscr_mod

  implicit none


  type(zqrm_dsmat_type)          :: a                    ! if copyout this is the source
                                                         ! matrix. If copyin it is the
                                                         ! destinantion matrix

  type(zqrm_dsmat_type)          :: b                    ! if copyout is the destination
                                                         ! matrix; if the source one.

  integer                        :: i, j, m, n, l        ! i and j are the first row
                                                         ! and column within the
                                                         ! matrix and m and n are the
                                                         ! number of rows and cols to
                                                         ! be copied. l is the number of
                                                         ! rows in the upper trapezoidal
                                                         ! part of the block (same meaning
                                                         ! as in tpqrt
  character                      :: inout, op
  integer                        :: rowmap(:,:)         ! block-row mapping from a to b
  integer, optional              :: colmap(:,:)         ! block-column mapping from a to b

  integer, optional              :: info

  type(qrm_dscr_type)            :: qrm_dscr

  ! error handling
  integer                        :: err
  character(len=*), parameter    :: name='qrm_dsmat_extadd'

  err = 0

  if(.not.b%inited) then
     err = 1000
     call qrm_error_print(err, name)
     goto 9999
  end if

  call qrm_dscr_init(qrm_dscr, nocuda=.true.)

  call zqrm_dsmat_extadd_async(qrm_dscr, a, b, i=i, j=j, m=m, n=n, l=l,  &
       inout=inout, op=op,                                               &
       rowmap=rowmap, colmap=colmap)

  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, err)

  call qrm_dscr_destroy(qrm_dscr)

9999 continue
  if(present(info)) info = err
  return

end subroutine zqrm_dsmat_extadd


subroutine zqrm_dsmat_extadd_async(qrm_dscr, a, b, i, j, m, n, l, inout, op, rowmap, colmap &
#if defined(have_starpu)
     ,hdl &
#endif
     )
  use zqrm_dsmat_mod, alias=>zqrm_dsmat_extadd_async
  use qrm_dscr_mod
  use qrm_mem_mod
  use iso_c_binding
  use zqrm_dsmat_task_mod

  implicit none

  type(qrm_dscr_type), target    :: qrm_dscr
  type(zqrm_dsmat_type)          :: a                    ! if copyout this is the source
                                                         ! matrix. If copyin it is the
                                                         ! destinantion matrix

  type(zqrm_dsmat_type)          :: b                    ! if copyout is the destination
                                                         ! matrix; if the source one.

  integer                        :: i, j, m, n, l        ! i and j are the first row
                                                         ! and column within the
                                                         ! matrix and m and n are the
                                                         ! number of rows and cols to
                                                         ! be copied. l is the number of
                                                         ! rows in the upper trapezoidal
                                                         ! part of the block (same meaning
                                                         ! as in tpqrt
  character                      :: inout, op
  integer                        :: rowmap(:,:)         ! block-row mapping from a to b
  integer, optional              :: colmap(:,:)         ! block-column mapping from a to b

#if defined(have_starpu)
  type(c_ptr), optional          :: hdl
#endif

  integer                        :: fbrow, fbcol, nbrow, nbcol, lbrow, lbcol
  integer                        :: bc, br, bm, bn, bi, bj, bl
  integer                        :: gi, gj, bcm, bcl

  ! error handling
  integer                        :: info
  character(len=*), parameter    :: name='qrm_dsmat_extadd_async'

  if(qrm_dscr%info.ne.0) return
  info = 0

  if(min(m,n).le.0) return

  if(.not.b%inited) then
     info = 1000
     call qrm_error_print(info, name)
     goto 9999
  end if


  fbrow = (i-1)/a%mb+1
  fbcol = (j-1)/a%mb+1

  lbcol = (j+n-2)/a%mb+1

  ! FIXME: add check for dimensions or automatically set them

  cols: do bc=fbcol, lbcol
     bj = max(1, j - (bc-1)*a%mb)
     bn = min(size(a%blocks(1,bc)%c,2), (j+n-1)-(bc-1)*a%mb) - bj+1

     gj = (bc-1)*a%mb+bj-j+1
     bcm = min( m-l+ gj +bn-1, m )
     if( gj.ge.l ) then
        bcl = 0
     else
        bcl = bcm-m+l-gj+1
     end if

     lbrow = (i+bcm-2)/a%mb+1
     rows: do br=fbrow, lbrow
        if(.not.qrm_allocated(a%blocks(br,bc)%c)) cycle rows
        bi = max(1, i - (br-1)*a%mb)
        bm = min(size(a%blocks(br,bc)%c,1), (i+bcm-1)-(br-1)*a%mb) - bi+1
        gi = (br-1)*a%mb+bi-i+1
        if((bcm-bcl)-gi+1 .ge. bm) then
           bl = 0
        else
           bl = bm - ( bcm -bcl -gi +1)
        end if
        if (bl.gt.bm) then
           bj = bj+bl-bm
           bn = bn-bl+bm
           bl = bm
        end if

        call zqrm_block_extadd_task(qrm_dscr, a, b, br, bc, i=bi, j=bj, m=bm, n=bn, l=bl, &
             inout=inout, op=op,                                                          &
             rowmap=rowmap, colmap=colmap                                                 &
#if defined(have_starpu)
             ,hdl=hdl                                                                     &
#endif
             )
     end do rows
  end do cols

9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return
end subroutine zqrm_dsmat_extadd_async
