!! ##############################################################################################
!!
!! Copyright 2012-2018 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains synchronous and asynchronous routines for
!! multiplying two matrices
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine zqrm_dsmat_gemm(transa, transb, alpha, a, b, beta, c, m, n, k, prio, info)

  use zqrm_dsmat_mod, s=>zqrm_dsmat_gemm
  use qrm_dscr_mod

  implicit none
  type(zqrm_dsmat_type)       :: a, b, c
  complex(r64)                   :: alpha, beta
  character                   :: transa, transb 
  integer, optional           :: m, n, k, prio
  integer, optional           :: info

  type(qrm_dscr_type)         :: qrm_dscr

  ! error handling
  integer                     :: err
  character(len=*), parameter :: name='qrm_dsmat_gemm'

  err = 0

  if((.not.b%inited) .or. (.not.a%inited) .or. (.not.c%inited)) then
     err = 1000
     call qrm_error_print(err, name)
     goto 9999
  end if
  
  call qrm_dscr_init(qrm_dscr)

  call zqrm_dsmat_gemm_async(qrm_dscr, transa, transb, alpha, a, b, beta, c, m, n, k, prio)

  ! wait for all the tasks in the descriptor
  call qrm_barrier(qrm_dscr, err)

  call qrm_dscr_destroy(qrm_dscr)

9999 continue
  if(present(info)) info = err
  return

end subroutine zqrm_dsmat_gemm

subroutine zqrm_dsmat_gemm_async(qrm_dscr, transa, transb, alpha, a, b, beta, c, m, n, k, prio)

  use zqrm_dsmat_mod, s=>zqrm_dsmat_gemm_async
  use qrm_dscr_mod
  use zqrm_dsmat_task_mod
  use qrm_string_mod
  use qrm_parameters_mod
  
  implicit none

  type(qrm_dscr_type)            :: qrm_dscr
  type(zqrm_dsmat_type), target  :: a, b, c
  complex(r64)                      :: alpha, beta
  character                      :: transa, transb 
  integer, optional              :: m, n, k, prio

  type(zqrm_block_type), pointer :: ail, blj, cij
  integer                        :: i, j, l, mb, nb, kb, iprio
  integer                        :: im, in, ik, mm, nn, kk
  complex(r64)                      :: ibeta
  character                      :: itransa, itransb 

  ! error handling
  integer                        :: info
  character(len=*), parameter    :: name='qrm_dsmat_gemm_async'

  if(qrm_dscr%info.ne.0) return
  info = 0

  if((.not.b%inited) .or. (.not.a%inited) .or. (.not.c%inited)) then
     info = 1000
     call qrm_error_print(info, name)
     goto 9999
  end if
  
  __QRM_IFPRESENT3(iprio, prio, 0)
  __QRM_IFPRESENT3(im, m, c%m)
  __QRM_IFPRESENT3(in, n, c%n)
  
  mb  = (im-1)/c%mb+1
  nb  = (in-1)/c%mb+1

  
  if((qrm_str_tolower(transa) .eq. qrm_conj_transp) .or. &
       (qrm_str_tolower(transa) .eq. qrm_transp)) then
     itransa = qrm_conj_transp
     __QRM_IFPRESENT3(ik, k, a%m)
  else
     itransa = qrm_no_transp
     __QRM_IFPRESENT3(ik, k, a%n)
  end if
  kb  = (ik-1)/a%mb+1

  if((qrm_str_tolower(transb) .eq. qrm_conj_transp) .or. &
       (qrm_str_tolower(transb) .eq. qrm_transp)) then
     itransb = qrm_conj_transp
  else
     itransb = qrm_no_transp
  end if
  
  do i=1, mb
     if(i.eq.mb) then; mm = im-(i-1)*c%mb; else; mm = c%mb; end if
     do j=1, nb
        cij => c%blocks(i,j)
        if(j.eq.nb) then; nn = in-(j-1)*c%mb; else; nn = c%mb; end if
        do l=1, kb
           if(l.eq.1) then
              ibeta=beta
           else
              ibeta=qrm_zone
           end if
           if(itransa.eq.qrm_no_transp) then
              ail => a%blocks(i,l)
           else
              ail => a%blocks(l,i)
           end if
           if(itransb.eq.qrm_no_transp) then
              blj => b%blocks(l,j)
           else
              blj => b%blocks(j,l)
           end if

           if(l.eq.kb) then; kk = ik-(l-1)*c%mb; else; kk = c%mb; end if

           if(qrm_allocated(ail%c) .and.      &
                qrm_allocated(blj%c) .and.    &
                qrm_allocated(cij%c))         &
                call zqrm_gemm_task(qrm_dscr, &
                     itransa, itransb,        &
                     mm, nn, kk,              &
                     alpha,                   &
                     ail,                     &
                     blj,                     &
                     ibeta,                   &
                     cij,                     &
                     iprio )

        end do
     end do
  end do

9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return

end subroutine zqrm_dsmat_gemm_async
