!! ##############################################################################################
!!
!! Copyright 2012-2018 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains synchronous and asynchronous routines for
!! computing the QR factorization of a triangle on top of a rectangle
!! or triangle
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine zqrm_dsmat_tpqr(a, b, t, ib, ts, work, prio, info)

  use zqrm_dsmat_mod, alias=>zqrm_dsmat_tpqr
  use qrm_dscr_mod
  implicit none

  type(zqrm_dsmat_type)       :: a, b, t
  type(zqrm_ws_type)          :: work
  integer                     :: ib
  character                   :: ts
  integer, optional           :: prio
  integer, optional           :: info

  type(qrm_dscr_type)         :: qrm_dscr

  ! error handling
  integer                     :: err
  character(len=*), parameter :: name='qrm_dsmat_tpqr'

  err = 0
  call qrm_dscr_init(qrm_dscr)

  call zqrm_dsmat_tpqr_async(qrm_dscr, a, b, t, ib, ts, work, prio)

  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, err)

  call qrm_dscr_destroy(qrm_dscr)

9999 continue
  if(present(info)) info = err
  return
end subroutine zqrm_dsmat_tpqr

subroutine zqrm_dsmat_tpqr_async(qrm_dscr, a, b, t, ib, ts, work, prio)

  use zqrm_dsmat_mod, alias=>zqrm_dsmat_tpqr_async
  use qrm_dscr_mod
  implicit none

  type(qrm_dscr_type)         :: qrm_dscr
  type(zqrm_dsmat_type)       :: a, b, t
  type(zqrm_ws_type)          :: work
  integer                     :: ib
  character                   :: ts
  integer, optional           :: prio

  integer                     :: i, j, k, m, nbr, nbc, nbe, iprio

  ! error handling
  integer                     :: info
  character(len=*), parameter :: name='qrm_dsmat_tpqr_async'

  if(qrm_dscr%info.ne.0) return
  info = 0

  if(min(a%m,a%n).eq.0) return

  __QRM_IFPRESENT3(iprio, prio, 0)
  
  nbr = size(b%blocks,1)
  nbc = size(a%blocks,2)
  nbe = min(nbr,nbc)

  do k=1, nbc
     if(ts.eq.qrm_square) then
        m = nbr
     else
        m = min(k-1,nbr)
     end if
     do i = 1, m
        call zqrm_hitpqrt(qrm_dscr,           &
             a%blocks(k,k),                   &
             b%blocks(i,k),                   &
             t%blocks(i,k),                   &
             a%nb, ib, qrm_square,            &
             work, iprio)
        do j=k+1, nbc
           call zqrm_hitpmqrt(qrm_dscr,       &
                b%blocks(i,k),                &
                t%blocks(i,k),                &
                a%blocks(k,j),                &
                b%blocks(i,j),                &
                a%nb, ib, qrm_square,         &
                work, iprio)
        end do
     end do
     if((ts.eq.qrm_triang) .and. (m.lt.nbr)) then
        i = k
        call zqrm_hitpqrt(qrm_dscr,           &
             a%blocks(k,k),                   &
             b%blocks(i,k),                   &
             t%blocks(i,k),                   &
             a%nb, ib, qrm_triang,            &
             work, iprio)
        do j=k+1, nbc
           call zqrm_hitpmqrt(qrm_dscr,       &
                b%blocks(i,k),                &
                t%blocks(i,k),                &
                a%blocks(k,j),                &
                b%blocks(i,j),                &
                a%nb, ib, qrm_triang,         &
                work, iprio)
        end do
     end if

  end do

9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return
end subroutine zqrm_dsmat_tpqr_async



subroutine zqrm_dsmat_init_t_tpqr(qrm_dsmat_a, qrm_dsmat_t, ib, ts, pin, seq, info)

#if defined (have_starpu)
  use fstarpu_mod
  use qrm_starpu_common_mod
#endif
  use qrm_parameters_mod
  use qrm_error_mod
  use qrm_mem_mod
  use zqrm_dsmat_mod, alias=>zqrm_dsmat_init_t_tpqr
  use qrm_dscr_mod

  implicit none

  type(zqrm_dsmat_type), intent(in)            :: qrm_dsmat_a
  type(zqrm_dsmat_type), intent(inout), target :: qrm_dsmat_t
  integer, intent(in)                          :: ib
  character                                    :: ts
  integer, optional                            :: info
  logical, optional                            :: pin, seq

  integer                                      :: i, j, is, js, parts, nbr, nbc, nb, m
  type(c_ptr)                                  :: bc_filter
  logical                                      :: iseq
  
  ! error handling
  integer                                      :: err
  character(len=*), parameter                  :: name='qrm_dsmat_init_t_tpqr'

  err = 0

  if(.not.qrm_dsmat_a%inited) return
  
  if(present(seq)) then
     iseq = seq
  else
     iseq = .false.
  end if
  
  nbr = size(qrm_dsmat_a%blocks,1)
  nbc = size(qrm_dsmat_a%blocks,2)
  nb  = qrm_dsmat_a%nb

  ! FIXME: add error checks

#if defined (have_starpu)
  if(.not.iseq) then
     bc_filter = qrm_get_bc_filter()
     call fstarpu_data_filter_set_filter_arg (bc_filter, nb)
  end if
#endif

  allocate(qrm_dsmat_t%blocks(nbr,nbc))

  do j=1, nbc
     if(ts.eq.qrm_square) then
        m = nbr
     else
        m = min(j,nbr)
     end if
     do i=1, m
        if(.not.qrm_allocated(qrm_dsmat_a%blocks(i,j)%c)) cycle

        is = size(qrm_dsmat_a%blocks(i,j)%c,1)
        js = size(qrm_dsmat_a%blocks(i,j)%c,2)

        call qrm_alloc(qrm_dsmat_t%blocks(i,j)%c, ib, js, err, pin)
        __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)

        qrm_dsmat_t%blocks(i,j)%c = qrm_zzero

#if defined (have_starpu)
        if(.not.iseq) then
           ! register the block to starpu
           call fstarpu_matrix_data_register(qrm_dsmat_t%blocks(i,j)%hdl, 0, &
                c_loc(qrm_dsmat_t%blocks(i,j)%c),                            &
                size(qrm_dsmat_t%blocks(i,j)%c,1),                           &
                size(qrm_dsmat_t%blocks(i,j)%c,1),                           &
                size(qrm_dsmat_t%blocks(i,j)%c,2),                           &
                c_sizeof(qrm_dsmat_t%blocks(i,j)%c(1,1)))
           if(nb.lt.js) then
              ! plan a dynamic partitioning
              parts = (js-1)/nb+1
              allocate(qrm_dsmat_t%blocks(i,j)%shdls(parts))
              call fstarpu_data_partition_plan(qrm_dsmat_t%blocks(i,j)%hdl, bc_filter, qrm_dsmat_t%blocks(i,j)%shdls)
           end if

        end if
#endif
        qrm_dsmat_t%blocks(i,j)%partitioned = .false.
     end do
  end do

  qrm_dsmat_t%inited = .true.

9999 continue 
  ! should do cleanup
  if(present(info)) info = err
  return
end subroutine zqrm_dsmat_init_t_tpqr
