!! ##############################################################################################
!!
!! Copyright 2012-2018 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains synchronous and asynchronous routines for
!! computing a triangular system solution
!!
!! ##############################################################################################

#include "qrm_common.h"


subroutine zqrm_dsmat_trsm(side, uplo, trans, diag, alpha, a, b, m, n, k, prio, info)

  use zqrm_dsmat_mod, alias => zqrm_dsmat_trsm
  use qrm_dscr_mod
  implicit none

  type(zqrm_dsmat_type)       :: a, b
  complex(r64)                :: alpha
  character                   :: side, uplo, trans, diag
  integer, optional           :: m, n, k, prio, info

  type(qrm_dscr_type)         :: qrm_dscr

  ! error handling
  integer                     :: err
  character(len=*), parameter :: name='qrm_dsmat_trsm'

  err = 0

  call qrm_dscr_init(qrm_dscr)

  call zqrm_dsmat_trsm_async(qrm_dscr, side, uplo, trans, diag, alpha, a, b, m, n, k, prio)

  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, err)

  call qrm_dscr_destroy(qrm_dscr)

9999 continue
  if(present(info)) info = err
  return

end subroutine zqrm_dsmat_trsm

subroutine zqrm_dsmat_trsm_async(qrm_dscr, side, uplo, trans, diag, alpha, a, b, m, n, k, prio)

  use zqrm_dsmat_mod, alias => zqrm_dsmat_trsm_async
  use qrm_dscr_mod
  use zqrm_dsmat_task_mod
  use qrm_parameters_mod
  implicit none

  type(qrm_dscr_type)         :: qrm_dscr
  type(zqrm_dsmat_type)       :: a, b
  complex(r64)                :: alpha
  character                   :: side, uplo, trans, diag
  integer, optional           :: prio, m, n, k

  integer                     :: i, j, l, iprio
  complex(r64)                :: ialpha
  integer                     :: im, in, ik, mm, nn, kk, mb, nb, kb

  ! error handling
  integer                     :: info
  character(len=*), parameter :: name='qrm_dsmat_trsm_async'

  if(qrm_dscr%info.ne.0) return
  info = 0

  __QRM_IFPRESENT3(iprio, prio, 0)
  __QRM_IFPRESENT3(im, m, a%m)
  __QRM_IFPRESENT3(in, n, b%n)
  __QRM_IFPRESENT3(ik, k, a%n)

  im = min(im,ik)
  
  mb  = (im-1)/a%mb+1
  nb  = (in-1)/b%mb+1
  kb  = (ik-1)/a%mb+1
  
  if(present(prio)) then
     iprio = prio
  else
     iprio = 0
  end if

  if(side.eq.qrm_left) then
     if(uplo.eq.qrm_upper) then
        if(trans.eq.qrm_no_transp) then
           do j=1, nb
              if(j.eq.nb) then
                 nn=in-(j-1)*b%mb
              else
                 nn=b%mb
              end if
              
              do l = kb, 1, -1
                 if(l.eq.kb) then
                    kk = ik-(l-1)*a%mb
                    ialpha = alpha
                 else
                    kk = a%mb
                    ialpha = qrm_zone
                 end if

                 ! b_j = ialpha*b_j/a_jj
                 if(l.le.mb) then
                    if(l.eq.mb) then
                       mm=im-(l-1)*a%mb
                    else
                       mm=a%mb
                    end if
                    call zqrm_trsm_task(qrm_dscr, &
                         qrm_left, qrm_upper,     &
                         qrm_no_transp, diag,     & 
                         mm, nn, kk,              &
                         ialpha,                  & 
                         a%blocks(l,l),           & 
                         b%blocks(l,j),           & 
                         iprio )
                 end if
                 do i=min(l-1, mb), 1, -1
                    ! b_i = ialpha*b_i - a_ij*b_j
                    if(i.eq.mb) then
                       mm=im-(i-1)*a%mb
                    else
                       mm=a%mb
                    end if
                    call zqrm_gemm_task(qrm_dscr,      &
                         qrm_no_transp, qrm_no_transp, &
                         mm, nn, kk,                   &
                         qrm_zmone,                    &
                         a%blocks(i,l),                &
                         b%blocks(l,j),                &
                         ialpha,                       &
                         b%blocks(i,j),                &
                         iprio)
                 end do

              end do
           end do
        else if(trans.eq.qrm_conj_transp) then
           do j=1, nb
              if(j.eq.nb) then
                 nn=in-(j-1)*b%mb
              else
                 nn=b%mb
              end if
              do i= 1, min(mb,kb)
                 if(i.eq.1) then
                    ialpha = alpha
                 else
                    ialpha = qrm_zone
                 end if

                 if(i.eq.mb) then
                    mm=im-(i-1)*a%mb
                 else
                    mm=a%mb
                 end if

                 if(i.eq.kb) then
                    kk = ik-(i-1)*a%mb
                 else
                    kk = a%mb
                 end if
                 ! b_i = ialpha*b_i/a_ii'
                 call zqrm_trsm_task(qrm_dscr, &
                      qrm_left, qrm_upper,     &
                      qrm_conj_transp, diag,   &   
                      mm, nn, kk,              &
                      ialpha,                  &
                      a%blocks(i,i),           &
                      b%blocks(i,j),           &
                      iprio)

                 do l=i+1, kb
                    if(l.eq.kb) then
                       kk = ik-(l-1)*a%mb
                       ialpha = alpha
                    else
                       kk = a%mb
                       ialpha = qrm_zone
                    end if
                    ! b_j = ialpha*b_j - a_ij'*b_i
                    call zqrm_gemm_task(qrm_dscr,        &
                         qrm_conj_transp, qrm_no_transp, &
                         kk, nn, mm,                     &
                         qrm_zmone,                      &
                         a%blocks(i,l),                  &
                         b%blocks(i,j),                  &
                         ialpha,                         &
                         b%blocks(l,j),                  &
                         iprio )
                 end do
              end do
           end do
        end if
     else
        write(*,'("TRSM with uplo=l not yet implemented")')
     end if
  else
     write(*,'("TRSM with side=r not yet implemented")')
  end if


9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return

end subroutine zqrm_dsmat_trsm_async

