!! ##############################################################################################
!!
!! Copyright 2012-2018 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file defines a task for summing two blocks
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine cqrm_block_axpy_task(qrm_dscr, alpha, a, b, ia, ja, ib, jb, m, n, l)
#if defined(have_starpu)
  use fstarpu_mod
  use cqrm_starpu_mod
#endif
  use qrm_dscr_mod
  use cqrm_dsmat_mod
  use qrm_string_mod
  use iso_c_binding
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none


  type(qrm_dscr_type), target    :: qrm_dscr
  complex(r32), target           :: alpha
  type(cqrm_block_type), target  :: a                    ! the source block
  type(cqrm_block_type), target  :: b                    ! the destination block
  integer, target                :: ia, ja, ib, jb       ! ia and ja are the staring row and col
                                                         ! in the A matrix matrix. ib and jb
                                                         ! the same for B
  integer, target                :: m, n, l              ! m and n are the
                                                         ! number of rows and cols to
                                                         ! be copied. l is the number of
                                                         ! rows in the upper trapezoidal
                                                         ! part of the block (same meaning
                                                         ! as in tpqrt
#if defined (have_starpu)
  type(c_ptr)                    :: sizeof
  integer, target                :: prio
  type(c_ptr), target            :: dscrpt
#endif
  integer(kind=8)                :: ts, te

  if(qrm_dscr%info.ne.0) return
  
#if defined (have_starpu)
  if(qrm_dscr%seq.eq.0) then

#if defined(dprec)
       sizeof = FSTARPU_SZ_REAL8
#elif defined(sprec)
       sizeof = FSTARPU_SZ_REAL4
#elif defined(zprec)
       sizeof = FSTARPU_SZ_COMPLEX8
#elif defined(cprec)
       sizeof = FSTARPU_SZ_COMPLEX4
#endif

     dscrpt = c_loc(qrm_dscr)

     prio = 0

     call fstarpu_insert_task((/ cqrm_block_axpy_cl,                &
          FSTARPU_VALUE, c_loc(dscrpt),       FSTARPU_SZ_C_PTR,     &
          FSTARPU_VALUE, c_loc(alpha),        sizeof,               &
          FSTARPU_VALUE, c_loc(ia),           FSTARPU_SZ_INTEGER,   &
          FSTARPU_VALUE, c_loc(ja),           FSTARPU_SZ_INTEGER,   &
          FSTARPU_VALUE, c_loc(ib),           FSTARPU_SZ_INTEGER,   &
          FSTARPU_VALUE, c_loc(jb),           FSTARPU_SZ_INTEGER,   &
          FSTARPU_VALUE, c_loc(m),            FSTARPU_SZ_INTEGER,   &
          FSTARPU_VALUE, c_loc(n),            FSTARPU_SZ_INTEGER,   &
          FSTARPU_VALUE, c_loc(l),            FSTARPU_SZ_INTEGER,   &
          FSTARPU_R, a%hdl,                                         &
          FSTARPU_RW, b%hdl,                                        &
          FSTARPU_PRIORITY, c_loc(prio),                            &
          FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),                   &
          C_NULL_PTR /))
          call fstarpu_task_wait_for_all()

  else
#endif
     __QRM_PROF_START(ts)
     call cqrm_axpy(alpha,                    &
                    a%c, size(a%c,1), ia, ja, &
                    b%c, size(b%c,1), ib, jb, &
                    m, n, l)
     __QRM_PROF_STOP(qrm_prof_axpy_task_,ts,te)
#if defined (have_starpu)
  end if
#endif

  return
end subroutine cqrm_block_axpy_task


#if defined (have_starpu)
recursive subroutine cqrm_starpu_block_axpy_cpu_func(buffers, cl_arg) bind(C)
  use fstarpu_mod
  use cqrm_dsmat_mod
  use qrm_dscr_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(c_ptr), value                :: cl_arg
  type(c_ptr), value                :: buffers

  integer, target                   :: ia, ja, ib, jb, m, n, l
  type(c_ptr), target               :: dscr_c
  type(qrm_dscr_type), pointer      :: qrm_dscr
  complex(r32), target              :: alpha
  integer                           :: ma, na, lda, mb, nb, ldb
  complex(r32), pointer             :: a(:,:), b(:,:)
  integer(kind=8)                   :: ts, te
  
  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c), c_loc(alpha), &
       c_loc(ia), c_loc(ja), c_loc(ib), c_loc(jb),                &
       c_loc(m), c_loc(n), c_loc(l) /))

  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)

  if(qrm_dscr%info.ne.0) return

  ma   = fstarpu_matrix_get_nx(buffers, 0)
  na   = fstarpu_matrix_get_ny(buffers, 0)
  lda  = fstarpu_matrix_get_ld(buffers, 0)
  mb   = fstarpu_matrix_get_nx(buffers, 1)
  nb   = fstarpu_matrix_get_ny(buffers, 1)
  ldb  = fstarpu_matrix_get_ld(buffers, 1)
  
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 0), a, shape=(/lda,na/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 1), b, shape=(/ldb,nb/))
  
  __QRM_PROF_START(ts)
  call cqrm_axpy(alpha,               &
                 a(1,1), lda, ia, ja, &
                 b(1,1), ldb, ib, jb, &
                 m, n, l)
  __QRM_PROF_STOP(qrm_prof_axpy_task_,ts,te)
  
  return

end subroutine cqrm_starpu_block_axpy_cpu_func
#endif





! subroutine cqrm_block_axpy_task(qrm_dscr, a, b, br, bc, i, j, m, n, l, alpha, rowmap, colmap, inout &
! #if defined(have_starpu)
!      ,hdl &
! #endif
!      )
! #if defined(have_starpu)
!   use fstarpu_mod
!   use cqrm_starpu_mod
! #endif
!   use qrm_dscr_mod
!   use cqrm_dsmat_mod
!   use qrm_string_mod
!   use iso_c_binding
! #if defined (perfprof)
!   use qrm_perfprof_mod
! #endif
!   implicit none


!   type(qrm_dscr_type), target    :: qrm_dscr
!   type(cqrm_dsmat_type), target  :: a                    ! if copyout this is the source
!                                                          ! matrix. If copyin it is the
!                                                          ! destinantion matrix

!   type(cqrm_dsmat_type), target  :: b                    ! if copyout is the destination
!                                                          ! matrix; if the source one.

!   integer, target                :: br, bc               ! block-row and block-column
!                                                          ! of the source/destination
!                                                          ! block in a
!   integer, optional              :: i, j, m, n, l        ! i and j are the first row
!                                                          ! and column within the
!                                                          ! matrix and m and n are the
!                                                          ! number of rows and cols to
!                                                          ! be copied. l is the number of
!                                                          ! rows in the upper trapezoidal
!                                                          ! part of the block (same meaning
!                                                          ! as in tpqrt
!   complex(r32), optional            :: alpha
!   character, optional            :: inout                ! if 'o' copies out (scatter); if
!                                                          ! 'i' copies in (gather)
!   integer, optional, target      :: rowmap(:), colmap(:) ! row and column mappings from a to b

! #if defined(have_starpu)
!   type(c_ptr), optional          :: hdl
! #endif

! #if defined (have_starpu)
!   integer, target                :: nh, prio
!   type(c_ptr), target            :: a_ptr, b_ptr, rmptr, cmptr
!   type(c_ptr)                    :: hdls, ptr, amode, bmode
!   type(c_ptr), target            :: dscrpt

!   integer                        :: abr, abc ! block row and col in a (just a copy of br and bc)
!   integer                        :: bbr, bbc ! block row and col in b
!   integer                        :: old_bbr, old_bbc
!   character, target              :: iout
!   complex(r32), target              :: ialpha
!   integer                        :: ai, aj, agi, agj ! local (to block) and global row and col indices in a
!   integer                        :: bi, bj, bgi, bgj ! local (to block) and global row and col indices in b
!   integer                        :: aroff, acoff ! row and column offsets in a
!   integer, target                :: ii, ij, im, in, il
!   integer, target                :: cms, rms
!   type(c_ptr)                    :: sizeof
! #endif
!   integer(kind=8)                :: ts, te


!   if(qrm_dscr%info.ne.0) return
  
! #if defined (have_starpu)

!   if(qrm_dscr%seq.eq.0) then

! #if defined(dprec)
!        sizeof = FSTARPU_SZ_REAL8
! #elif defined(sprec)
!        sizeof = FSTARPU_SZ_REAL4
! #elif defined(zprec)
!        sizeof = FSTARPU_SZ_COMPLEX8
! #elif defined(cprec)
!        sizeof = FSTARPU_SZ_COMPLEX4
! #endif
     
!      prio = 0
!      a_ptr = c_loc(a)
!      b_ptr = c_loc(b)
     
!      __QRM_IFPRESENT3(ii, i, 1)
!      __QRM_IFPRESENT3(ij, j, 1)
!      __QRM_IFPRESENT3(im, m, size(a%blocks(br,bc)%c,1)-ii+1)
!      __QRM_IFPRESENT3(in, n, size(a%blocks(br,bc)%c,2)-ij+1)
!      __QRM_IFPRESENT3(il, l, 0)
!      __QRM_IFPRESENT3(iout, inout, 'o')
!      __QRM_IFPRESENT3(ialpha, alpha, qrm_cone)
     
!      if(present(rowmap)) then
!         rmptr = c_loc(rowmap(1))
!         rms   = size(rowmap)
!      else
!         rmptr = c_null_ptr
!         rms   = 0
!      end if
     
!      if(present(colmap)) then
!         cmptr = c_loc(colmap(1))
!         cms   = size(colmap)
!      else
!         cmptr = c_null_ptr
!         cms   = 0
!      end if
     
!      if(qrm_str_tolower(iout).eq.'o') then
!         ! if copyout (scatter) a is read and b written
!         amode = FSTARPU_R
!         bmode = FSTARPU_RW
!      else
!         ! if copyin (gather) a is written and b read
!         amode = FSTARPU_RW
!         bmode = FSTARPU_R
!      end if

!      abr = br ! just a copy to avoid confusion with blocks ob b
!      abc = bc ! just a copy to avoid confusion with blocks ob b

!      aroff = (abr-1)*a%mb  ! row-offset in a
!      acoff = (abc-1)*a%mb  ! col-offset in a

!      old_bbc = 0
!      old_bbr = 0

!      nh   = 0

!      dscrpt = c_loc(qrm_dscr)

!      hdls = fstarpu_data_descr_array_alloc(size(b%blocks,1)*size(b%blocks,2)+1)
  
!      ! add source block handle
!      if(present(hdl)) then
!         call fstarpu_data_descr_array_set(hdls, 0, &
!              hdl,                                  &
!              amode)
!      else
!         call fstarpu_data_descr_array_set(hdls, 0, &
!              a%blocks(br,bc)%hdl,             &
!              amode)
!      end if
        
!      cols: do aj = ij+in-1, ij, -1
!         agj = acoff + aj                               ! global a col index of the current coefficient
!         __QRM_IFPRESENT4(bgj,colmap,colmap(agj),agj)     ! global b col index of the current coefficient
!         if(bgj.lt.0) cycle cols                        ! if colmap(agj) is negative, this col has to be skipped
!         bbc = (bgj-1)/b%mb+1                           ! block-col in b     of the current coefficient
!         if(bbc.eq.old_bbc) cycle cols
!         old_bbc = bbc
!         old_bbr = 0
        
!         rows: do ai = ii, ii+im-1
!            if(ai-ii .gt. aj-ij+im-il) cycle cols
!            agi = aroff + ai                            ! global a row index of the current coefficient
!            __QRM_IFPRESENT4(bgi,rowmap,rowmap(agi),agi)  ! global b row index of the current coefficient
!            if(bgi.lt.0) cycle rows                     ! if rowmap(agj) is negative, this row has to be skipped
!            bbr = (bgi-1)/b%mb+1                        ! block-row in b     of the current coefficient
!            if(bbr.eq.old_bbr) cycle rows
!            old_bbr = bbr

!            if(.not. qrm_allocated(b%blocks(bbr,bbc)%c)) cycle rows
!            nh = nh+1
!            call fstarpu_data_descr_array_set(hdls, nh,                      &
!                 b%blocks(bbr,bbc)%hdl, &
!                 bmode)
!         end do rows
!      end do cols

!      nh = nh+1

!      call fstarpu_insert_task((/ cqrm_block_axpy_cl,                &
!           FSTARPU_VALUE, c_loc(dscrpt),       FSTARPU_SZ_C_PTR,     &
!           FSTARPU_VALUE, c_loc(br),           FSTARPU_SZ_C_INT,     &
!           FSTARPU_VALUE, c_loc(bc),           FSTARPU_SZ_C_INT,     &
!           FSTARPU_VALUE, c_loc(a_ptr),        FSTARPU_SZ_C_PTR,     &
!           FSTARPU_VALUE, c_loc(b_ptr),        FSTARPU_SZ_C_PTR,     &
!           FSTARPU_VALUE, c_loc(ii),           FSTARPU_SZ_C_INT,     &
!           FSTARPU_VALUE, c_loc(ij),           FSTARPU_SZ_C_INT,     &
!           FSTARPU_VALUE, c_loc(im),           FSTARPU_SZ_C_INT,     &
!           FSTARPU_VALUE, c_loc(in),           FSTARPU_SZ_C_INT,     &
!           FSTARPU_VALUE, c_loc(il),           FSTARPU_SZ_C_INT,     &
!           FSTARPU_VALUE, c_loc(ialpha),       sizeof,               &
!           FSTARPU_VALUE, c_loc(rmptr),        FSTARPU_SZ_C_PTR,     &
!           FSTARPU_VALUE, c_loc(cmptr),        FSTARPU_SZ_C_PTR,     &
!           FSTARPU_VALUE, c_loc(rms),          FSTARPU_SZ_C_INT,     &
!           FSTARPU_VALUE, c_loc(cms),          FSTARPU_SZ_C_INT,     &
!           FSTARPU_VALUE, c_loc(iout),         FSTARPU_SZ_CHARACTER, &
!           FSTARPU_DATA_MODE_ARRAY, hdls, c_loc(nh),                 &
!           FSTARPU_PRIORITY, c_loc(prio),                            &
!           FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),                   &
!           C_NULL_PTR /))
!           ! call fstarpu_task_wait_for_all()


!      call fstarpu_data_descr_array_free(hdls)
!   else
! #endif
!      __QRM_PROF_START(ts)
!      ! write(*,*)'axpy ',br,bc
!      call cqrm_block_axpy(a, b, br, bc, i=i, j=j, m=m, n=n, l=l, &
!           alpha=alpha, rowmap=rowmap, colmap=colmap, inout=inout)
!      ! call cqrm_block_extadd(a, b, br, bc, i, j, m, n, l, &
!           ! rowmap, colmap)
!      __QRM_PROF_STOP(qrm_prof_axpy_task_,ts,te)
! #if defined (have_starpu)
!   end if
! #endif

!   return
! end subroutine cqrm_block_axpy_task


! #if defined (have_starpu)
! recursive subroutine cqrm_starpu_block_axpy_cpu_func(buffers, cl_arg) bind(C)
!   use fstarpu_mod
!   use cqrm_dsmat_mod
! #if defined (perfprof)
!   use qrm_perfprof_mod
! #endif
!   implicit none

!   type(c_ptr), value             :: cl_arg
!   type(c_ptr), value             :: buffers

!   integer, target                :: br, bc, i, j, m, n, l, rms, cms
!   type(c_ptr), target            :: a_ptr, b_ptr, rmptr, cmptr
!   type(cqrm_dsmat_type), pointer :: a, b
!   character, target              :: inout
!   complex(r32), target              :: alpha
!   integer, pointer               :: rowmap(:), colmap(:)
!   type(c_ptr), target            :: dscr_c
!   type(qrm_dscr_type), pointer   :: qrm_dscr
!   integer(kind=8)                :: ts, te
  
!   __QRM_PROF_START(ts)
!   call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c),        &
!        c_loc(br), c_loc(bc),                              &
!        c_loc(a_ptr), c_loc(b_ptr),                        &
!        c_loc(i), c_loc(j), c_loc(m), c_loc(n), c_loc(l),  &
!        c_loc(alpha) ,                                     &
!        c_loc(rmptr), c_loc(cmptr), c_loc(rms), c_loc(cms),&
!        c_loc(inout) /))
  
!   ! until StarPU has a proper error handling method we have to check
!   ! manually
!   call c_f_pointer(dscr_c, qrm_dscr)
!   if(qrm_dscr%info.ne.0) return

!   call c_f_pointer(a_ptr, a)
!   call c_f_pointer(b_ptr, b)

!   if(c_associated(rmptr)) then
!      call c_f_pointer(rmptr, rowmap, shape=(/rms/))
     
!      if(c_associated(cmptr)) then
!         call c_f_pointer(cmptr, colmap, shape=(/cms/))
!         call cqrm_block_axpy(a, b, br, bc, i=i, j=j, m=m, n=n, l=l, &
!              alpha=alpha, rowmap=rowmap, colmap=colmap, inout=inout)
!      else
!         call cqrm_block_axpy(a, b, br, bc, i=i, j=j, m=m, n=n, l=l, &
!              alpha=alpha, rowmap=rowmap, inout=inout)
!      end if
!   else
!      if(c_associated(cmptr)) then
!         call c_f_pointer(cmptr, colmap, shape=(/cms/))
!         call cqrm_block_axpy(a, b, br, bc, i=i, j=j, m=m, n=n, l=l, &
!              alpha=alpha, colmap=colmap, inout=inout)
!      else
!         call cqrm_block_axpy(a, b, br, bc, i=i, j=j, m=m, n=n, l=l, &
!              alpha=alpha, inout=inout)
!      end if
!   end if
!   __QRM_PROF_STOP(qrm_prof_axpy_task_,ts,te)

!   return

! end subroutine cqrm_starpu_block_axpy_cpu_func
! #endif
