!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine cqrm_potrf_task(qrm_dscr, uplo, m, k, a, prio)

#if defined(have_starpu)
  use fstarpu_mod
  use cqrm_starpu_mod
  use qrm_pthread_mod
#endif
  use qrm_dscr_mod
  use cqrm_dsmat_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(qrm_dscr_type), target :: qrm_dscr
  character, target           :: uplo
  integer, target             :: m, k
  type(cqrm_block_type)       :: a
  integer, target             :: prio

  integer, target             :: potrfprio
  type(c_ptr)                 :: sizeof
  integer                     :: lda, i, info
  integer(kind=8)             :: ts, te
#if defined (have_starpu)
  type(c_ptr), target         :: dscrpt
#endif

  if(qrm_dscr%info.ne.0) return


#if defined (have_starpu)
  if(qrm_dscr%seq.eq.0) then

     if(qrm_dscr%ngpu>0) then
        potrfprio = potrf_hp_prio_
     else
        potrfprio = prio + potrf_prio_
     end if

     dscrpt = c_loc(qrm_dscr)

     call fstarpu_insert_task((/ cqrm_potrf_cl,                &
          FSTARPU_VALUE, c_loc(dscrpt), FSTARPU_SZ_C_PTR,      &
          FSTARPU_VALUE, c_loc(uplo),   FSTARPU_SZ_CHARACTER,  &
          FSTARPU_VALUE, c_loc(m),      FSTARPU_SZ_C_INT,      &
          FSTARPU_VALUE, c_loc(k),      FSTARPU_SZ_C_INT,      &
          FSTARPU_RW, a%hdl,                                   &
          FSTARPU_PRIORITY, c_loc(potrfprio),                  &
          FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),              &
          C_NULL_PTR /))

  else
#endif

     ! make a call to trsm
     lda  = size(a%c,1)

     __QRM_PROF_START(ts)
     call cqrm_potrf(uplo, m, k, &
          a%c(1,1), lda, info)
     __QRM_PROF_STOP(qrm_prof_potrf_task_,ts,te)

     if(info.gt.0) then
        call qrm_atomic_cas(qrm_dscr%info, 0, 37)
        call qrm_error_print(37,'qrm_potrf')
     end if

#if defined (have_starpu)
  end if
#endif

  return

end subroutine cqrm_potrf_task


#if defined (have_starpu)
recursive subroutine cqrm_starpu_potrf_cpu_func(buffers, cl_arg) bind(C)
  use qrm_parameters_mod
  use fstarpu_mod
  use qrm_dscr_mod
  use qrm_pthread_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(c_ptr), value           :: cl_arg
  type(c_ptr), value           :: buffers

  integer(c_int), target       :: i, m, k
  integer                      :: na, lda, info
  complex(r32), pointer        :: a(:,:)
  character, target            :: uplo
  type(c_ptr), target          :: dscr_c
  type(qrm_dscr_type), pointer :: qrm_dscr
  integer(kind=8)              :: ts, te

  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c), &
       c_loc(uplo),                   &
       c_loc(m),                                   &
       c_loc(k)/))

  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return

  na   = fstarpu_matrix_get_ny(buffers, 0)
  lda  = fstarpu_matrix_get_ld(buffers, 0)

  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 0), a, shape=(/lda,na/))

  __QRM_PROF_START(ts)
  call cqrm_potrf(uplo, m, k, &
       a(1,1), lda, info)
  __QRM_PROF_STOP(qrm_prof_potrf_task_,ts,te)

  if(info.gt.0) then
     call qrm_atomic_cas(qrm_dscr%info, 0, 37)
     call qrm_error_print(37,'qrm_potrf')
  end if

  return
end subroutine cqrm_starpu_potrf_cpu_func
#endif
