!! ##############################################################################################
!!
!! Copyright 2012-2018 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!! This file contains a routine that updates a sum of squares with the
!! square of the frobenius norm of a block
!!
!! ##############################################################################################
#include "qrm_common.h"

subroutine dqrm_block_nrm_task(qrm_dscr, a, m, n, ssq)

  use iso_c_binding
  use qrm_common_mod
#if defined(have_starpu)
  use fstarpu_mod
  use dqrm_starpu_mod
  use qrm_starpu_common_mod
#endif
  use qrm_string_mod
  use qrm_dscr_mod
  use dqrm_dsmat_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(qrm_dscr_type), target   :: qrm_dscr
  type(dqrm_block_type), target :: a
  integer, target               :: m, n
  type(dqrm_ws_type)            :: ssq
  
  integer                       :: j
#if defined(have_starpu)
  type(c_ptr), target           :: dscrpt
#endif
  integer(kind=8)               :: ts, te

  ! until StarPU has a proper error handling method we have to check
  ! manually
  if(qrm_dscr%info.ne.0) return

#if defined (have_starpu)
  dscrpt = c_loc(qrm_dscr)
  call fstarpu_insert_task( (/                                                &
                              dqrm_block_nrm_cl,                              &
                              FSTARPU_VALUE, c_loc(dscrpt), FSTARPU_SZ_C_PTR, &
                              FSTARPU_VALUE, c_loc(m), FSTARPU_SZ_INTEGER,    &
                              FSTARPU_VALUE, c_loc(n), FSTARPU_SZ_INTEGER,    &
                              FSTARPU_R, a%hdl,                               &
                              FSTARPU_REDUX, ssq%hdl,                         &
                              FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),         &
                              C_NULL_PTR /) )
  call fstarpu_task_wait_for_all()

#else
  __QRM_PROF_START(ts)
  do j=1, n
     call dlassq(m, a%c(1,j), 1, ssq%c(1,1), ssq%c(2,1))
  end do
  __QRM_PROF_STOP(qrm_prof_nrm_task_,ts,te)
#endif

  return
end subroutine dqrm_block_nrm_task

#if defined (have_starpu)
recursive subroutine dqrm_starpu_block_nrm_cpu_func(buffers, cl_args) bind(C)

  use qrm_parameters_mod
  use fstarpu_mod
  use qrm_string_mod
  use qrm_dscr_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(c_ptr), value           :: cl_args
  type(c_ptr), value           :: buffers
  type(c_ptr), target          :: dscr_c
  type(qrm_dscr_type), pointer :: qrm_dscr

  integer                      :: ma, na, lda
  integer                      :: i, j
  integer(c_int)               :: worker_id, comm_rank = -1
  integer, target              :: m, n
  real(r64), pointer        :: a(:,:)
  real(r64), pointer           :: ssq(:)
  integer(kind=8)              :: ts, te

  call fstarpu_unpack_arg( cl_args,(/ c_loc(dscr_c), c_loc(m), c_loc(n) /))

  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return

  ma   = fstarpu_matrix_get_nx(buffers, 0)
  na   = fstarpu_matrix_get_ny(buffers, 0)
  lda  = fstarpu_matrix_get_ld(buffers, 0)

  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 0), a, shape=(/lda,na/))
  call c_f_pointer(fstarpu_vector_get_ptr(buffers, 1), ssq, shape=(/2/))

  __QRM_PROF_START(ts)
  do j=1, n
     call dlassq(m, a(1,j), 1, ssq(1), ssq(2))
  end do
  __QRM_PROF_STOP(qrm_prof_nrm_task_,ts,te)

  ! write(*,*)'in nrm task ',m, n, ssq
  
  return
end subroutine dqrm_starpu_block_nrm_cpu_func

! This is the reduction routine for combining local, partial results
recursive subroutine dqrm_starpu_lassq_red_cpu_func (buffers, cl_args) bind(C)
  use qrm_parameters_mod
  use iso_c_binding
  use fstarpu_mod
  implicit none
  
  type(c_ptr), value, intent(in) :: buffers, cl_args ! cl_args is unused
  real(r64), pointer             :: ssqs(:), ssqd(:)
  real(r64)                      :: dold(2)
  
  call c_f_pointer(fstarpu_vector_get_ptr(buffers, 0), ssqd, shape=(/2/))
  call c_f_pointer(fstarpu_vector_get_ptr(buffers, 1), ssqs, shape=(/2/))
  
  if(max(ssqs(1), ssqd(1)).eq.0) goto 9999
  
  if(ssqs(1).gt.ssqd(1)) then
     ssqd(2) = ssqs(2) + ssqd(2)*((ssqd(1)/ssqs(1))**2)
     ssqd(1) = ssqs(1)
  else
     ssqd(2) = ssqd(2) + ssqs(2)*((ssqs(1)/ssqd(1))**2)
  end if
  
9999 continue
  
  ! write(*,*)'in nrm redu ',ssqs, dold, ssqd
  return
end subroutine dqrm_starpu_lassq_red_cpu_func

! This routine is for initializing the reduction buffers
recursive subroutine dqrm_starpu_lassq_ini_cpu_func (buffers, cl_args) bind(C)
    use iso_c_binding
    use fstarpu_mod
    use qrm_parameters_mod
    implicit none

    type(c_ptr), value, intent(in) :: buffers, cl_args ! cl_args is unused
    integer(c_int), target         :: arg
    integer, target                :: na, i
    real(r64), pointer             :: ssq(:)
    type(c_ptr)                    :: aptr
    
    call c_f_pointer(fstarpu_vector_get_ptr(buffers, 0), ssq, shape=(/2/))

    ssq(1) = qrm_dzero
    ssq(2) = qrm_done

    return
end subroutine dqrm_starpu_lassq_ini_cpu_func
#endif
