!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine dqrm_higemqrt_task(qrm_dscr, trans, a, t, c, k, j, nb, ib, work, prio)

#if defined(have_starpu)
  use fstarpu_mod
  use dqrm_starpu_mod
  use qrm_starpu_common_mod
#endif
  use dqrm_dsmat_mod
  use qrm_dscr_mod
  use qrm_mem_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif

  implicit none

  type(qrm_dscr_type), target   :: qrm_dscr
  character, target             :: trans
  type(dqrm_block_type), target :: a, t, c
  integer                       :: k, j, nb
  type(dqrm_ws_type)            :: work
  integer, target               :: ib
  integer                       :: prio

  integer, target               :: i, jj, gemprio, kk
  type(c_ptr)                   :: hdlt, hdlc
  type(c_ptr), target           :: dscrpt, strpt
  integer                       :: mc, nc, ldc, nt, ldt, info
  integer(kind=8)               :: ts, te

  if(qrm_dscr%info.ne.0) return
  
  if(t%partitioned) then
     ! necessarily means also c is paritioned
     i    = (k-1)*nb+1
  else
     i    = 1
  end if

#if defined (have_starpu)
  if(qrm_dscr%seq.eq.0) then

     ! submit a geqrt task
     if(t%partitioned) then
        hdlt = t%shdls(k)
     else
        hdlt = t%hdl
     end if

     if(c%partitioned) then
        hdlc = c%shdls(j)
        if(qrm_dscr%ngpu>0) then
           gemprio = gemqrt_i_hp_prio_
        else
           gemprio = prio + gemqrt_i_prio_
        end if
     else
        hdlc = c%hdl
        if(qrm_dscr%ngpu>0) then
           gemprio = gemqrt_o_hp_prio_
        else
           gemprio = prio + gemqrt_o_prio_
        end if
     end if

     dscrpt = c_loc(qrm_dscr)
     if(qrm_allocated(a%stair)) then
        strpt = c_loc(a%stair(i))
     else
        strpt = c_null_ptr
     end if

     call fstarpu_insert_task((/ dqrm_gemqrt_cl,             &
          FSTARPU_VALUE, c_loc(dscrpt), FSTARPU_SZ_C_PTR,    &
          FSTARPU_VALUE, c_loc(i),  FSTARPU_SZ_C_INT,        &
          FSTARPU_VALUE, c_loc(ib), FSTARPU_SZ_C_INT,        &
          FSTARPU_VALUE, c_loc(trans), FSTARPU_SZ_CHARACTER, &
          FSTARPU_VALUE, c_loc(strpt), FSTARPU_SZ_C_PTR,     &
          FSTARPU_R,  hdlt,                                  &
          FSTARPU_RW, hdlc,                                  &
          FSTARPU_SCRATCH, work%hdl,                         &
          FSTARPU_PRIORITY, c_loc(gemprio),                  &
          FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),            &
          C_NULL_PTR /))
     ! call fstarpu_task_wait_for_all()
  else
#endif

     ! make a simple call to lapack gemqrt
     jj  = (j-1)*nb+1
     mc  = size(c%c,1) ! TODO: implement qrm_size
     if(c%partitioned) then
        nc  = min(nb,size(c%c,2)-jj+1)
     else
        nc  = size(c%c,2)
     end if
     ldc = size(c%c,1)
     if(t%partitioned) then
        nt  = min(nb,size(t%c,2)-i+1)
     else
        nt  = size(t%c,2)
     end if
     ldt = size(t%c,1)
     kk  = min(nt,mc-i+1)

     __QRM_PROF_START(ts)
     ! move this if inside the kernel
     if(qrm_allocated(a%stair)) then
        call dqrm_gemqrt(qrm_left, trans, &
             mc, nc, kk, ib,              &
             a%stair(1),  i,              &
             t%c(ib+1,i), ldt,            &
             t%c(1,i), ldt,               &
             c%c(1,jj), ldc,              &
             work%c(1,1), info)
     else
        call dqrm_gemqrt(qrm_left, trans, &
             mc, nc, kk, ib,              &
             -1, i,                       &
             t%c(ib+1,i), ldt,            &
             t%c(1,i), ldt,               &
             c%c(1,jj), ldc,              &
             work%c(1,1), info)
     end if
     __QRM_PROF_STOP(qrm_prof_gemqrt_task_,ts,te)

#if defined (have_starpu)
  end if
#endif

  return
end subroutine dqrm_higemqrt_task

#if defined (have_starpu)
recursive subroutine dqrm_starpu_gemqrt_cpu_func(buffers, cl_arg) bind(C)
  use qrm_parameters_mod
  use qrm_dscr_mod
  use fstarpu_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(c_ptr), value           :: cl_arg
  type(c_ptr), value           :: buffers

  integer(c_int), target       :: i, ii, jj, ib
  character, target            :: trans
  type(c_ptr), target          :: dscr_c, stair_c
  integer                      :: mv, nv, ldv, mc, nc, ldc, mt, nt, ldt, nw, ldw, info, k
  real(r64), pointer        :: v(:,:), t(:,:), c(:,:), w(:,:)
  integer, pointer             :: stair(:)
  type(qrm_dscr_type), pointer :: qrm_dscr
  integer(kind=8)              :: ts, te


  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c), c_loc(i), c_loc(ib), c_loc(trans), c_loc(stair_c) /))

  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return

  mt   = fstarpu_matrix_get_nx(buffers, 0)
  nt   = fstarpu_matrix_get_ny(buffers, 0)
  ldt  = fstarpu_matrix_get_ld(buffers, 0)
  mc   = fstarpu_matrix_get_nx(buffers, 1)
  nc   = fstarpu_matrix_get_ny(buffers, 1)
  ldc  = fstarpu_matrix_get_ld(buffers, 1)
  nw   = fstarpu_matrix_get_ny(buffers, 2)
  ldw  = fstarpu_matrix_get_ld(buffers, 2)

  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 0), t, shape=(/ldt,nt/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 1), c, shape=(/ldc,nc/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 2), w, shape=(/ldw,nw/))

  k   = min(nt,mc-i+1)

  __QRM_PROF_START(ts)
  if(c_associated(stair_c)) then
     call c_f_pointer(stair_c, stair, shape=(/nt/))
     call dqrm_gemqrt(qrm_left, trans, & 
          mc, nc, k, ib,               & 
          stair(1),  i,                & 
          t(ib+1,1), ldt,              & 
          t(1,1), ldt,                 & 
          c(1,1), ldc,                 & 
          w(1,1), info)
  else
     call dqrm_gemqrt(qrm_left, trans, &
          mc, nc, k, ib,               &
          -1,  i,                      &
          t(ib+1,1), ldt,              &
          t(1,1), ldt,                 &
          c(1,1), ldc,                 &
          w(1,1), info)
  end if
  __QRM_PROF_STOP(qrm_prof_gemqrt_task_,ts,te)

  ! TODO: add check for return code in info

  return

end subroutine dqrm_starpu_gemqrt_cpu_func

#if defined (have_cuda)
recursive subroutine dqrm_starpu_gemqrt_cuda_func(buffers, cl_arg) bind(C)
  use qrm_parameters_mod
  use qrm_dscr_mod
  use fstarpu_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(c_ptr), value           :: cl_arg
  type(c_ptr), value           :: buffers

  integer(c_int), target       :: i, ib, iib
  character, target            :: trans
  type(c_ptr), target          :: dscr_c, stair_c, tp, cp, wp, vp, handlep
  integer                      :: ldv, mc, nc, ldc, mt, nt, ldt, nw, ldw, info, id, k, lw
  type(qrm_dscr_type), pointer :: qrm_dscr
  integer(kind=8)              :: ts, te

  interface
     subroutine dqrm_cuda_gemqrt(handlep, &
          side, trans,                    &
          m, n, k, nb,                    &
          stair, i,                       &
          t, ldt,                         &
          c, ldc,                         &
          w, ldw) bind(C)
       use iso_c_binding
       type(c_ptr), value       :: handlep, t, c, w, stair
       character(c_char), value :: side, trans
       integer, value           :: m, n, k, nb, i, ldt, ldc, ldw
     end subroutine dqrm_cuda_gemqrt
  end interface

  
  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c), c_loc(i), c_loc(ib), c_loc(trans), c_loc(stair_c) /))

  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return

  mt   = fstarpu_matrix_get_nx(buffers, 0)
  nt   = fstarpu_matrix_get_ny(buffers, 0)
  ldt  = fstarpu_matrix_get_ld(buffers, 0)
  mc   = fstarpu_matrix_get_nx(buffers, 1)
  nc   = fstarpu_matrix_get_ny(buffers, 1)
  ldc  = fstarpu_matrix_get_ld(buffers, 1)
  nw   = fstarpu_matrix_get_ny(buffers, 2)
  ldw  = fstarpu_matrix_get_ld(buffers, 2)
  lw   = nw*ldw

  tp = fstarpu_matrix_get_ptr(buffers, 0)
  cp = fstarpu_matrix_get_ptr(buffers, 1)
  wp = fstarpu_matrix_get_ptr(buffers, 2)

  id      = fstarpu_worker_get_id();
  handlep = qrm_dscr%cuhandles(id+1)
  
  k   = min(nt,mc-i+1)

  __QRM_PROF_START(ts)
  call dqrm_cuda_gemqrt(handlep, &
       qrm_left, trans,          &
       mc, nc, k, ib,            &
       stair_c, i,               &
       tp, ldt,                  &
       cp, ldc,                  &
       wp, lw)
  __QRM_PROF_STOP(qrm_prof_gemqrt_task_,ts,te)
  
  ! TODO: add check for return code in info

  return

end subroutine dqrm_starpu_gemqrt_cuda_func
#endif
#endif
