!! ##############################################################################################
!!
!! Copyright 2012-2018 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file defines a task for (extend-)summing a block into a matrix
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine sqrm_block_extadd_task(qrm_dscr, a, b, br, bc, i, j, m, n, l, inout, op, rowmap, colmap &
#if defined(have_starpu)
     ,hdl &
#endif
     )
#if defined(have_starpu)
  use fstarpu_mod
  use sqrm_starpu_mod
#endif
  use qrm_dscr_mod
  use sqrm_dsmat_mod
  use qrm_string_mod
  use iso_c_binding
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none


  type(qrm_dscr_type), target    :: qrm_dscr
  type(sqrm_dsmat_type), target  :: a                    ! if copyout this is the source
                                                         ! matrix. If copyin it is the
                                                         ! destinantion matrix

  type(sqrm_dsmat_type), target  :: b                    ! if copyout is the destination
                                                         ! matrix; if the source one.

  integer, target                :: br, bc               ! block-row and block-column
                                                         ! of the source/destination
                                                         ! block in a
  integer, target                :: i, j, m, n, l        ! i and j are the first row
                                                         ! and column within the
                                                         ! matrix and m and n are the
                                                         ! number of rows and cols to
                                                         ! be copied. l is the number of
                                                         ! rows in the upper trapezoidal
                                                         ! part of the block (same meaning
                                                         ! as in tpqrt
  character, target              :: inout, op
  integer, target                :: rowmap(:,:)         ! block-row mapping from a to b
  integer, optional, target      :: colmap(:,:)         ! block-column mapping from a to b

#if defined(have_starpu)
  type(c_ptr), optional          :: hdl
#endif

#if defined (have_starpu)
  integer, target                :: nh, prio
  type(c_ptr), target            :: a_ptr, b_ptr, rmptr, cmptr
  type(c_ptr)                    :: hdls, ptr, amode, bmode
  type(c_ptr), target            :: dscrpt

  integer                        :: abr, abc ! block row and col in a (just a copy of br and bc)
  integer                        :: bbr, bbc ! block row and col in b
  integer                        :: old_bbr, old_bbc
  integer                        :: ai, aj, agi, agj ! local (to block) and global row and col indices in a
  integer                        :: bi, bj, bgi, bgj ! local (to block) and global row and col indices in b
  integer                        :: aroff, acoff ! row and column offsets in a
  integer, target                :: cms, rms
#endif
  integer(kind=8)                :: ts, te


  if(qrm_dscr%info.ne.0) return

#if defined (have_starpu)

  if(qrm_dscr%seq.eq.0) then

     prio = 0
     a_ptr = c_loc(a)
     b_ptr = c_loc(b)

     rmptr = c_loc(rowmap(1,1))
     rms   = size(rowmap,1)

     if(present(colmap)) then
        cmptr = c_loc(colmap(1,1))
        cms   = size(colmap,1)
     else
        cmptr = c_null_ptr
        cms   = 0
     end if

     if(qrm_str_tolower(inout).eq.'o') then
        ! if copyout (scatter) a is read and b written
        amode = FSTARPU_R
        bmode = FSTARPU_RW .ior. FSTARPU_COMMUTE
     else
        ! if copyin (gather) a is written and b read
        amode = FSTARPU_RW .ior. FSTARPU_COMMUTE
        bmode = FSTARPU_R
     end if

     abr = br ! just a copy to avoid confusion with blocks of b
     abc = bc ! just a copy to avoid confusion with blocks of b

     aroff = (abr-1)*a%mb  ! row-offset in a
     acoff = (abc-1)*a%mb  ! col-offset in a

     old_bbc = 0
     old_bbr = 0

     nh   = 0

     dscrpt = c_loc(qrm_dscr)

     hdls = fstarpu_data_descr_array_alloc(size(b%blocks,1)*size(b%blocks,2)+1)

     ! add source block handle
     if(present(hdl)) then
        call fstarpu_data_descr_array_set(hdls, 0, &
             hdl,                                  &
             amode)
     else
        call fstarpu_data_descr_array_set(hdls, 0, &
             a%blocks(br,bc)%hdl,             &
             amode)
     end if

     cols: do aj = j+n-1, j, -1
        agj = acoff + aj                               ! global a col index of the current coefficient
        if(present(colmap)) then
           bbc = colmap(agj,2)                         ! block-col in b     of the current coefficient
        else
           bbc = (agj-1)/b%mb+1                        ! block-col in b     of the current coefficient
        end if
        if(bbc.eq.old_bbc) cycle cols
        old_bbc = bbc
        old_bbr = 0

        rows: do ai = i, i+m-1
           if(ai-i .gt. aj-j+m-l) cycle cols
           agi = aroff + ai                            ! global a row index of the current coefficient
           bbr = rowmap(agi,2)                         ! block-row in b     of the current coefficient
           if(bbr.eq.old_bbr) cycle rows
           old_bbr = bbr

           if(.not. qrm_allocated(b%blocks(bbr,bbc)%c)) cycle rows
           nh = nh+1

           call fstarpu_data_descr_array_set(hdls, nh,                      &
                b%blocks(bbr,bbc)%hdl, &
                bmode)
        end do rows
     end do cols

     nh = nh+1

     call fstarpu_insert_task((/ sqrm_block_extadd_cl,              &
          FSTARPU_VALUE, c_loc(dscrpt),       FSTARPU_SZ_C_PTR,     &
          FSTARPU_VALUE, c_loc(br),           FSTARPU_SZ_C_INT,     &
          FSTARPU_VALUE, c_loc(bc),           FSTARPU_SZ_C_INT,     &
          FSTARPU_VALUE, c_loc(a_ptr),        FSTARPU_SZ_C_PTR,     &
          FSTARPU_VALUE, c_loc(b_ptr),        FSTARPU_SZ_C_PTR,     &
          FSTARPU_VALUE, c_loc(i),            FSTARPU_SZ_C_INT,     &
          FSTARPU_VALUE, c_loc(j),            FSTARPU_SZ_C_INT,     &
          FSTARPU_VALUE, c_loc(m),            FSTARPU_SZ_C_INT,     &
          FSTARPU_VALUE, c_loc(n),            FSTARPU_SZ_C_INT,     &
          FSTARPU_VALUE, c_loc(l),            FSTARPU_SZ_C_INT,     &
          FSTARPU_VALUE, c_loc(inout),        FSTARPU_SZ_CHARACTER, &
          FSTARPU_VALUE, c_loc(op),           FSTARPU_SZ_CHARACTER, &
          FSTARPU_VALUE, c_loc(rmptr),        FSTARPU_SZ_C_PTR,     &
          FSTARPU_VALUE, c_loc(cmptr),        FSTARPU_SZ_C_PTR,     &
          FSTARPU_VALUE, c_loc(rms),          FSTARPU_SZ_C_INT,     &
          FSTARPU_VALUE, c_loc(cms),          FSTARPU_SZ_C_INT,     &
          FSTARPU_DATA_MODE_ARRAY, hdls, c_loc(nh),                 &
          FSTARPU_PRIORITY, c_loc(prio),                            &
          FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),                   &
          C_NULL_PTR /))
          ! call fstarpu_task_wait_for_all()


     call fstarpu_data_descr_array_free(hdls)
  else
#endif
     __QRM_PROF_START(ts)
     call sqrm_block_extadd(a, b, br, bc, i, j, m, n, l, &
          inout, op,                                     &
          rowmap=rowmap, colmap=colmap)
     __QRM_PROF_STOP(qrm_prof_extadd_task_,ts,te)
#if defined (have_starpu)
  end if
#endif

  return
end subroutine sqrm_block_extadd_task


#if defined (have_starpu)
recursive subroutine sqrm_starpu_block_extadd_cpu_func(buffers, cl_arg) bind(C)
  use fstarpu_mod
  use sqrm_dsmat_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(c_ptr), value             :: cl_arg
  type(c_ptr), value             :: buffers

  integer, target                :: br, bc, i, j, m, n, l, rms, cms
  type(c_ptr), target            :: a_ptr, b_ptr, rmptr, cmptr
  type(sqrm_dsmat_type), pointer :: a, b
  integer, pointer               :: rowmap(:,:), colmap(:,:)
  type(c_ptr), target            :: dscr_c
  type(qrm_dscr_type), pointer   :: qrm_dscr
  character, target              :: inout, op
  integer(kind=8)                :: ts, te

  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c),        &
       c_loc(br), c_loc(bc),                              &
       c_loc(a_ptr), c_loc(b_ptr),                        &
       c_loc(i), c_loc(j), c_loc(m), c_loc(n), c_loc(l),  &
       c_loc(inout), c_loc(op),                           &
       c_loc(rmptr), c_loc(cmptr), c_loc(rms), c_loc(cms) &
       /))

  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return

  call c_f_pointer(a_ptr, a)
  call c_f_pointer(b_ptr, b)

  call c_f_pointer(rmptr, rowmap, shape=(/rms,3/))

  __QRM_PROF_START(ts)
  if(c_associated(cmptr)) then
     call c_f_pointer(cmptr, colmap, shape=(/cms,3/))
     call sqrm_block_extadd(a, b, br, bc, i=i, j=j, m=m, n=n, l=l, &
          inout=inout, op=op,                                      &
          rowmap=rowmap, colmap=colmap)
  else
     call sqrm_block_extadd(a, b, br, bc, i=i, j=j, m=m, n=n, l=l, &
          inout=inout, op=op,                                      &
          rowmap=rowmap)
  end if
  __QRM_PROF_STOP(qrm_prof_extadd_task_,ts,te)

  return

end subroutine sqrm_starpu_block_extadd_cpu_func
#endif
