!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine sqrm_trmm_task(qrm_dscr, side, uplo, transa, diag, m, n, k, alpha, a, b, prio)

#if defined(have_starpu)
  use fstarpu_mod
  use sqrm_starpu_mod
  use qrm_starpu_common_mod
#endif
  use qrm_parameters_mod
  use qrm_dscr_mod
  use sqrm_dsmat_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(qrm_dscr_type), target :: qrm_dscr
  character, target           :: side, uplo, transa, diag
  integer, target             :: m, n, k
  real(r32), target        :: alpha
  type(sqrm_block_type)       :: a, b
  integer, target             :: prio

  integer, target             :: trmmprio
  type(c_ptr)                 :: sizeof
  integer                     :: lda, ldb, i
  real(r32)                :: ialpha
  integer(kind=8)             :: ts, te
#if defined (have_starpu)
  type(c_ptr), target         :: dscrpt
#endif
  
  if(qrm_dscr%info.ne.0) return

#if defined (have_starpu)
  if(qrm_dscr%seq.eq.0) then
  
#if defined(dprec)
     sizeof = FSTARPU_SZ_REAL8
#elif defined(sprec)
     sizeof = FSTARPU_SZ_REAL4
#elif defined(zprec)
     sizeof = FSTARPU_SZ_COMPLEX8
#elif defined(cprec)
     sizeof = FSTARPU_SZ_COMPLEX4
#endif

     if(qrm_dscr%ngpu>0) then
        trmmprio = trmm_hp_prio_
     else
        trmmprio = prio + trmm_prio_
     end if

     dscrpt = c_loc(qrm_dscr)
     
     call fstarpu_insert_task((/ sqrm_trmm_cl,                 &
          FSTARPU_VALUE, c_loc(dscrpt), FSTARPU_SZ_C_PTR,      &
          FSTARPU_VALUE, c_loc(side),   FSTARPU_SZ_CHARACTER,  &
          FSTARPU_VALUE, c_loc(uplo),   FSTARPU_SZ_CHARACTER,  &
          FSTARPU_VALUE, c_loc(transa), FSTARPU_SZ_CHARACTER,  &
          FSTARPU_VALUE, c_loc(diag),   FSTARPU_SZ_CHARACTER,  &
          FSTARPU_VALUE, c_loc(m),      FSTARPU_SZ_C_INT,      &
          FSTARPU_VALUE, c_loc(n),      FSTARPU_SZ_C_INT,      &
          FSTARPU_VALUE, c_loc(k),      FSTARPU_SZ_C_INT,      &
          FSTARPU_VALUE, c_loc(alpha),  sizeof,                &
          FSTARPU_R, a%hdl,                                    &
          FSTARPU_RW, b%hdl,                                   &
          FSTARPU_PRIORITY, c_loc(trmmprio),                   &
          FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),              &
          C_NULL_PTR /))

  else
#endif

     ! make a call to trmm
     lda  = size(a%c,1)
     ldb  = size(b%c,1)

     __QRM_PROF_START(ts)
     call sqrm_trmm(qrm_left, qrm_upper, &
          transa, diag,                  &
          m, n, k, alpha,                &
          a%c(1,1), lda,                 &
          b%c(1,1), ldb)
     __QRM_PROF_STOP(qrm_prof_trmm_task_,ts,te)

#if defined (have_starpu)
  end if
#endif

  return

end subroutine sqrm_trmm_task


#if defined (have_starpu)
recursive subroutine sqrm_starpu_trmm_cpu_func(buffers, cl_arg) bind(C)
  use qrm_parameters_mod
  use fstarpu_mod
  use qrm_dscr_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(c_ptr), value           :: cl_arg
  type(c_ptr), value           :: buffers
                                
  integer(c_int), target       :: i, m, n, k, l, ib
  integer                      :: na, lda, nb, ldb, info, eunit
  real(r32), pointer        :: a(:,:), b(:,:)
  character, target            :: side, uplo, diag, transa
  real(r32), target         :: alpha, ialpha
  type(c_ptr), target          :: dscr_c
  type(qrm_dscr_type), pointer :: qrm_dscr
  integer(kind=8)              :: ts, te

  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c), &
       c_loc(side), c_loc(uplo),                   &
       c_loc(transa), c_loc(diag),                 &
       c_loc(m),                                   & 
       c_loc(n),                                   & 
       c_loc(k),                                   & 
       c_loc(alpha) /))

  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return
  
  na   = fstarpu_matrix_get_ny(buffers, 0)
  lda  = fstarpu_matrix_get_ld(buffers, 0)
  nb   = fstarpu_matrix_get_ny(buffers, 1)
  ldb  = fstarpu_matrix_get_ld(buffers, 1)

  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 0), a, shape=(/lda,na/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 1), b, shape=(/ldb,nb/))

  __QRM_PROF_START(ts)
  call sqrm_trmm(qrm_left, qrm_upper, &
       transa, diag,                  &
       m, n, k, alpha,                &
       a, lda,                        &
       b, ldb)
  __QRM_PROF_STOP(qrm_prof_trmm_task_,ts,te)

  return
end subroutine sqrm_starpu_trmm_cpu_func
#endif
