!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine zqrm_higeqrt_task(qrm_dscr, a, t, k, nb, ib, work, prio)

#if defined(have_starpu)
  use fstarpu_mod
  use zqrm_starpu_mod
  use qrm_starpu_common_mod
#endif
  use zqrm_dsmat_mod
  use qrm_dscr_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif

  implicit none

  type(qrm_dscr_type), target   :: qrm_dscr
  type(zqrm_block_type), target :: a, t
  integer                       :: k, nb
  type(zqrm_ws_type)            :: work
  integer, target               :: prio, ib
  integer(kind=8)               :: ts, te


#if defined (have_starpu)
  type(c_ptr)                   :: hdla, hdlt
  type(c_ptr), target           :: dscrpt, strpt
#endif
  integer, target               :: i, geprio
  integer                       :: ma, na, lda, ldt, info, j, li

  if(qrm_dscr%info.ne.0) return

  if(a%partitioned) then
     i    = (k-1)*nb+1
  else
     i    = 1
  end if

#if defined (have_starpu)
  if(qrm_dscr%seq.eq.0) then

     ! submit a geqrt task
     if(a%partitioned) then
        hdla = a%shdls(k)
        hdlt = t%shdls(k)
     else
        hdla = a%hdl
        hdlt = t%hdl
     end if

     if(qrm_dscr%ngpu>0) then
        if(qrm_allocated(a%stair)) then
           geprio = 0
        else
           geprio = geqrt_hp_prio_
        end if
     else
        geprio = prio + geqrt_prio_
     end if

     dscrpt = c_loc(qrm_dscr)
     if(qrm_allocated(a%stair)) then
        strpt = c_loc(a%stair(i))
     else
        strpt = c_null_ptr
     end if

     call fstarpu_insert_task((/ zqrm_geqrt_cl,           &
          FSTARPU_VALUE, c_loc(dscrpt), FSTARPU_SZ_C_PTR, &
          FSTARPU_VALUE, c_loc(i),  FSTARPU_SZ_C_INT,     &
          FSTARPU_VALUE, c_loc(ib), FSTARPU_SZ_C_INT,     &
          FSTARPU_VALUE, c_loc(strpt), FSTARPU_SZ_C_PTR,  &
          FSTARPU_RW, hdla,                               &
          FSTARPU_RW, hdlt,                               &
          FSTARPU_SCRATCH, work%hdl,                      &
          FSTARPU_PRIORITY, c_loc(geprio),                &
          FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),         &
          C_NULL_PTR /))
     ! call fstarpu_task_wait_for_all()
  else
#endif

     ! make a simple call to lapack geqrt
     ma = size(a%c,1)
     if(a%partitioned) then
        na = min(nb,size(a%c,2)-i+1)
     else
        na = size(a%c,2)
     end if
     lda = size(a%c,1)
     ldt = size(t%c,1)

     __QRM_PROF_START(ts)
     if(allocated(a%stair)) then
        call zqrm_geqrt(ma, na, ib,  &
             a%stair(i), i,          &
             a%c(1,i), lda,          &
             t%c(1,i), ldt,          &
             work%c, info)
     else
        call zqrm_geqrt(ma, na, ib,  &
             -1, i,                  &
             a%c(1,i), lda,          &
             t%c(1,i), ldt,          &
             work%c, info)
     end if
     __QRM_PROF_STOP(qrm_prof_geqrt_task_,ts,te)

#if defined (have_starpu)
  end if
#endif

  return
end subroutine zqrm_higeqrt_task

#if defined (have_starpu)
recursive subroutine zqrm_starpu_geqrt_cpu_func(buffers, cl_arg) bind(C)
  use qrm_parameters_mod
  use qrm_dscr_mod
  use fstarpu_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(c_ptr), value           :: cl_arg
  type(c_ptr), value           :: buffers

  integer(c_int), target       :: i, ib
  type(c_ptr), target          :: dscr_c, stair_c
  integer                      :: ma, na, lda, nt, ldt, nw, ldw, info, j, li, ii, jj
  complex(r64), pointer        :: a(:,:), t(:,:), w(:,:)
  integer, pointer             :: stair(:)
  type(qrm_dscr_type), pointer :: qrm_dscr
  integer(kind=8)              :: ts, te

  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c), c_loc(i), c_loc(ib), c_loc(stair_c) /))

  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return

  nw   = fstarpu_matrix_get_ny(buffers, 2)
  ldw  = fstarpu_matrix_get_ld(buffers, 2)
  ma   = fstarpu_matrix_get_nx(buffers, 0)
  na   = fstarpu_matrix_get_ny(buffers, 0)
  lda  = fstarpu_matrix_get_ld(buffers, 0)
  nt   = fstarpu_matrix_get_ny(buffers, 1)
  ldt  = fstarpu_matrix_get_ld(buffers, 1)

  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 0), a, shape=(/lda,na/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 1), t, shape=(/ldt,nt/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 2), w, shape=(/ldw,nw/))

  __QRM_PROF_START(ts)
  if(c_associated(stair_c)) then

     call c_f_pointer(stair_c, stair, shape=(/na/))
     call zqrm_geqrt(ma, na, ib,   &
          stair(1),  i ,           &
          a(1,1)  , lda,           &
          t(1,1)  , ldt,           &
          w(1,1)  , info)
  else
     call zqrm_geqrt(ma, na, ib,   &
          -1,  i,                  &
          a(1,1), lda,             &
          t(1,1), ldt,             &
          w(1,1), info)
  end if
  __QRM_PROF_STOP(qrm_prof_geqrt_task_,ts,te)

  ! TODO: add check for return code in info

  return
end subroutine zqrm_starpu_geqrt_cpu_func


#if defined (have_cuda)
recursive subroutine zqrm_starpu_geqrt_cuda_func(buffers, cl_arg) bind(C)
  use qrm_parameters_mod
  use qrm_dscr_mod
  use fstarpu_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(c_ptr), value           :: cl_arg
  type(c_ptr), value           :: buffers

  integer(c_int), target       :: i, ib
  type(c_ptr), target          :: dscr_c, stair_c, handlep, ap, tp, wp
  integer                      :: ma, na, lda, nt, ldt, nw, ldw, info, id, lwd
  type(qrm_dscr_type), pointer :: qrm_dscr
  integer(kind=8)              :: ts, te

  interface
     subroutine zqrm_cuda_geqrt(handlep, &
          m, n, nb, i,                   &    
          a, lda,                        &    
          t, ldt,                        &    
          wd, lwd) bind(C)
       use iso_c_binding
       type(c_ptr), value       :: handlep, a, t, wd
       integer, value           :: m, n, nb, i, lda, ldt, lwd
     end subroutine zqrm_cuda_geqrt
  end interface

  
  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c), c_loc(i), c_loc(ib), c_loc(stair_c) /))

  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return

  ma   = fstarpu_matrix_get_nx(buffers, 0)
  na   = fstarpu_matrix_get_ny(buffers, 0)
  lda  = fstarpu_matrix_get_ld(buffers, 0)
  nt   = fstarpu_matrix_get_ny(buffers, 1)
  ldt  = fstarpu_matrix_get_ld(buffers, 1)
  nw   = fstarpu_matrix_get_ny(buffers, 2)
  ldw  = fstarpu_matrix_get_ld(buffers, 2)
  lwd  = nw*ldw
  ap   = fstarpu_matrix_get_ptr(buffers, 0)
  tp   = fstarpu_matrix_get_ptr(buffers, 1)
  wp   = fstarpu_matrix_get_ptr(buffers, 2)
  
  id      = fstarpu_worker_get_id();
  handlep = qrm_dscr%cuhandles(id+1)

  __QRM_PROF_START(ts)
  call zqrm_cuda_geqrt(handlep, &
       ma, na, ib, i,           &
       ap, lda,                 &
       tp, ldt,                 &
       wp, lwd)
  __QRM_PROF_STOP(qrm_prof_geqrt_task_,ts,te)
  ! TODO: add check for return code in info
  
  return
end subroutine zqrm_starpu_geqrt_cuda_func
#endif


#endif
