!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!!
!! This file contains the module which holds the data types used during the factorization.
!!
!! ##############################################################################################


#include "qrm_common.h"

!> @brief This module contains the definition of a dense matrix
!!
module zqrm_dsmat_mod
#if defined (have_starpu)
  use fstarpu_mod
#endif
  use qrm_error_mod
  use qrm_mem_mod
  use qrm_dscr_mod
  use qrm_parameters_mod
  

  !> @brief This type defines a workspace type
  type zqrm_ws_type
     !> the coefficients
     complex(r64), pointer                 :: c(:,:) => null()
#if defined (have_starpu)
     type(c_ptr)                        :: hdl = c_null_ptr
#endif
  end type zqrm_ws_type


  !> @brief This type defines a block type
  type zqrm_block_type
     !> the coefficients
     complex(r64), allocatable             :: c(:,:)
     integer, allocatable               :: stair(:)
     logical                            :: partitioned=.false.
#if defined (have_starpu)
     !> the handle
     type(c_ptr)                        :: hdl = c_null_ptr
     type(c_ptr), allocatable           :: shdls(:)
#endif
  end type zqrm_block_type


  type zqrm_dsmat_type
     integer                            :: m, n, mb, nb
     integer                            :: d              ! used to count the number of zeros along the diagonal after factorization
     type(zqrm_block_type), allocatable :: blocks(:,:)
     logical                            :: inited=.false. ! FIXME: see if needed
  end type zqrm_dsmat_type

  interface qrm_block_partition
     module procedure zqrm_block_partition1, zqrm_block_partition2
  end interface qrm_block_partition

  interface zqrm_block_partition
     module procedure zqrm_block_partition1, zqrm_block_partition2
  end interface zqrm_block_partition

  interface qrm_block_unpartition
     module procedure zqrm_block_unpartition1, zqrm_block_unpartition2
  end interface qrm_block_unpartition

  interface zqrm_block_unpartition
     module procedure zqrm_block_unpartition1, zqrm_block_unpartition2
  end interface zqrm_block_unpartition

  interface qrm_block_destroy
     module procedure zqrm_block_destroy
  end interface qrm_block_destroy

  interface qrm_dsmat_init
     module procedure zqrm_dsmat_init
  end interface qrm_dsmat_init

  interface qrm_dsmat_destroy
     module procedure zqrm_dsmat_destroy
  end interface qrm_dsmat_destroy

  interface qrm_dsmat_read
     module procedure zqrm_dsmat_read
  end interface qrm_dsmat_read

  interface qrm_dsmat_print
     module procedure zqrm_dsmat_print
  end interface qrm_dsmat_print

  interface qrm_block_print
     module procedure zqrm_block_print
  end interface qrm_block_print

  interface qrm_ws_init
     module procedure zqrm_ws_init
  end interface qrm_ws_init

  interface qrm_ws_destroy
     module procedure zqrm_ws_destroy
  end interface qrm_ws_destroy

  interface qrm_block_extadd
     module procedure zqrm_block_extadd
  end interface qrm_block_extadd
  
  interface qrm_dsmat_copy
     subroutine zqrm_dsmat_copy(a, b, ia, ja, ib, jb, m, n, l, transp, info)
       use qrm_dscr_mod
       import :: zqrm_dsmat_type
       type(zqrm_dsmat_type)          :: a
       type(zqrm_dsmat_type)          :: b
       integer, optional              :: ia, ja, ib, jb
       integer, optional              :: m, n, l
       character, optional            :: transp
       integer, optional              :: info
     end subroutine zqrm_dsmat_copy
  end interface qrm_dsmat_copy

  interface qrm_dsmat_copy_async
     subroutine zqrm_dsmat_copy_async(qrm_dscr, a, b, ia, ja, ib, jb, m, n, l, transp)
       use qrm_dscr_mod
       import :: zqrm_dsmat_type
       type(qrm_dscr_type), target    :: qrm_dscr
       type(zqrm_dsmat_type)          :: a
       type(zqrm_dsmat_type)          :: b
       integer, optional              :: ia, ja, ib, jb
       integer, optional              :: m, n, l
       character, optional            :: transp
     end subroutine zqrm_dsmat_copy_async
  end interface qrm_dsmat_copy_async

  interface qrm_dsmat_axpy
     subroutine zqrm_dsmat_axpy(a, b, ia, ja, ib, jb, m, n, l, alpha, info)
       use qrm_dscr_mod
       use qrm_parameters_mod
       import :: zqrm_dsmat_type
       type(zqrm_dsmat_type)          :: a
       type(zqrm_dsmat_type)          :: b
       integer, optional              :: ia, ja, ib, jb
       integer, optional              :: m, n, l
       complex(r64), optional         :: alpha
       integer, optional              :: info
     end subroutine zqrm_dsmat_axpy
  end interface qrm_dsmat_axpy

  interface qrm_dsmat_axpy_async
     subroutine zqrm_dsmat_axpy_async(qrm_dscr, a, b, ia, ja, ib, jb, m, n, l, alpha)
       use qrm_dscr_mod
       use qrm_parameters_mod
       import :: zqrm_dsmat_type
       type(qrm_dscr_type), target    :: qrm_dscr
       type(zqrm_dsmat_type)          :: a
       type(zqrm_dsmat_type)          :: b
       integer, optional              :: ia, ja, ib, jb
       integer, optional              :: m, n, l
       complex(r64), optional         :: alpha
     end subroutine zqrm_dsmat_axpy_async
  end interface qrm_dsmat_axpy_async

  interface qrm_dsmat_extadd
     subroutine zqrm_dsmat_extadd(a, b, i, j, m, n, l, iout, op, rowmap, colmap, inout, info)
       use qrm_dscr_mod
       import :: zqrm_dsmat_type
       type(zqrm_dsmat_type)          :: a
       type(zqrm_dsmat_type)          :: b
       integer                        :: i, j, m, n, l
       character                      :: inout, op
       integer                        :: rowmap(:,:)
       integer, optional              :: colmap(:,:)
       integer                        :: info
     end subroutine zqrm_dsmat_extadd
  end interface qrm_dsmat_extadd

  interface qrm_dsmat_extadd_async
     subroutine zqrm_dsmat_extadd_async(qrm_dscr, a, b, i, j, m, n, l, inout, op, rowmap, colmap &
#if defined(have_starpu)
          , hdl &
#endif
          )
       use qrm_dscr_mod
       import :: zqrm_dsmat_type
       type(qrm_dscr_type), target    :: qrm_dscr
       type(zqrm_dsmat_type)          :: a
       type(zqrm_dsmat_type)          :: b
       integer                        :: i, j, m, n, l
       character                      :: inout, op
       integer                        :: rowmap(:,:)
       integer, optional              :: colmap(:,:)
#if defined(have_starpu)
       type(c_ptr), optional          :: hdl
#endif
     end subroutine zqrm_dsmat_extadd_async
  end interface qrm_dsmat_extadd_async


  interface qrm_dsmat_nrm
     subroutine zqrm_dsmat_nrm(qrm_dsmat, nrm, m, n, info)
       use qrm_dscr_mod
       use qrm_parameters_mod
       import :: zqrm_dsmat_type
       type(zqrm_dsmat_type), target  :: qrm_dsmat
       real(r64)                      :: nrm
       integer, optional              :: m, n, info
     end subroutine zqrm_dsmat_nrm
  end interface qrm_dsmat_nrm

  interface qrm_dsmat_nrm_async
     subroutine zqrm_dsmat_nrm_async(qrm_dscr, qrm_dsmat, nrm, m, n)
       use qrm_dscr_mod
       use qrm_parameters_mod
       import :: zqrm_dsmat_type
       type(qrm_dscr_type), target    :: qrm_dscr
       type(zqrm_dsmat_type), target  :: qrm_dsmat
       real(r64)                      :: nrm
       integer, optional              :: m, n
     end subroutine zqrm_dsmat_nrm_async
  end interface qrm_dsmat_nrm_async


  interface qrm_dsmat_init_t
     subroutine zqrm_dsmat_init_t(qrm_dsmat_a, qrm_dsmat_t, ib, bh, pin, seq, info)
       import :: zqrm_dsmat_type
       type(zqrm_dsmat_type), intent(in)            :: qrm_dsmat_a
       type(zqrm_dsmat_type), intent(inout), target :: qrm_dsmat_t
       integer, intent(in)                          :: ib, bh
       logical, optional                            :: pin, seq
       integer, optional                            :: info
     end subroutine zqrm_dsmat_init_t
     subroutine zqrm_dsmat_init_t_tpqr(qrm_dsmat_a, qrm_dsmat_t, ib, ts, pin, seq, info)
       import :: zqrm_dsmat_type
       type(zqrm_dsmat_type), intent(in)            :: qrm_dsmat_a
       type(zqrm_dsmat_type), intent(inout), target :: qrm_dsmat_t
       integer, intent(in)                          :: ib
       character                                    :: ts
       logical, optional                            :: pin, seq
       integer, optional                            :: info
     end subroutine zqrm_dsmat_init_t_tpqr
  end interface qrm_dsmat_init_t

  interface qrm_dsmat_tpqr
     subroutine zqrm_dsmat_tpqr(a, b, t, ib, ts, work, prio, info)
       import :: zqrm_dsmat_type, zqrm_ws_type
       type(zqrm_dsmat_type)       :: a, b, t
       type(zqrm_ws_type)          :: work
       integer                     :: ib
       character                   :: ts
       integer, optional           :: prio
       integer, optional           :: info
     end subroutine zqrm_dsmat_tpqr
  end interface qrm_dsmat_tpqr

  interface qrm_dsmat_tpqr_async
     subroutine zqrm_dsmat_tpqr_async(qrm_dscr, a, b, t, ib, ts, work, prio)
       use qrm_dscr_mod
       import :: zqrm_dsmat_type, zqrm_ws_type
       type(qrm_dscr_type)         :: qrm_dscr
       type(zqrm_dsmat_type)       :: a, b, t
       type(zqrm_ws_type)          :: work
       integer                     :: ib
       character                   :: ts
       integer, optional           :: prio
     end subroutine zqrm_dsmat_tpqr_async
  end interface qrm_dsmat_tpqr_async

  interface qrm_dsmat_tpmqr
     subroutine zqrm_dsmat_tpmqr(v, a, b, t, ib, ts, work, prio, info)
       import :: zqrm_dsmat_type, zqrm_ws_type
       type(zqrm_dsmat_type)       :: v, a, b, t
       type(zqrm_ws_type)          :: work
       integer                     :: ib
       character                   :: ts
       integer, optional           :: prio
       integer, optional           :: info
     end subroutine zqrm_dsmat_tpmqr
  end interface qrm_dsmat_tpmqr

  interface qrm_dsmat_tpmqr_async
     subroutine zqrm_dsmat_tpmqr_async(qrm_dscr, v, a, b, t, ib, ts, work, prio)
       use qrm_dscr_mod
       import :: zqrm_dsmat_type, zqrm_ws_type
       type(qrm_dscr_type)         :: qrm_dscr
       type(zqrm_dsmat_type)       :: v, a, b, t
       type(zqrm_ws_type)          :: work
       integer                     :: ib
       character                   :: ts
       integer, optional           :: prio
     end subroutine zqrm_dsmat_tpmqr_async
  end interface qrm_dsmat_tpmqr_async


  interface qrm_dsmat_geqr
     subroutine zqrm_dsmat_geqr(a, t, ib, bh, work, prio, prepart, info)
       import :: zqrm_dsmat_type, zqrm_ws_type
       type(zqrm_dsmat_type)       :: a, t
       type(zqrm_ws_type)          :: work
       integer                     :: ib, bh
       integer, optional           :: prio
       integer, optional           :: prepart
       integer, optional           :: info
     end subroutine zqrm_dsmat_geqr
  end interface qrm_dsmat_geqr

  interface qrm_dsmat_geqr_async
     subroutine zqrm_dsmat_geqr_async(qrm_dscr, a, t, ib, bh, work, prio, prepart)
       use qrm_dscr_mod
       import :: zqrm_dsmat_type, zqrm_ws_type
       type(qrm_dscr_type)         :: qrm_dscr
       type(zqrm_dsmat_type)       :: a, t
       type(zqrm_ws_type)          :: work
       integer                     :: ib, bh
       integer, optional           :: prio
       integer, optional           :: prepart
     end subroutine zqrm_dsmat_geqr_async
  end interface qrm_dsmat_geqr_async

  interface qrm_dsmat_gemqr
     subroutine zqrm_dsmat_gemqr(trans, a, t, x, ib, bh, work, prio, info)
       use qrm_dscr_mod
       import :: zqrm_dsmat_type, zqrm_ws_type
       character                   :: trans
       type(zqrm_dsmat_type)       :: a, t, x
       type(zqrm_ws_type)          :: work
       integer                     :: ib, bh
       integer, optional           :: prio
       integer, optional           :: info
     end subroutine zqrm_dsmat_gemqr
  end interface qrm_dsmat_gemqr

  interface qrm_dsmat_gemqr_async
     subroutine zqrm_dsmat_gemqr_async(qrm_dscr, trans, a, t, x, ib, bh, work, prio)
       use qrm_dscr_mod
       import :: zqrm_dsmat_type, zqrm_ws_type
       type(qrm_dscr_type)         :: qrm_dscr
       character                   :: trans
       type(zqrm_dsmat_type)       :: a, t, x
       type(zqrm_ws_type)          :: work
       integer                     :: ib, bh
       integer, optional           :: prio
     end subroutine zqrm_dsmat_gemqr_async
  end interface qrm_dsmat_gemqr_async

  interface qrm_dsmat_trsm
     subroutine zqrm_dsmat_trsm(side, uplo, trans, diag, alpha, a, b, m, n, k, prio, info)
       use qrm_dscr_mod
       use qrm_parameters_mod
       import :: zqrm_dsmat_type
       type(zqrm_dsmat_type)       :: a, b
       complex(r64)                :: alpha
       character                   :: side, uplo, trans, diag
       integer, optional           :: m, n, k, prio
       integer, optional           :: info
     end subroutine zqrm_dsmat_trsm
  end interface qrm_dsmat_trsm

  interface qrm_dsmat_trsm_async
     subroutine zqrm_dsmat_trsm_async(qrm_dscr, side, uplo, trans, diag, alpha, a, b, m, n, k, prio)
       use qrm_parameters_mod
       use qrm_dscr_mod
       import :: zqrm_dsmat_type
       type(qrm_dscr_type)         :: qrm_dscr
       type(zqrm_dsmat_type)       :: a, b
       complex(r64)                :: alpha
       character                   :: side, uplo, trans, diag
       integer, optional           :: m, n, k, prio
     end subroutine zqrm_dsmat_trsm_async
  end interface qrm_dsmat_trsm_async

  interface qrm_dsmat_trmm
     subroutine zqrm_dsmat_trmm(side, uplo, trans, diag, alpha, a, b, m, n, k, prio, info)
       use qrm_dscr_mod
       use qrm_parameters_mod
       import :: zqrm_dsmat_type
       type(zqrm_dsmat_type)       :: a, b
       complex(r64)                :: alpha
       character                   :: side, uplo, trans, diag
       integer, optional           :: m, n, k, prio
       integer, optional           :: info
     end subroutine zqrm_dsmat_trmm
  end interface qrm_dsmat_trmm

  interface qrm_dsmat_trmm_async
     subroutine zqrm_dsmat_trmm_async(qrm_dscr, side, uplo, trans, diag, alpha, a, b, m, n, k, prio)
       use qrm_parameters_mod
       use qrm_dscr_mod
       import :: zqrm_dsmat_type
       type(qrm_dscr_type)         :: qrm_dscr
       type(zqrm_dsmat_type)       :: a, b
       complex(r64)                :: alpha
       character                   :: side, uplo, trans, diag
       integer, optional           :: m, n, k, prio
     end subroutine zqrm_dsmat_trmm_async
  end interface qrm_dsmat_trmm_async

  interface qrm_dsmat_gemm
     subroutine zqrm_dsmat_gemm(transa, transb, alpha, a, b, beta, c, m, n, k, prio, info)
       use qrm_parameters_mod
       use qrm_dscr_mod
       import :: zqrm_dsmat_type
       type(zqrm_dsmat_type)       :: a, b, c
       complex(r64)                :: alpha, beta
       character                   :: transa, transb
       integer, optional           :: m, n, k, prio
       integer, optional           :: info
     end subroutine zqrm_dsmat_gemm
  end interface qrm_dsmat_gemm

  interface qrm_dsmat_gemm_async
     subroutine zqrm_dsmat_gemm_async(qrm_dscr, transa, transb, alpha, a, b, beta, c, m, n, k, prio)
       use qrm_parameters_mod
       use qrm_dscr_mod
       import :: zqrm_dsmat_type
       type(qrm_dscr_type)            :: qrm_dscr
       type(zqrm_dsmat_type), target  :: a, b, c
       complex(r64)                   :: alpha, beta
       character                      :: transa, transb
       integer, optional              :: m, n, k, prio
     end subroutine zqrm_dsmat_gemm_async
  end interface qrm_dsmat_gemm_async

  interface qrm_dsmat_trdcn
     subroutine zqrm_dsmat_trdcn(a, d, eps, n, info)
       use qrm_parameters_mod
       import :: zqrm_dsmat_type
       type(zqrm_dsmat_type)       :: a
       integer                     :: d
       integer, optional           :: n, info
       real(r64)                   :: eps
     end subroutine zqrm_dsmat_trdcn
  end interface qrm_dsmat_trdcn

  interface qrm_dsmat_trdcn_async
     subroutine zqrm_dsmat_trdcn_async(qrm_dscr, a, d, eps, n)
       use qrm_parameters_mod
       use qrm_dscr_mod
       import :: zqrm_dsmat_type
       type(qrm_dscr_type)         :: qrm_dscr
       type(zqrm_dsmat_type)       :: a
       integer                     :: d
       integer, optional           :: n
       real(r64)                   :: eps
     end subroutine zqrm_dsmat_trdcn_async
  end interface qrm_dsmat_trdcn_async


  interface qrm_dsmat_potrf
     subroutine zqrm_dsmat_potrf(uplo, a, m, k, prio, info)
       import :: zqrm_dsmat_type
       type(zqrm_dsmat_type)       :: a
       character                   :: uplo
       integer, optional           :: m, k, prio, info
     end subroutine zqrm_dsmat_potrf
  end interface qrm_dsmat_potrf

  interface qrm_dsmat_potrf_async
     subroutine zqrm_dsmat_potrf_async(qrm_dscr, uplo, a, m, k, prio)
       use qrm_dscr_mod
       import :: zqrm_dsmat_type
       type(qrm_dscr_type)         :: qrm_dscr
       type(zqrm_dsmat_type)       :: a
       character                   :: uplo
       integer, optional           :: m, k, prio
     end subroutine zqrm_dsmat_potrf_async
  end interface qrm_dsmat_potrf_async



  interface qrm_higeqrt
     subroutine zqrm_higeqrt(qrm_dscr, a, t, nb, ib, work, prio)
       use qrm_dscr_mod
       import :: zqrm_block_type, zqrm_ws_type
       type(qrm_dscr_type)   :: qrm_dscr
       type(zqrm_block_type) :: a, t
       integer               :: nb, ib
       type(zqrm_ws_type)    :: work
       integer               :: prio
     end subroutine zqrm_higeqrt
  end interface qrm_higeqrt

  interface qrm_higemqrt
     subroutine zqrm_higemqrt(qrm_dscr, a, t, c, nb, ib, work, prio)
       use qrm_dscr_mod
       import :: zqrm_block_type, zqrm_ws_type
       type(qrm_dscr_type)   :: qrm_dscr
       type(zqrm_block_type) :: a, t, c
       type(zqrm_ws_type)    :: work
       integer               :: nb, ib
       integer               :: prio
     end subroutine zqrm_higemqrt
  end interface qrm_higemqrt

  interface qrm_hitpqrt
     subroutine zqrm_hitpqrt(qrm_dscr, a1, a2, t, nb, ib, ts, work, prio)
       use qrm_dscr_mod
       import :: zqrm_block_type, zqrm_ws_type
       type(qrm_dscr_type)   :: qrm_dscr
       type(zqrm_block_type) :: a1, a2, t
       type(zqrm_ws_type)    :: work
       integer               :: nb, ib
       character             :: ts
       integer               :: prio
     end subroutine zqrm_hitpqrt
  end interface qrm_hitpqrt

  interface qrm_hitmpqrt
     subroutine zqrm_hitpmqrt(qrm_dscr, v, t, c1, c2, nb, ib, ts, work, prio)
       use qrm_dscr_mod
       import :: zqrm_block_type, zqrm_ws_type
       type(qrm_dscr_type)   :: qrm_dscr
       type(zqrm_block_type) :: v, t, c1, c2
       type(zqrm_ws_type)    :: work
       integer               :: nb, ib
       character             :: ts
       integer               :: prio
     end subroutine zqrm_hitpmqrt
  end interface qrm_hitmpqrt

contains

  subroutine zqrm_block_destroy(b, pin, seq, info)

    type(zqrm_block_type) :: b
    logical, optional     :: pin, seq
    integer, optional     :: info

    logical               :: iseq
    integer, target       :: nparts

    if(.not.qrm_allocated(b%c)) return

    if(present(seq)) then
       iseq = seq
    else
       iseq = .false.
    end if

#if defined (have_starpu)
    if(.not.iseq) then
       if(allocated(b%shdls)) then
          ! a partition was planned and thus it has to be cleaned up
          nparts = size(b%shdls)
          call fstarpu_data_partition_clean (b%hdl,nparts,b%shdls)
          deallocate(b%shdls)
       end if
       if(c_associated(b%hdl)) then
          call fstarpu_data_unregister(b%hdl)
          b%hdl = c_null_ptr
       end if
    end if
#endif

    call qrm_dealloc(b%c, pin=pin)

    call qrm_dealloc(b%stair)

    return
  end subroutine zqrm_block_destroy


  subroutine zqrm_block_unregister_async(b, info)

    type(zqrm_block_type) :: b
    integer, optional     :: info

    logical               :: iseq
    integer, target       :: nparts

#if defined (have_starpu)
    if(allocated(b%shdls)) then
       ! a partition was planned and thus it has to be cleaned up
       nparts = size(b%shdls)
       call fstarpu_data_partition_clean (b%hdl,nparts,b%shdls)
       deallocate(b%shdls)
    end if
    if(c_associated(b%hdl)) then
       call fstarpu_data_unregister_submit(b%hdl)
       b%hdl = c_null_ptr
    end if
#endif

    return
  end subroutine zqrm_block_unregister_async



  subroutine zqrm_dsmat_init(qrm_dsmat, m, n, mb, nb, pin, stair, init, ivalue, imat, seq, uplo, info)
    use qrm_parameters_mod
#if defined (have_starpu)
    use fstarpu_mod
    use zqrm_starpu_mod
#endif
    implicit none

    type(zqrm_dsmat_type), target  :: qrm_dsmat
    integer                        :: m, n, mb, nb
    integer, optional              :: info, stair(:)
    logical, optional              :: pin, seq
    character, optional            :: init, uplo
    complex(r64), optional         :: ivalue, imat(:,:)

    integer                        :: nbr, nbc, i, j, k, is, js, fr, fc, parts, ib, ie
    type(c_ptr)                    :: bc_filter
    logical                        :: izero, iseq
    integer, save                  :: iseed(4)=(/1, 1, 1, 1/)
    character                      :: iinit, iuplo

    ! error handling
    integer                        :: err
    character(len=*), parameter    :: name='qrm_dsmat_init'

    err = 0

    ! if(mod(mb,nb).ne.0) then
       ! write(*,*)'Error!'
       ! err = 1000 ! FIXME: proper error code
       ! goto 9999
    ! end if

    __QRM_IFPRESENT3(iseq, seq, .false.)
    __QRM_IFPRESENT3(iinit, init, qrm_no)
    __QRM_IFPRESENT3(iuplo, uplo, qrm_no)

    qrm_dsmat%m  = m
    qrm_dsmat%n  = n
    qrm_dsmat%mb = mb
    qrm_dsmat%nb = nb

    nbr = (m-1)/mb+1
    nbc = (n-1)/mb+1

#if defined (have_starpu)
    if(.not.iseq) then
       bc_filter = qrm_get_bc_filter()
       call fstarpu_data_filter_set_filter_arg (bc_filter, nb)
       call zqrm_starpu_init_codelets()
    end if
#endif
    allocate(qrm_dsmat%blocks(nbr, nbc))

    cols: do j=1, nbc
       fc = (j-1)*mb+1
       if(iuplo.eq.qrm_upper) then
          ib=1; ie=j
       else if(iuplo.eq.qrm_lower) then
          ib=j; ie=nbr
       else
          ib=1; ie=nbr
       end if
       rows: do i=ib, ie
          fr = (i-1)*mb+1

          ! if we get below the stair move to the next col
          is = min(mb,m-fr+1)
          js = min(mb,n-fc+1)

          if(present(stair)) then
             ! if we're below the staircase, then move to the next column
             if(fr .gt. stair(fc+js-1)) cycle cols
             ! if this block is traversed by the staircase, allocate its stair
             if((stair(fc).lt.fr+is-1) .and. (stair(fc+js-1).ge.fr)) then
                call qrm_alloc(qrm_dsmat%blocks(i,j)%stair, js, err)
                do k=1, js
                   qrm_dsmat%blocks(i,j)%stair(k) = min(max(stair(fc+k-1)-fr+1, 0),is)
                end do
             end if
          end if

          ! allocate the block
          call qrm_alloc(qrm_dsmat%blocks(i,j)%c, is, js, err, pin)
          __QRM_INFO_CHECK(err, name, 'qrm_alloc', 9999)

          ! init out the block if required
          if(iinit .ne. qrm_no) then
             if(qrm_allocated(qrm_dsmat%blocks(i,j)%stair)) then
                qrm_dsmat%blocks(i,j)%c = qrm_zzero
                do k=1, js
                   if(iinit.eq.'r') then
                      call zlarnv(2, iseed(1), &
                           min(is, qrm_dsmat%blocks(i,j)%stair(k)), &
                           qrm_dsmat%blocks(i,j)%c(1, k))
                   else if(iinit.eq.'v') then
                      qrm_dsmat%blocks(i,j)%c(1:qrm_dsmat%blocks(i,j)%stair(k), k) = ivalue
                   else if(iinit.eq.'m') then
                      qrm_dsmat%blocks(i,j)%c(1:qrm_dsmat%blocks(i,j)%stair(k), k) = &
                           imat(fr:fr+qrm_dsmat%blocks(i,j)%stair(k)-1, fc+k-1)
                   end if
                end do
             else
                if(iinit.eq.'r') then
                   call zlarnv(2, iseed(1), is*js, qrm_dsmat%blocks(i,j)%c(1, 1))
                else if(iinit.eq.'v') then
                   qrm_dsmat%blocks(i,j)%c = ivalue
                else if(iinit.eq.'m') then
                   qrm_dsmat%blocks(i,j)%c = imat(fr:min(fr+mb-1,m), fc:min(fc+mb-1,n))
                end if
             end if
          end if

#if defined (have_starpu)
          if(.not.iseq) then
             ! register the block to starpu
             call fstarpu_matrix_data_register(qrm_dsmat%blocks(i,j)%hdl, 0,           &
                  c_loc(qrm_dsmat%blocks(i,j)%c),         &
                  size(qrm_dsmat%blocks(i,j)%c,1),        &
                  size(qrm_dsmat%blocks(i,j)%c,1),        &
                  size(qrm_dsmat%blocks(i,j)%c,2),        &
                  c_sizeof(qrm_dsmat%blocks(i,j)%c(1,1)))
             if(nb.lt.js) then
                ! plan a dynamic partitioning
                parts = (js-1)/nb+1
                allocate(qrm_dsmat%blocks(i,j)%shdls(parts))
                call fstarpu_data_partition_plan(qrm_dsmat%blocks(i,j)%hdl, &
                     bc_filter, qrm_dsmat%blocks(i,j)%shdls)
             end if
          end if
#endif
          qrm_dsmat%blocks(i,j)%partitioned = .false.
       end do rows
    end do cols

    qrm_dsmat%inited = .true.

9999 continue
    if(present(info)) info = err
    return
  end subroutine zqrm_dsmat_init


  subroutine zqrm_dsmat_destroy(qrm_dsmat, pin, seq)

    type(zqrm_dsmat_type), target  :: qrm_dsmat
    logical, optional              :: pin, seq

    integer :: i, j

    if(.not.qrm_dsmat%inited) return

    do j=1, size(qrm_dsmat%blocks,2)
       do i=1, size(qrm_dsmat%blocks,1)
          call zqrm_block_destroy(qrm_dsmat%blocks(i,j), pin=pin, seq=seq)
       end do
    end do

    if(allocated(qrm_dsmat%blocks)) deallocate(qrm_dsmat%blocks)

    qrm_dsmat%inited = .false.

    return
  end subroutine zqrm_dsmat_destroy


  subroutine zqrm_ws_init(work, m, n, info)
    implicit none
    type(zqrm_ws_type), target :: work
    integer                    :: m, n
    integer, optional          :: info

    complex(r64) :: x

#if defined (have_starpu)
    call fstarpu_matrix_data_register(work%hdl, -1, &
         c_null_ptr,                                              &
         m, &
         m, &
         n, &
         c_sizeof(x))
#else
    call qrm_alloc(work%c, m, n, info=info)
#endif

    return
  end subroutine zqrm_ws_init


  subroutine zqrm_ws_destroy(w, seq, info)

    type(zqrm_ws_type)    :: w
    logical, optional     :: seq
    integer, optional     :: info

    logical               :: iseq

    if(present(seq)) then
       iseq = seq
    else
       iseq = .false.
    end if

#if defined (have_starpu)
    if(.not.iseq) then
       if(c_associated(w%hdl)) then
          call fstarpu_data_unregister(w%hdl)
          w%hdl = c_null_ptr
       end if
    end if
#endif

    call qrm_dealloc(w%c)

    return
  end subroutine zqrm_ws_destroy

  subroutine zqrm_block_extadd(a, b, br, bc, i, j, m, n, l, inout, op, rowmap, colmap)
    use qrm_string_mod
    implicit none

    type(zqrm_dsmat_type)  :: a                    ! if copyout this is the
                                                   ! matrix containing the source
                                                   ! block. If copyin it contains
                                                   ! the destinantion block

    type(zqrm_dsmat_type)  :: b                    ! if copyout contains the
                                                   ! destination blocks; if
                                                   ! copyin contains the source
                                                   ! blocks
    integer                :: br, bc               ! block-row and block-column
                                                   ! of the source/destination
                                                   ! block in a
    integer                :: i, j, m, n, l        ! i and j are the first row
                                                   ! and column within the (bc,br)
                                                   ! block and m and n are the
                                                   ! number of rows and cols to
                                                   ! be copied. l is the number of
                                                   ! rows in the upper trapezoidal
                                                   ! part of the block (same meaning
                                                   ! as in tpqrt
    character              :: inout, op
    integer                :: rowmap(:,:)          ! block-row mapping from a to b
    integer, optional      :: colmap(:,:)          ! block-col mapping from a to b


    integer                :: abr, abc ! block row and col in a (just a copy of br and bc)
    integer                :: bbr, bbc ! block row and col in b

    logical                :: iout
    integer                :: ai, aj, agi, agj ! local (to block) and global row and col indices in a
    integer                :: bi, bj, bgi, bgj ! local (to block) and global row and col indices in b

    integer                :: aroff, acoff ! row and column offsets in a

    if(.not. qrm_allocated(a%blocks(br,bc)%c)) return

    abr = br ! just a copy to avoid confusion with blocks ob b
    abc = bc ! just a copy to avoid confusion with blocks ob b

    aroff = (abr-1)*a%mb  ! row-offset in a
    acoff = (abc-1)*a%mb  ! col-offset in a


    ! This is very ugly duplicate code but this is done to minimize
    ! the "if"s in inner loops
    if(op.eq.'a') then
       if(inout.eq.'o') then
          if(present(colmap)) then
             colsaoc: do aj = j, j+n-1
                agj = acoff + aj                               ! global a col index of the current coefficient
                bbc = colmap(agj,2)                           ! block-col in b     of the current coefficient
                bj  = colmap(agj,3)                           ! col index within block-col bbc of b

                rowsaoc: do ai = i, min(i+aj-j+m-l,i+m-1)
                   agi = aroff + ai                            ! global a row index of the current coefficient
                   bbr = rowmap(agi,2)                        ! block-row in b     of the current coefficient
                   bi  = rowmap(agi,3)                        ! row index within block-row bbr of b

                   b%blocks(bbr,bbc)%c(bi,bj) = b%blocks(bbr,bbc)%c(bi,bj)+a%blocks(abr,abc)%c(ai,aj)

                end do rowsaoc
             end do colsaoc
          else
             colsao: do aj = j, j+n-1
                agj = acoff + aj                               ! global a col index of the current coefficient
                bgj = agj                                      ! global b col index of the current coefficient
                bbc = (bgj-1)/b%mb+1                           ! block-col in b     of the current coefficient
                bj  = mod(bgj-1,b%mb) + 1                      ! col index within block-col bbc of b

                rowsao: do ai = i, min(i+aj-j+m-l,i+m-1)
                   agi = aroff + ai                            ! global a row index of the current coefficient
                   bbr = rowmap(agi,2)                        ! block-row in b     of the current coefficient
                   bi  = rowmap(agi,3)                        ! row index within block-row bbr of b

                   b%blocks(bbr,bbc)%c(bi,bj) = b%blocks(bbr,bbc)%c(bi,bj)+a%blocks(abr,abc)%c(ai,aj)

                end do rowsao
             end do colsao
          end if
       else if(inout.eq.'i') then
          if(present(colmap)) then
             colscaic: do aj = j, j+n-1
                agj = acoff + aj                               ! global a col index of the current coefficient
                bbc = colmap(agj,2)                           ! block-col in b     of the current coefficient
                bj  = colmap(agj,3)                           ! col index within block-col bbc of b

                rowscaic: do ai = i, min(i+aj-j+m-l,i+m-1)
                   agi = aroff + ai                            ! global a row index of the current coefficient
                   bbr = rowmap(agi,2)                        ! block-row in b     of the current coefficient
                   bi  = rowmap(agi,3)                        ! row index within block-row bbr of b

                   a%blocks(abr,abc)%c(ai,aj) = a%blocks(abr,abc)%c(ai,aj) + b%blocks(bbr,bbc)%c(bi,bj)

                end do rowscaic
             end do colscaic
          else
             colsai: do aj = j, j+n-1
                agj = acoff + aj                               ! global a col index of the current coefficient
                bgj = agj                                      ! global b col index of the current coefficient
                bbc = (bgj-1)/b%mb+1                           ! block-col in b     of the current coefficient
                bj  = mod(bgj-1,b%mb) + 1                      ! col index within block-col bbc of b

                rowsai: do ai = i, min(i+aj-j+m-l,i+m-1)
                   agi = aroff + ai                            ! global a row index of the current coefficient
                   bbr = rowmap(agi,2)                        ! block-row in b     of the current coefficient
                   bi  = rowmap(agi,3)                        ! row index within block-row bbr of b

                   a%blocks(abr,abc)%c(ai,aj) = a%blocks(abr,abc)%c(ai,aj) + b%blocks(bbr,bbc)%c(bi,bj)

                end do rowsai
             end do colsai
          end if
       end if

    else if(op.eq.'c') then
       if(inout.eq.'o') then
          if(present(colmap)) then
             colscoc: do aj = j, j+n-1
                agj = acoff + aj                               ! global a col index of the current coefficient
                bbc = colmap(agj,2)                           ! block-col in b     of the current coefficient
                bj  = colmap(agj,3)                           ! col index within block-col bbc of b

                rowscoc: do ai = i, min(i+aj-j+m-l,i+m-1)
                   agi = aroff + ai                            ! global a row index of the current coefficient
                   bbr = rowmap(agi,2)                        ! block-row in b     of the current coefficient
                   bi  = rowmap(agi,3)                        ! row index within block-row bbr of b

                   b%blocks(bbr,bbc)%c(bi,bj) = a%blocks(abr,abc)%c(ai,aj)

                end do rowscoc
             end do colscoc
          else
             colsco: do aj = j, j+n-1
                agj = acoff + aj                               ! global a col index of the current coefficient
                bgj = agj                                      ! global b col index of the current coefficient
                bbc = (bgj-1)/b%mb+1                           ! block-col in b     of the current coefficient
                bj  = mod(bgj-1,b%mb) + 1                      ! col index within block-col bbc of b

                rowsco: do ai = i, min(i+aj-j+m-l,i+m-1)
                   agi = aroff + ai                            ! global a row index of the current coefficient
                   bbr = rowmap(agi,2)                        ! block-row in b     of the current coefficient
                   bi  = rowmap(agi,3)                        ! row index within block-row bbr of b

                   b%blocks(bbr,bbc)%c(bi,bj) = a%blocks(abr,abc)%c(ai,aj)

                end do rowsco
             end do colsco
          end if
       else if(inout.eq.'i') then
          if(present(colmap)) then
             colscic: do aj = j, j+n-1
                agj = acoff + aj                               ! global a col index of the current coefficient
                bbc = colmap(agj,2)                           ! block-col in b     of the current coefficient
                bj  = colmap(agj,3)                           ! col index within block-col bbc of b

                rowscic: do ai = i, min(i+aj-j+m-l,i+m-1)
                   agi = aroff + ai                            ! global a row index of the current coefficient
                   bbr = rowmap(agi,2)                        ! block-row in b     of the current coefficient
                   bi  = rowmap(agi,3)                        ! row index within block-row bbr of b

                   a%blocks(abr,abc)%c(ai,aj) = b%blocks(bbr,bbc)%c(bi,bj)

                end do rowscic
             end do colscic
          else
             colsci: do aj = j, j+n-1
                agj = acoff + aj                               ! global a col index of the current coefficient
                bgj = agj                                      ! global b col index of the current coefficient
                bbc = (bgj-1)/b%mb+1                           ! block-col in b     of the current coefficient
                bj  = mod(bgj-1,b%mb) + 1                      ! col index within block-col bbc of b

                rowsci: do ai = i, min(i+aj-j+m-l,i+m-1)
                   agi = aroff + ai                            ! global a row index of the current coefficient
                   bbr = rowmap(agi,2)                        ! block-row in b     of the current coefficient
                   bi  = rowmap(agi,3)                        ! row index within block-row bbr of b

                   a%blocks(abr,abc)%c(ai,aj) = b%blocks(bbr,bbc)%c(bi,bj)

                end do rowsci
             end do colsci
          end if
       end if

    end if

    return
end subroutine zqrm_block_extadd


  subroutine zqrm_block_partition1(a)
    implicit none

    type(zqrm_block_type) :: a
    integer               :: parts

#if defined (have_starpu)
    if(qrm_allocated(a%c)) then
       if(.not.a%partitioned) then
          if(allocated(a%shdls)) then
             parts = size(a%shdls)
             call fstarpu_data_partition_submit(a%hdl, parts, a%shdls)
             a%partitioned = .true.
          end if
       end if
    end if
#endif

    return
  end subroutine zqrm_block_partition1

  subroutine zqrm_block_partition2(a, b)
    implicit none

    type(zqrm_block_type) :: a, b
    integer               :: parts

#if defined (have_starpu)
    if(qrm_allocated(a%c)) then
       if(.not.a%partitioned) then
          if(allocated(a%shdls)) then
             parts = size(a%shdls)
             call fstarpu_data_partition_submit(a%hdl, parts, a%shdls)
             a%partitioned = .true.
          end if
       end if
    end if
#endif

#if defined (have_starpu)
    if(qrm_allocated(b%c)) then
       if(.not.b%partitioned) then
          if(allocated(b%shdls)) then
             parts = size(b%shdls)
             call fstarpu_data_partition_submit(b%hdl, parts, b%shdls)
             b%partitioned = .true.
          end if
       end if
    end if
#endif

    return
  end subroutine zqrm_block_partition2

  subroutine zqrm_block_unpartition1(a)
    implicit none

    type(zqrm_block_type) :: a
    integer               :: parts

#if defined (have_starpu)
    if(qrm_allocated(a%c)) then
       if(a%partitioned) then
          parts = size(a%shdls)
          call fstarpu_data_unpartition_submit(a%hdl, parts, a%shdls, 0)
          a%partitioned = .false.
       end if
    end if
#endif

    return
  end subroutine zqrm_block_unpartition1

  subroutine zqrm_block_unpartition2(a, b)
    implicit none

    type(zqrm_block_type) :: a, b
    integer               :: parts

#if defined (have_starpu)
    if(qrm_allocated(a%c)) then
       if(a%partitioned) then
          parts = size(a%shdls)
          call fstarpu_data_unpartition_submit(a%hdl, parts, a%shdls, 0)
          a%partitioned = .false.
       end if
    end if
#endif

#if defined (have_starpu)
    if(qrm_allocated(b%c)) then
       if(b%partitioned) then
          parts = size(b%shdls)
          call fstarpu_data_unpartition_submit(b%hdl, parts, b%shdls, 0)
          b%partitioned = .false.
       end if
    end if
#endif

    return
  end subroutine zqrm_block_unpartition2


  subroutine zqrm_block_print(qrm_block, fm, unit)
    implicit none

    type(zqrm_block_type)  :: qrm_block
    character(len=*)       :: fm
    integer, optional      :: unit

    integer :: i, j, iunit, l
    character(len=:), allocatable :: zfm, dfm

    if(present(unit)) then
       iunit = unit
    else
       iunit = 6
    end if

    l = len(fm)

    if(fm(1:1).eq.'f') then
       i = 2
    else if(fm(1:1).eq.'e') then
       if(fm(2:2).eq.'s') then
          i = 3
       else
          i = 2
       end if
    else
       write(*,'("Invalid format. Returning.")')
       return
    end if

    do j=i, l
       if(fm(j:j).eq.'.') exit
    end do
    ! if(j.eq.l+1) j=j-1

    zfm = "("//fm(i:j-1)//"x,x"//")"
    dfm = "("//fm//",x)"


    do i=1, size(qrm_block%c,1)
       do j=1, size(qrm_block%c,2)
          write(iunit,dfm, advance='no')qrm_block%c(i,j)
       end do
       write(iunit,'(" ")')
    end do
    return
  end subroutine zqrm_block_print

  subroutine zqrm_dsmat_print(qrm_dsmat, fm, triu, tril, unit, zero, m, absval, ir)
    use qrm_parameters_mod
    implicit none

    type(zqrm_dsmat_type) :: qrm_dsmat
    character(len=*)       :: fm
    logical, optional      :: triu, tril, zero, absval, ir
    integer, optional      :: unit, m

    logical                :: itriu, itril, izero, iabsval, iir
    integer                :: i, j, ii, jj, br, bc, zlen, l, iunit, im
    real                   :: datalen
    character              :: w
    character(len=:), allocatable :: zfm, dfm

    __QRM_IFPRESENT3(itriu,triu,.false.)
    __QRM_IFPRESENT3(itril,tril,.false.)
    __QRM_IFPRESENT3(iunit,unit,6)
    __QRM_IFPRESENT3(izero,zero,.false.)
    __QRM_IFPRESENT3(im,m,qrm_dsmat%m)
    __QRM_IFPRESENT3(iabsval,absval,.false.)
    __QRM_IFPRESENT3(iir,ir,.false.)

    l = len(fm)

    if(fm(1:1).eq.'f') then
       i = 2
    else if(fm(1:1).eq.'e') then
       if(fm(2:2).eq.'s') then
          i = 3
       else
          i = 2
       end if
    else
       write(*,'("Invalid format. Returning.")')
       return
    end if

    do j=i, l
       if(fm(j:j).eq.'.') exit
    end do
    ! if(j.eq.l+1) j=j-1

#if defined(dprec) || defined (sprec)
    zfm = "("//fm(i:j-1)//"x,x"//")"
    dfm = "("//fm//",x)"
#elif defined(zprec) || defined (cprec)
    if(iir) then
       w   = 'r' ! real part first
       zfm = "("//fm(i:j-1)//"x,x"//")"
       dfm = "("//fm//",x)"
    else
       zfm = "("//fm(i:j-1)//"x"//fm(i:j-1)//"x,4x"//")"
       dfm = "(""("","//fm//","","","//fm//","")"",x)"
    end if
#endif

#if defined (have_starpu)
    do i = 1, size(qrm_dsmat%blocks, 1)
       do j = 1, size(qrm_dsmat%blocks, 2)
          if (c_associated(qrm_dsmat%blocks(i,j)%hdl)) then
             call fstarpu_data_acquire(qrm_dsmat%blocks(i,j)%hdl, FSTARPU_R)
          end if
       end do
    end do
#endif

    
10 continue
    
    rows: do i=1, im !qrm_dsmat%m
       cols: do j=1, qrm_dsmat%n
          if(itriu) then
             if (i.gt.min(qrm_dsmat%m,qrm_dsmat%n)) cycle rows
             if(i.gt.j) then
                if(izero) then
#if defined(dprec) || defined (sprec)
                   write(iunit,dfm, advance='no')qrm_dzero
#elif defined(zprec) || defined (cprec)
                   if(iir) then
                      write(iunit,dfm, advance='no')qrm_dzero
                   else
                      write(iunit,dfm, advance='no')qrm_dzero,qrm_dzero
                   end if
#endif
                else
                   write(iunit,zfm, advance='no')
                end if
                cycle cols
             end if
          end if

          if(itril) then
             if (i.le.qrm_dsmat%m-min(qrm_dsmat%m,qrm_dsmat%n)) cycle rows
             if (i-min(qrm_dsmat%m,qrm_dsmat%n).lt.j) then
                write(iunit,zfm, advance='no')
                exit cols
             end if
          end if

          br = (i-1)/qrm_dsmat%mb+1
          bc = (j-1)/qrm_dsmat%mb+1

          if(.not.qrm_allocated(qrm_dsmat%blocks(br,bc)%c)) then
             if(izero) then
#if defined(dprec) || defined (sprec)
                   write(iunit,dfm, advance='no')qrm_dzero
#elif defined(zprec) || defined (cprec)
                   if(iir) then
                      write(iunit,dfm, advance='no')qrm_dzero
                   else
                      write(iunit,dfm, advance='no')qrm_dzero,qrm_dzero
                   end if
#endif
             else
                write(iunit,zfm, advance='no')
             end if
             cycle cols
          end if

          ii = i-(br-1)*qrm_dsmat%mb
          jj = j-(bc-1)*qrm_dsmat%mb

          if(qrm_allocated(qrm_dsmat%blocks(br,bc)%stair)) then
             if(ii.gt.qrm_dsmat%blocks(br,bc)%stair(jj)) then
                if(izero) then
#if defined(dprec) || defined (sprec)
                   write(iunit,dfm, advance='no')qrm_dzero
#elif defined(zprec) || defined (cprec)
                   if(iir) then
                      write(iunit,dfm, advance='no')qrm_dzero
                   else
                      write(iunit,dfm, advance='no')qrm_dzero,qrm_dzero
                   end if
#endif
                else
                   write(iunit,zfm, advance='no')
                end if
                cycle cols
             end if
          end if

          if(qrm_allocated(qrm_dsmat%blocks(br,bc)%c)) then
             if(ii.le.size(qrm_dsmat%blocks(br,bc)%c,1) .and. &
                  jj.le.size(qrm_dsmat%blocks(br,bc)%c,2)) then
                if(iabsval) then
#if defined(dprec) || defined (sprec)
                   write(iunit,dfm, advance='no')abs(qrm_dsmat%blocks(br,bc)%c(ii,jj))
#elif defined(zprec) || defined (cprec)
                   if(iir) then
                      if(w.eq.'r') then
                         write(iunit,dfm, advance='no')&
                              abs(real(qrm_dsmat%blocks(br,bc)%c(ii,jj)))
                      else if(w.eq.'i') then
                         write(iunit,dfm, advance='no')&
                              abs(aimag(qrm_dsmat%blocks(br,bc)%c(ii,jj)))
                      end if
                   else
                      write(iunit,dfm, advance='no')&
                           abs(real(qrm_dsmat%blocks(br,bc)%c(ii,jj))), &
                           abs(aimag(qrm_dsmat%blocks(br,bc)%c(ii,jj)))
                   end if
#endif
                   write(iunit,dfm, advance='no')
                else
#if defined(dprec) || defined (sprec)
                   write(iunit,dfm, advance='no')qrm_dsmat%blocks(br,bc)%c(ii,jj)
#elif defined(zprec) || defined (cprec)
                   if(iir) then
                      if(w.eq.'r') then
                         write(iunit,dfm, advance='no')&
                              real(qrm_dsmat%blocks(br,bc)%c(ii,jj))
                      else if(w.eq.'i') then
                         write(iunit,dfm, advance='no')&
                              aimag(qrm_dsmat%blocks(br,bc)%c(ii,jj))
                      end if
                   else
                      write(iunit,dfm, advance='no')&
                           real(qrm_dsmat%blocks(br,bc)%c(ii,jj)), &
                           aimag(qrm_dsmat%blocks(br,bc)%c(ii,jj))
                   end if
#endif
                end if
             else
                write(iunit,zfm, advance='no')
             end if
          else
             write(iunit,zfm, advance='no')
          end if
       end do cols
       if((.not.itriu) .or. (i.le.min(qrm_dsmat%m,qrm_dsmat%n))) write(iunit,'(" ")')
    end do rows

#if defined(zprec) || defined (cprec)
    if(w.eq.'r') then
       write(iunit,'(30("-"))')
       w = 'i'
       goto 10
    end if
#endif

#if defined (have_starpu)
    do i = 1, size(qrm_dsmat%blocks, 1)
       do j = 1, size(qrm_dsmat%blocks, 2)
          if (c_associated(qrm_dsmat%blocks(i,j)%hdl)) then
             call fstarpu_data_release(qrm_dsmat%blocks(i,j)%hdl)
          end if
       end do
    end do
#endif
    
    return

  end subroutine zqrm_dsmat_print


  subroutine zqrm_dsmat_read(qrm_dsmat, a)

    implicit none

    type(zqrm_dsmat_type) :: qrm_dsmat
    complex(r64)             :: a(:,:)

    integer :: i, j

    if((size(a,1).lt.qrm_dsmat%m) .or. (size(a,2).lt.qrm_dsmat%n)) then
       write(*,'("Insufficient a size in qrm_dsmat_read")')
       return
    end if


    do i = 1, size(qrm_dsmat%blocks, 1)
       do j = 1, size(qrm_dsmat%blocks, 2)

          if (qrm_allocated(qrm_dsmat%blocks(i,j)%c)) then
#if defined (have_starpu)
             call fstarpu_data_acquire(qrm_dsmat%blocks(i,j)%hdl, FSTARPU_R)
#endif
             ! Copy in a
             a((i-1)*qrm_dsmat%mb+1:min(i*qrm_dsmat%mb,qrm_dsmat%m), &
               (j-1)*qrm_dsmat%mb+1:min(j*qrm_dsmat%mb,qrm_dsmat%n)) &
               = qrm_dsmat%blocks(i,j)%c
#if defined (have_starpu)
             call fstarpu_data_release(qrm_dsmat%blocks(i,j)%hdl)
#endif
          end if
       end do
    end do

    return

  end subroutine zqrm_dsmat_read









end module zqrm_dsmat_mod
