!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_clean_front.F90
!! This file contains the routines for cleaning of a front
!!
!!
!! ##############################################################################################


#include "qrm_common.h"

!> @brief This routine performs the cleaning of a front.

!> Cleaning a front means saving the parts corresponding tot he R and
!! Q factors, and then freeing all the memory that is not needed
!! anymore
!!
!! @param[in,out] qrm_spfct the whole problem. this obviously contains
!!                        the fornt to be cleaned
!!
!! @param[in,out] front the front to be cleaned
!!
subroutine cqrm_clean_front(qrm_spfct, front, info)

  use qrm_mem_mod
  use cqrm_spfct_mod
  use cqrm_fdata_mod
  use cqrm_factorization_mod, protect => cqrm_clean_front  
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(cqrm_spfct_type), target   :: qrm_spfct
  type(cqrm_front_type)           :: front
  integer, optional               :: info
  
  integer                         :: i, j, p, c, row, ne, npiv
  integer(kind=8)                 :: peak, pers
  logical                         :: storeh, storer
  type(qrm_adata_type), pointer   :: adata
  type(cqrm_fdata_type), pointer  :: fdata
  type(cqrm_front_type), pointer  :: cfront
  integer(kind=8)                :: ts, te
  
  ! error management
  integer                         :: err
  character(len=*), parameter     :: name='qrm_clean_front'

  __QRM_PROF_START(ts)

  err = 0
  
  if(min(front%m,front%n).le.0) then
     goto 9999
  end if
  ! to make things easier
  adata       => qrm_spfct%adata
  fdata       => qrm_spfct%fdata
  
  storer = qrm_spfct%icntl(qrm_keeph_).ge.0
  storeh = qrm_spfct%icntl(qrm_keeph_).ge.1
  
  ! tmpmem = qrm_tot_mem
  
  if(err.eq.0) call qrm_dealloc(front%aiptr, err)
  if(err.eq.0) call qrm_dealloc(front%ajcn,  err)
  if(err.eq.0) call qrm_dealloc(front%aval,  err)
  __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)
  
  if(.not.storer) then
     call qrm_dealloc(front%cols)
     call qrm_dealloc(front%rows)
     call qrm_dealloc(front%colmap)
     call qrm_dealloc(front%rowmap)
     call qrm_dealloc(front%stair)
  end if

  ! write(*,'(i5,2x,i5," clean -- ",i10,2x,i10,2x,f7.2,2x,i10,2x,i10)')front%num, qrm_spfct%adata%small(front%num), &
       ! & tmpmem-qrm_tot_mem, &
       ! & qrm_spfct%adata%csize(front%num), &
       ! & 100.d0*real(qrm_spfct%adata%csize(front%num)-(tmpmem-qrm_tot_mem))/real(tmpmem-qrm_tot_mem), &
       ! & qrm_tot_mem, qrm_max_mem
  
  if(qrm_spfct%adata%small(front%num).ge.0) then
     call qrm_facto_mem_get(qrm_spfct%fdata%ma, -qrm_spfct%adata%csize(front%num))
  end if

  call qrm_atomic_add(qrm_spfct%gstats(qrm_nnz_r_), front%rsize)
  call qrm_atomic_add(qrm_spfct%gstats(qrm_nnz_h_), front%hsize)

  
9999 continue
  __QRM_PROF_STOP(qrm_prof_clean_front_task_,ts,te)

  if(present(info)) info = err
  return

end subroutine cqrm_clean_front


