!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_do_subtree.F90
!! This file contains the routines that take care of factorizing sequentially an entire subtree
!!
!!
!! ##############################################################################################


#include "qrm_common.h"

!> @brief This subroutine does the sequential factorization of an
!! entire subtree
!!
!! @param[in,out] qrm_spfct the whole problem data structure
!!
!! @param[in] root the root of the subtree
!! 
!! @param[in,out] flops a counter for the flops which is updated
!! 
subroutine cqrm_do_subtree(qrm_spfct, root, flops, info)
  
  use iso_c_binding
  use qrm_mem_mod
  use cqrm_spfct_mod
  use cqrm_utils_mod
  use qrm_common_mod
  use cqrm_factorization_mod, protect => cqrm_do_subtree
  use qrm_dscr_mod
  use cqrm_dsmat_mod
  use qrm_parameters_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  
  implicit none
  
  type(cqrm_spfct_type), target  :: qrm_spfct
  type(cqrm_front_type), target  :: root
  real(r32)                      :: flops
  integer, optional              :: info
  
  type(cqrm_front_type), pointer :: child, ffront, front
  type(qrm_adata_type), pointer  :: adata
  type(cqrm_fdata_type), pointer :: fdata
  
  integer                        :: node, p, c, i, j, cnt, f, m, n, k, ofsa, ib, br, bc, d
  integer(kind=8)                :: asize, csize, hsize, rsize
  logical                        :: storer, storeh
  type(qrm_dscr_type)            :: qrm_seq_dscr
  type(cqrm_ws_type)             :: work
  real(r32)                      :: eps
  integer(kind=8)                :: ts, te

  
  ! error management
  integer                        :: err
  character(len=*), parameter    :: name='qrm_do_subtree'

  __QRM_PROF_START(ts)

  err = 0

  ! simplify
  adata => qrm_spfct%adata
  fdata => qrm_spfct%fdata

  eps = qrm_spfct%rcntl(qrm_rd_eps_)  
  
  allocate(work%c(adata%rc(root%num),qrm_spfct%icntl(qrm_mb_)))
  cnt = 0

  storer = qrm_spfct%icntl(qrm_keeph_).ge.0
  storeh = qrm_spfct%icntl(qrm_keeph_).ge.1

  ! pointer to the first leaf of the sequential subtree
  node = adata%small(root%num)

  call qrm_dscr_init(qrm_seq_dscr, seq=.true.)
  
  subtree: do
     f = adata%torder(node)
     front => fdata%front_list(f)

     ! assemble
     if(front%num.ne.root%num) call cqrm_activate_front(qrm_spfct, front, info=err)
     __QRM_INFO_CHECK(err, name,'qrm_activate_front',9999)

     do br=1, front%nr
        do bc=1, front%nc
           if(qrm_allocated(front%f%blocks(br,bc)%c)) then
              call cqrm_init_block(qrm_spfct, front%num, br, bc)
           end if
        end do
     end do
     
     call cqrm_init_front(qrm_spfct, front, err)
     __QRM_INFO_CHECK(err, name,'qrm_init_front',9999)

     do i = adata%childptr(front%num), adata%childptr(front%num+1)-1
        child => fdata%front_list(adata%child(i))

        call cqrm_assemble_front(qrm_seq_dscr, qrm_spfct, front, child)

        do br =  1, child%nr
           do bc = 1, child%nc
              if(qrm_allocated(child%f%blocks(br,bc)%c)) &
                   call cqrm_clean_block(qrm_spfct, child%num, br, bc)
           end do
        end do
        call cqrm_clean_front(qrm_spfct, child, err)
     end do
     
     if(qrm_size(work%c).lt.front%n*front%mb) then
        deallocate(work%c)
        allocate(work%c(front%mb,front%n))
     end if

     if(qrm_spfct%sym.gt.0) then
        call cqrm_dsmat_potrf_async(qrm_seq_dscr, &
             'u', front%f, front%m, front%npiv)              
     else
        call cqrm_dsmat_geqr_async(qrm_seq_dscr,  &
             front%f, front%t,                    &
             front%ib, front%bh,                  &
             work)
     end if

     if(eps.ne.qrm_szero) then
        call cqrm_dsmat_trdcn_async(qrm_seq_dscr, front%f, &
             fdata%rd, eps, &
             front%npiv)
  
     end if
     
     if(front%num.eq.root%num) exit subtree
     node = node+1

  end do subtree

  err = qrm_seq_dscr%info
  
  ! if root is not the only node in the tree, then we have to release
  ! csize(l) memory, where l is the first leaf of the subtree (see the
  ! comments in qrm_compute_memory.F90
  if(adata%torder(adata%small(root%num)) .ne. root%num) then
     call qrm_facto_mem_get(fdata%ma, -adata%csize(adata%torder(adata%small(root%num))))
  end if

9999 continue ! error management
  if(associated(work%c)) then
     deallocate(work%c)
     nullify(work%c)
  end if

  __QRM_PROF_STOP(qrm_prof_do_subtree_,ts,te)

  if(present(info)) info = err
  return

end subroutine cqrm_do_subtree







