!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_factorization_task_mod.F90
!! This file contains the module which holds interfaces of the multifrontal factorization tasks
!!
!!
!! ##############################################################################################


#include "qrm_common.h"

! #define prnt_tsk

module cqrm_factorization_tasks_mod

  use iso_c_binding



  interface qrm_do_subtree_task
     module procedure cqrm_do_subtree_task
  end interface qrm_do_subtree_task

  interface qrm_init_front_task
     module procedure cqrm_init_front_task
  end interface qrm_init_front_task

  interface qrm_init_block_task
     module procedure cqrm_init_block_task
  end interface qrm_init_block_task

  interface qrm_clean_block_task
     module procedure cqrm_clean_block_task
  end interface qrm_clean_block_task

  interface qrm_clean_front_task
     module procedure cqrm_clean_front_task
  end interface qrm_clean_front_task

contains


  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !!  do_subtree
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine cqrm_do_subtree_task(qrm_dscr, qrm_spfct, rootnum, prio)
#if defined (have_starpu)
    use fstarpu_mod
    use cqrm_starpu_mod
#endif
    use cqrm_spfct_mod
    use cqrm_fdata_mod
    use iso_c_binding
    use qrm_dscr_mod
    use cqrm_factorization_mod

    implicit none

    type(qrm_dscr_type), target         :: qrm_dscr
    type(cqrm_spfct_type), target       :: qrm_spfct
    integer, target                     :: rootnum
    integer                             :: prio

    type(cqrm_front_type), pointer      :: father, front
    integer                             :: f
    type(c_ptr), target                 :: qrm_spfct_c, father_sym_handle_c, dscrpt
    real(kind=c_float), target          :: flops_c
    real(r32)                           :: rflops
    integer, target                     :: dosubprio

    ! error management
    integer                             :: info
    character(len=*), parameter         :: name='qrm_do_subtree_task'

    if(qrm_dscr%info.ne.0) return
    info = 0

    front => qrm_spfct%fdata%front_list(rootnum)

#if defined(have_starpu)
  
    if(qrm_dscr%ngpu>0) then
       dosubprio = dosub_hp_prio_
    else
       dosubprio = prio
    end if
    
    qrm_spfct_c = c_loc(qrm_spfct)
    dscrpt      = c_loc(qrm_dscr)

    call fstarpu_insert_task((/ cqrm_do_subtree_cl,                   &
         FSTARPU_VALUE    , c_loc(dscrpt)      , FSTARPU_SZ_C_PTR,    &
         FSTARPU_VALUE    , c_loc(qrm_spfct_c) , FSTARPU_SZ_C_PTR,    &
         FSTARPU_VALUE    , c_loc(rootnum)     , FSTARPU_SZ_C_INT,    &
         FSTARPU_RW       , front%sym_handle,                         &
         FSTARPU_PRIORITY , c_loc(dosubprio),                         &
         FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),                      &
         C_NULL_PTR /))

#else
    call cqrm_do_subtree(qrm_spfct, front, rflops, info)
    if(info.ne.0) qrm_dscr%info = info
#endif

    return

  end subroutine cqrm_do_subtree_task


#if defined (have_starpu)
  recursive subroutine cqrm_do_subtree_cpu_func(buffers, cl_arg) bind(C)
    use qrm_common_mod
    use cqrm_spfct_mod
    use fstarpu_mod
    use iso_c_binding
    use qrm_dscr_mod
    use cqrm_factorization_mod
    implicit none

    type(c_ptr), value             :: cl_arg
    type(c_ptr), value             :: buffers

    type(c_ptr), target            :: qrm_spfct_c, qrm_dscr_c
    integer(c_int), target         :: rootnum
    real(r32), target              :: flops

    type(qrm_dscr_type), pointer   :: qrm_dscr
    type(cqrm_spfct_type), pointer :: qrm_spfct
    type(cqrm_front_type), pointer :: root
    integer                        :: info

    call fstarpu_unpack_arg(cl_arg,(/ c_loc(qrm_dscr_c), c_loc(qrm_spfct_c), c_loc(rootnum) /))

    call c_f_pointer(qrm_dscr_c, qrm_dscr)
    ! until StarPU has a proper error handling method we have to check
    ! manually
    if(qrm_dscr%info.ne.0) return

    call c_f_pointer(qrm_spfct_c, qrm_spfct)

    root => qrm_spfct%fdata%front_list(rootnum)

    call cqrm_do_subtree(qrm_spfct, root, flops, info)
    if(info.ne.0) qrm_dscr%info=info

#if defined(prnt_tsk)
    write(*,'("Subt  --  inode: ",i4)')rootnum
#endif    

    return

  end subroutine cqrm_do_subtree_cpu_func
#endif


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  init
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine cqrm_init_front_task(qrm_dscr, qrm_spfct, fnum, prio)
    ! the Intel compler bugs here when used with vectorization options
    !DIR$ NOOPTIMIZE

    use iso_c_binding
    use cqrm_fdata_mod
    use cqrm_spfct_mod
    use qrm_dscr_mod
    use cqrm_factorization_mod
#if defined(have_starpu)
    use fstarpu_mod
    use cqrm_starpu_mod
#endif
    implicit none

    type(qrm_dscr_type), target    :: qrm_dscr
    type(cqrm_spfct_type), target  :: qrm_spfct
    integer, target                :: fnum
    integer                        :: prio

    type(cqrm_front_type), pointer :: front
    type(c_ptr), target            :: qrm_spfct_c, dscrpt
#if defined (have_starpu)
    type(c_ptr)                    :: hdls
    integer, target                :: nc
    integer                        :: c, i
    integer, target                :: initfprio
#endif  

    ! error management
    integer                        :: info
    character(len=*), parameter    :: name='qrm_init_front_task'

    if(qrm_dscr%info.ne.0) return
    info = 0

    front => qrm_spfct%fdata%front_list(fnum)

#if defined(have_starpu)

    if(qrm_dscr%ngpu>0) then
       initfprio = initf_hp_prio_
    else
       initfprio = prio
    end if

    qrm_spfct_c = c_loc(qrm_spfct)
    dscrpt      = c_loc(qrm_dscr)

    nc = qrm_spfct%adata%childptr(fnum+1)-qrm_spfct%adata%childptr(fnum)
    hdls = fstarpu_data_descr_array_alloc(nc+1)

    call fstarpu_data_descr_array_set(hdls, 0, &
         front%sym_handle, FSTARPU_RW)

    do i=1, nc
       c = qrm_spfct%adata%child(i-1 + qrm_spfct%adata%childptr(fnum))
       call fstarpu_data_descr_array_set(hdls, i, &
            qrm_spfct%fdata%front_list(c)%sym_handle, FSTARPU_R)
    end do

    nc = nc+1

    call fstarpu_insert_task((/ cqrm_init_front_cl,                   &
         FSTARPU_VALUE    , c_loc(dscrpt)      , FSTARPU_SZ_C_PTR,    &
         FSTARPU_VALUE    , c_loc(qrm_spfct_c) , FSTARPU_SZ_C_PTR,    &
         FSTARPU_VALUE    , c_loc(fnum)        , FSTARPU_SZ_C_INT,    &
         FSTARPU_DATA_MODE_ARRAY, hdls, c_loc(nc),                    &
         FSTARPU_PRIORITY , c_loc(initfprio),                         &
         FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),                      &
         C_NULL_PTR /))

    call fstarpu_data_descr_array_free(hdls)

#else
    call cqrm_init_front(qrm_spfct, front, info)
    if(info.ne.0) qrm_dscr%info = info
#endif

    return

  end subroutine cqrm_init_front_task


#if defined (have_starpu)
  recursive subroutine cqrm_init_front_cpu_func(buffers, cl_arg) bind(C)
    use qrm_common_mod
    use cqrm_spfct_mod
    use fstarpu_mod
    use iso_c_binding
    use qrm_dscr_mod
    use cqrm_factorization_mod
    implicit none

    type(c_ptr), value             :: cl_arg
    type(c_ptr), value             :: buffers

    type(c_ptr), target            :: qrm_spfct_c, qrm_dscr_c
    integer(c_int), target         :: fnum
    type(cqrm_front_type), pointer :: front

    type(qrm_dscr_type), pointer   :: qrm_dscr
    type(cqrm_spfct_type), pointer :: qrm_spfct
    integer                        :: info

    call fstarpu_unpack_arg(cl_arg,(/ c_loc(qrm_dscr_c), c_loc(qrm_spfct_c), c_loc(fnum) /))

    ! until StarPU has a proper error handling method we have to check
    ! manually
    call c_f_pointer(qrm_dscr_c, qrm_dscr)
    if(qrm_dscr%info.ne.0) return

    call c_f_pointer(qrm_spfct_c, qrm_spfct)

    front => qrm_spfct%fdata%front_list(fnum)    

    call cqrm_init_front(qrm_spfct, front, info)
    if(info.ne.0) qrm_dscr%info=info

#if defined(prnt_tsk)
    write(*,'("Init  --  inode: ",i4)')fnum
#endif    

    return
  end subroutine cqrm_init_front_cpu_func
#endif


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !!  clean
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  subroutine cqrm_clean_front_task(qrm_dscr, qrm_spfct, fnum, prio)
#if defined (have_starpu)
    use fstarpu_mod
    use cqrm_starpu_mod
#endif
    use cqrm_spfct_mod
    use cqrm_fdata_mod
    use iso_c_binding
    use qrm_dscr_mod
    use cqrm_factorization_mod

    implicit none

    type(qrm_dscr_type), target         :: qrm_dscr
    type(cqrm_spfct_type), target       :: qrm_spfct
    integer, target                     :: fnum
    integer                             :: prio

    type(cqrm_front_type), pointer      :: front
    type(c_ptr), target                 :: qrm_spfct_c, dscrpt
    real(r32)                           :: rflops
    integer, target                     :: cleanfprio

    ! error management
    integer                             :: info
    character(len=*), parameter         :: name='qrm_clean_front_task'

    if(qrm_dscr%info.ne.0) return

    front => qrm_spfct%fdata%front_list(fnum)

#if defined(have_starpu)

    if(qrm_dscr%ngpu>0) then
       cleanfprio = cleanf_hp_prio_
    else
       cleanfprio = prio
    end if

    qrm_spfct_c = c_loc(qrm_spfct)
    dscrpt      = c_loc(qrm_dscr)

    call fstarpu_insert_task((/ cqrm_clean_front_cl,                  &
         FSTARPU_VALUE    , c_loc(dscrpt)      , FSTARPU_SZ_C_PTR,    &
         FSTARPU_VALUE    , c_loc(qrm_spfct_c) , FSTARPU_SZ_C_PTR,    &
         FSTARPU_VALUE    , c_loc(fnum)        , FSTARPU_SZ_C_INT,    &
         FSTARPU_RW       , front%sym_handle,                         &
         FSTARPU_PRIORITY , c_loc(cleanfprio),                        &
         FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),                      &
         C_NULL_PTR /))

#else
    info = 0
    call cqrm_clean_front(qrm_spfct, front, info)
    if (info.ne.0) qrm_dscr%info = info
#endif

    return

  end subroutine cqrm_clean_front_task


#if defined (have_starpu)
  recursive subroutine cqrm_clean_front_cpu_func(buffers, cl_arg) bind(C)
    use qrm_common_mod
    use cqrm_spfct_mod
    use fstarpu_mod
    use iso_c_binding
    use qrm_dscr_mod
    use cqrm_factorization_mod
    implicit none

    type(c_ptr), value             :: cl_arg
    type(c_ptr), value             :: buffers

    type(c_ptr), target            :: qrm_spfct_c, qrm_dscr_c
    integer(c_int), target         :: fnum
    real(r32), target              :: flops
    type(cqrm_front_type), pointer :: front

    type(qrm_dscr_type), pointer   :: qrm_dscr
    type(cqrm_spfct_type), pointer :: qrm_spfct
    type(cqrm_front_type), pointer :: root
    integer                        :: info

    call fstarpu_unpack_arg(cl_arg,(/ c_loc(qrm_dscr_c), c_loc(qrm_spfct_c), c_loc(fnum) /))

    ! until StarPU has a proper error handling method we have to check
    ! manually
    call c_f_pointer(qrm_dscr_c, qrm_dscr)
    if(qrm_dscr%info.ne.0) return

    call c_f_pointer(qrm_spfct_c, qrm_spfct)

    front => qrm_spfct%fdata%front_list(fnum)

    call cqrm_clean_front(qrm_spfct, front, info)
    if(info.ne.0) qrm_dscr%info=info ! TODO: set it atomically

#if defined(prnt_tsk)
    write(*,'("Clean  --  inode: ",i4)')fnum
#endif    

    return

  end subroutine cqrm_clean_front_cpu_func
#endif


!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  Init block
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


  subroutine cqrm_init_block_task(qrm_dscr, qrm_spfct, fnum, br, bc, prio)
    use iso_c_binding
    use cqrm_spfct_mod
    use cqrm_fdata_mod
    use qrm_common_mod
    use qrm_dscr_mod
    use cqrm_factorization_mod
#if defined(have_starpu)
    use fstarpu_mod
    use cqrm_starpu_mod
#endif
    implicit none

    type(qrm_dscr_type), target    :: qrm_dscr
    type(cqrm_spfct_type), target  :: qrm_spfct
    integer, target                :: br, bc, fnum
    integer                        :: prio

#if defined (have_starpu)
    type(cqrm_front_type), pointer :: front
    type(c_ptr), target            :: qrm_spfct_c, dscrpt
    integer, target                :: initbprio
#endif
    
    if(qrm_dscr%info.ne.0) return

#if defined (have_starpu)
    front => qrm_spfct%fdata%front_list(fnum)
    
    qrm_spfct_c = c_loc(qrm_spfct)
    dscrpt      = c_loc(qrm_dscr)

    if(qrm_dscr%ngpu>0) then
       initbprio = initb_hp_prio_
    else
       initbprio = prio
    end if

    call fstarpu_insert_task((/ cqrm_init_block_cl,                   &
         FSTARPU_VALUE    , c_loc(dscrpt)      , FSTARPU_SZ_C_PTR,    &
         FSTARPU_VALUE    , c_loc(qrm_spfct_c) , FSTARPU_SZ_C_PTR,    &
         FSTARPU_VALUE    , c_loc(br)          , FSTARPU_SZ_C_INT,    &
         FSTARPU_VALUE    , c_loc(bc)          , FSTARPU_SZ_C_INT,    &
         FSTARPU_VALUE    , c_loc(fnum)        , FSTARPU_SZ_C_INT,    &
         FSTARPU_R        , front%sym_handle,                         &
         FSTARPU_RW       , front%f%blocks(br,bc)%hdl,                &
         FSTARPU_PRIORITY , c_loc(initbprio),                         &
         FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),                      &
         C_NULL_PTR /))

#else
    call cqrm_init_block(qrm_spfct, fnum, br, bc)
#endif

    return 

  end subroutine cqrm_init_block_task


#if defined (have_starpu)
  recursive subroutine cqrm_init_block_cpu_func(buffers, cl_arg) bind(C)
    use qrm_common_mod
    use cqrm_spfct_mod
    use fstarpu_mod
    use iso_c_binding
    use qrm_dscr_mod
    use cqrm_factorization_mod
    implicit none

    type(c_ptr), value             :: cl_arg
    type(c_ptr), value             :: buffers

    type(c_ptr), target            :: qrm_spfct_c, qrm_dscr_c
    integer(c_int), target         :: fnum, br, bc

    type(qrm_dscr_type), pointer   :: qrm_dscr
    type(cqrm_spfct_type), pointer :: qrm_spfct
    integer                        :: info

    call fstarpu_unpack_arg(cl_arg,(/ c_loc(qrm_dscr_c), c_loc(qrm_spfct_c), c_loc(br), c_loc(bc), c_loc(fnum) /))

    ! until StarPU has a proper error handling method we have to check
    ! manually
    call c_f_pointer(qrm_dscr_c, qrm_dscr)
    if(qrm_dscr%info.ne.0) return

    call c_f_pointer(qrm_spfct_c, qrm_spfct)

    call cqrm_init_block(qrm_spfct, fnum, br, bc)
    
#if defined(prnt_tsk)
    write(*,'("Init block  --  inode: ",i4,"  (",i4,",",i4,")")')fnum,br,bc
#endif    

    return

  end subroutine cqrm_init_block_cpu_func
#endif



!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!!  Clean block
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


  subroutine cqrm_clean_block_task(qrm_dscr, qrm_spfct, fnum, br, bc, prio)
    use iso_c_binding
    use cqrm_spfct_mod
    use cqrm_fdata_mod
    use qrm_common_mod
    use qrm_dscr_mod
    use cqrm_factorization_mod
#if defined(have_starpu)
    use fstarpu_mod
    use cqrm_starpu_mod
#endif
    implicit none

    type(qrm_dscr_type), target    :: qrm_dscr
    type(cqrm_spfct_type), target  :: qrm_spfct
    integer, target                :: br, bc, fnum
    integer                        :: prio

    integer                        :: info
    type(cqrm_front_type), pointer :: front
#if defined (have_starpu)
    type(c_ptr), target            :: qrm_spfct_c, dscrpt
    integer, target                :: cleanbprio
    integer(c_int), target         :: nh
    type(c_ptr)                    :: hdls
#endif

    if(qrm_dscr%info.ne.0) return
    info = 0
    
    front => qrm_spfct%fdata%front_list(fnum)
    
#if defined (have_starpu)
    qrm_spfct_c = c_loc(qrm_spfct)
    dscrpt      = c_loc(qrm_dscr)

    if(qrm_dscr%ngpu>0) then
       cleanbprio = cleanb_hp_prio_
    else
       cleanbprio = prio
    end if

    hdls = fstarpu_data_descr_array_alloc(4)

    nh = 0
    
    if(front%small.eq.0) then
       ! add front handle
       call fstarpu_data_descr_array_set(hdls, nh, &
            front%sym_handle, FSTARPU_R)
       nh = nh+1 

       ! add block handle
       call fstarpu_data_descr_array_set(hdls, nh, &
            front%f%blocks(br,bc)%hdl, FSTARPU_RW)
       nh = nh+1

       if(qrm_spfct%sym.eq.0) then
          if(br.ge.bc) then
             ! add t handle
             call fstarpu_data_descr_array_set(hdls, nh, &
                  front%t%blocks(br,bc)%hdl, FSTARPU_RW)
             nh = nh+1
          end if
          
          if(size(front%t%blocks,2).ge.front%np+bc) then
             ! if((br.ne.bc) .and. (mod(br-bc,front%bh).eq.0)) then
             if(qrm_allocated(front%t%blocks(br,front%np+bc)%c)) then
                ! add other t handle
                call fstarpu_data_descr_array_set(hdls, nh, &
                     front%t%blocks(br,front%np+bc)%hdl, FSTARPU_RW)
                nh = nh+1
             end if
          end if
       end if
    else
       ! add front handle in RW mode
       call fstarpu_data_descr_array_set(hdls, nh, &
            front%sym_handle, FSTARPU_RW)
       nh = nh+1 

    end if
    
    call fstarpu_insert_task((/ cqrm_clean_block_cl,                  &
         FSTARPU_VALUE    , c_loc(dscrpt)      , FSTARPU_SZ_C_PTR,    &
         FSTARPU_VALUE    , c_loc(qrm_spfct_c) , FSTARPU_SZ_C_PTR,    &
         FSTARPU_VALUE    , c_loc(br)          , FSTARPU_SZ_C_INT,    &
         FSTARPU_VALUE    , c_loc(bc)          , FSTARPU_SZ_C_INT,    &
         FSTARPU_VALUE    , c_loc(fnum)        , FSTARPU_SZ_C_INT,    &
         FSTARPU_DATA_MODE_ARRAY, hdls, c_loc(nh),                    &
         FSTARPU_PRIORITY , c_loc(cleanbprio),                        &
         FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),                      &
         C_NULL_PTR /))

    call fstarpu_data_descr_array_free(hdls)

#else
    call cqrm_clean_block(qrm_spfct, fnum, br, bc, info)
    if (info.ne.0) qrm_dscr%info = info
#endif

    return 

  end subroutine cqrm_clean_block_task


#if defined (have_starpu)
  recursive subroutine cqrm_clean_block_cpu_func(buffers, cl_arg) bind(C)
    use qrm_common_mod
    use cqrm_spfct_mod
    use fstarpu_mod
    use iso_c_binding
    use qrm_dscr_mod
    use cqrm_factorization_mod
    implicit none

    type(c_ptr), value             :: cl_arg
    type(c_ptr), value             :: buffers

    type(c_ptr), target            :: qrm_spfct_c, qrm_dscr_c
    integer(c_int), target         :: fnum, br, bc

    type(qrm_dscr_type), pointer   :: qrm_dscr
    type(cqrm_spfct_type), pointer :: qrm_spfct
    integer                        :: info

    call fstarpu_unpack_arg(cl_arg,(/  c_loc(qrm_dscr_c), c_loc(qrm_spfct_c), c_loc(br), c_loc(bc), c_loc(fnum) /))

    ! until StarPU has a proper error handling method we have to check
    ! manually
    call c_f_pointer(qrm_dscr_c, qrm_dscr)
    if(qrm_dscr%info.ne.0) return

    call c_f_pointer(qrm_spfct_c, qrm_spfct)

    call cqrm_clean_block(qrm_spfct, fnum, br, bc, info)
    if (info.ne.0) qrm_dscr%info = info ! TODO: set atomically
    
#if defined(prnt_tsk)
    write(*,'("Clean block  --  inode: ",i4,"  (",i4,",",i4,")")')fnum,br,bc
#endif    

    return

  end subroutine cqrm_clean_block_cpu_func
#endif


  
end module cqrm_factorization_tasks_mod
