!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_clean_block.F90
!! This file contains a routine which performs the cleaning of a block in a front
!!
!!
!! ##############################################################################################


#include "qrm_common.h"


subroutine sqrm_clean_block(qrm_spfct, fnum, br, bc, info)
  use qrm_mem_mod
  use sqrm_spfct_mod
  use sqrm_fdata_mod
  use sqrm_utils_mod
  use qrm_common_mod
  use qrm_sort_mod
  use sqrm_factorization_mod, protect => sqrm_clean_block
  use iso_c_binding
#if defined (perfprof)
  use qrm_perfprof_mod
#endif

  implicit none

  type(sqrm_spfct_type), target  :: qrm_spfct
  integer                        :: fnum, br, bc
  integer, optional              :: info

  integer                        :: i, j, fc, fr, lc, lr
  integer                        :: mr, nr, mh, nh, pinth
  integer(kind=8)                :: nzr, nzh
  logical                        :: storeh, storer, inh, inr, inc
  integer(kind=8)                :: ts, te
  logical                        :: dopin

  real(r32), allocatable      :: tmp(:,:)

  type(sqrm_front_type), pointer :: front

  ! error management
  integer                        :: err
  character(len=*), parameter    :: name='qrm_clean_block'

  __QRM_PROF_START(ts)

  err = 0

  front => qrm_spfct%fdata%front_list(fnum)

  ! TODO: error checking in allocs

  if( (front%n .le. 0) .or. (front%m .le. 0)) then
     goto 9999
  end if

  storer = qrm_spfct%icntl(qrm_keeph_).ge.0
  if(qrm_spfct%sym.gt.0) then
     storeh = .false.
  else
     storeh = qrm_spfct%icntl(qrm_keeph_).ge.1
  end if


  fc = (bc-1)*front%mb+1
  lc = min(bc*front%mb, front%n)
  fr = (br-1)*front%mb+1
  lr = min(fr+size(front%f%blocks(br,bc)%c,1)-1, front%m)

  inr = storer .and. ((bc.ge.br) .and. (fr.le.front%npiv))
  inh = storeh .and. (bc.le.br)
  inc = (lc.gt.fr) .and. (lr.gt.front%npiv)

  call qrm_get(qrm_spfct, 'qrm_pinth', pinth)
  dopin = (min(front%m,front%n).ge.pinth).and.(front%small.eq.0)

#if defined (have_cuda)
  if(dopin) then
     call qrm_unpin_mem(c_loc(front%f%blocks(br,bc)%c))

     if(qrm_spfct%sym.eq.0) then
        if((size(front%t%blocks,2).ge.bc) .and. &
             qrm_allocated(front%t%blocks(br,bc)%c)) then
           call qrm_unpin_mem(c_loc(front%t%blocks(br,bc)%c))
        end if
        if((size(front%t%blocks,2).ge.front%np+bc) .and. &
             qrm_allocated(front%t%blocks(br,front%np+bc)%c)) then
           call qrm_unpin_mem(c_loc(front%t%blocks(br,front%np+bc)%c))
        end if
     end if
  end if
#endif

  if(inr) then

     mr  = min(size(front%f%blocks(br,bc)%c,1), front%npiv-fr+1)
     nr  = size(front%f%blocks(br,bc)%c,2)
     nzr = mr*nr


     if (fc.eq.fr) nzr=nzr- ((mr*(mr-1))/2)
     call qrm_atomic_add(front%rsize, +nzr)

     if((.not.inh) .and. (inc)) then
        ! reallocate
        call qrm_alloc(tmp, mr, nr)
        tmp(1:mr,1:nr) = front%f%blocks(br,bc)%c(1:mr,1:nr)
        call qrm_dealloc(front%f%blocks(br,bc)%c)
        call qrm_move_alloc(tmp, front%f%blocks(br,bc)%c)
     end if

  end if

  if(inh) then
     mh  = size(front%f%blocks(br,bc)%c,1)
     nh  = size(front%f%blocks(br,bc)%c,2)
     if(inc) then
        nzh = mh*(mh+1)/2
     else
        nzh = mh*nh
     end if
     call qrm_atomic_add(front%hsize, +nzh)

#if defined (have_cuda)
     if((mod(br-bc,front%bh).eq.0) .and. &
         (front%small.eq.0)        .and. &
         (.not.allocated(front%f%blocks(br,bc)%stair))) then
        ! zeros were added below the diagonal of this block to allow
        ! for GEMMS on GPU. Copy back the coefficients
        do j=1, min(nh,mh)
           front%f%blocks(br,bc)%c(j+1:mh,j) = front%t%blocks(br,bc)%c(front%ib+j+1:front%ib+mh,j)
        end do
     end if
#endif
  end if

  if((.not.inh) .and. (.not.inr)) then
     ! deallocate
     call qrm_dealloc(front%f%blocks(br,bc)%c)
     if(qrm_spfct%sym.eq.0) then
        if(br.ge.bc) call qrm_dealloc(front%t%blocks(br,bc)%c)
        if(size(front%t%blocks,2).ge.front%np+bc) call qrm_dealloc(front%t%blocks(br,front%np+bc)%c)
     end if
  end if

#if defined(have_starpu)
  if(front%small.eq.0) then
     call sqrm_block_unregister_async(front%f%blocks(br,bc))
     if(qrm_spfct%sym.eq.0) then
        if(br.ge.bc) then
           call sqrm_block_unregister_async(front%t%blocks(br,bc))
           if(size(front%t%blocks,2).ge.front%np+bc) then
              call sqrm_block_unregister_async(front%t%blocks(br,front%np+bc))
           end if
        end if
     end if
  end if
#endif

9999 continue
  __QRM_PROF_STOP(qrm_prof_clean_block_task_,ts,te)

  if(present(info)) info = err
  return

end subroutine sqrm_clean_block
