!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_factorization_core.F90
!! This file holds the core of the QR numerical factorization.
!!
!!
!! ##############################################################################################


#include "qrm_common.h"


!> @brief This is the main factorization routine. It performs the factorization of all
!! the fronts that have been assigned to the node (MPI task). The whole process
!! is OpenMP parallel.

!> This factorization is completely dynamic and results in an out of order execution
!! of computational and symbolic tasks. Each front is cut into block-columns. Computational
!! tasks are scheduled on a per-front or per-block-column basis according to a graph of
!! dependencies; each task is pushed on the queue of the thread who owns the related front.
!! The ownership of a front is fixed at the moment where the front is activated. If a thread
!! runs out of tasks to perform (i.e., its queue is empty) it can steal tasks from other
!! threads (see the @link qrm_task_mod @endlink module for the details).
!!
!! Six different types of tasks can be executed by a thread:
!! - panel    : the panel factorization of a column in a front
!! - update   : the update of a block-column wrt to a panel in a front
!! - activate : activation of a front. this operation consists in computing the full
!!              structure of a front and allocating all the memory areas needed for the
!!              subsequent factorization
!! - assemble : this operation consists in assemblying one block-column in the contribution
!!              block of a front into the father node
!! - clean    : cleans up a front, i.e., it deallocates all the memory areas that are no more
!!              needed
!!
!! @param[in,out] qrm_spfct This is the main data structure associated
!!                to a problem. The @link sqrm_factorization_core @endlink
!!                assumes that the analysis has been done and thus that the
!!                qrm_spfct%adata field contains meaningful stuff
!!

subroutine sqrm_factorization_core(qrm_dscr, qrm_spfct)

  use iso_c_binding
  use sqrm_spfct_mod
  use qrm_mem_mod
  use qrm_error_mod
  use qrm_common_mod
  use qrm_dscr_mod
  use sqrm_factorization_mod, protect => sqrm_factorization_core
  use sqrm_factorization_tasks_mod
  use sqrm_dsmat_mod 
  implicit none

  type(qrm_dscr_type)            :: qrm_dscr
  type(sqrm_spfct_type), target  :: qrm_spfct

  integer                        :: i, j, p, c, node, fnum, ret, d
  integer, allocatable           :: work(:)
  type(qrm_adata_type), pointer  :: adata
  type(sqrm_fdata_type), pointer :: fdata
  type(sqrm_front_type), pointer :: front, child
  integer                        :: bc, br, prio

  real(r32)                      :: eps
  
  ! error management
  integer                        :: info
  character(len=*), parameter    :: name='qrm_factorization_core'

  if (qrm_dscr%info.ne.0) return
  
  allocate(work(qrm_spfct%n))

  
  ! simplify
  adata => qrm_spfct%adata
  fdata => qrm_spfct%fdata

  eps = qrm_spfct%rcntl(qrm_rd_eps_)  
  
  fdata%rd = 0
  front => null()
  info = 0

  ! loop over all nodes
  main: do node=1, adata%nnodes
     ! until a proper error analysis is implemented in starpu we have
     ! to check whether an error was raised by some task in order to
     ! stop further submissions
     if(qrm_dscr%info.ne.0) goto 9999

     fnum = adata%torder(node)

     if(adata%small(fnum) .lt. 0) cycle
     prio = (adata%nnodes - node + 1)*4
     front => fdata%front_list(fnum)

     if(adata%small(fnum).gt.0) then
        call qrm_facto_mem_get(fdata%ma, adata%asize(fnum))
        call sqrm_activate_front(qrm_spfct, front, work, info)
        if(info.ne.0) qrm_dscr%info = info ! TODO: set atomically
        call sqrm_do_subtree_task(qrm_dscr, qrm_spfct, fnum, prio=huge(1))
! #if defined(have_starpu)
        ! call fstarpu_task_wait_for_all()
! #endif
        __QRM_INFO_CHECK(qrm_dscr%info, name,'qrm_do_subtree_task',9999)
     else
        call qrm_facto_mem_get(fdata%ma, adata%asize(front%num))
        call sqrm_activate_front(qrm_spfct, front, work, info)
        if(info.ne.0) qrm_dscr%info = info ! TODO: set atomically
        __QRM_INFO_CHECK(qrm_dscr%info, name,'qrm_activate_front',9999)

! #if defined(have_starpu)
        ! call fstarpu_task_wait_for_all()
! #endif
        call sqrm_init_front_task(qrm_dscr, qrm_spfct, front%num, prio=huge(1))
        __QRM_INFO_CHECK(qrm_dscr%info, name,'qrm_init_task',9999)
        
        do br=1, front%nr
           do bc=1, front%nc
              if(qrm_allocated(front%f%blocks(br,bc)%c)) then
                 call sqrm_init_block_task(qrm_dscr, qrm_spfct, front%num, br, bc, prio=-1000)
                 __QRM_INFO_CHECK(qrm_dscr%info, name,'qrm_init_task',9999)
! #if defined(have_starpu)
                 ! call fstarpu_task_wait_for_all()
! #endif
              end if
           end do
        end do
! #if defined(have_starpu)
        ! call fstarpu_task_wait_for_all()
! #endif

        do i = adata%childptr(front%num), adata%childptr(front%num+1)-1
           child => fdata%front_list(adata%child(i))
           ! assemble child into front
           if(child%npiv.lt.child%m) then
              call sqrm_assemble_front(qrm_dscr, qrm_spfct, front, child)
           end if

           do br =  1, child%nr
              do bc = 1, child%nc
                 if(qrm_allocated(child%f%blocks(br,bc)%c)) &
                 ! call fstarpu_task_wait_for_all()
                 call sqrm_clean_block_task(qrm_dscr, qrm_spfct, child%num, br, bc, prio=huge(1))
! #if defined(have_starpu)
                 ! call fstarpu_task_wait_for_all()
! #endif
              end do
           end do
           
           ! once the child is assembled, it can be cleaned up
           call sqrm_clean_front_task(qrm_dscr, qrm_spfct, child%num, prio=huge(1))
           __QRM_INFO_CHECK(qrm_dscr%info, name,'qrm_clean_task',9999)
           ! call fstarpu_task_wait_for_all()
        end do

        ! call fstarpu_task_wait_for_all()

        if(qrm_spfct%sym.gt.0) then
           call sqrm_dsmat_potrf_async(qrm_dscr, &
                'u', front%f, front%m, front%npiv, &
                prio=prio)
        else
           call sqrm_dsmat_geqr_async(qrm_dscr, &
                front%f, front%t,               &
                front%ib, front%bh,             &
                fdata%work, prio=prio)
        end if

        ! call fstarpu_task_wait_for_all()
        
        if(eps.ne.qrm_szero) then
           call sqrm_dsmat_trdcn_async(      &
                qrm_dscr, front%f, fdata%rd, &
                eps,                         &
                front%npiv)
        end if
        
! #if defined(have_starpu)
        ! call fstarpu_task_wait_for_all()
! #endif
     end if

  end do main

  fnum = adata%torder(adata%nnodes)
  call sqrm_clean_front_task(qrm_dscr, qrm_spfct, fnum, prio=huge(1))

9999 continue
  ! cleanup and return
#if defined (have_starpu)
  call fstarpu_data_unregister_submit(fdata%work%hdl)
  fdata%work%hdl = c_null_ptr
#else
  deallocate(fdata%work%c)
#endif

  deallocate(work)
  
  return

end subroutine sqrm_factorization_core

