!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_factorization_init.F90
!! This file contains a subroutine that initializes the factorization
!!
!!
!! ##############################################################################################


#include "qrm_common.h" 
!> @brief This subroutine initializes the data structures needed for
!! the actual factorization.

!> The main task achieved by this routine is the creation of what we
!! call (in mumps terminology) the arrowheads. Basically it builds a
!! list of @link zqrm_fdata_mod::zqrm_front_type @endlink elements
!! (each one corresponding to one front) and associates to each of
!! them the related coefficients of the original matrix in CSR
!! format. This coefficients will be assembled into the front matrix
!! at the moment of its activation (this is done by the @link
!! zqrm_init_front @endlink routine).
!!
!! param[in]     qrm_mat the input sparse matrix
!!
!! param[in,out] qrm_spfct the sparse factorization object
!!


subroutine zqrm_factorization_init(qrm_dscr, qrm_mat, qrm_spfct, transp)

  use qrm_mem_mod
  use zqrm_spmat_mod
  use zqrm_spfct_mod
  use zqrm_fdata_mod
  use zqrm_utils_mod
  use qrm_string_mod
  use qrm_dscr_mod
  use qrm_common_mod
  use qrm_error_mod
  use qrm_parameters_mod
  use zqrm_factorization_mod, protect => zqrm_factorization_init
#if defined (have_starpu)
  use qrm_starpu_common_mod
  use fstarpu_mod
#endif
#if defined (perfprof)
  use qrm_perfprof_mod
#endif

  implicit none
  
  type(qrm_dscr_type), target      :: qrm_dscr
  type(zqrm_spmat_type), target    :: qrm_mat
  type(zqrm_spfct_type), target    :: qrm_spfct
  character, intent(in)            :: transp
  
  integer                          :: nworker, node
  type(zqrm_front_type), pointer   :: front
#if defined (have_starpu)
  type(c_ptr), target              :: qrm_mat_c, qrm_spfct_c, dscrpt
  character(kind=c_char), target   :: transp_c
#endif
  ! error management
  integer                          :: err
  character(len=*), parameter      :: name='qrm_factorization_init'

  err = 0
  
  if(allocated(qrm_spfct%fdata)) then
     call zqrm_fdata_cleanup(qrm_spfct%fdata, err)
     __QRM_INFO_CHECK(err, name,'qrm_fdata_cleanup',9999)
  else
     call qrm_fdata_init(qrm_spfct%fdata, err)
     __QRM_INFO_CHECK(err, name,'qrm_fdata_init',9999)
  end if  

  allocate(qrm_spfct%fdata%front_list(qrm_spfct%adata%nnodes))

#if defined(have_starpu)

  transp_c = qrm_str_tolower(transp)

  do node=1, qrm_spfct%adata%nnodes
     front => qrm_spfct%fdata%front_list(node)
     if(qrm_spfct%adata%small(node) .ge. 0) then
        call fstarpu_void_data_register(front%sym_handle)
     else
        front%sym_handle = c_null_ptr
     end if
  end do
  nworker = fstarpu_worker_get_count()
  
  call zqrm_ws_init(qrm_spfct%fdata%work, qrm_spfct%icntl(qrm_mb_),   &
       2*qrm_spfct%icntl(qrm_mb_))
  
  qrm_mat_c   = c_loc(qrm_mat)
  qrm_spfct_c = c_loc(qrm_spfct)
  dscrpt      = c_loc(qrm_dscr)

  call fstarpu_insert_task((/ zqrm_factinit_cl,                      &
       FSTARPU_VALUE    , c_loc(dscrpt)      , FSTARPU_SZ_C_PTR,     &
       FSTARPU_VALUE    , c_loc(qrm_mat_c)   , FSTARPU_SZ_C_PTR,     &
       FSTARPU_VALUE    , c_loc(qrm_spfct_c) , FSTARPU_SZ_C_PTR,     &
       FSTARPU_VALUE    , c_loc(transp_c)    , FSTARPU_SZ_CHARACTER, &
       FSTARPU_RW       , qrm_spfct%adata%hdl,                       &
       FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),                       &
       C_NULL_PTR /))

#else
  nworker = 1
  allocate(qrm_spfct%fdata%work%c(qrm_spfct%icntl(qrm_mb_),2*qrm_spfct%icntl(qrm_mb_)))

  call zqrm_factorization_init_core(qrm_mat, qrm_spfct, transp, err)
  __QRM_INFO_CHECK(err, name,'qrm_factorization_init_core',9999)

#endif

9999 continue
  call qrm_error_set(qrm_dscr%info, err)
  return

end subroutine zqrm_factorization_init



#if defined (have_starpu)
recursive subroutine zqrm_factorization_init_cpu_func(buffers, cl_arg) bind(C)
    use qrm_common_mod
    use zqrm_spmat_mod
    use zqrm_spfct_mod
    use fstarpu_mod
    use iso_c_binding
    use qrm_dscr_mod
    use zqrm_analysis_mod
    use zqrm_factorization_mod
    implicit none

    type(c_ptr), value             :: cl_arg
    type(c_ptr), value             :: buffers

    type(c_ptr), target            :: qrm_mat_c, qrm_spfct_c, qrm_dscr_c

    type(qrm_dscr_type), pointer   :: qrm_dscr
    type(zqrm_spmat_type), pointer :: qrm_mat
    type(zqrm_spfct_type), pointer :: qrm_spfct
    character(kind=c_char), target :: transp
    integer                        :: info

    call fstarpu_unpack_arg(cl_arg,(/ &
         c_loc(qrm_dscr_c),           &
         c_loc(qrm_mat_c),            &
         c_loc(qrm_spfct_c),          &
         c_loc(transp)                &
         /))

    ! until StarPU has a proper error handling method we have to check
    ! manually
    call c_f_pointer(qrm_dscr_c, qrm_dscr)
    if(qrm_dscr%info.ne.0) return

    call c_f_pointer(qrm_mat_c, qrm_mat)
    call c_f_pointer(qrm_spfct_c, qrm_spfct)

    call zqrm_factorization_init_core(qrm_mat, qrm_spfct, transp, info)
    call qrm_error_set(qrm_dscr%info, info)

    return

  end subroutine zqrm_factorization_init_cpu_func
#endif


subroutine zqrm_factorization_init_core(qrm_mat, qrm_spfct, transp, info)
 
  use qrm_mem_mod
  use zqrm_spmat_mod
  use zqrm_spfct_mod
  use zqrm_fdata_mod
  use zqrm_utils_mod
  use qrm_common_mod
  use qrm_error_mod
  use zqrm_factorization_mod, protect => zqrm_factorization_init_core
#if defined (have_starpu)
  use fstarpu_mod
#endif
#if defined (perfprof)
  use qrm_perfprof_mod
#endif

  implicit none
  
  type(zqrm_spmat_type)          :: qrm_mat
  type(zqrm_spfct_type), target  :: qrm_spfct
  character, intent(in)          :: transp
  integer, optional              :: info
  
  integer                        :: f, nrowsf, i, j, roff, r, c, mnv, mnj, t
  integer                        :: lrow, itmp, k, m, n, cnt, nr, node, nworker
  integer, allocatable           :: icperm(:), irperm(:), gcolmap(:), rmap(:), row_to_frow(:), rcnt(:), vcnt(:)
  type(zqrm_front_type), pointer :: front
  type(qrm_adata_type), pointer  :: adata
  type(zqrm_fdata_type), pointer :: fdata
  type(zqrm_spmat_type)          :: mat_tmp
  integer(kind=qrm_mem_kind)     :: avail_mem, size_arrowheads
  real(r32)                      :: mem_relax
  complex(r64)                   :: td
  logical                        :: flip
  integer(kind=8)                :: ts, te, tr, ts1, ts2, ts3, te1, te2, te3
  
  ! error management
  integer                        :: err
  character(len=*), parameter    :: name='qrm_factorization_init_core'

  err = 0

  __QRM_PROF_START(ts)
  __QRM_PROF_START(ts1)
  
  adata => qrm_spfct%adata
  fdata => qrm_spfct%fdata

  if(transp .eq. qrm_conj_transp) then
     n = qrm_mat%m
     m = qrm_mat%n
  else
     m = qrm_mat%m
     n = qrm_mat%n
  end if

  ! rmap contains a mapping of the rows onto fronts
  if(err.eq.0) call qrm_alloc(rmap, m)
  if(err.eq.0) call qrm_alloc(row_to_frow, m)
  if(err.eq.0) call qrm_alloc(rcnt, m)
  if(err.eq.0) call qrm_alloc(vcnt, adata%nnodes)
  if(err.eq.0) call qrm_alloc(icperm, n, err)
  if(err.eq.0) call qrm_alloc(irperm, m, err)
  __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)

  rcnt = 0
  vcnt = 0
  
  do i=1, n
     icperm(adata%cperm(i)) = i
  end do
  
  if(qrm_mat%sym.gt.0) then
     call qrm_alloc(adata%rperm, size(adata%cperm))
     adata%rperm = adata%cperm
  end if
  
  do i=1, m
     irperm(adata%rperm(i)) = i
  end do

  ! build the row mapping
  roff = 1
  do f = 1, adata%nnodes

     if(qrm_mat%sym.gt.0) then
        nrowsf    = adata%cp_ptr(f+1) - adata%cp_ptr(f)
     else
        nrowsf    =  adata%stair(f) - roff+1
     end if

     ! for all the rows belonging to this front
     do i = roff, roff+nrowsf-1
        rmap(adata%rperm(i)) = f
        row_to_frow(adata%rperm(i)) = i-roff+1
     end do
     roff = roff + nrowsf
  end do

  ! first pass to count
  do k=1, qrm_mat%nz
     j = qrm_mat%jcn(k)
     i = qrm_mat%irn(k)

     if(qrm_mat%sym.gt.0) then
        if(irperm(i).gt.icperm(j)) then
           r = j; c = i
        else
           r = i; c = j
        end if
     else if(qrm_mat%sym.eq.0) then
        if (transp .eq. qrm_conj_transp) then
           r = j; c = i
        else
           r = i; c = j
        end if
     end if
     
     if((c.le.0) .or. (c.gt.n) .or. (r.le.0) .or. (r.gt.m) ) cycle

     f = rmap(r)
     ! count the number of values per row
     rcnt(r) = rcnt(r)+1
     ! count the number of nonzeros from the original matrix in front f
     vcnt(f) = vcnt(f)+1
  end do

  __QRM_PROF_STOP(qrm_prof_factinit1_,ts1,te1)
  __QRM_PROF_START(ts2)

  size_arrowheads = 0
  fdata%nfronts = adata%nnodes
  
  roff = 1
  i    = 1

  ! pass to init and prepare pointers
  do f = 1, adata%nnodes
     front      => fdata%front_list(f)

     if(qrm_mat%sym.gt.0) then
        nrowsf    = adata%cp_ptr(f+1) - adata%cp_ptr(f)
     else
        nrowsf    =  adata%stair(f) - roff+1
     end if

     front%anrows = nrowsf
     front%num    = f
     front%small  =  adata%small(f)
     front%n      =  adata%rc(f)

     if(qrm_spfct%sym.gt.0) then
        front%m     =  front%n
     else
        front%m     =  adata%nfrows(f)
     end if

     ! The number of pivots in f
     front%npiv = min(min(adata%cp_ptr(f+1)-adata%cp_ptr(f),front%m),front%n)

     if(qrm_spfct%sym.gt.0) then
        ! The number of eliminations to be performed on the front f
        front%ne   = front%npiv
     else
        ! The number of eliminations to be performed on the front f
        front%ne   = min(front%m,front%n)
     end if

     front%rsize =  0
     front%hsize =  0  

     if( (front%n .le. 0) .or. (front%m .le. 0)) then
        ! nothing to do here. Mark the front as done
        front%np = 0
        front%nr = 0
        front%nc = 0

     else
        front%ib   =  min(qrm_spfct%icntl(qrm_ib_),front%ne)
        if(front%small.eq.0) then
           if(qrm_spfct%icntl(qrm_mb_).le.0) then
              front%mb   = max(front%m,front%n)
           else
              front%mb   =  min(qrm_spfct%icntl(qrm_mb_),max(front%m,front%n))
           end if
           front%nb   =  min(min(qrm_spfct%icntl(qrm_nb_),front%n),front%mb)
        else
           ! no blocking is done on small fronts because they are treated
           ! sequentially FIXME: 2D partitioning is currently used on small fronts
           front%mb   =  min(qrm_spfct%icntl(qrm_mb_),max(front%m,front%n))
           front%nb   =  front%mb
        end if
        
        ! The number of panel columns in f
        front%np   = max((front%ne-1)/front%mb + 1,0)
        ! The number of block-rows in f
        front%nr   = (front%m-1)/front%mb + 1
        ! The number of block-columns in f
        front%nc   = (front%n-1)/front%mb + 1
        
        if(qrm_spfct%icntl(qrm_bh_).le.0) then
           front%bh = front%nr
        else
           front%bh = qrm_spfct%icntl(qrm_bh_)
        end if
     end if

     
     call qrm_alloc(front%aiptr, max(nrowsf+1,2), err)
     __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)

     size_arrowheads = size_arrowheads + max(nrowsf+1,2)*qrm_sizeof_i
     front%aiptr(1:2) = 1

     call qrm_aalloc(front%ajcn,  vcnt(f), err)
     call qrm_aalloc(front%aval,  vcnt(f), err)
     __QRM_INFO_CHECK(err, name,'qrm_aalloc',9999)

     size_arrowheads = size_arrowheads + vcnt(f)*qrm_sizeof_i + vcnt(f)*qrm_sizeof_z

     front%aiptr(1:2) = 1
     do i = 1, nrowsf-1
        front%aiptr(i+2) = front%aiptr(i+1)+rcnt(adata%rperm(roff+i-1))
     end do

     roff = roff + nrowsf
  end do

  ! pass to fill
  do k=1, qrm_mat%nz
     j = qrm_mat%jcn(k)
     i = qrm_mat%irn(k)

     if(qrm_mat%sym.gt.0) then
        flip = irperm(i).gt.icperm(j)
     else
        flip = transp.eq.qrm_conj_transp
     end if

     if (flip) then
        r = j; c = i
     else
        r = i; c = j
     end if
     
     if((c.le.0) .or. (c.gt.n) .or. (r.le.0) .or. (r.gt.m) ) cycle

     f     = rmap(r)
     lrow  = row_to_frow(r)
     front => fdata%front_list(f)
     front%ajcn(front%aiptr(lrow+1)) = c
     if (flip) then
        front%aval(front%aiptr(lrow+1)) = conjg(qrm_mat%val(k))
     else
        front%aval(front%aiptr(lrow+1)) = qrm_mat%val(k)
     end if
     front%aiptr(lrow+1) = front%aiptr(lrow+1)+1
  end do

  __QRM_PROF_STOP(qrm_prof_factinit2_,ts2,te2)
  __QRM_PROF_START(ts3)

  ! here we rely on the fact that m>=n
  call qrm_move_alloc(rcnt, gcolmap)
  ! pass to translate indices into mapping
  do f = 1, adata%nnodes
     front => fdata%front_list(f)
     do j=1, front%n
        k = adata%fcol(adata%fcol_ptr(f)+j-1)
        gcolmap(k)=j
     end do
     do i=1, front%anrows
        mnv = huge(1)
        mnj = 0
        do j= front%aiptr(i), front%aiptr(i+1)-1
           front%ajcn(j) = gcolmap(front%ajcn(j))
           if(front%ajcn(j).lt.mnv) then
              mnv = front%ajcn(j)
              mnj = j
           end if
        end do
        
        ! swap first coefficient of a row with the one who has the
        ! lowest index. This speeds up the creation of the staircase
        ! in the front activation because we know that the first
        ! coefficient in a row is really the leftmost
        if(mnj.gt.1) then
           t                          = front%ajcn(front%aiptr(i))
           td                         = front%aval(front%aiptr(i))
           front%ajcn(front%aiptr(i)) = front%ajcn(mnj)
           front%aval(front%aiptr(i)) = front%aval(mnj)
           front%ajcn(mnj)            = t
           front%aval(mnj)            = td
        end if
     end do

  end do

  __QRM_PROF_STOP(qrm_prof_factinit3_,ts3,te3)
  
  call qrm_get(qrm_spfct,'qrm_mem_relax', mem_relax)
  if(mem_relax.lt.0) then
     avail_mem = huge(avail_mem)
  else
     avail_mem = ceiling(real(qrm_spfct%gstats(qrm_e_facto_mempeak_),r32)*mem_relax,kind=qrm_mem_kind)
  end if
  
  call qrm_facto_mem_init(fdata%ma, avail_mem)
  ! reserve arrowheads memory
  call qrm_facto_mem_get(fdata%ma, size_arrowheads)
  
  qrm_spfct%gstats(qrm_nnz_r_) = 0
  qrm_spfct%gstats(qrm_nnz_h_) = 0

9999 continue 
  if(err.ne.0) call zqrm_fdata_destroy(qrm_spfct%fdata)

  __QRM_PROF_STOP(qrm_prof_factinit_,ts,te)

  
  if(present(info)) info = err
  return
 
end subroutine zqrm_factorization_init_core
