!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!>  qrm_spfct_geqrs.F90
!! This file contains a subroutine for computing the least-squares solution of a problem
!!
!!
!! ##############################################################################################

#include "qrm_common.h"

!> This routine computes the least-squares solution of a problem

!> This routine computed the least-squares solution of an overdetermined system
!! Ax=b with multiple RHSs.
!!
!! [in]     qrm_spfct a qrm_spfct_type data which contains the factorization result.
!!                        On output the original data will be unchanged.
!! [in,out]       b the RHSs. A 2D array of leading dimension qrm_spfct%m. On output
!!                        it will contain Q'*b
!! [out]          x the solution, i.e., R\Q'*b
!!
!! [in] transp        a character saying whether to solve against A or A'

subroutine cqrm_spfct_geqrs2d(qrm_spfct, b, x, transp, info)
  use cqrm_spmat_mod
  use cqrm_spfct_mod
  use qrm_error_mod
  use cqrm_analysis_mod
  use cqrm_factorization_mod
  use cqrm_solve_mod
  use cqrm_sdata_mod
  use qrm_dscr_mod
  implicit none

  type(cqrm_spfct_type)               :: qrm_spfct
  complex(r32), intent(inout), target :: b(:,:)
  complex(r32), intent(out), target   :: x(:,:)
  character, optional                 :: transp
  integer, optional                   :: info

  type(cqrm_sdata_type), allocatable  :: qrm_sdata_x(:), qrm_sdata_b(:)
  integer                             :: i, nb, nrhs, n
  type(qrm_dscr_type)                 :: qrm_dscr
  character                           :: itransp
  
  ! error management
  integer                             :: err
  character(len=*), parameter         :: name='qrm_spfct_geqrs'

  err = 0

  __QRM_PRNT_DBG('("Entering the spfct_geqrs driver")')

  __QRM_IFPRESENT3(itransp, transp, qrm_no_transp)
  
  ! if(qrm_spfct%m .ge. qrm_spfct%n) then
     ! transp = qrm_no_transp
  ! else
     ! transp = qrm_conj_transp
  ! end if

  if(itransp.eq.qrm_no_transp) then
     ! check rhs and solution sizes
     if( (size(x,1).lt.qrm_spfct%n) .or. (size(b,1).lt.qrm_spfct%m) ) then
        __QRM_INFO_SET(err,36,name,9999)
     end if
  else if(itransp.eq.qrm_conj_transp) then
     ! check rhs and solution sizes
     if( (size(x,1).lt.qrm_spfct%m) .or. (size(b,1).lt.qrm_spfct%n) ) then
        __QRM_INFO_SET(err,36,name,9999)
     end if
  end if

  ! init the descriptor
  call qrm_dscr_init(qrm_dscr)

  ! blocking to deal with multiple rhs
  call qrm_get(qrm_spfct, 'qrm_rhsnb', nb)
  nrhs = size(b,2)
  if(nb.le.0) nb = nrhs

  allocate(qrm_sdata_b((nrhs-1)/nb+1))
  allocate(qrm_sdata_x((nrhs-1)/nb+1))

  do i=1, (nrhs-1)/nb+1
     call cqrm_sdata_init(qrm_sdata_b(i), b(:,(i-1)*nb+1:min(i*nb,nrhs)))
     call cqrm_sdata_init(qrm_sdata_x(i), x(:,(i-1)*nb+1:min(i*nb,nrhs)))
     ! if(qrm_spfct%m .ge. qrm_spfct%n) then
     if(itransp.eq.qrm_no_transp) then
        call cqrm_spfct_unmqr_async(qrm_dscr, qrm_spfct, qrm_conj_transp, qrm_sdata_b(i))
        call cqrm_spfct_trsm_async(qrm_dscr, qrm_spfct, qrm_no_transp, qrm_sdata_b(i), qrm_sdata_x(i))
     else if(itransp.eq.qrm_conj_transp) then
        call cqrm_spfct_trsm_async(qrm_dscr, qrm_spfct, qrm_conj_transp, qrm_sdata_b(i), qrm_sdata_x(i))
        call cqrm_spfct_unmqr_async(qrm_dscr, qrm_spfct, qrm_no_transp, qrm_sdata_x(i))
     end if
  end do

  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, err)

  call qrm_dscr_destroy(qrm_dscr)

  do i=1, (nrhs-1)/nb+1
     call cqrm_sdata_destroy(qrm_sdata_b(i))
     call cqrm_sdata_destroy(qrm_sdata_x(i))
  end do

9999 continue
  if(present(info)) info = err
  return

end subroutine cqrm_spfct_geqrs2d




!> This routine computes the least-squares solution of a problem

!> This routine computed the least-squares solution of an overdetermined system
!! Ax=b with a single RHS.
!!
!! [in]     qrm_spfct a qrm_spfct_type data which contains the factorization result.
!!                        On output the original data will be unchanged.
!! [in,out]       b the RHSs. A 1D array of leading dimension qrm_spfct%m. On output
!!                        it will contain Q'*b
!! [out]          x the solution, i.e., R\Q'*b
!!
subroutine cqrm_spfct_geqrs1d(qrm_spfct, b, x, transp, info)
  use cqrm_spfct_mod
  use qrm_error_mod
  use cqrm_utils_mod
  use cqrm_methods_mod, savesym => cqrm_spfct_geqrs1d
  implicit none

  type(cqrm_spfct_type)           :: qrm_spfct
  complex(r32), intent(inout)     :: b(:)
  complex(r32), intent(out)       :: x(:)
  character, optional             :: transp
  integer, optional               :: info

  complex(r32), pointer           :: pntb(:,:), pntx(:,:)
  integer                         :: n
  ! error management
  integer                         :: err
  character(len=*), parameter     :: name='qrm_spfct_geqrs1d'

  n = size(b,1)
  call cqrm_remap_pnt(b, pntb, n)
  n = size(x,1)
  call cqrm_remap_pnt(x, pntx, n)

  call cqrm_spfct_geqrs2d(qrm_spfct, pntb, pntx, transp, info)

  return

end subroutine cqrm_spfct_geqrs1d
