!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_spfct_potrs.F90
!! This file contains a subroutine for computing the least-squares solution of a problem
!!
!!
!! ##############################################################################################

#include "qrm_common.h"

!> @brief This routine computes the solution of a SPD problem

!> This routine computed the solution of a SPD system
!! Ax=b with multiple RHSs.
!!
!! @param[in]   qrm_spfct a qrm_spfct_type data which contains the
!!                        result of the factorization.
!!                        On output the original data will be unchanged.
!! @param[in,out]       b the RHSs. A 2D array of leading dimension qrm_spfct%m.
!!                        On output b is modified
!! @param[out]          x the solution, i.e., R'\R\b
!!
subroutine cqrm_spfct_potrs2d(qrm_spfct, b, x, info)
  use cqrm_spmat_mod
  use cqrm_spfct_mod
  use qrm_error_mod
  use cqrm_analysis_mod
  use cqrm_factorization_mod
  use cqrm_solve_mod
  use cqrm_sdata_mod
  use qrm_dscr_mod
  implicit none

  type(cqrm_spfct_type)               :: qrm_spfct
  complex(r32), intent(inout), target :: b(:,:)
  complex(r32), intent(out), target   :: x(:,:)
  integer, optional                   :: info

  type(cqrm_sdata_type), allocatable  :: qrm_sdata_x(:), qrm_sdata_b(:)
  integer                             :: i, nb, nrhs, n
  type(qrm_dscr_type)                 :: qrm_dscr

  ! error management
  integer                            :: err
  character(len=*), parameter        :: name='qrm_spfct_potrs'

  err = 0

  __QRM_PRNT_DBG('("Entering the spfct_potrs")')

  if(qrm_spfct%n .ne. qrm_spfct%m) then
     err = 31
     call qrm_error_print(err, name,ied=(/qrm_spfct%m,qrm_spfct%n/))
     goto 9999
  end if

  ! init the descriptor
  call qrm_dscr_init(qrm_dscr)

  ! blocking to deal with multiple rhs
  call qrm_get(qrm_spfct, 'qrm_rhsnb', nb)
  nrhs = size(b,2)
  if(nb.le.0) nb = nrhs

  allocate(qrm_sdata_b((nrhs-1)/nb+1))
  allocate(qrm_sdata_x((nrhs-1)/nb+1))
  x = b
  
  do i=1, (nrhs-1)/nb+1
     call cqrm_sdata_init(qrm_sdata_b(i), b(:,(i-1)*nb+1:min(i*nb,nrhs)))
     call cqrm_sdata_init(qrm_sdata_x(i), x(:,(i-1)*nb+1:min(i*nb,nrhs)))
     call cqrm_spfct_trsm_async(qrm_dscr, qrm_spfct, qrm_conj_transp, qrm_sdata_x(i), qrm_sdata_b(i))
     call cqrm_spfct_trsm_async(qrm_dscr, qrm_spfct, qrm_no_transp, qrm_sdata_b(i), qrm_sdata_x(i))
  end do

  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, err)
  
  call qrm_dscr_destroy(qrm_dscr)

  do i=1, (nrhs-1)/nb+1
     call cqrm_sdata_destroy(qrm_sdata_b(i))
     call cqrm_sdata_destroy(qrm_sdata_x(i))
  end do

9999 continue
  if(present(info)) info = err
  return

end subroutine cqrm_spfct_potrs2d




!> @brief This routine computes the solution of a sparse SPD problem

!> This routine computed the solution of a sparse SPD system
!! Ax=b with a single RHS.
!! 
!! @param[in]   qrm_spfct a qrm_spfct_type data which contains the
!!                        result of the factorization.
!!                        On output the original data will be unchanged.
!! @param[in,out]       b the RHSs. A 1D array of leading dimension qrm_spfct%m. On output
!!                        it will be modified
!! @param[out]          x the solution, i.e., R'\R\b
!!
subroutine cqrm_spfct_potrs1d(qrm_spfct, b, x, info)
  use cqrm_spfct_mod
  use qrm_error_mod
  use cqrm_utils_mod
  use cqrm_methods_mod, savesym => cqrm_spfct_potrs1d
  implicit none

  type(cqrm_spfct_type)           :: qrm_spfct
  complex(r32), intent(inout)     :: b(:)
  complex(r32), intent(out)       :: x(:)
  integer, optional               :: info

  complex(r32), pointer           :: pntb(:,:), pntx(:,:)
  integer                         :: n
  ! error management
  integer                         :: err
  character(len=*), parameter     :: name='qrm_least_squares1d'

  n = size(b,1)
  call cqrm_remap_pnt(b, pntb, n)
  n = size(x,1)
  call cqrm_remap_pnt(x, pntx, n)

  call cqrm_spfct_potrs2d(qrm_spfct, pntb, pntx, info)

  return

end subroutine cqrm_spfct_potrs1d
