!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> qrm_spmat_backslash.F90
!! This file contains a subroutine for computing the least-squares solution of a problem
!!
!!
!! ##############################################################################################

#include "qrm_common.h"

!>  This routine computes the solution of a sparse problem

!> This routine computed the least-squares solution of a sparse system
!! Ax=b with multiple RHSs.
!!
!! [in]     qrm_spmat a qrm_spmat_type data which contains the input matrix.
!!                        On output the original data will be unchanged and
!!                        the result of the analysis and factorization phases will
!!                        be stored in the adata and fdata fields, respectively.
!! [in,out]       b the RHSs. A 2D array of leading dimension qrm_spmat%m.
!!                        On output b is modified
!! [out]          x the solution
!!
subroutine cqrm_spmat_backslash2d(qrm_spmat, b, x, transp, cperm, info)
  use qrm_parameters_mod
  use cqrm_spmat_mod
  use qrm_error_mod
  use cqrm_methods_mod, savesym => cqrm_spmat_backslash2d
  implicit none

  type(cqrm_spmat_type)               :: qrm_spmat
  complex(r32), intent(inout), target :: b(:,:)
  complex(r32), intent(out), target   :: x(:,:)
  integer, optional, target           :: cperm(:)
  character, optional                 :: transp
  integer, optional                   :: info

  ! error management
  integer                             :: err
  character(len=*), parameter         :: name='qrm_spmat_backslash'

  err = 0

  __QRM_PRNT_DBG('("Entering the spmat_backslash")')

  if(qrm_spmat%sym.gt.0) then
     call cqrm_spmat_posv2d(qrm_spmat, b, x, cperm, err)
  else
     call cqrm_spmat_gels2d(qrm_spmat, b, x, transp, cperm, err)
     ! if(qrm_spmat%m.ge.qrm_spmat%n) then
        ! call cqrm_least_squares2d(qrm_spmat, b, x, cperm, err)
     ! else
        ! call cqrm_min_norm2d(qrm_spmat, b, x, cperm, err)
     ! end if
  end if
  
9999 continue
  if(present(info)) info = err
  return

end subroutine cqrm_spmat_backslash2d




!>  This routine computes the solution of a sparse problem

!> This routine computed the solution of a sparse system
!! Ax=b with a single RHS.
!!
!! [in]     qrm_spmat a qrm_spmat_type data which contains the input matrix.
!!                        On output the original data will be unchanged and
!!                        the result of the analysis and factorization phases will
!!                        be stored in the adata and fdata fields, respectively.
!! [in,out]           b the RHSs. A 1D array of leading dimension qrm_spmat%m. On output
!!                        it will be modified
!! [out]              x the solution
!!
subroutine cqrm_spmat_backslash1d(qrm_spmat, b, x, transp, cperm, info)
  use cqrm_spmat_mod
  use qrm_error_mod
  use cqrm_utils_mod
  use cqrm_methods_mod, savesym => cqrm_spmat_backslash1d
  implicit none

  type(cqrm_spmat_type)           :: qrm_spmat
  complex(r32), intent(inout)     :: b(:)
  complex(r32), intent(out)       :: x(:)
  character, optional             :: transp
  integer, optional               :: cperm(:)
  integer, optional               :: info

  complex(r32), pointer           :: pntb(:,:), pntx(:,:)
  integer                         :: n
  ! error management
  integer                         :: err
  character(len=*), parameter     :: name='qrm_spmat_backslash1d'

  n = size(b,1)
  call cqrm_remap_pnt(b, pntb, n)
  n = size(x,1)
  call cqrm_remap_pnt(x, pntx, n)

  call cqrm_spmat_backslash2d(qrm_spmat, pntb, pntx, transp, cperm, info)

  return

end subroutine cqrm_spmat_backslash1d
