!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> qrm_residual_norm.F90
!! This file contains a subroutine for computing the scaled norm of the residual
!!
!!
!! ##############################################################################################

#include "qrm_common.h"

!> This routine computes the scaled norm of multiple residuals

!> This routine computes the norm of the scaled residual, i.e., ||r||/||A||
!! for a single RHS.
!!
!! [in]     qrm_spmat a qrm_spmat_type data which contains the input matrix.
!!
!! [in,out]         b the RHSs. A 1D array of leading dimension qrm_spmat%m. On output
!!                        it will contain the residual
!!
!! [in]             x the solution
!!
!! [out]          nrm the residual norm
!!
subroutine dqrm_residual_norm2d(qrm_spmat, b, x, nrm, transp, info)
  use qrm_mem_mod
  use dqrm_spmat_mod
  use qrm_error_mod
  use dqrm_utils_mod
  use qrm_parameters_mod
  implicit none

  type(dqrm_spmat_type)  :: qrm_spmat
  real(r64)           :: b(:,:), x(:,:)
  real(r64)              :: nrm(:)
  character, optional    :: transp
  integer, optional      :: info

  real(r64), allocatable :: nrmb(:), nrmx(:)
  real(r64)              :: nrma
  integer                :: nrhs
  character              :: itransp

  ! error management
  integer                         :: err
  character(len=*), parameter     :: name='dqrm_residual_norm'

  err = 0

  __QRM_IFPRESENT3(itransp, transp, qrm_no_transp)
  
  nrhs = min(size(b,2),size(x,2))

  call qrm_alloc(nrmb, nrhs, err)
  call qrm_alloc(nrmx, nrhs, err)
  __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)

  call dqrm_vecnrm(b(:,1:nrhs), qrm_spmat%m, 'i', nrmb)
  call dqrm_vecnrm(x(:,1:nrhs), qrm_spmat%n, 'i', nrmx)

  ! compute the residual
  call dqrm_spmat_mv(qrm_spmat, itransp, qrm_dmone, x, qrm_done, b)

  call dqrm_spmat_nrm(qrm_spmat, 'i', nrma)
  call dqrm_vecnrm(b, qrm_spmat%m, 'i', nrm)

  nrmb = nrmb + nrma*nrmx

  nrm = nrm/nrmb

  call qrm_dealloc(nrmx)
  call qrm_dealloc(nrmb)

9999 continue ! error management
  if(present(info)) info = err
  return

end subroutine dqrm_residual_norm2d



!> This routine computes the scaled norm of the residual

!> This routine computes the norm of the scaled residual, i.e., ||r||/||A||
!! for a single RHS.
!!
!! [in]     qrm_spmat a qrm_spmat_type data which contains the input matrix.
!!
!! [in,out]         b the RHSs. A 1D array of leading dimension qrm_spmat%m. On output
!!                        it will contain the residual
!!
!! [in]             x the solution
!!
!! [out]          nrm the residual norm
!!
subroutine dqrm_residual_norm1d(qrm_spmat, b, x, nrm, transp, info)
  use dqrm_spmat_mod
  use qrm_error_mod
  use dqrm_utils_mod
  use qrm_parameters_mod
  implicit none

  type(dqrm_spmat_type) :: qrm_spmat
  real(r64)          :: b(:), x(:)
  real(r64)             :: nrm
  character, optional   :: transp
  integer, optional     :: info

  real(r64)             :: nrmb, nrmx
  real(r64)             :: nrma
  character             :: itransp

  ! error management
  integer                         :: err
  character(len=*), parameter     :: name='dqrm_residual_norm'

  __QRM_IFPRESENT3(itransp, transp, qrm_no_transp)

  err = 0

  call dqrm_vecnrm(b, qrm_spmat%m, 'i', nrmb)
  call dqrm_vecnrm(x, qrm_spmat%n, 'i', nrmx)

  ! compute the residual
  call dqrm_spmat_mv(qrm_spmat, itransp, qrm_dmone, x, qrm_done, b)
  call dqrm_spmat_nrm(qrm_spmat, 'i', nrma)
  call dqrm_vecnrm(b, qrm_spmat%m, 'i', nrm)

  nrmb = nrmb + nrma*nrmx
  nrm = nrm/nrmb

9999 continue ! error management
  if(present(info)) info = err
  return

end subroutine dqrm_residual_norm1d
