!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!>  qrm_spmat_gels.F90
!! This file contains a subroutine for computing the least-squares solution of a problem
!!
!!
!! ##############################################################################################

#include "qrm_common.h"

!> This routine computes the least-squares solution of a problem

!> This routine computed the least-squares solution of an overdetermined system
!! Ax=b with multiple RHSs.
!!
!! [in]     qrm_spmat a qrm_spmat_type data which contains the input matrix.
!!                        On output the original data will be unchanged and
!!                        the result of the analysis and factorization phases will
!!                        be stored in the adata and fdata fields, respectively.
!! [in,out]       b the RHSs. A 2D array of leading dimension qrm_spmat%m. On output
!!                        it will contain Q'*b
!! [out]          x the solution, i.e., R\Q'*b
!!
subroutine dqrm_spmat_gels2d(qrm_spmat, b, x, transp, cperm, info)
  use dqrm_spmat_mod
  use dqrm_spfct_mod
  use qrm_error_mod
  use dqrm_analysis_mod
  use dqrm_factorization_mod
  use dqrm_solve_mod
  use dqrm_sdata_mod
  use qrm_dscr_mod
  implicit none

  type(dqrm_spmat_type)               :: qrm_spmat
  real(r64), intent(inout), target :: b(:,:)
  real(r64), intent(out), target   :: x(:,:)
  character, optional                 :: transp
  integer, optional, target           :: cperm(:)
  integer, optional                   :: info

  type(dqrm_sdata_type), allocatable  :: qrm_sdata_x(:), qrm_sdata_b(:)
  integer                             :: i, nb, nrhs, n
  type(qrm_dscr_type)                 :: qrm_dscr
  type(dqrm_spfct_type)               :: qrm_spfct
  character                           :: itransp, ftransp, stransp
  
  ! error management
  integer                             :: err
  character(len=*), parameter         :: name='qrm_spmat_gels'

  err = 0

  __QRM_PRNT_DBG('("Entering the spmat_gels driver")')

  call dqrm_spfct_init(qrm_spfct, qrm_spmat, err)
  __QRM_INFO_CHECK(err, name, 'qrm_spfct_init', 9999)

  if(present(cperm)) then
     qrm_spfct%cperm_in => cperm
     call qrm_set(qrm_spfct, 'qrm_ordering', qrm_given_)
  end if

  __QRM_IFPRESENT3(itransp, transp, qrm_no_transp)

  if(itransp.eq.qrm_no_transp) then
     ! check rhs and solution sizes
     if( (size(x,1).lt.qrm_spmat%n) .or. (size(b,1).lt.qrm_spmat%m) ) then
        __QRM_INFO_SET(err,36,name,9999)
     end if
  else if(itransp.eq.qrm_transp) then
     ! check rhs and solution sizes
     if( (size(x,1).lt.qrm_spmat%m) .or. (size(b,1).lt.qrm_spmat%n) ) then
        __QRM_INFO_SET(err,36,name,9999)
     end if
  end if
  
  if(qrm_spmat%m.ge.qrm_spmat%n) then
     ftransp = qrm_no_transp
     stransp = itransp
  else
     ftransp = qrm_transp
     if(itransp.eq.qrm_no_transp) then
        stransp = qrm_transp
     else
        stransp = qrm_no_transp
     end if
  end if

  
  ! init the descriptor
  call qrm_dscr_init(qrm_dscr)

  call dqrm_analyse_async(qrm_dscr, qrm_spmat, qrm_spfct, ftransp)
  call dqrm_factorize_async(qrm_dscr, qrm_spmat, qrm_spfct, ftransp)

  ! blocking to deal with multiple rhs
  call qrm_get(qrm_spfct, 'qrm_rhsnb', nb)
  nrhs = size(b,2)
  if(nb.le.0) nb = nrhs

  allocate(qrm_sdata_b((nrhs-1)/nb+1))
  allocate(qrm_sdata_x((nrhs-1)/nb+1))

  do i=1, (nrhs-1)/nb+1
     call dqrm_sdata_init(qrm_sdata_b(i), b(:,(i-1)*nb+1:min(i*nb,nrhs)))
     call dqrm_sdata_init(qrm_sdata_x(i), x(:,(i-1)*nb+1:min(i*nb,nrhs)))
     if(stransp.eq.qrm_no_transp) then
        call dqrm_spfct_unmqr_async(qrm_dscr, qrm_spfct, qrm_transp, qrm_sdata_b(i))
        call dqrm_spfct_trsm_async(qrm_dscr, qrm_spfct, qrm_no_transp, qrm_sdata_b(i), qrm_sdata_x(i))
     else if(stransp.eq.qrm_transp) then
        call dqrm_spfct_trsm_async(qrm_dscr, qrm_spfct, qrm_transp, qrm_sdata_b(i), qrm_sdata_x(i))
        call dqrm_spfct_unmqr_async(qrm_dscr, qrm_spfct, qrm_no_transp, qrm_sdata_x(i))
     end if
  end do

  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, err)

  call qrm_dscr_destroy(qrm_dscr)

  do i=1, (nrhs-1)/nb+1
     call dqrm_sdata_destroy(qrm_sdata_b(i))
     call dqrm_sdata_destroy(qrm_sdata_x(i))
  end do

  call dqrm_spfct_destroy(qrm_spfct, err)
  __QRM_INFO_CHECK(err, name, 'qrm_spfct_destroy', 9999)

9999 continue
  if(present(info)) info = err
  return

end subroutine dqrm_spmat_gels2d




!> This routine computes the least-squares solution of a problem

!> This routine computed the least-squares solution of an overdetermined system
!! Ax=b with a single RHS.
!!
!! [in]     qrm_spmat a qrm_spmat_type data which contains the input matrix.
!!                        On output the original data will be unchanged and
!!                        the result of the analysis and factorization phases will
!!                        be stored in the adata and fdata fields, respectively.
!! [in,out]       b the RHSs. A 1D array of leading dimension qrm_spmat%m. On output
!!                        it will contain Q'*b
!! [out]          x the solution, i.e., R\Q'*b
!!
subroutine dqrm_spmat_gels1d(qrm_spmat, b, x, transp, cperm, info)
  use dqrm_spmat_mod
  use qrm_error_mod
  use dqrm_utils_mod
  use dqrm_methods_mod, savesym => dqrm_spmat_gels1d
  implicit none

  type(dqrm_spmat_type)           :: qrm_spmat
  real(r64), intent(inout)     :: b(:)
  real(r64), intent(out)       :: x(:)
  character, optional             :: transp
  integer, optional               :: cperm(:)
  integer, optional               :: info

  real(r64), pointer           :: pntb(:,:), pntx(:,:)
  integer                         :: n
  ! error management
  integer                         :: err
  character(len=*), parameter     :: name='qrm_spmat_gels1d'

  n = size(b,1)
  call dqrm_remap_pnt(b, pntb, n)
  n = size(x,1)
  call dqrm_remap_pnt(x, pntx, n)

  call dqrm_spmat_gels2d(qrm_spmat, pntb, pntx, transp, cperm, info)

  return

end subroutine dqrm_spmat_gels1d
