!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_get_r.F90
!! This file contains a subroutine which extracts the R factor
!! resulting from the factorization in sparse form
!!
!!
!! ##############################################################################################

#include "qrm_common.h"

!> @brief This routine computes the scaled norm of the product A'*r for multiple residuals

!> This routine computes the norm of the scaled product A'*r
!! for a single RHS.
!!
!! @param[in]     qrm_mat a qrm_spmat_type data which contains the R factor to be extracted
!!
!! @param[out]    r  a qrm_spmat_type data which contains the extracted R factor 
!!
subroutine sqrm_get_r(qrm_spfct, r, info)
  use qrm_common_mod
  use qrm_error_mod
  use sqrm_spmat_mod
  use sqrm_spfct_mod
  use qrm_mem_mod
  implicit none

  type(sqrm_spfct_type), target  :: qrm_spfct
  type(sqrm_spmat_type)          :: r
  integer, optional              :: info

  type(sqrm_front_type), pointer :: front
  integer                        :: cnt, f, i, j, k, m, row, col, rbcnt, rtcnt, br, bc

  ! error management
  integer                          :: err
  character(len=*), parameter      :: name='qrm_get_r'

  err = 0
  r%nz = qrm_spfct%gstats(qrm_nnz_r_)
  r%m  = qrm_spfct%m
  r%n  = qrm_spfct%n

  if(err.eq.0) call qrm_alloc(r%irn, r%nz, err)
  if(err.eq.0) call qrm_alloc(r%jcn, r%nz, err)
  if(err.eq.0) call qrm_alloc(r%val, r%nz, err)
  __QRM_INFO_CHECK(err, name, 'qrm_alloc', 9999)

  rtcnt = 1
  rbcnt = min(r%m,r%n)+1
  cnt   = 1
  do f = 1, qrm_spfct%adata%nnodes 
     front => qrm_spfct%fdata%front_list(f)

     if(.not.allocated(front%f%blocks)) cycle

     do row=1, front%npiv
        br = (row-1)/front%mb+1
        i = mod(row-1,front%mb)+1
        do col=row, front%n
           bc = (col-1)/front%mb+1
           j = mod(col-1,front%mb)+1
           r%irn(cnt) = front%rows(row)
           r%jcn(cnt) = front%cols(col)
           r%val(cnt) = front%f%blocks(br,bc)%c(i,j)
           cnt=cnt+1
        end do
     end do
  end do

  r%nz = cnt-1
  if(err.eq.0) call qrm_realloc(r%irn, r%nz, info=err, copy=.true.)
  if(err.eq.0) call qrm_realloc(r%jcn, r%nz, info=err, copy=.true.)
  if(err.eq.0) call qrm_realloc(r%val, r%nz, info=err, copy=.true.)
  __QRM_INFO_CHECK(err, name, 'qrm_realloc', 9999)

  if(present(info)) info = err
  return
  
9999 continue
  ! cleanup and return
  call qrm_dealloc(r%irn)
  call qrm_dealloc(r%jcn)
  call qrm_dealloc(r%val)
  if(present(info)) info = err

  return
  
end subroutine sqrm_get_r
