!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!!
!! This file contains a module that defines the data type for storing
!! the results of the solve phase (not used at the moment)
!!
!! ##############################################################################################


!> This module contains the definition of all the data related to the
!! solution phase. 
!! 
module dqrm_sdata_mod
  use iso_c_binding
  use dqrm_fdata_mod
#if defined (have_starpu)
  use fstarpu_mod
#endif


  ! Type: dqrm_rhs_type
  !
  ! This type defines a data structure containing a rhs block
  !
  type dqrm_rhs_type
     type(dqrm_dsmat_type)    :: rhs  ! the small RHSs assembled on each front
#if defined (have_starpu)            
     type(c_ptr)              :: hdl  ! symbolic handle
#endif
  end type dqrm_rhs_type

  ! Type: dqrm_sdata_type
  !
  ! This type defines a data structure containing a rhs block
  !
  type dqrm_sdata_type
     real(r64), pointer                 :: p(:,:)
#if defined (have_starpu)            
     type(c_ptr)                        :: hdl
#endif
     type(dqrm_rhs_type), allocatable   :: front_rhs(:)
     type(dqrm_ws_type)                 :: work
  end type dqrm_sdata_type

  interface qrm_sdata_init
     module procedure dqrm_sdata_init2d, dqrm_sdata_init1d
  end interface qrm_sdata_init

  interface dqrm_sdata_init
     module procedure dqrm_sdata_init2d, dqrm_sdata_init1d
  end interface dqrm_sdata_init

  interface qrm_sdata_destroy
     module procedure dqrm_sdata_destroy
  end interface qrm_sdata_destroy

  interface qrm_sync_sdata
     module procedure  dqrm_sync_sdata
  end interface qrm_sync_sdata

  interface qrm_sync_data
     module procedure :: dqrm_sync_sdata
  end interface qrm_sync_data  
  
contains
  
  subroutine dqrm_sdata_init2d(qrm_sdata, x)
    use qrm_mem_mod
    implicit none

    type(dqrm_sdata_type) :: qrm_sdata
    real(r64), target     :: x(:,:)
    
    qrm_sdata%p => x

#if defined (have_starpu)
    call fstarpu_matrix_data_register(qrm_sdata%hdl, 0, &
         c_loc(qrm_sdata%p),                       &
         size(qrm_sdata%p,1),                           &
         size(qrm_sdata%p,1),                           &
         size(qrm_sdata%p,2),                           &
         c_sizeof(qrm_sdata%p(1,1)))
#endif

    if(allocated(qrm_sdata%front_rhs)) deallocate(qrm_sdata%front_rhs)
    call dqrm_ws_destroy(qrm_sdata%work)

    return
  end subroutine dqrm_sdata_init2d

  subroutine dqrm_sdata_init1d(qrm_sdata, x)
    use qrm_mem_mod
    use dqrm_utils_mod
    implicit none

    type(dqrm_sdata_type) :: qrm_sdata
    real(r64), target     :: x(:)
    real(r64), pointer    :: pnt(:,:)
    
    ! Turn the 1D pointer into a 2D one and call the 2D method below
    call dqrm_remap_pnt(x, pnt, size(x,1))    

    call dqrm_sdata_init2d(qrm_sdata, pnt)

  end subroutine dqrm_sdata_init1d

  
  
  
  subroutine dqrm_sdata_destroy(qrm_sdata)
    use qrm_mem_mod
    implicit none

    type(dqrm_sdata_type) :: qrm_sdata
    integer               :: i

#if defined (have_starpu)
    call fstarpu_data_unregister(qrm_sdata%hdl)
    qrm_sdata%hdl = c_null_ptr
#endif

    if(allocated(qrm_sdata%front_rhs)) then
       do i=1, size(qrm_sdata%front_rhs,1)
          call dqrm_dsmat_destroy(qrm_sdata%front_rhs(i)%rhs, seq=.true.)
          
#if defined (have_starpu)
          if(c_associated(qrm_sdata%front_rhs(i)%hdl)) &
               call fstarpu_data_unregister(qrm_sdata%front_rhs(i)%hdl)
#endif
       end do
       deallocate(qrm_sdata%front_rhs)
    end if

    call dqrm_ws_destroy(qrm_sdata%work)
    
    nullify(qrm_sdata%p)

    return
  end subroutine dqrm_sdata_destroy

  subroutine dqrm_sync_sdata(qrm_sdata)
    use qrm_common_mod
    use qrm_error_mod
    implicit none
    
    type(dqrm_sdata_type), target  :: qrm_sdata
    
    integer :: inode, node
    
#if defined(have_starpu)
    if(allocated(qrm_sdata%front_rhs)) then
       do inode = 1, size(qrm_sdata%front_rhs,1)
          if(c_associated(qrm_sdata%front_rhs(inode)%hdl)) then
             call fstarpu_data_acquire(qrm_sdata%front_rhs(inode)%hdl, FSTARPU_R)
             call fstarpu_data_release(qrm_sdata%front_rhs(inode)%hdl)
          end if
       end do
    end if
#endif
    return
    
  end subroutine dqrm_sync_sdata
  

end module dqrm_sdata_mod
