!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################

#include "qrm_common.h"

module qrm_dscr_mod
  use iso_c_binding
  use qrm_error_mod
  use qrm_pthread_mod
  use qrm_parameters_mod
  
  ! IMPORTANT: this type is bind(C) because we also have a C
  ! equivalent. Any modification made to this type must also be made
  ! to the corresponding C type defined in qrm_starpu.h
  type, bind(C) :: qrm_dscr_type
     !> @brief The error status
     integer(c_int)           :: info
     integer(c_int)           :: ncpu
     integer(c_int)           :: ngpu
     integer(c_int)           :: seq  ! if .ne.0 this is a sequential descriptor
#if defined (have_starpu)
     !> @brief A lock to prevent simultaneous writing of the error
     type(qrm_pthread_mutex)  :: mutex
     integer(c_int)           :: ctx
#if defined (have_cuda)
     type(c_ptr)              :: cuhandles(256)
#endif
#endif
  end type qrm_dscr_type

  interface qrm_barrier
     module procedure qrm_barrier, qrm_barrier_dscr
  end interface qrm_barrier

  ! integer :: qrm_ncpu=-1, qrm_ngpu=-1


  type(qrm_dscr_type) :: qrm_seq_dscr_ = qrm_dscr_type(info=0, ncpu=1, ngpu=0, seq=1 &
#if defined (have_starpu)
       , ctx=1, mutex=qrm_pthread_mutex(c_null_ptr) &
#if defined (have_cuda)
       , cuhandles=c_null_ptr &
#endif
#endif
       )

contains

  subroutine qrm_dscr_init(qrm_dscr, nocuda, seq)
#if defined (have_starpu)
    use fstarpu_mod
    use qrm_starpu_common_mod
#endif
    use qrm_string_mod
    implicit none

    type(qrm_dscr_type)  :: qrm_dscr
    logical, optional    :: nocuda, seq
    
#if defined (have_starpu)
    integer              :: nworkers, i
    integer, allocatable :: workers(:)
    character(kind=c_char,len=11), target :: ctx_policy_hp     = C_CHAR_"heteroprio"//C_NULL_CHAR
    character(kind=c_char,len=6) , target :: ctx_policy_eager  = C_CHAR_"eager"//C_NULL_CHAR
    character(kind=c_char,len=6) , target :: ctx_policy_dmdas  = C_CHAR_"dmdas"//C_NULL_CHAR
    character(kind=c_char,len=4) , target :: ctx_policy_lws    = C_CHAR_"lws"//C_NULL_CHAR
    character(kind=c_char,len=3) , target :: ctx_policy_ws     = C_CHAR_"ws"//C_NULL_CHAR
    character(kind=c_char,len=5) , target :: ctx_policy_prio   = C_CHAR_"prio"//C_NULL_CHAR
    type(c_ptr) :: ptr
#if defined (have_cuda)
    type(c_ptr)  :: hp_init_ptr
    interface
       subroutine qrm_heteroprio_init_ctx(ctx) bind(c)
         integer, value :: ctx
       end subroutine qrm_heteroprio_init_ctx
    end interface
    interface
       function qrm_get_hp_init_ptr() result(hp_init_ptr) bind(c)
         use iso_c_binding
         type(c_ptr) :: hp_init_ptr
       end function qrm_get_hp_init_ptr
    end interface
#endif
#endif
    
    qrm_dscr%info = 0
    qrm_dscr%seq  = 0

    if(present(seq)) then
       if(seq) then
          qrm_dscr%seq  = 1
          qrm_dscr%ncpu = 1
          return
       end if
    end if
#if defined (have_starpu)
    call qrm_pthread_mutex_init(qrm_dscr%mutex)

    
    nworkers      = fstarpu_worker_get_count()
    qrm_dscr%ncpu = fstarpu_cpu_worker_get_count()
    if(present(nocuda)) then
       if(nocuda) then
          qrm_dscr%ngpu=0
       else
          qrm_dscr%ngpu = fstarpu_cuda_worker_get_count()
       end if
    else
#if defined (have_cuda)
       qrm_dscr%ngpu = fstarpu_cuda_worker_get_count()
#else
       qrm_dscr%ngpu = 0
#endif
    end if

    allocate(workers(nworkers))

    if(qrm_dscr%ngpu.eq.0) then
       i = fstarpu_worker_get_ids_by_type(FSTARPU_CPU_WORKER, workers, qrm_dscr%ncpu)

       call strtoptr(ctx_policy_lws, ptr)
       qrm_dscr%ctx = fstarpu_sched_ctx_create(workers, qrm_dscr%ncpu, &
            C_CHAR_"qrm_ctx"//C_NULL_CHAR, &
            (/ FSTARPU_SCHED_CTX_POLICY_NAME, ptr, &
            c_null_ptr /))
#if defined (have_cuda)
    else
       ! i = fstarpu_worker_get_ids_by_type(FSTARPU_ANY_WORKER, workers, nworkers)
       do i=1, nworkers
          workers(i) = i-1
       end do
       hp_init_ptr = qrm_get_hp_init_ptr()
       call strtoptr(ctx_policy_hp, ptr)
       qrm_dscr%ctx = fstarpu_sched_ctx_create(workers, nworkers, &
            C_CHAR_"qrm_ctx"//C_NULL_CHAR, &
            (/ FSTARPU_SCHED_CTX_POLICY_NAME, ptr, &
            FSTARPU_SCHED_CTX_POLICY_INIT, hp_init_ptr, &
            c_null_ptr /))
       ! call strtoptr(ctx_policy_dmdas, ptr)
       ! qrm_dscr%ctx = fstarpu_sched_ctx_create(workers, nworkers, &
            ! C_CHAR_"qrm_ctx"//C_NULL_CHAR, &
            ! (/ FSTARPU_SCHED_CTX_POLICY_NAME, ptr, &
            ! c_null_ptr /))
#endif
    end if
    deallocate(workers)

#if defined (have_cuda)
    ! Initialize cuBLAS handles
    call starpu_f_init_cuhandles(qrm_dscr%cuhandles)
#endif

#endif

    return

  end subroutine qrm_dscr_init

  subroutine qrm_dscr_destroy(qrm_dscr)

#if defined (have_starpu)
    use fstarpu_mod
    use qrm_starpu_common_mod
#endif
    implicit none
    
    type(qrm_dscr_type) :: qrm_dscr

#if defined (have_starpu)
#if defined (have_cuda)
    call starpu_f_destroy_cuhandles(qrm_dscr%cuhandles)
#endif
    call qrm_pthread_mutex_destroy(qrm_dscr%mutex)
    call fstarpu_sched_ctx_delete(qrm_dscr%ctx)
#endif
    qrm_dscr%info = 0
    
    return

  end subroutine qrm_dscr_destroy

  subroutine qrm_init(ncpu, ngpu, info)
    use qrm_memhandling_mod
    use qrm_parameters_mod
    use iso_c_binding
#if defined (have_starpu)
    use fstarpu_mod
#endif
#if defined (perfprof)
    use qrm_perfprof_mod
#endif
    implicit none

    integer, optional           :: ncpu, ngpu, info

    integer                     :: ret, ierr, len
    character(LEN=10)           :: str
#if defined (have_starpu)
    type(c_ptr)                 :: qrm_starpu_conf
#endif
      
    ! error management
    integer                     :: err
    character(len=*), parameter :: name='qrm_init'
    
    err = 0

    call qrm_parameters_reset()

#if defined (have_starpu)
    qrm_ncpu = qrm_default_ncpu_
    if(present(ncpu)) then
       if(ncpu.gt.0) qrm_ncpu = ncpu
    end if
    
#if defined (have_cuda)
    qrm_ngpu = qrm_default_ngpu_
    if(present(ngpu)) then
       if(ngpu.ge.0) qrm_ngpu = ngpu
    end if
#endif
#endif
    
#if defined (have_starpu)
    call fstarpu_fxt_autostart_profiling(0)
    qrm_starpu_conf = fstarpu_conf_allocate()
    call fstarpu_conf_init(qrm_starpu_conf)
    call fstarpu_conf_set_ncpu(qrm_starpu_conf, max(qrm_ncpu,1))
#if defined (have_cuda)
    call fstarpu_conf_set_ncuda(qrm_starpu_conf, max(qrm_ngpu,0))
#endif
    ret = fstarpu_init(qrm_starpu_conf)
    if(ret.ne.0) then
       err = 31
       call qrm_error_print(err, name)
    end if
    call fstarpu_conf_free(qrm_starpu_conf)

#endif

#if defined (perfprof)
    call qrm_perfprof_init()
#endif
    
    qrm_tot_mem = 0
    qrm_max_mem = 0
    
    if(present(info)) info = err
    return
  end subroutine qrm_init


  subroutine qrm_finalize()
    use qrm_memhandling_mod
#if defined (perfprof)
    use qrm_perfprof_mod
#endif

#if defined (have_starpu)
    use qrm_starpu_common_mod
    use fstarpu_mod

    call qrm_starpu_destroy_codelets()
    call fstarpu_shutdown()
    
#endif
#if defined (perfprof)
    call qrm_perfprof_finalize()
#endif

    qrm_ncpu = -1
    qrm_ngpu = -1
    
    return
  end subroutine qrm_finalize

  subroutine qrm_barrier()
#if defined (have_starpu)
    use fstarpu_mod
#endif
    implicit none    

#if defined (have_starpu)
    call fstarpu_task_wait_for_all()
#endif
    
    return
  end subroutine qrm_barrier

  subroutine qrm_start_profiling()
#if defined (have_starpu)
    use fstarpu_mod
    call fstarpu_fxt_start_profiling()
#endif
    return
  end subroutine qrm_start_profiling

  subroutine qrm_stop_profiling()
#if defined (have_starpu)
    use fstarpu_mod
    call fstarpu_fxt_stop_profiling()
#endif
    return
  end subroutine qrm_stop_profiling
    
  
  subroutine qrm_barrier_dscr(qrm_dscr, info)
#if defined (have_starpu)
    use fstarpu_mod 
#endif
    implicit none    

    type(qrm_dscr_type) :: qrm_dscr
    integer, optional   :: info

#if defined (have_starpu)
    call fstarpu_task_wait_for_all_in_ctx(qrm_dscr%ctx)
#endif

    if(present(info)) info = qrm_dscr%info
    
    return
  end subroutine qrm_barrier_dscr


  subroutine qrm_status_set(qrm_dscr, code, where, ied, aed)
    use qrm_error_mod
    use qrm_parameters_mod
    implicit none
    
    type(qrm_dscr_type)             :: qrm_dscr
    integer                         :: code
    character(len=*)                :: where
    integer, optional               :: ied(:)
    character(len=*), optional      :: aed

#if defined (have_starpu)
    call qrm_pthread_mutex_lock(qrm_dscr%mutex)
#endif
    if(qrm_dscr%info.eq.0) then
       qrm_dscr%info = code
       call qrm_error_print(code, where, ied, aed)
    end if
#if defined (have_starpu)
    call qrm_pthread_mutex_unlock(qrm_dscr%mutex)
#endif
    return
  end subroutine qrm_status_set
  
  function qrm_get_num_threads(qrm_dscr) result(nworkers)
    use qrm_error_mod
    use qrm_parameters_mod
#if defined (have_starpu)
    use fstarpu_mod
    use qrm_starpu_common_mod
#endif
    implicit none
    
    type(qrm_dscr_type), optional :: qrm_dscr
    integer                       :: nworkers
    
    if(present(qrm_dscr)) then
#if defined (have_starpu)
       nworkers = fstarpu_sched_ctx_get_nworkers(qrm_dscr%ctx)
#else             
       nworkers = 1
#endif
    else
#if defined (have_starpu)
       nworkers = fstarpu_cpu_worker_get_count()
#else
       nworkers = 1
#endif
    end if
    
    return
  end function qrm_get_num_threads
  
  
end module qrm_dscr_mod
