!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!!
!!  
!! This file contains the implementation of a sparse factorization
!! object which is meant to store all the information related to a
!! sparse factorization (control parameters, symbolic analysis,
!! factors, tree, stats etc.)
!!
!! ##############################################################################################


#include "qrm_common.h"


module sqrm_spfct_mod
  use qrm_adata_mod
  use sqrm_fdata_mod
  use qrm_parameters_mod

  type sqrm_spfct_type
     !> global number of rows and cols and nz in the original sparse matrix
     integer                            :: m, n, nz
     !> Symmetry of the matrix/facto; can be 0 for unsym, >0 for sym
     integer                            :: sym
     !> an integer array containing control parameters
     !! expressed as an integer value. Undocumented entries are
     !! assigned as nicntl, nicntl-1 etc.
     !!
     !! Meaning:
     !! - icntl(qrm_ordering_)     : the ordering method to be used:
     !!             - qrm_auto_    : automatic choice
     !!             - qrm_natural_ : natural order
     !!             - qrm_given_   : given order
     !!             - qrm_colamd_  : COLAMD
     !!             - qrm_metis_   : METIS
     !!             - qrm_scotch_  : SCOTCH
     !! - icntl(qrm_minamalg_)     : minimum node size for amalgamation. Node i can be amalgamated
     !!               to its father iff they both have less than icntl(qrm_minamalg_) variables
     !!               and the fill in introduced is less than rcntl(1).
     !! - icntl(qrm_nb_)           : the block size to be used during factorization
     !! - inctl(qrm_keeph_)        : whether H (the Householder vectors) has to be stored or not
     !!             - -1  : both R and H are discarded (hidden feature, not documented)
     !!             -  0  : (qrm_no_) only H is discarded
     !!             -  1  : (qrm_yes_) nothing is discarded
     !! - inctl(qrm_ib_)           : the internal blocking size to reduce flops
     !! - icntl(qrm_rhsnb_)        : the blocking parameter to handle multiple right-hand-sides
     !! - icntl(qrm_nlz_)          : the minimum number of subtrees in L0 is this times the
     !!                              number of threads
     !! - icntl(qrm_cnode_)        : the number of cores per NUMA node (or cache, whatever)
     !! - icntl(qrm_bh_)           : in CA trees for the QR facto of a front, this is the number
     !!                              of leaves in each flat subtree
     integer                            :: icntl(nicntl)=0
     !> an float precision array containing control parameters
     !! expressed as a real value. Undocumented entries are
     !! assigned as nrcntl, nrcntl-1 etc.
     !!
     !! Meaning:
     !! - rcntl(qrm_amalgth_=1)      : fill-in threshold for amalgamation. Node i can be amalgamated
     !!                                to its father iff they both have less than icntl(3) variables
     !!                                and the fill in introduced is less than rcntl(1).
     !! - rcntl(qrm_rweight_=nrcntl) : subtrees with relative weight below this threashold will
     !!                                be put in L0
     !! - rcntl(qrm_mem_relax_=2)    : the relaxation parameter for the factorization memory
     !! - rcntl(qrm_rd_eps_=3)       : the threshold for detecting rank deficiencies
     real(r32)                          :: rcntl(nrcntl)=0.d0
     !> an array containing local global stats. some of its content will only
     !! be relevant on the master node. Meaning:
     !! - gstats(1) = estimated total number of flops
     !! - gstats(2) = estimated total number of nonzeroes in R (estimated)
     !! - gstats(3) = estimated total number of nonzeroes in H (estimated)
     !! - gstats(4) = actual total number of flops
     !! - gstats(5) = actual total number of nonzeroes in R (estimated)
     !! - gstats(6) = actual total number of nonzeroes in H (estimated)
     !! - gstats(7) = estimated peak memory consumption
     integer(i64)                       :: gstats(10)=0
     !> A pointer to an array containing a column permutation provided
     !! by the user
     integer, pointer,    dimension(:)  :: cperm_in => null()
     !> a  qrm_adata_mod::qrm_adata_type  data which is meant to
     !> contain all the data related to the analysis phase
     type(qrm_adata_type), allocatable  :: adata
     !> a  sqrm_fdata_mod::sqrm_fdata_type  data which is meant to
     !> contain all the data related to the factorization phase
     type(sqrm_fdata_type), allocatable :: fdata
  end type sqrm_spfct_type


  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! Subroutine interfaces
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  

  !>  Generif interface for the  ::sqrm_cntl_init  routine
  interface qrm_spfct_init
     module procedure sqrm_spfct_init
  end interface

  !>  Generif interface for the  ::sqrm_spfct_destroy  routine
  interface qrm_spfct_destroy
     module procedure sqrm_spfct_destroy
  end interface


  !>  Generic interface for the
  !!  sqrm_spfct_set_i4 and
  !!  sqrm_spfct_set_r4
  !!
  interface qrm_set
     module procedure sqrm_spfct_set_i4, sqrm_spfct_set_r4
  end interface

  !>  Generif interface for the
  !!  sqrm_spfct_get_i4,
  !!  sqrm_spfct_get_i8 and
  !!  sqrm_spfct_get_r4
  !!
  interface qrm_get
     module procedure sqrm_spfct_get_i4, sqrm_spfct_get_i8, sqrm_spfct_get_r4
  end interface

  interface qrm_sync_data
     module procedure :: sqrm_spfct_sync
  end interface qrm_sync_data

  interface qrm_spfct_sync
     module procedure  sqrm_spfct_sync
  end interface qrm_spfct_sync
  

contains

  !>  This subroutine initializes a qrm_spfct_type instance setting
  !! default values into the control parameters
  !!
  !! @param[in,out] qrm_spfct The factorization object to be initialized
  !!
  subroutine sqrm_spfct_init(qrm_spfct, qrm_spmat, info)

    use qrm_common_mod
    use qrm_parameters_mod
    use sqrm_spmat_mod
    implicit none
    type(sqrm_spfct_type), intent(inout) :: qrm_spfct
    type(sqrm_spmat_type), intent(inout) :: qrm_spmat
    integer, optional                    :: info

    ! error management
    integer                              :: err
    character(len=*), parameter          :: name='sqrm_spfct_init'

    err = 0
    
    ! set default values for icntl and rcntl
    qrm_spfct%icntl(qrm_ordering_)      = qrm_default_ordering_
    qrm_spfct%icntl(qrm_minamalg_)      = qrm_default_minamalg_
    qrm_spfct%icntl(qrm_mb_)            = qrm_default_mb_
    if(qrm_default_nb_.eq.qrm_noinit_) then
       qrm_spfct%icntl(qrm_nb_)         = qrm_default_mb_
    else
       qrm_spfct%icntl(qrm_nb_)         = min(qrm_default_nb_,qrm_default_mb_)
    end if

    qrm_spfct%icntl(qrm_ib_)            = min(qrm_default_ib_,qrm_spfct%icntl(qrm_nb_))
    ! find the first value higer than ib which is a divisor of nb
    do
       if(mod(qrm_spfct%icntl(qrm_nb_),qrm_spfct%icntl(qrm_ib_)).eq.0) exit
       qrm_spfct%icntl(qrm_ib_) = qrm_spfct%icntl(qrm_ib_)+1
    end do

    qrm_spfct%icntl(qrm_bh_)            = qrm_default_bh_
    qrm_spfct%icntl(qrm_keeph_)         = qrm_default_keeph_
    qrm_spfct%icntl(qrm_rhsnb_)         = qrm_default_rhsnb_
    qrm_spfct%icntl(qrm_nlz_)           = qrm_default_nlz_
    qrm_spfct%icntl(qrm_cnode_)         = qrm_default_cnode_
    qrm_spfct%icntl(qrm_sing_)          = qrm_default_sing_
    qrm_spfct%icntl(qrm_pinth_)         = qrm_default_pinth_

    qrm_spfct%rcntl(qrm_amalgth_)       = qrm_default_amalgth_
    qrm_spfct%rcntl(qrm_rweight_)       = qrm_default_rweight_
    qrm_spfct%rcntl(qrm_mem_relax_)     = qrm_default_mem_relax_
    qrm_spfct%rcntl(qrm_rd_eps_)        = qrm_default_rd_eps_

    ! qrm_spfct%m                         = qrm_spmat%m
    ! qrm_spfct%n                         = qrm_spmat%n
    qrm_spfct%sym                       = qrm_spmat%sym
    
    call sqrm_spfct_check(qrm_spfct, err)

    if(present(info)) info = err
    return

  end subroutine sqrm_spfct_init


  !>  This subroutine destroyes a qrm_spfct instance
  !!
  !! @param[in,out] qrm_spfct the factorization object to be destroyed
  !!
  subroutine sqrm_spfct_destroy(qrm_spfct, info)

    implicit none

    type(sqrm_spfct_type)       :: qrm_spfct
    integer, optional           :: info

    ! error management
    integer                     :: err
    character(len=*), parameter :: name='sqrm_spfct_cleanup'

    err = 0

    if(allocated(qrm_spfct%adata)) then
       call qrm_adata_destroy(qrm_spfct%adata, err)
       __QRM_INFO_CHECK(err, name, 'qrm_adata_destroy', 9999)
    end if
    if(allocated(qrm_spfct%fdata)) then
       call sqrm_fdata_destroy(qrm_spfct%fdata, err)
       __QRM_INFO_CHECK(err, name, 'qrm_fdata_destroy', 9999)
    end if

9999 continue

    if(present(info)) info = err
    return

  end subroutine sqrm_spfct_destroy



  ! The following subroutine set or get control parameters from the
  ! cntl or rcntl control arrays. All the set and get routines are
  ! gathered under the same, overloaded interface, respectively


  !>  This subroutine is meant to set the integer control parameters
  !!
  !! @param[in,out] qrm_spfct The qrm_spfct instance concerned by the setting
  !!
  !! @param[in] string a string describing the parameter to be
  !!            set. Accepted values are:
  !!            - "qrm_ordering" : to set a method for the fill-reducing
  !!                               column permutation. Accepted values are:
  !!                               - qrm_auto_=0    for automatic choice
  !!                               - qrm_natural_=1 for natural ordering
  !!                               - qrm_given_=2   for given ordering (through
  !!                                                the qrm_spfct%cperm_in pointer)
  !!                               - qrm_colamd_=3  for COLAMD
  !!                               - qrm_metis_=4   for METIS
  !!                               - qrm_scotch_=5  for SCOTCH
  !!            - "qrm_minamalg" : fronts whose size is smaller than this will be
  !!                               systematically amalgamated to their father
  !!            - "qrm_nb" : the block-size that defines the granularity of parallel tasks
  !!            - "qrm_ib" : the block-size for computations
  !!            - "qrm_rhsnb" : the block-size for grouping RHSs
  !!            - "qrm_keeph" : whether to store or not the Householder vectors. Accepted
  !!                            values are qrm_yes_ and qrm_no_
  !!            - "qrm_sing" : whether or not to detect the presence of singletons. Accepted
  !!                           values are qrm_yes_ and qrm_no_
  !!            - "qrm_nlz"  : the number of subtrees in L0 will be at least this times
  !!                           the number of threads
  !!
  !! @param[in] ival Any of the accepted values described above
  !!
  subroutine sqrm_spfct_set_i4(qrm_spfct, string, ival, info)
    use qrm_common_mod
    use qrm_string_mod
    use qrm_error_mod
    implicit none

    type(sqrm_spfct_type)           :: qrm_spfct
    character(len=*)                :: string
    integer                         :: ival
    integer, optional               :: info

    character(len=len(string))      :: istring

    ! error management
    integer                         :: err
    character(len=*), parameter     :: name='sqrm_spfct_set_i4'

    err = 0

    istring = qrm_str_tolower(string)
    if(index(istring,'qrm_ordering') .eq. 1) then
       qrm_spfct%icntl(qrm_ordering_) = ival
    else if (index(istring,'qrm_minamalg') .eq. 1) then
       qrm_spfct%icntl(qrm_minamalg_) = ival
    else if (index(istring,'qrm_mb') .eq. 1) then
       qrm_spfct%icntl(qrm_mb_) = ival
    else if (index(istring,'qrm_nb') .eq. 1) then
       qrm_spfct%icntl(qrm_nb_) = ival
    else if (index(istring,'qrm_ib') .eq. 1) then
       qrm_spfct%icntl(qrm_ib_) = ival
    else if (index(istring,'qrm_bh') .eq. 1) then
       qrm_spfct%icntl(qrm_bh_) = ival
    else if (index(istring,'qrm_rhsnb') .eq. 1) then
       qrm_spfct%icntl(qrm_rhsnb_) = ival
    else if (index(istring,'qrm_keeph') .eq. 1) then
       select case (ival)
       case (:-1)
          qrm_spfct%icntl(qrm_keeph_) = -1
       case (0)
          qrm_spfct%icntl(qrm_keeph_) = 0
       case (1:)
          qrm_spfct%icntl(qrm_keeph_) = 1
       end select
    else if (index(istring,'qrm_nlz') .eq. 1) then
       qrm_spfct%icntl(qrm_nlz_) = ival
    else if (index(istring,'qrm_pinth') .eq. 1) then
       qrm_spfct%icntl(qrm_pinth_) = ival
    else
       err = 23
       call qrm_error_print(err, name, aed=string)
       goto 9999
    end if

9999 continue
    if(present(info)) info = err
    return

  end subroutine sqrm_spfct_set_i4


  !>  This subroutine is meant to set the real control parameters
  !!
  !! @param[in,out] qrm_spfct The qrm_spfct instance concerned by the setting
  !!
  !! @param[in] string a string describing the parameter to be
  !!            set. Accepted values are:
  !!            - "qrm_amalgth" : the threshold that controls the amalgamation.
  !!                              A higher threshold means more fill-in but also
  !!                              more BLAS-3. Any real value is accepted
  !!
  !! @param[in] rval Any of the accepted values described above
  !!
  subroutine sqrm_spfct_set_r4(qrm_spfct, string, rval, info)

    use qrm_common_mod
    use qrm_string_mod
    use qrm_error_mod
    implicit none

    type(sqrm_spfct_type)           :: qrm_spfct
    character(len=*)                :: string
    real(r32)                       :: rval
    integer, optional               :: info

    character(len=len(string))      :: istring

    ! error management
    integer                         :: err
    character(len=*), parameter     :: name='sqrm_spfct_set_r4'

    err = 0

    istring = qrm_str_tolower(string)

    if(index(istring,'qrm_amalgth') .eq. 1) then
       qrm_spfct%rcntl(qrm_amalgth_) = rval
    else if(index(istring,'qrm_rweight') .eq. 1) then
       qrm_spfct%rcntl(qrm_rweight_) = rval
    else if(index(istring,'qrm_mem_relax') .eq. 1) then
       qrm_spfct%rcntl(qrm_mem_relax_) = rval
    else if(index(istring,'qrm_rd_eps') .eq. 1) then
       qrm_spfct%rcntl(qrm_rd_eps_) = rval
    else
       err = 23
       call qrm_error_print(err, name, aed=string)
       goto 9999
    end if

9999 continue

    if(present(info)) info = err
    return

  end subroutine sqrm_spfct_set_r4


  !>  Gets the values of an integer control parameter.
  !!        This is the dual of the  ::sqrm_pseti 
  !!        routine; the parameters and accepted values are the same.
  subroutine sqrm_spfct_get_i4(qrm_spfct, string, ival, info)
    use qrm_common_mod
    use qrm_string_mod
    use qrm_error_mod
    implicit none

    type(sqrm_spfct_type)           :: qrm_spfct
    character(len=*)                :: string
    integer                         :: ival
    integer, optional               :: info

    character(len=len(string))      :: istring
    integer(kind=8)                 :: iival

    ! error management
    integer                         :: err
    character(len=*), parameter     :: name='sqrm_spfct_get_i4'

    call sqrm_spfct_get_i8(qrm_spfct, string, iival, info)

    ival = iival

    return
    
  end subroutine sqrm_spfct_get_i4
  
  !>  Gets the values of an integer control parameter.
  !!        This is the dual of the  ::sqrm_pseti 
  !!        routine; the parameters and accepted values are the same.
  subroutine sqrm_spfct_get_i8(qrm_spfct, string, ival, info)
    use qrm_common_mod
    use qrm_string_mod
    use qrm_error_mod
    implicit none

    type(sqrm_spfct_type)           :: qrm_spfct
    character(len=*  )              :: string
    integer(kind=8)                 :: ival
    integer, optional               :: info

    character(len=len(string))      :: istring

    ! error management
    integer                         :: err
    character(len=*), parameter     :: name='sqrm_spfct_get_i8'

    err = 0

    istring = qrm_str_tolower(string)

    if(index(istring,'qrm_ordering') .eq. 1) then
       ival = qrm_spfct%icntl(qrm_ordering_)
    else if (index(istring,'qrm_minamalg') .eq. 1) then
       ival = qrm_spfct%icntl(qrm_minamalg_)
    else if (index(istring,'qrm_nb') .eq. 1) then
       ival = qrm_spfct%icntl(qrm_nb_)
    else if (index(istring,'qrm_mb') .eq. 1) then
       ival = qrm_spfct%icntl(qrm_mb_)
    else if (index(istring,'qrm_ib') .eq. 1) then
       ival = qrm_spfct%icntl(qrm_ib_)
    else if (index(istring,'qrm_bh') .eq. 1) then
       ival = qrm_spfct%icntl(qrm_bh_)
    else if (index(istring,'qrm_rhsnb') .eq. 1) then
       ival = qrm_spfct%icntl(qrm_rhsnb_)
    else if (index(istring,'qrm_keeph') .eq. 1) then
       ival = qrm_spfct%icntl(qrm_keeph_)
    else if (index(istring,'qrm_nlz') .eq. 1) then
       ival = qrm_spfct%icntl(qrm_nlz_)
    else if (index(istring,'qrm_pinth') .eq. 1) then
       ival = qrm_spfct%icntl(qrm_pinth_)
    else if (index(istring,'qrm_e_nnz_r') .eq. 1) then
       ival = qrm_spfct%gstats(qrm_e_nnz_r_)
    else if (index(istring,'qrm_e_nnz_h') .eq. 1) then
       ival = qrm_spfct%gstats(qrm_e_nnz_h_)
    else if (index(istring,'qrm_e_facto_flops') .eq. 1) then
       ival = qrm_spfct%gstats(qrm_e_facto_flops_)
    else if (index(istring,'qrm_e_facto_mempeak') .eq. 1) then
       ival = qrm_spfct%gstats(qrm_e_facto_mempeak_)
    else if (index(istring,'qrm_nnz_r') .eq. 1) then
       ival = qrm_spfct%gstats(qrm_nnz_r_)
    else if (index(istring,'qrm_nnz_h') .eq. 1) then
       ival = qrm_spfct%gstats(qrm_nnz_h_)
    else if (index(istring,'qrm_facto_flops') .eq. 1) then
       ival = qrm_spfct%gstats(qrm_facto_flops_)
    else if (index(istring,'qrm_rd_num') .eq. 1) then
       ival = qrm_spfct%gstats(qrm_rd_num_)
    else
       err = 23
       call qrm_error_print(err, name, aed=string)
       goto 9999
    end if

9999 continue

    if(present(info)) info = err
    return

  end subroutine sqrm_spfct_get_i8



  !>  Gets the values of a real control parameter.
  !!        This is the dual of the  ::sqrm_psetr 
  !!        routine; the parameters and accepted values are the same.
  !!
  subroutine sqrm_spfct_get_r4(qrm_spfct, string, rval, info)

    use qrm_common_mod
    use qrm_string_mod
    use qrm_error_mod
    implicit none

    type(sqrm_spfct_type)           :: qrm_spfct
    character(len=*)                :: string
    real(r32)                       :: rval
    integer, optional               :: info

    character(len=len(string))      :: istring

    ! error management
    integer                         :: err
    character(len=*), parameter     :: name='sqrm_spfct_get_r4'

    err = 0

    istring = qrm_str_tolower(string)

    if(index(istring,'qrm_amalgth') .eq. 1) then
       rval = qrm_spfct%rcntl(qrm_amalgth_)
    else if(index(istring,'qrm_rweight') .eq. 1) then
       rval = qrm_spfct%rcntl(qrm_rweight_)
    else if(index(istring,'qrm_mem_relax') .eq. 1) then
       rval = qrm_spfct%rcntl(qrm_mem_relax_)
    else if(index(istring,'qrm_rd_eps') .eq. 1) then
       rval = qrm_spfct%rcntl(qrm_rd_eps_)
    else
       err = 1
       call qrm_error_print(err, name, aed=string)
       goto 9999
    end if

9999 continue

    if(present(info)) info = err
    return

  end subroutine sqrm_spfct_get_r4


  !>  Check the compatibility and correctness of icntl and rcntl
  !! parameters
  subroutine sqrm_spfct_check(qrm_spfct, op, info)
    use qrm_common_mod
    use qrm_string_mod
    use qrm_error_mod
    use qrm_parameters_mod
    implicit none

    type(sqrm_spfct_type)           :: qrm_spfct
    integer, optional               :: op
    integer, optional               :: info

    integer                         :: iop
    integer, allocatable            :: ied(:)
    
    ! error management
    integer                         :: err
    character(len=*), parameter     :: name='sqrm_spfct_check'

    err = 0

    if(present(op)) then
       iop = op
    else
       iop = qrm_allop_
    end if

    if((iop.eq.qrm_allop_) .or. (iop.eq.qrm_analyse_)) then

       ! all the potential cases of conflict with the orderings
       select case(qrm_spfct%icntl(qrm_ordering_))
       case(:-1,6:)
          __QRM_INFO_SET_I(err, 9, name, (/qrm_spfct%icntl(qrm_ordering_),0,0,0,0/), 9999)
       case (qrm_given_)
          if(qrm_spfct%icntl(qrm_sing_) .eq. qrm_yes_) then
             __QRM_INFO_SET_I(err, 27, name, (/qrm_ordering_,qrm_sing_/), 9999)
          end if
       end select

       if(qrm_spfct%sym.eq.0) then
          ! all the potential cases of conflict with the orderings
          select case(qrm_spfct%icntl(qrm_nb_))
          case(:-1)
             ied = (/qrm_spfct%icntl(qrm_mb_), &
                  qrm_spfct%icntl(qrm_nb_),qrm_spfct%icntl(qrm_ib_)/)
             __QRM_INFO_SET_I(err, 28, name, ied, 9999)
          case default
             if(qrm_spfct%icntl(qrm_nb_) .lt. qrm_spfct%icntl(qrm_ib_)) then
                __QRM_INFO_SET_I(err, 27, name, (/qrm_nb_,qrm_ib_/), 9999)
             end if
          end select

          if(  (mod(qrm_spfct%icntl(qrm_mb_),qrm_spfct%icntl(qrm_nb_)).gt.0) .or. &
               (mod(qrm_spfct%icntl(qrm_nb_),qrm_spfct%icntl(qrm_ib_)).gt.0) .or. &
               ( (qrm_spfct%icntl(qrm_mb_).gt.0) .and. (qrm_spfct%icntl(qrm_mb_) .lt. qrm_spfct%icntl(qrm_nb_)))) then
             ied = (/qrm_spfct%icntl(qrm_mb_), &
                  qrm_spfct%icntl(qrm_nb_),qrm_spfct%icntl(qrm_ib_)/)
             __QRM_INFO_SET_I(err, 28, name, ied, 9999)
          end if
          select case(qrm_spfct%icntl(qrm_ib_))
          case(:-1)
             ied=(/qrm_spfct%icntl(qrm_ib_)/)
             __QRM_INFO_SET_I(err, 28, name, ied, 9999)
          end select
       end if
    end if

9999 continue

    if(present(info)) info = err
    return

  end subroutine sqrm_spfct_check


  subroutine sqrm_spfct_sync(qrm_spfct)
    use qrm_common_mod
    use qrm_error_mod
    use sqrm_fdata_mod
    use qrm_mem_mod
#if defined(have_starpu)
    use fstarpu_mod
#endif
    implicit none

    type(sqrm_spfct_type), target  :: qrm_spfct

    type(sqrm_front_type), pointer :: front
    type(qrm_adata_type), pointer  :: adata
    type(sqrm_fdata_type), pointer :: fdata

    integer :: inode, node

    adata => qrm_spfct%adata
    fdata => qrm_spfct%fdata

#if defined(have_starpu)
    if(allocated(qrm_spfct%adata)) then
       if(c_associated(adata%hdl)) then
          call fstarpu_data_acquire(adata%hdl, FSTARPU_R)
          call fstarpu_data_release(adata%hdl)
       end if
    end if

    if(allocated(qrm_spfct%fdata)) then
       if(allocated(fdata%front_list)) then
          do inode = 1, adata%nnodes
             node = adata%torder(inode)
             if(adata%small(node) .ge. 0) then
                front => fdata%front_list(node)
                if(c_associated(front%sym_handle)) then
                   call fstarpu_data_acquire(front%sym_handle, FSTARPU_R)
                   call fstarpu_data_release(front%sym_handle)
                end if
             end if
          end do
       end if
    end if
#endif
    return

  end subroutine sqrm_spfct_sync
  

  
end module sqrm_spfct_mod
