!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!!
!! This file contains a module with generic interfaces for a number of auxiliary tools
!!
!! ##############################################################################################


!> This module contains generic interfaces for a number of auxiliary tools
module sqrm_utils_mod
  use iso_c_binding

  !> Generic interface for the @link ::sqrm_build_laplace7d @endlink
  !> routine
  interface qrm_matgen_7pt
     subroutine sqrm_matgen_7pt(qrm_spmat, nx, ny, nz, sym)
       use sqrm_spmat_mod
       integer               :: nx, ny, nz
       type(sqrm_spmat_type) :: qrm_spmat
       integer, optional     :: sym
     end subroutine sqrm_matgen_7pt
  end interface qrm_matgen_7pt
  
  !> Generic interface for the @link ::sqrm_build_laplace7d @endlink
  !> routine
  interface qrm_matgen_mitt3d
     subroutine sqrm_matgen_mitt3d(qrm_spmat, n)
       use sqrm_spmat_mod
       integer               :: n
       type(sqrm_spmat_type) :: qrm_spmat
     end subroutine sqrm_matgen_mitt3d
  end interface qrm_matgen_mitt3d

  ! Generic interface for the sqrm_tikhonov routine
  interface qrm_tikhonov
     subroutine sqrm_tikhonov(qrm_spmat, gamma)
       use sqrm_spmat_mod
       type(sqrm_spmat_type)  :: qrm_spmat
       real(r32)              :: gamma
     end subroutine sqrm_tikhonov
  end interface qrm_tikhonov

  !> Generic interface for the @link ::sqrm_readmat @endlink
  !> routine
  interface qrm_readmat
     subroutine sqrm_readmat(matfile, qrm_mat, fakec, info)
       use sqrm_spmat_mod
       character, intent(in)                :: matfile*30
       type(sqrm_spmat_type), intent(inout) :: qrm_mat
       logical, optional                    :: fakec
       integer, optional                    :: info
     end subroutine sqrm_readmat
  end interface qrm_readmat


  interface qrm_writemat
     subroutine sqrm_writemat(matfile, qrm_mat, info)
       use sqrm_spmat_mod
       character(len=*), intent(in)              :: matfile
       type(sqrm_spmat_type), intent(inout)      :: qrm_mat
       integer, optional                         :: info
     end subroutine sqrm_writemat
  end interface qrm_writemat
  
  !> Generic interface for the @link ::sqrm_vecnrm2d @endlink
  !> and @link ::sqrm_vecnrm1d @endlink routines
  interface qrm_vecnrm
     procedure :: sqrm_vecnrm2d, sqrm_vecnrm1d
  end interface qrm_vecnrm

  !> Generic interface for the @link ::sqrm_vecnrm2d @endlink
  !> and @link ::sqrm_vecnrm1d @endlink routines
  interface sqrm_vecnrm
     subroutine sqrm_vecnrm2d(vec, n, ntype, nrm, info)
       use qrm_parameters_mod
       real(r32), intent(in)  :: vec(:,:)
       real(r32)              :: nrm(:)
       integer, intent(in)    :: n
       character              :: ntype
       integer, optional      :: info
     end subroutine sqrm_vecnrm2d
     subroutine sqrm_vecnrm1d(vec, n, ntype, nrm, info)
       use qrm_parameters_mod
       real(r32), intent(in)  :: vec(:)
       real(r32)              :: nrm
       integer, intent(in)    :: n
       character              :: ntype
       integer, optional      :: info
     end subroutine sqrm_vecnrm1d
  end interface sqrm_vecnrm

  !> Generic interface for the @link ::sqrm_remap_pnt @endlink
  !> routine
  interface qrm_remap_pnt
     subroutine sqrm_remap_pnt(arr1d, pnt2d, n)
       use qrm_parameters_mod
       integer :: n
       real(r32), target  :: arr1d(1:n)
       real(r32), pointer :: pnt2d(:,:)
     end subroutine sqrm_remap_pnt
  end interface

  interface qrm_get_front_mem
     module procedure sqrm_get_front_mem_unsym, sqrm_get_front_mem_sym
  end interface qrm_get_front_mem
  
  interface qrm_get_front_flops
     module procedure sqrm_get_front_flops
  end interface qrm_get_front_flops
  

contains

  subroutine sqrm_get_front_mem_unsym(m, n, mb, nb, ib, bh, np, amem, cmem, hsize, rsize, storer, storeh, stair)
    use qrm_adata_mod
    use qrm_common_mod
    use qrm_parameters_mod
    implicit none

    integer              :: m, n, mb, nb, ib, np, bh
    ! if stair is not present we assume that the node is small
    integer, optional    :: stair(:)
    integer(kind=8)      :: amem, cmem, hsize, rsize
    logical              :: storer, storeh
    
    integer(kind=8)      :: ne, cm, cn, cne, cnp, c, ibh, iib, imb, inb
    integer(kind=8)      :: nn, mm, i, j, k, nc, nr

    if(mb.lt.0) then
       imb = m
    else
       imb = min(mb,m)
    end if
    ne = min(m,n)
    inb = min(nb,n)
    iib = min(int(ib,kind=8),int(ne,kind=8))
    ibh = bh
        
    hsize = 0
    rsize = 0
    amem  = 0
    cmem  = 0

    if(min(m,n).le.0) return

    nc = (n-1)/inb + 1
    nr = (m-1)/imb + 1
    if(ibh.le.0) ibh = nr
    
    ! count only real/complex data for the moment
    ! copy/paste the allocation loop in qrm_activate_front

    cols: do j=1, nc
       nn = min(inb,int(n-(j-1)*inb,kind=8))

       if(present(stair)) then
          if((j-1)*inb .lt. ne) hsize = hsize + nn*(stair((j-1)*inb+nn)-(j-1)*inb)
       else
          if((j-1)*inb .lt. ne) hsize = hsize + nn*(m-(j-1)*inb)
       end if
       rsize = rsize + nn*min(int(np,kind=8),(j-1)*inb+nn)

       rows: do i=1, nr
          if(present(stair)) then
             mm = min(int(stair(min(j*inb,int(n,kind=8))) - (i-1)*imb,kind=8), imb)
          else
             mm = min(m - (i-1)*imb, imb)
          end if
          if(mm .le. 0) cycle cols
          amem = amem + mm*nn
          if((i-1)*imb+mm .ge. (j-1)*inb+1) then
             k = ((j-1)*inb)/imb+1
             if(mod(i-k,ibh).eq.0) then
                ! For those blocks where _geqrt is done, after the
                ! _geqrt operation the v vectors are copied into the
                ! front%v block along with the T matrices. This trick
                ! allows us to remove a false dependency between _gemqrt
                ! and _tpqrt tasks. Padding is added to avoid TRSMs on
                ! the GPU and thus the T matrices are literally stored
                ! on top of the V blocks:
                !   |TTTT|
                !   |V000|
                !   |VV00|
                !   |VVV0|
                !   |VVVV|
                hsize = hsize + iib*nn
                if(.not.present(stair)) then
                   mm = iib
                else
                   mm = mm - max((j-1)*inb +1 - (i-1)*imb,1_8)+1 + iib
                end if
                amem = amem + mm*nn
             end if
             if(present(stair))  then
                hsize = hsize + iib*nn
                amem = amem + iib*nn
             end if
          end if
       end do rows
    end do cols

    if(storeh.and.storer) then
       cmem = amem - (hsize+rsize)
    else if (storer) then
       cmem = amem - rsize
    else
       cmem = amem
    end if
    
    ! add one temp block for the clean
    amem = amem+2*imb*imb
    ! add one temp block for the clean
    cmem = cmem+2*imb*imb

    ! hsize and rsize are expressed in # of entries whereas
    ! amem and cmem are expressed in Bytes
    amem = amem*qrm_sizeof_s
    cmem = cmem*qrm_sizeof_s

    ! cols
    amem = amem+n*qrm_sizeof_i
    ! rows
    amem = amem+m*qrm_sizeof_i
    ! stair
    amem = amem+(n+1)*qrm_sizeof_i
    ! colmap
    amem = amem+3*n*qrm_sizeof_i
    ! rowmap
    amem = amem+3*ne*qrm_sizeof_i

    if(.not.storer) then
       ! cols
       cmem = cmem+n*qrm_sizeof_i
       ! rows
       cmem = cmem+m*qrm_sizeof_i
       ! stair
       cmem = cmem+(n+1)*qrm_sizeof_i
       ! colmap
       cmem = cmem+3*n*qrm_sizeof_i
       ! rowmap
       cmem = cmem+3*ne*qrm_sizeof_i
    end if
    
    ! the amount of memory freed by the clean routine may be
    ! negaitve. This is due to the fact that there may be some overlap
    ! between the H and R blocks. This is just a quick fix.
    if(cmem.lt.0) then
       amem = amem-cmem
       cmem = 0
    end if

    return
  end subroutine sqrm_get_front_mem_unsym

  subroutine sqrm_get_front_mem_sym(m, mb, np, amem, cmem, rsize, storer)
    use qrm_adata_mod
    use qrm_common_mod
    use qrm_parameters_mod
    implicit none

    integer              :: m, mb, np
    ! if stair is not present we assume that the node is small
    integer(kind=8)      :: amem, cmem, rsize
    logical              :: storer
    
    integer(kind=8)      :: imb
    integer(kind=8)      :: nn, mm, i, j, k, nc, nr

    if(mb.lt.0) then
       imb = m
    else
       imb = min(mb,m)
    end if
        
    rsize = 0
    amem  = 0
    cmem  = 0

    if(m.le.0) return

    nr = (m-1)/imb + 1
    
    ! count only real/complex data for the moment
    ! copy/paste the allocation loop in qrm_activate_front

    rows: do i=1, nr
       mm = min(m - (i-1)*imb, imb)

       amem = amem + mm*(m - (i-1)*imb)

       if(i.le.((np-1)/imb+1)) then
          nn = min(np - (i-1)*imb, imb)
          rsize = rsize + nn*(m - (i-1)*imb)
       end if
    end do rows

    if (storer) then
       cmem = amem - rsize
    else
       cmem = amem
    end if
    
    ! add one temp block for the clean
    amem = amem+2*imb*imb
    ! add one temp block for the clean
    cmem = cmem+2*imb*imb

    ! rsize is expressed in # of entries whereas
    ! amem and cmem are expressed in Bytes
    amem = amem*qrm_sizeof_s
    cmem = cmem*qrm_sizeof_s

    ! rows
    amem = amem+m*qrm_sizeof_i
    ! rowmap
    amem = amem+3*m*qrm_sizeof_i
    ! colmap
    amem = amem+3*m*qrm_sizeof_i

    if(.not.storer) then
       ! cols
       cmem = cmem+m*qrm_sizeof_i
       ! rowmap
       cmem = cmem+3*m*qrm_sizeof_i
       ! colmap
       cmem = cmem+3*m*qrm_sizeof_i
    end if
    
    ! the amount of memory freed by the clean routine may be
    ! negaitve. This is due to the fact that there may be some overlap
    ! between the H and R blocks. This is just a quick fix.
    if(cmem.lt.0) then
       amem = amem-cmem
       cmem = 0
    end if

    return
  end subroutine sqrm_get_front_mem_sym



  subroutine sqrm_get_front_flops(m, n, stair, mb, nb, ib, flops)
    use qrm_common_mod
    implicit none
    integer :: m, n, mb, nb, ib
    integer :: stair(:)
    integer(kind=8) :: flops

    integer :: j, ne
    integer :: nn, mm, kk

    ne = min(m,n)
    flops = 0

    if(min(m,n).le.0) return
        
    do j=1, ne, ib
       kk = min(ib,ne-j+1)
       mm = stair(j+kk-1) - j+1
       flops = flops + qrm_count_realflops(mm, kk, kk, 'panel')
       nn = n - (j+kk) + 1
       if(nn.gt.0) flops = flops + qrm_count_realflops(mm, nn, kk, 'update')
    end do

  end subroutine sqrm_get_front_flops


  function qrm_compute_task_flops(optype, m, n, k, l, stair, ofs)
    implicit none

    integer         :: m, n, k, l, ofs, stair(:)
    character       :: optype*(*)
    integer(kind=8) :: qrm_compute_task_flops
    
    integer         :: i, j, im, in
    integer(kind=8) :: flops

    flops = 0

    select case(optype)
    case ('geqrt')
       i = ofs
       do j=1, min(n,m-ofs+1)
          if(stair(1).lt.0) then
             im = m-i+1
          else
             im = min(stair(j),m)-i+1
          end if
          if(im.le.0) cycle
          in = n-j
          flops = flops + im*(3+4*in)
          i = i+1
       end do
    case ('gemqrt')
       i = ofs
       do j=1, min(m-ofs+1,k)
          if(stair(1).lt.0) then
             im = m-i+1
          else
             im = min(stair(j),m)-i+1
          end if
          if(im.le.0) cycle
          flops = flops + im*4*n
          i = i+1
       end do
    case ('tpqrt')
       im = 0
       i  = 1
       do j=1, n
          if(stair(1) .lt. 0) then
             ! in this case we can have any l
             if(l.eq.0) then
                im = m
             else
                im = m-l+min(j,l)
             end if
          else
             ! In this case the block is traversed by stair and tuhs we
             ! only hande the case l==0
             im = min(stair(j),m)
          end if
          in = n-j
          flops = flops + (im+1)*(3+4*in)
       end do
    case ('tpmqrt')
       im = 0
       do j=1, k
          if(stair(1) .lt. 0) then
             ! in this case we can have any l
             if(l.eq.0) then
                im = m
             else
                im = m-l+min(j,l)
             end if
          else
             ! In this case the block is traversed by stair and tuhs we
             ! only hande the case l==0
             im = min(stair(j),m)
          end if
          flops = flops + (im+1)*4*n
       end do
    end select

    if(flops.lt.0) write(*,*)'Error in flopcount'

    qrm_compute_task_flops = flops
    
    return
  end function qrm_compute_task_flops


  function qrm_compute_task_flops_ib(optype, m, n, k, l, ib, stair, ofs)
    implicit none

    integer         :: m, n, k, l, ofs, ib, stair(:)
    character       :: optype*(*)
    integer(kind=8) :: qrm_compute_task_flops_ib
    
    integer         :: i, j
    integer(kind=8) :: flops, im, in, ik, il, fp, ft, fu

    flops = 0
    fp = 0
    ft = 0
    fu = 0
    
    select case(optype)
    case ('geqrt')
       i = ofs
       do j=1, min(n,m-ofs+1), ib
          ik = min(ib, min(n,m-ofs+1)-j+1)
          if(stair(1).lt.0) then
             im = m-i+1
          else
             im = min(stair(j+ik-1),m)-i+1
          end if
          if(im.le.0) cycle
          in = n-(j+ik-1)
          ! Panel
          if(im.gt.ik) then
             flops = flops + 2*ik*ik*int(real(im)-real(ik)/3.0)
             fp = fp + 2*ik*ik*int(real(im)-real(ik)/3.0)
          else
             flops = flops + 2*im*im*int(real(ik)-real(im)/3.0)
             fp = fp + 2*im*im*int(real(ik)-real(im)/3.0)
          end if
          ! compute T
          flops = flops + im*(min(im,ik)**2)
          ft = ft + im*(min(im,ik)**2)
          !update
          flops = flops + 4*im*in*ik - ik*ik*in
          fu = fu + 4*im*in*ik - ik*ik*in
          i = i+ik
       end do
       ! write(*,'(" ")')
       ! write(*,'(i10,2x,i10,2x,i10)')fp,ft,fu
    case ('gemqrt')
       i = ofs
       in = n
       do j=1, min(k,m-ofs+1), ib
          ik = min(ib, min(k,m-ofs+1)-j+1)
          if(stair(1).lt.0) then
             im = m-i+1
          else
             im = min(stair(j+ik-1),m)-i+1
          end if
          if(im.le.0) cycle
          !update
          flops = flops + 4*im*in*ik - ik*ik*in
          i = i+ik
       end do
    case ('tpqrt')
       im = 0
       i  = 1
       do j=1, n, ib
          ik = min(ib, n-j+1)
          if(stair(1) .lt. 0) then
             ! in this case we can have any l
             if(l.eq.0) then
                im = m
                il = 0
             else
                im = m-l+min(j+ik-1,l)
                il = im - (m-l+min(j,l))
             end if
          else
             ! In this case the block is traversed by stair and thus we
             ! only hande the case l==0
             im = min(stair(j+ik-1),m)
             il = 0
          end if
          in = n-(j+ik-1)
          ! panel
          flops = flops + 2*im*ik*ik - (il**3)
          fp = fp + 2*im*ik*ik - (il**3)
          ! T
          flops = flops + im*ik*ik+(ik**3)/3 - (il**3)
          ft = ft + im*ik*ik+(ik**3)/3 - (il**3)
          ! update
          flops = flops + 4*im*in*ik + ik*ik*in - 2*(il**3)
          fu = fu + 4*im*in*ik + ik*ik*in - 2*(il**3)
       end do
       ! write(*,'(" ")')
       ! write(*,'(i10,2x,i10,2x,i10)')fp,ft,fu
    case ('tpmqrt')
       im = 0
       do j=1, k, ib
          ik = min(ib, k-j+1)
          if(stair(1) .lt. 0) then
             ! in this case we can have any l
             if(l.eq.0) then
                im = m
                il = 0
             else
                im = m-l+min(j+ik-1,l)
                il = im - (m-l+min(j,l))
             end if
          else
             ! In this case the block is traversed by stair and thus we
             ! only hande the case l==0
             im = min(stair(j+ik-1),m)
             il = 0
          end if
          in = n
          ! update
          flops = flops + 4*im*in*ik + ik*ik*in - 2*(il**3)
       end do
    end select

    if(flops.lt.0) write(*,*)'Error in flopcount'

    qrm_compute_task_flops_ib = flops
    
    return
  end function qrm_compute_task_flops_ib

  
end module sqrm_utils_mod







