!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_cintface.F90
!! This file contains the C interface for qr_mumps.
!!
!! Author: Alfredo Buttari
!!
!! ##############################################################################################


#include "qrm_common.h"

!> This module contains the definition of the qr_mumps C interface.
module zqrm_c_interface
  use iso_c_binding
  use zqrm_spmat_mod
  use qrm_string_mod
  
  !> This is the main qr_mumps data type which is meant to hold
  !! all the information related to a matrix. It is equivalent to the
  !! zqrm_spmat_mod Fortran type.
  type, bind(c) :: zqrm_spmat_type_c
     !> This array contains the list of row indices of the nonzeroes
     !! in the matrix.
     type(c_ptr)      :: irn
     !> This array contains the list of column indices of the nonzeroes
     !! in the matrix.
     type(c_ptr)      :: jcn
     !> This array contains the list of values of the nonzeroes
     !! in the matrix.
     type(c_ptr)      :: val
     !> The number of rows in the matrix.
     integer(c_int)   :: m
     !> The number of columns in the matrix.
     integer(c_int)   :: n
     !> The number of nonzeroes in the matrix.
     integer(c_int)   :: nz
     !> The symmetry of the matrix
     integer(c_int)   :: sym
     !> The handle to the zqrm_spmat_mod instance which will be used
     !! internally by qr_mumps
     type(c_ptr)      :: h
  end type zqrm_spmat_type_c
  
  !> This is the main qr_mumps facto data type which is meant to hold
  !! all the information related to a facto. It is equivalent to the
  !! zqrm_spfct_mod Fortran type.
  type, bind(c) :: zqrm_spfct_type_c
     integer(c_int)        :: m
     integer(c_int)        :: n
     integer(c_int)        :: nz
     integer(c_int)        :: sym
     !> A pointer to an array containing a column permutation provided
     !! by the user
     type(c_ptr)           :: cperm_in
     !> The integer control parameters
     integer(c_int)        :: icntl(20)
     !> The real control parameters
     real(c_float)         :: rcntl(10)
     !> The stats resulting from various operations
     integer(c_long_long)  :: gstats(10)
     !> The handle to the zqrm_spmat_mod
     !! instance which will be used internally by qr_mumps
     type(c_ptr)           :: h
  end type zqrm_spfct_type_c
  
contains

  !> C equivalent of the zqrm_spmat_init routine
  function zqrm_spmat_init_c(qrm_spmat_c) result(info) bind(c)
    use zqrm_spmat_mod
    implicit none

    type(zqrm_spmat_type_c)        :: qrm_spmat_c
    integer(c_int)                 :: info
    type(zqrm_spmat_type), pointer :: fmat

    allocate(fmat)
    
    call zqrm_spmat_init(fmat, info)
    qrm_spmat_c%h      = c_loc(fmat)
    nullify(fmat)
    
    return

  end function zqrm_spmat_init_c


  !> C equivalent of the zqrm_spmat_destroy routine
  function zqrm_spmat_destroy_c(qrm_spmat_c) result(info) bind(c)
    use zqrm_spmat_mod
    implicit none
    type(zqrm_spmat_type_c)        :: qrm_spmat_c
    integer(c_int)                 :: info

    type(zqrm_spmat_type), pointer :: fmat

    call c_f_pointer(qrm_spmat_c%h, fmat)
    call zqrm_spmat_destroy(fmat, info=info)

    deallocate(fmat)
    qrm_spmat_c%h      = c_null_ptr

    return

  end function zqrm_spmat_destroy_c

  
  !> C equivalent of the zqrm_spmat_init routine
  function zqrm_spfct_init_c(qrm_spfct_c, qrm_spmat_c) result(info) bind(c)
    use zqrm_spfct_mod
    implicit none

    type(zqrm_spfct_type_c)        :: qrm_spfct_c
    type(zqrm_spmat_type_c)        :: qrm_spmat_c
    integer(c_int)                 :: info
    type(zqrm_spfct_type), pointer :: ffct
    type(zqrm_spmat_type), pointer :: fmat

    allocate(ffct)
    call c_f_pointer(qrm_spmat_c%h, fmat)

    fmat%m   = qrm_spmat_c%m   
    fmat%n   = qrm_spmat_c%n   
    fmat%nz  = qrm_spmat_c%nz  
    fmat%sym = qrm_spmat_c%sym
    
    call zqrm_spfct_init(ffct, fmat, info)
    
    qrm_spfct_c%m      = ffct%m
    qrm_spfct_c%n      = ffct%n
    qrm_spfct_c%nz     = ffct%nz
    qrm_spfct_c%sym    = ffct%sym
    qrm_spfct_c%icntl  = ffct%icntl
    qrm_spfct_c%rcntl  = ffct%rcntl
    qrm_spfct_c%gstats = ffct%gstats

    qrm_spfct_c%h      = c_loc(ffct)
    nullify(ffct)
    
    return

  end function zqrm_spfct_init_c


  !> C equivalent of the zqrm_spmat_destroy routine
  function zqrm_spfct_destroy_c(qrm_spfct_c) result(info) bind(c)
    use zqrm_spfct_mod
    implicit none
    type(zqrm_spfct_type_c)        :: qrm_spfct_c
    integer(c_int)                 :: info

    type(zqrm_spfct_type), pointer :: ffct

    call c_f_pointer(qrm_spfct_c%h, ffct)
    call zqrm_spfct_destroy(ffct, info=info)

    qrm_spfct_c%m      = 0
    qrm_spfct_c%n      = 0
    qrm_spfct_c%nz     = 0
    qrm_spfct_c%sym    = 0
    qrm_spfct_c%icntl  = ffct%icntl
    qrm_spfct_c%rcntl  = ffct%rcntl
    qrm_spfct_c%gstats = ffct%gstats

    deallocate(ffct)
    qrm_spfct_c%h      = c_null_ptr

    return

  end function zqrm_spfct_destroy_c


  !> C equivalent of the zqrm_analyse routine
  function zqrm_analyse_c(qrm_spmat_c, qrm_spfct_c, transp) result(info) bind(c)
    use zqrm_analysis_mod
    use zqrm_spfct_mod
    implicit none
    type(zqrm_spmat_type_c)        :: qrm_spmat_c
    type(zqrm_spfct_type_c)        :: qrm_spfct_c
    character(kind=c_char), value  :: transp
    integer(c_int)                 :: info

    type(zqrm_spmat_type), pointer :: fmat
    type(zqrm_spfct_type), pointer :: ffct

    call c_f_pointer(qrm_spmat_c%h, fmat)
    call c_f_pointer(qrm_spfct_c%h, ffct)

    fmat%m   = qrm_spmat_c%m   
    fmat%n   = qrm_spmat_c%n   
    fmat%nz  = qrm_spmat_c%nz  
    fmat%sym = qrm_spmat_c%sym

    call c_f_pointer(qrm_spmat_c%irn     , fmat%irn     ,(/qrm_spmat_c%nz/))
    call c_f_pointer(qrm_spmat_c%jcn     , fmat%jcn     ,(/qrm_spmat_c%nz/))
    call c_f_pointer(qrm_spmat_c%val     , fmat%val     ,(/qrm_spmat_c%nz/))
    call c_f_pointer(qrm_spfct_c%cperm_in, ffct%cperm_in,(/qrm_spmat_c%n/))

    ffct%m      = qrm_spfct_c%m   
    ffct%n      = qrm_spfct_c%n   
    ffct%nz     = qrm_spfct_c%nz  
    ffct%sym    = qrm_spfct_c%sym 
    ffct%icntl  = qrm_spfct_c%icntl 
    ffct%rcntl  = qrm_spfct_c%rcntl 

    call zqrm_analyse(fmat, ffct, transp, info)

    qrm_spfct_c%m  =  ffct%m      
    qrm_spfct_c%n  =  ffct%n      
    qrm_spfct_c%nz =  ffct%nz     

    qrm_spfct_c%gstats = ffct%gstats

    return
  end function zqrm_analyse_c

  !> C equivalent of the zqrm_factorize routine
  function zqrm_factorize_c(qrm_spmat_c, qrm_spfct_c, transp) result(info) bind(c)
    use zqrm_factorization_mod 
    use zqrm_spfct_mod
    implicit none

    type(zqrm_spmat_type_c)        :: qrm_spmat_c
    type(zqrm_spfct_type_c)        :: qrm_spfct_c
    character(kind=c_char), value  :: transp
    integer(c_int)                 :: info

    type(zqrm_spmat_type), pointer :: fmat
    type(zqrm_spfct_type), pointer :: ffct

    call c_f_pointer(qrm_spmat_c%h, fmat)
    call c_f_pointer(qrm_spfct_c%h, ffct)

    fmat%m   = qrm_spmat_c%m   
    fmat%n   = qrm_spmat_c%n   
    fmat%nz  = qrm_spmat_c%nz  
    fmat%sym = qrm_spmat_c%sym

    call c_f_pointer(qrm_spmat_c%irn, fmat%irn,(/qrm_spmat_c%nz/))
    call c_f_pointer(qrm_spmat_c%jcn, fmat%jcn,(/qrm_spmat_c%nz/))
    call c_f_pointer(qrm_spmat_c%val, fmat%val,(/qrm_spmat_c%nz/))

    ffct%m      = qrm_spfct_c%m   
    ffct%n      = qrm_spfct_c%n   
    ffct%nz     = qrm_spfct_c%nz  
    ffct%sym    = qrm_spfct_c%sym 
    ffct%icntl  = qrm_spfct_c%icntl 
    ffct%rcntl  = qrm_spfct_c%rcntl 

    call zqrm_factorize(fmat, ffct, transp, info)

    qrm_spfct_c%gstats = ffct%gstats

    return
  end function zqrm_factorize_c


  !> C equivalent of the zqrm_spfct_trsm routine
  function zqrm_solve_c(qrm_spfct_c, transp, b, x, nrhs) result(info) bind(c)
    use zqrm_solve_mod
    use zqrm_spfct_mod
    implicit none

    type(zqrm_spfct_type_c)        :: qrm_spfct_c
    character(kind=c_char), value  :: transp
    type(c_ptr), value             :: b, x
    integer(c_int), value          :: nrhs
    integer(c_int)                 :: info

    info = zqrm_spfct_trsm_c(qrm_spfct_c, transp, b, x, nrhs)

    return
  end function zqrm_solve_c
  
  !> C equivalent of the zqrm_spfct_trsm routine
  function zqrm_spfct_trsm_c(qrm_spfct_c, transp, b, x, nrhs) result(info) bind(c)
    use zqrm_solve_mod
    use zqrm_spfct_mod
    implicit none

    type(zqrm_spfct_type_c)        :: qrm_spfct_c
    character(kind=c_char), value  :: transp
    type(c_ptr), value             :: b, x
    integer(c_int), value          :: nrhs
    integer(c_int)                 :: info

    complex(r64), pointer          :: ib(:,:), ix(:,:)
    type(zqrm_spfct_type), pointer :: ffct

    call c_f_pointer(qrm_spfct_c%h, ffct)

    if(transp.eq.qrm_no_transp) then
       call c_f_pointer(b, ib,(/ffct%m, nrhs/))
       call c_f_pointer(x, ix,(/ffct%n, nrhs/))
    else if(transp.eq.qrm_conj_transp) then
       call c_f_pointer(b, ib,(/ffct%n, nrhs/))
       call c_f_pointer(x, ix,(/ffct%m, nrhs/))
    end if
    
    ffct%m      = qrm_spfct_c%m   
    ffct%n      = qrm_spfct_c%n   
    ffct%nz     = qrm_spfct_c%nz  
    ffct%sym    = qrm_spfct_c%sym 
    ffct%icntl  = qrm_spfct_c%icntl 
    ffct%rcntl  = qrm_spfct_c%rcntl 

    call zqrm_spfct_trsm(ffct, transp, ib, ix, info)
    
    qrm_spfct_c%gstats = ffct%gstats

    return
  end function zqrm_spfct_trsm_c


  !> C equivalent of the zqrm_apply routine
  function zqrm_apply_c(qrm_spfct_c, transp, b, nrhs) result(info) bind(c)
    use zqrm_spfct_mod
    use zqrm_solve_mod
    implicit none
    type(zqrm_spfct_type_c)        :: qrm_spfct_c
    character(kind=c_char), value  :: transp
    type(c_ptr), value             :: b
    integer(c_int), value          :: nrhs
    integer(c_int)                 :: info

    info = zqrm_spfct_unmqr_c(qrm_spfct_c, transp, b, nrhs)

    return

  end function zqrm_apply_c
    
  !> C equivalent of the zqrm_spfct_zunmqr routine
  function zqrm_spfct_unmqr_c(qrm_spfct_c, transp, b, nrhs) result(info) bind(c)
    use zqrm_spfct_mod
    use zqrm_solve_mod
    implicit none
    type(zqrm_spfct_type_c)        :: qrm_spfct_c
    character(kind=c_char), value  :: transp
    type(c_ptr), value             :: b
    integer(c_int), value          :: nrhs
    integer(c_int)                 :: info

    complex(r64), pointer          :: ib(:,:)
    type(zqrm_spfct_type), pointer :: ffct

    call c_f_pointer(qrm_spfct_c%h, ffct)

    call c_f_pointer(b, ib,(/ffct%m, nrhs/))

    ffct%m     = qrm_spfct_c%m   
    ffct%n     = qrm_spfct_c%n   
    ffct%nz    = qrm_spfct_c%nz  
    ffct%sym   = qrm_spfct_c%sym 
    ffct%icntl = qrm_spfct_c%icntl 
    ffct%rcntl = qrm_spfct_c%rcntl 

    call zqrm_spfct_unmqr(ffct, transp, ib, info)
    
    qrm_spfct_c%gstats = ffct%gstats

    return
  end function zqrm_spfct_unmqr_c



  !> C equivalent of the qrm_spmat_mv
  !> routine
  subroutine zqrm_spmat_mv_c(qrm_spmat_c, transp, alpha, x, beta, y, nrhs) bind(c)
    use zqrm_utils_mod
    implicit none
    type(zqrm_spmat_type_c)        :: qrm_spmat_c
    character(kind=c_char), value  :: transp
    complex(r64), value            :: alpha, beta
    type(c_ptr), value             :: x, y
    integer(c_int), value          :: nrhs

    complex(r64), pointer          :: ix(:,:), iy(:,:)
    type(zqrm_spmat_type), pointer :: fmat

    call c_f_pointer(qrm_spmat_c%h, fmat)

    fmat%m  = qrm_spmat_c%m   
    fmat%n  = qrm_spmat_c%n   
    fmat%nz = qrm_spmat_c%nz  
    fmat%sym = qrm_spmat_c%sym

    call c_f_pointer(qrm_spmat_c%irn, fmat%irn,(/qrm_spmat_c%nz/))
    call c_f_pointer(qrm_spmat_c%jcn, fmat%jcn,(/qrm_spmat_c%nz/))
    call c_f_pointer(qrm_spmat_c%val, fmat%val,(/qrm_spmat_c%nz/))

    if(transp.eq.qrm_no_transp) then
       call c_f_pointer(x, ix,(/qrm_spmat_c%n, nrhs/))
       call c_f_pointer(y, iy,(/qrm_spmat_c%m, nrhs/))
    else if (transp.eq.qrm_conj_transp) then
       call c_f_pointer(x, ix,(/qrm_spmat_c%m, nrhs/))
       call c_f_pointer(y, iy,(/qrm_spmat_c%n, nrhs/))
    end if

    call zqrm_spmat_mv(fmat, transp, alpha, ix, beta, iy)
    
    return
  end subroutine zqrm_spmat_mv_c
  

  !> @brief C equivalent of the @link ::zqrm_matnrm @endlink
  !> routine
  function zqrm_spmat_nrm_c(qrm_spmat_c, ntype, nrm) result(info) bind(c)
    use zqrm_utils_mod
    implicit none
    type(zqrm_spmat_type_c)        :: qrm_spmat_c
    character(kind=c_char), value  :: ntype
    real(r64)                      :: nrm
    integer(c_int)                 :: info

    type(zqrm_spmat_type), pointer :: fmat

    call c_f_pointer(qrm_spmat_c%h, fmat)

    fmat%m   = qrm_spmat_c%m   
    fmat%n   = qrm_spmat_c%n   
    fmat%nz  = qrm_spmat_c%nz  
    fmat%sym = qrm_spmat_c%sym

    call c_f_pointer(qrm_spmat_c%irn, fmat%irn,(/qrm_spmat_c%nz/))
    call c_f_pointer(qrm_spmat_c%jcn, fmat%jcn,(/qrm_spmat_c%nz/))
    call c_f_pointer(qrm_spmat_c%val, fmat%val,(/qrm_spmat_c%nz/))

    call zqrm_spmat_nrm(fmat, ntype, nrm, info)

    return
  end function zqrm_spmat_nrm_c

  !> C equivalent of the zqrm_vecnrm
  !> routine
  function zqrm_vecnrm_c(x, n, nrhs, ntype, nrm) result(info) bind(c)
    use zqrm_utils_mod
    implicit none
    type(c_ptr), value            :: x
    integer(c_int), value         :: nrhs
    integer(c_int), value         :: n
    character(kind=c_char), value :: ntype
    type(c_ptr), value            :: nrm
    integer(c_int)                :: info

    complex(r64), pointer         :: ix(:,:)
    real(r64), pointer            :: inrm(:)

    call c_f_pointer(x, ix,(/n,nrhs/))
    call c_f_pointer(nrm, inrm,(/nrhs/))

    call zqrm_vecnrm(ix, n, ntype, inrm, info)

    return
  end function zqrm_vecnrm_c

  !> C equivalent of the zqrm_spbackslash routine
  function zqrm_spbackslash_c(qrm_spmat_c, b, x, nrhs, transp) result(info) bind(c)
    use zqrm_spmat_mod
    use zqrm_methods_mod
    implicit none
    type(zqrm_spmat_type_c)        :: qrm_spmat_c
    type(c_ptr), value             :: b, x
    integer(c_int), value          :: nrhs
    character(kind=c_char), value  :: transp
    integer(c_int)                 :: info

    info = zqrm_spmat_backslash_c(qrm_spmat_c, b, x, nrhs, transp)

    return
  end function zqrm_spbackslash_c
    
  !> C equivalent of the zqrm_spmat_backslash routine
  function zqrm_spmat_backslash_c(qrm_spmat_c, b, x, nrhs, transp) result(info) bind(c)
    use zqrm_spmat_mod
    use zqrm_methods_mod
    implicit none
    type(zqrm_spmat_type_c)        :: qrm_spmat_c
    type(c_ptr), value             :: b, x
    integer(c_int), value          :: nrhs
    character(kind=c_char), value  :: transp
    integer(c_int)                 :: info
 
    complex(r64), pointer          :: ib(:,:), ix(:,:)

    type(zqrm_spmat_type), pointer :: fmat

    call c_f_pointer(qrm_spmat_c%h, fmat)

    fmat%m   = qrm_spmat_c%m   
    fmat%n   = qrm_spmat_c%n   
    fmat%nz  = qrm_spmat_c%nz  
    fmat%sym = qrm_spmat_c%sym

    call c_f_pointer(qrm_spmat_c%irn, fmat%irn,(/qrm_spmat_c%nz/))
    call c_f_pointer(qrm_spmat_c%jcn, fmat%jcn,(/qrm_spmat_c%nz/))
    call c_f_pointer(qrm_spmat_c%val, fmat%val,(/qrm_spmat_c%nz/))

    if(transp.eq.qrm_no_transp) then
       call c_f_pointer(b, ib,(/qrm_spmat_c%m, nrhs/))
       call c_f_pointer(x, ix,(/qrm_spmat_c%n, nrhs/))
    else if(transp.eq.qrm_conj_transp) then
       call c_f_pointer(b, ib,(/qrm_spmat_c%n, nrhs/))
       call c_f_pointer(x, ix,(/qrm_spmat_c%m, nrhs/))
    end if
    
    call qrm_spmat_backslash(fmat, ib, ix, transp=transp, info=info)
    
    return
  end function zqrm_spmat_backslash_c

  !> C equivalent of the zqrm_spmat_backslash routine
  function zqrm_spfct_backslash_c(qrm_spfct_c, b, x, nrhs, transp) result(info) bind(c)
    use zqrm_spfct_mod
    use zqrm_methods_mod
    implicit none
    type(zqrm_spfct_type_c)        :: qrm_spfct_c
    type(c_ptr), value             :: b, x
    integer(c_int), value          :: nrhs
    character(kind=c_char), value  :: transp
    integer(c_int)                 :: info
 
    complex(r64), pointer          :: ib(:,:), ix(:,:)

    type(zqrm_spfct_type), pointer :: ffct

    call c_f_pointer(qrm_spfct_c%h, ffct)

    if(transp.eq.qrm_no_transp) then
       call c_f_pointer(b, ib,(/ffct%m, nrhs/))
       call c_f_pointer(x, ix,(/ffct%n, nrhs/))
    else if(transp.eq.qrm_conj_transp) then
       call c_f_pointer(b, ib,(/ffct%n, nrhs/))
       call c_f_pointer(x, ix,(/ffct%m, nrhs/))
    end if
       
    ffct%m      = qrm_spfct_c%m   
    ffct%n      = qrm_spfct_c%n   
    ffct%nz     = qrm_spfct_c%nz  
    ffct%sym    = qrm_spfct_c%sym 
    ffct%icntl  = qrm_spfct_c%icntl 
    ffct%rcntl  = qrm_spfct_c%rcntl 

    call qrm_spfct_backslash(ffct, ib, ix, transp=transp, info=info)

    qrm_spfct_c%gstats = ffct%gstats

    return
  end function zqrm_spfct_backslash_c

  !> C equivalent of the zqrm_least_squares routine
  function zqrm_spposv_c(qrm_spmat_c, b, x, nrhs) result(info) bind(c)
    use zqrm_spmat_mod
    use zqrm_methods_mod
    implicit none
    type(zqrm_spmat_type_c)        :: qrm_spmat_c
    type(c_ptr), value             :: b, x
    integer(c_int), value          :: nrhs
    integer(c_int)                 :: info
 
    complex(r64), pointer          :: ib(:,:), ix(:,:)

    type(zqrm_spmat_type), pointer :: fmat

    call c_f_pointer(qrm_spmat_c%h, fmat)

    fmat%m   = qrm_spmat_c%m   
    fmat%n   = qrm_spmat_c%n   
    fmat%nz  = qrm_spmat_c%nz  
    fmat%sym = qrm_spmat_c%sym

    call c_f_pointer(qrm_spmat_c%irn, fmat%irn,(/qrm_spmat_c%nz/))
    call c_f_pointer(qrm_spmat_c%jcn, fmat%jcn,(/qrm_spmat_c%nz/))
    call c_f_pointer(qrm_spmat_c%val, fmat%val,(/qrm_spmat_c%nz/))

    call c_f_pointer(b, ib,(/qrm_spmat_c%m, nrhs/))
    call c_f_pointer(x, ix,(/qrm_spmat_c%n, nrhs/))

    call qrm_spposv(fmat, ib, ix, info=info)
    
    return
  end function zqrm_spposv_c

  
  !> C equivalent of the zqrm_least_squares routine
  function zqrm_least_squares_c(qrm_spmat_c, b, x, nrhs, transp) result(info) bind(c)
    use zqrm_spmat_mod
    use zqrm_methods_mod
    implicit none
    type(zqrm_spmat_type_c)        :: qrm_spmat_c
    type(c_ptr), value             :: b, x
    integer(c_int), value          :: nrhs
    character(kind=c_char), value  :: transp
    integer(c_int)                 :: info
 
    complex(r64), pointer          :: ib(:,:), ix(:,:)

    type(zqrm_spmat_type), pointer :: fmat

    call c_f_pointer(qrm_spmat_c%h, fmat)

    fmat%m   = qrm_spmat_c%m   
    fmat%n   = qrm_spmat_c%n   
    fmat%nz  = qrm_spmat_c%nz  
    fmat%sym = qrm_spmat_c%sym

    call c_f_pointer(qrm_spmat_c%irn, fmat%irn,(/qrm_spmat_c%nz/))
    call c_f_pointer(qrm_spmat_c%jcn, fmat%jcn,(/qrm_spmat_c%nz/))
    call c_f_pointer(qrm_spmat_c%val, fmat%val,(/qrm_spmat_c%nz/))

    if(transp.eq.qrm_no_transp) then
       call c_f_pointer(b, ib,(/fmat%m, nrhs/))
       call c_f_pointer(x, ix,(/fmat%n, nrhs/))
    else if(transp.eq.qrm_conj_transp) then
       call c_f_pointer(b, ib,(/fmat%n, nrhs/))
       call c_f_pointer(x, ix,(/fmat%m, nrhs/))
    end if

    call qrm_least_squares(fmat, ib, ix, transp=transp, info=info)
    
    return
  end function zqrm_least_squares_c

  !> C equivalent of the zqrm_min_norm routine
  function zqrm_min_norm_c(qrm_spmat_c, b, x, nrhs, transp) result(info) bind(c)
    use zqrm_spmat_mod
    use zqrm_methods_mod
    implicit none
    type(zqrm_spmat_type_c)        :: qrm_spmat_c
    type(c_ptr), value             :: b, x
    integer(c_int), value          :: nrhs
    character(kind=c_char), value  :: transp
    integer(c_int)                 :: info

    complex(r64), pointer          :: ib(:,:), ix(:,:)

    type(zqrm_spmat_type), pointer :: fmat

    call c_f_pointer(qrm_spmat_c%h, fmat)

    fmat%m   = qrm_spmat_c%m   
    fmat%n   = qrm_spmat_c%n   
    fmat%nz  = qrm_spmat_c%nz  
    fmat%sym = qrm_spmat_c%sym

    call c_f_pointer(qrm_spmat_c%irn, fmat%irn,(/qrm_spmat_c%nz/))
    call c_f_pointer(qrm_spmat_c%jcn, fmat%jcn,(/qrm_spmat_c%nz/))
    call c_f_pointer(qrm_spmat_c%val, fmat%val,(/qrm_spmat_c%nz/))

    if(transp.eq.qrm_no_transp) then
       call c_f_pointer(b, ib,(/fmat%m, nrhs/))
       call c_f_pointer(x, ix,(/fmat%n, nrhs/))
    else if(transp.eq.qrm_conj_transp) then
       call c_f_pointer(b, ib,(/fmat%n, nrhs/))
       call c_f_pointer(x, ix,(/fmat%m, nrhs/))
    end if

    call qrm_min_norm(fmat, ib, ix, transp=transp, info=info)
    
    return
  end function zqrm_min_norm_c


  !> C equivalent of the zqrm_residual_norm routine
  function zqrm_residual_norm_c(qrm_spmat_c, b, x, nrhs, nrm, transp) result(info) bind(c)
    use zqrm_spmat_mod
    use zqrm_methods_mod
    implicit none
    type(zqrm_spmat_type_c)        :: qrm_spmat_c
    integer(c_int), value          :: nrhs
    type(c_ptr), value             :: b, x
    type(c_ptr), value             :: nrm
    character(kind=c_char), value  :: transp
    integer(c_int)                 :: info

    complex(r64), pointer          :: ib(:,:), ix(:,:)
    real(r64), pointer             :: inrm(:)

    type(zqrm_spmat_type), pointer :: fmat
    call c_f_pointer(qrm_spmat_c%h, fmat)

    fmat%m   = qrm_spmat_c%m   
    fmat%n   = qrm_spmat_c%n   
    fmat%nz  = qrm_spmat_c%nz  
    fmat%sym = qrm_spmat_c%sym

    call c_f_pointer(qrm_spmat_c%irn, fmat%irn,(/qrm_spmat_c%nz/))
    call c_f_pointer(qrm_spmat_c%jcn, fmat%jcn,(/qrm_spmat_c%nz/))
    call c_f_pointer(qrm_spmat_c%val, fmat%val,(/qrm_spmat_c%nz/))

    if(transp.eq.qrm_no_transp) then
       call c_f_pointer(b, ib,(/fmat%m, nrhs/))
       call c_f_pointer(x, ix,(/fmat%n, nrhs/))
    else if(transp.eq.qrm_conj_transp) then
       call c_f_pointer(b, ib,(/fmat%n, nrhs/))
       call c_f_pointer(x, ix,(/fmat%m, nrhs/))
    end if
    call c_f_pointer(nrm, inrm,(/nrhs/))

    call qrm_residual_norm(fmat, ib, ix, inrm, transp=transp, info=info)
    
    return
  end function zqrm_residual_norm_c

  !> C equivalent of the zqrm_residual_orth routine
  function zqrm_residual_orth_c(qrm_spmat_c, r, nrhs, nrm, transp) result(info) bind(c)
    use zqrm_spmat_mod
    use zqrm_methods_mod
    implicit none
    type(zqrm_spmat_type_c)        :: qrm_spmat_c
    type(c_ptr), value             :: r
    integer(c_int), value          :: nrhs
    type(c_ptr), value             :: nrm
    character(kind=c_char), value  :: transp
    integer(c_int)                 :: info

    complex(r64), pointer          :: ir(:,:)
    real(r64), pointer             :: inrm(:)

    type(zqrm_spmat_type), pointer :: fmat

    call c_f_pointer(qrm_spmat_c%h, fmat)

    fmat%m   = qrm_spmat_c%m   
    fmat%n   = qrm_spmat_c%n   
    fmat%nz  = qrm_spmat_c%nz  
    fmat%sym = qrm_spmat_c%sym

    call c_f_pointer(qrm_spmat_c%irn, fmat%irn,(/qrm_spmat_c%nz/))
    call c_f_pointer(qrm_spmat_c%jcn, fmat%jcn,(/qrm_spmat_c%nz/))
    call c_f_pointer(qrm_spmat_c%val, fmat%val,(/qrm_spmat_c%nz/))

    if(transp.eq.qrm_no_transp) then
       call c_f_pointer(r, ir,(/qrm_spmat_c%m,nrhs/))
    else if(transp.eq.qrm_conj_transp) then
       call c_f_pointer(r, ir,(/qrm_spmat_c%n,nrhs/))
    end if

    call c_f_pointer(nrm, inrm,(/nrhs/))

    call qrm_residual_orth(fmat, ir, inrm, transp=transp, info=info)
    
    return
  end function zqrm_residual_orth_c


  !> C equivalent of the zqrm_spfct_set_i4
  !> routine (only for global, integer type)
  function zqrm_spfct_set_i4_c(qrm_spfct_c, string, val) result(info) bind(c)
    use zqrm_spfct_mod
    implicit none
    type(zqrm_spfct_type_c) :: qrm_spfct_c
    character(kind=c_char)  :: string(*)
    integer(c_int), value   :: val
    integer(c_int)          :: info
    
    character(len=:), allocatable  :: a
    type(zqrm_spfct_type), pointer :: ffct

    call c_f_pointer(qrm_spfct_c%h, ffct)
    
    a = qrm_c2f_string(string)
    
    call zqrm_spfct_set_i4(ffct, a, val, info)

    qrm_spfct_c%icntl = ffct%icntl  
    qrm_spfct_c%rcntl = ffct%rcntl 

    return
    
  end function zqrm_spfct_set_i4_c


  !> C equivalent of the zqrm_spfct_set_i4
  !> routine (only for global, integer type)
  function zqrm_spfct_set_r4_c(qrm_spfct_c, string, val) result(info) bind(c)
    use zqrm_spfct_mod
    implicit none
    type(zqrm_spfct_type_c) :: qrm_spfct_c
    character(kind=c_char)  :: string(*)
    real(c_float), value    :: val
    integer(c_int)          :: info
    
    character(len=:), allocatable  :: a
    type(zqrm_spfct_type), pointer :: ffct

    call c_f_pointer(qrm_spfct_c%h, ffct)
    
    a = qrm_c2f_string(string)
    
    call zqrm_spfct_set_r4(ffct, a, val, info)

    qrm_spfct_c%icntl = ffct%icntl  
    qrm_spfct_c%rcntl = ffct%rcntl 

    return
    
  end function zqrm_spfct_set_r4_c


  !> C equivalent of the zqrm_spfct_geti routine (only for global, integer type)
  function zqrm_spfct_get_i4_c(qrm_spfct_c, string, val) result(info) bind(c)
    use zqrm_spfct_mod
    implicit none
    type(zqrm_spfct_type_c) :: qrm_spfct_c
    character(kind=c_char)  :: string(*)
    integer(c_int)          :: val
    integer(c_int)          :: info
    
    character(len=:), allocatable  :: a
    type(zqrm_spfct_type), pointer :: ffct
    
    call c_f_pointer(qrm_spfct_c%h, ffct)
    
    a = qrm_c2f_string(string)

    ffct%icntl = qrm_spfct_c%icntl 
    ffct%rcntl = qrm_spfct_c%rcntl 
    
    call zqrm_spfct_get_i4(ffct, a, val, info)

    return
    
  end function zqrm_spfct_get_i4_c

  !> C equivalent of the zqrm_spfct_geti routine (only for global, integer type)
  function zqrm_spfct_get_r4_c(qrm_spfct_c, string, val) result(info) bind(c)
    use zqrm_spfct_mod
    implicit none
    type(zqrm_spfct_type_c) :: qrm_spfct_c
    character(kind=c_char)  :: string(*)
    real(c_float)           :: val
    integer(c_int)          :: info
    
    character(len=:), allocatable  :: a
    type(zqrm_spfct_type), pointer :: ffct
    
    call c_f_pointer(qrm_spfct_c%h, ffct)
    
    a = qrm_c2f_string(string)

    ffct%icntl = qrm_spfct_c%icntl 
    ffct%rcntl = qrm_spfct_c%rcntl 
    
    call zqrm_spfct_get_r4(ffct, a, val, info)

    return
    
  end function zqrm_spfct_get_r4_c

  !> C equivalent of the zqrm_spfct_get_i8 routine (only for global, integer type)
  function zqrm_spfct_get_i8_c(qrm_spfct_c, string, val) result(info) bind(c)
    use zqrm_spfct_mod
    implicit none
    type(zqrm_spfct_type_c) :: qrm_spfct_c
    character(kind=c_char)  :: string(*)
    integer(c_long_long)    :: val
    integer(c_int)          :: info
    
    character(len=:), allocatable  :: a
    type(zqrm_spfct_type), pointer :: ffct
    
    call c_f_pointer(qrm_spfct_c%h, ffct)
    
    a = qrm_c2f_string(string)

    ffct%icntl = qrm_spfct_c%icntl 
    ffct%rcntl = qrm_spfct_c%rcntl 
    
    call zqrm_spfct_get_i8(ffct, a, val, info)

    return
    
  end function zqrm_spfct_get_i8_c


end module zqrm_c_interface

