!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> qrm_spfct_trsm2d.F90
!! This file contains a routine that solves for R or R' against multiple vectors
!!
!!
!! ##############################################################################################


! #include "prec.h"
#include "qrm_common.h"



!> @brief This function solves for R or R' against a single vector
!!
!! [in] qrm_fct   the main qrm data structure after factorization.
!!
!! [in] transp    a string saying whether R or R^T will be solved
!!                      for. Only the first character is important.
!!
!! [in]     b     a 1d array containing the RHS vector
!!
!! [out]    x     a 1d array containing the solution vector
!!
subroutine zqrm_spfct_trsm1d(qrm_fct, transp, b, x, info)

  use zqrm_spfct_mod
  use qrm_error_mod
  use zqrm_fdata_mod
  use qrm_string_mod
  use zqrm_utils_mod
  use zqrm_solve_mod, protect => zqrm_spfct_trsm1d
  use qrm_error_mod
  implicit none

  type(zqrm_spfct_type), target  :: qrm_fct
  complex(r64), intent(in)       :: b(:)
  complex(r64), intent(out)      :: x(:)
  character(len=*)               :: transp
  integer, optional              :: info

  complex(r64), pointer          :: pntb(:,:), pntx(:,:)
  integer                        :: n, i

  n = size(b,1)
  call zqrm_remap_pnt(b, pntb, n)
  n = size(x,1)
  call zqrm_remap_pnt(x, pntx, n)

  call zqrm_spfct_trsm2d(qrm_fct, transp, pntb, pntx, info)

  return

end subroutine zqrm_spfct_trsm1d




!> @brief This function solves for R or R' against multiple vectors
!!
!! [in] qrm_fct   the main qrm data structure after factorization.
!!
!! [in] transp    a string saying whether R or R^T will be solved
!!                      for. Only the first character is important.
!!
!! [in]     b     a 2d array containing the RHS vectors
!!
!! [out]    x     a 2d array containing the solution vectors
!!
subroutine zqrm_spfct_trsm2d(qrm_fct, transp, b, x, info)

  use zqrm_spfct_mod
  use qrm_error_mod
  use zqrm_fdata_mod
  use zqrm_sdata_mod
  use qrm_string_mod
  use zqrm_utils_mod
  use qrm_dscr_mod
  use zqrm_solve_mod, protect => zqrm_spfct_trsm2d
  implicit none

  type(zqrm_spfct_type), target      :: qrm_fct
  complex(r64), intent(inout), target   :: b(:,:)
  complex(r64), intent(out), target     :: x(:,:)
  character(len=*)                   :: transp
  integer, optional                  :: info

  type(qrm_dscr_type)                :: qrm_dscr
  integer                            :: i, nb, nrhs, be, en, in
  integer                            :: node, inode, ret, keeph
  type(zqrm_front_type), pointer     :: front
  type(zqrm_sdata_type), allocatable :: qrm_sdata_b(:), qrm_sdata_x(:)


  ! error management
  integer                            :: err
  character(len=*), parameter        :: name='qrm_spfct_trsm'

 err = 0

 ! immediately check if the facto was done. Otherwise push an error and return
 if(.not.allocated(qrm_fct%fdata)) then
    err = 14
    call qrm_error_print(err, name)
    goto 9999
 else
    if(.not. qrm_fct%fdata%ok) then
       err = 14
       call qrm_error_print(err, name)
       goto 9999
    end if
 end if

 call qrm_get(qrm_fct, 'qrm_keeph', keeph)
 if(keeph .lt. 0) then
    err = 30
    call qrm_error_print(err, name)
    goto 9999
 end if

 ! blocking to deal with multiple rhs
 call qrm_get(qrm_fct, 'qrm_rhsnb', nb)
 nrhs = size(b,2)
 if(nb.le.0) nb = nrhs

 allocate(qrm_sdata_b((nrhs-1)/nb+1))
 allocate(qrm_sdata_x((nrhs-1)/nb+1))

 ! init the descriptor
 call qrm_dscr_init(qrm_dscr)

 do i=1, (nrhs-1)/nb+1
    call zqrm_sdata_init(qrm_sdata_b(i), b(:,(i-1)*nb+1:min(i*nb,nrhs)))
    call zqrm_sdata_init(qrm_sdata_x(i), x(:,(i-1)*nb+1:min(i*nb,nrhs)))
    call zqrm_spfct_trsm_async(qrm_dscr, qrm_fct, transp, qrm_sdata_b(i), qrm_sdata_x(i))
 end do

 call qrm_barrier(qrm_dscr, err)
 __QRM_INFO_CHECK(err, name,'qrm_spfct_trsm_async',9998)

 call qrm_dscr_destroy(qrm_dscr)

9998 continue 
 do i=1, (nrhs-1)/nb+1
    call zqrm_sdata_destroy(qrm_sdata_b(i))
    call zqrm_sdata_destroy(qrm_sdata_x(i))
 end do
 deallocate(qrm_sdata_b, qrm_sdata_x)

9999 continue 
 if(present(info)) info = err
  return

end subroutine zqrm_spfct_trsm2d


subroutine zqrm_spfct_trsm_async(qrm_dscr, qrm_fct, transp, qrm_sdata_b, qrm_sdata_x)

  use zqrm_spfct_mod
  use qrm_error_mod
  use zqrm_fdata_mod
  use zqrm_sdata_mod
  use qrm_string_mod
  use zqrm_utils_mod
  use qrm_dscr_mod
  use zqrm_solve_mod, protect => zqrm_spfct_trsm_async
#if defined (have_starpu)
  use zqrm_starpu_mod
  use fstarpu_mod
#endif
  implicit none

  type(qrm_dscr_type)                        :: qrm_dscr
  type(zqrm_spfct_type), intent(in), target  :: qrm_fct
  character(len=*), intent(in)               :: transp
  type(zqrm_sdata_type)                      :: qrm_sdata_b, qrm_sdata_x

  integer                                    :: i, nb, nrhs, be, en, in
  integer                                    :: node, inode, ret
  type(qrm_adata_type), pointer              :: adata


  ! error management
  integer                                    :: info
  character(len=*), parameter                :: name='qrm_spfct_trsm_async'

  if(qrm_dscr%info .ne. 0) return

  info = 0
 
  ! simplify
  adata => qrm_fct%adata
 
  nb = size(qrm_sdata_b%p,2)
 
  ! whether to go top-down or bottom-up
  if(qrm_str_tolower(transp(1:1)) .eq. qrm_conj_transp) then
     be=1; en=adata%nnodes; in=1
  else
     be=adata%nnodes; en=1; in=-1
  end if
 
  ! do not reallocate if already done
  if(.not.allocated(qrm_sdata_x%front_rhs)) then
     allocate(qrm_sdata_x%front_rhs(adata%nnodes))
#if defined (have_starpu)
     qrm_sdata_x%front_rhs(:)%hdl = c_null_ptr
#endif
  end if
  if(.not.allocated(qrm_sdata_b%front_rhs)) then
     allocate(qrm_sdata_b%front_rhs(adata%nnodes))
#if defined (have_starpu)
     qrm_sdata_b%front_rhs(:)%hdl = c_null_ptr
#endif
  end if
 

#if defined (have_starpu)
  do node=1, adata%nnodes
     if(adata%small(node) .ge. 0) then
        if(.not.c_associated(qrm_sdata_x%front_rhs(node)%hdl))  call fstarpu_void_data_register(qrm_sdata_x%front_rhs(node)%hdl)
        if(.not.c_associated(qrm_sdata_b%front_rhs(node)%hdl))  call fstarpu_void_data_register(qrm_sdata_b%front_rhs(node)%hdl)
     end if
  end do
#endif
 
  ! inode is the node index (i.e., the i-th node in the give
  ! postorder) while node is its actual ID
  tree: do inode=be, en, in
     node = adata%torder(inode)
     if(adata%small(node) .lt. 0) cycle tree

     call zqrm_node_trsm_task(qrm_dscr, transp, qrm_fct, node, qrm_sdata_b, qrm_sdata_x)
     __QRM_INFO_CHECK(qrm_dscr%info, name,'qrm_node_trsm_task',9999)
  end do tree
  
9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return
  
end subroutine zqrm_spfct_trsm_async
