!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_spfct_unmqr2d.F90
!! This file contains a routine that applies Q or Q' to a set of vectors
!!
!!
!! ##############################################################################################


#include "qrm_common.h"

!> @brief This function applies Q or Q^T to a single vector
!! 
!! @param[in] qrm_mat   the main qrm data structure after factorization.
!!
!! @param[in] transp    a string saying whether Q or Q^T will be applied. Only the
!!                      first character is important.
!! 
!! @param[in,out] b     a 1d array containing the vector to which Q will 
!!                      be applied. 
!!
subroutine zqrm_spfct_unmqr1d(qrm_spfct, transp, b, info)

  use zqrm_spfct_mod
  use qrm_common_mod
  use qrm_string_mod
  use zqrm_solve_mod, protect => zqrm_spfct_unmqr1d
  use zqrm_utils_mod
  implicit none

  type(zqrm_spfct_type), intent(in), target  :: qrm_spfct
  complex(r64), intent(inout)                   :: b(:)
  character(len=*), intent(in)               :: transp
  integer, optional                          :: info
  
  complex(r64), pointer                         :: pnt(:,:)

  ! Turn the 1D pointer into a 2D one and call the 2D method below
  call zqrm_remap_pnt(b, pnt, size(b,1))

  call zqrm_spfct_unmqr2d(qrm_spfct, transp, pnt, info)

  return
   
end subroutine zqrm_spfct_unmqr1d



!> @brief This function applies Q or Q^T to a set of vectors
!! 
!! @param[in] qrm_spfct   the main qrm data structure after factorization.
!!
!! @param[in] transp    a string saying whether Q or Q^T will be applied. Only the
!!                      first character is important.
!! 
!! @param[in,out] b     a 2d array containing the vectors to which Q will 
!!                      be applied. 
!!

subroutine zqrm_spfct_unmqr2d(qrm_spfct, transp, b, info)

  use zqrm_spfct_mod
  use qrm_common_mod
  use qrm_string_mod
  use zqrm_sdata_mod
  use zqrm_solve_mod, protect => zqrm_spfct_unmqr2d
  use qrm_dscr_mod
  implicit none

  type(zqrm_spfct_type), intent(in), target  :: qrm_spfct
  complex(r64), intent(inout), target           :: b(:,:)
  character(len=*), intent(in)               :: transp
  integer, optional                          :: info

  type(zqrm_sdata_type), allocatable         :: qrm_sdatas(:)
  integer                                    :: nb, nrhs, i, keeph
  integer                                    :: n, inode, node, be, en, in, ret
  type(qrm_dscr_type)                        :: qrm_dscr

  ! error management
  integer                                    :: err
  character(len=*), parameter                :: name='qrm_spfct_unmqr'
  
  err = 0

  ! immediately check if the facto was done. Otherwise push an error and return
  if(.not.allocated(qrm_spfct%fdata)) then
     err = 14
     call qrm_error_print(err, name)
     goto 9999
  else
     if(.not. qrm_spfct%fdata%ok) then
        err = 14
        call qrm_error_print(err, name)
        goto 9999
     end if
  end if

  call qrm_get(qrm_spfct, 'qrm_keeph', keeph)
  if(keeph .ne. qrm_yes_) then
     err = 30
     call qrm_error_print(err, name)
     goto 9999
  end if

  ! blocking to deal with multiple rhs
  call qrm_get(qrm_spfct, 'qrm_rhsnb', nb)
  nrhs = size(b,2)
  if(nb.le.0) nb = nrhs

  allocate(qrm_sdatas((nrhs-1)/nb+1))

  ! init the descriptor
  call qrm_dscr_init(qrm_dscr)

  do i=1, (nrhs-1)/nb+1
     call zqrm_sdata_init(qrm_sdatas(i), b(:,(i-1)*nb+1:min(i*nb,nrhs)))
     call zqrm_spfct_unmqr_async(qrm_dscr, qrm_spfct, transp, qrm_sdatas(i))
  end do

  call qrm_barrier(qrm_dscr, err)
  __QRM_INFO_CHECK(err, name,'qrm_spfct_unmqr_async',9999)

  call qrm_dscr_destroy(qrm_dscr)

9998 continue
  ! cleanup and return
  do i=1, (nrhs-1)/nb+1
     call zqrm_sdata_destroy(qrm_sdatas(i))
  end do
  deallocate(qrm_sdatas)

9999 continue
  if(present(info)) info = err
  return
  
end subroutine zqrm_spfct_unmqr2d



subroutine zqrm_spfct_unmqr_async(qrm_dscr, qrm_spfct, transp, qrm_sdata)

  use zqrm_spfct_mod
  use qrm_common_mod
  use qrm_string_mod
  use zqrm_sdata_mod
  use qrm_dscr_mod
  use zqrm_solve_mod, protect => zqrm_spfct_unmqr_async
  use iso_c_binding
#if defined (have_starpu)
  use zqrm_starpu_mod
  use fstarpu_mod
#endif
  implicit none

  type(qrm_dscr_type)                        :: qrm_dscr
  type(zqrm_spfct_type), intent(in), target  :: qrm_spfct
  character(len=*), intent(in)               :: transp
  type(zqrm_sdata_type)                      :: qrm_sdata

  integer                                    :: i, keeph, n, inode
  integer                                    :: node, be, en, in, ret, nb
  type(qrm_adata_type), pointer              :: adata
  
  
  ! error management
  integer                                    :: info
  character(len=*), parameter                :: name='qrm_spfct_unmqr_async'
  
  info = 0

  if(qrm_dscr%info .ne. 0) return
  
  ! simplify 
  adata => qrm_spfct%adata

  nb = size(qrm_sdata%p,2)


  
  ! whether to go top-down or bottom-up
  if(qrm_str_tolower(transp(1:1)) .eq. qrm_conj_transp) then
     be=1; en=adata%nnodes; in=1
  else
     be=adata%nnodes; en=1; in=-1
  end if


  if(.not.allocated(qrm_sdata%front_rhs)) then
     allocate(qrm_sdata%front_rhs(adata%nnodes))
#if defined (have_starpu)
     qrm_sdata%front_rhs(:)%hdl = c_null_ptr
#endif
  end if

#if defined (have_starpu)
  call zqrm_ws_init(qrm_sdata%work, nb, qrm_spfct%icntl(qrm_nb_))
  
  do node=1, adata%nnodes
     if(adata%small(node) .ge. 0) then
        if(.not.c_associated(qrm_sdata%front_rhs(node)%hdl)) then
           call fstarpu_void_data_register(qrm_sdata%front_rhs(node)%hdl)
        end if
     end if
  end do
#else
    ! allocate workspace
  call qrm_alloc(qrm_sdata%work%c, nb, qrm_spfct%icntl(qrm_nb_), info)
  __QRM_INFO_CHECK(info, name,'qrm_alloc',9999)
#endif

  ! inode is the node index (i.e., the i-th node in the give
  ! postorder) while node is its actual ID
  tree: do inode=be, en, in
     node = adata%torder(inode)
     if(adata%small(node) .lt. 0) cycle tree

     call zqrm_node_unmqr_task(qrm_dscr, transp, qrm_spfct, node, qrm_sdata)
     __QRM_INFO_CHECK(qrm_dscr%info, name,'qrm_node_unmqr_task',9999)
  end do tree
  
9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return
  
end subroutine zqrm_spfct_unmqr_async


