!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_spmat_convert.F90
!! This file contains routines to convert qrm_spmat_type matrices from one format to another
!!
!!
!! ##############################################################################################

#include "qrm_common.h"


!> This subroutine converts an input matrix into a different
!! storage format. Optionally the values may be ignored
!! (this comes handy during the analysis)
!!
!! @param[in] in_mat      the input matrix
!!
!! @param[in,out] out_mat the output matrix in fmt format
!!
!! @param[in] fmt         the format of the output matrix
!!
!! @param[in] values      (optional) if values=.true. the output matrix will include
!!                        numerical values, otherwise only the structure
!!
!! @param[in] dups        (optional) remove duplicates. If
!!                        values=.true. duplicate values will be summed otherwise
!!                        they will be suppressed.
!!
!! @param[in] uplo        (optional) if the matrix is symmetric and uplo.eq.'u' ('l')
!!                        only the upper (lower) triangular part will be copied in out_mat.
!!                        If coefficients of in_mat are in the lower (upper) triangular part,
!!                        they will be transposed to the upper (lower) part.
!!                        
subroutine dqrm_spmat_convert(in_mat, out_mat, fmt, values, dups, uplo, info)
  use dqrm_spmat_mod, savesym => dqrm_spmat_convert
  use qrm_error_mod
  implicit none

  type(dqrm_spmat_type), intent(in)  :: in_mat
  type(dqrm_spmat_type)              :: out_mat
  character, intent(in)              :: fmt*(*)
  logical, optional                  :: values, dups
  character, optional                :: uplo
  integer, optional                  :: info

  integer                            :: err
  character(len=*), parameter        :: name='qrm_spmat_convert'

  err = 0

  select case(in_mat%fmt)
  case('csc')
     select case(fmt)
     case('csr')
        call dqrm_csc_to_csr(in_mat, out_mat, values, dups, uplo, err)
        __QRM_INFO_CHECK(err, name, 'qrm_csc_to_csr', 9999)
     case default
        ! format not known. set info and jump to 9999
        err = 1
        call qrm_error_print(err, name, aed=fmt)
        goto 9999
     end select
  case('coo')
     select case(fmt)
     case('csc')
        call dqrm_coo_to_csc(in_mat, out_mat, values, dups, uplo, err)
        __QRM_INFO_CHECK(err, name, 'qrm_coo_to_csc', 9999)
     case('csr')
        call dqrm_coo_to_csr(in_mat, out_mat, values, dups, uplo, err)
        __QRM_INFO_CHECK(err, name, 'qrm_coo_to_csr', 9999)
     case default
        ! format not known. set info and jump to 9999
        err = 1
        call qrm_error_print(err, name, aed=fmt)
        goto 9999
     end select
  case default
     ! format not known. set info and jump to 9999
     info = 1
     call qrm_error_print(info, name, aed=fmt)
     goto 9999
  end select

9999 continue
  if(present(info)) info = err
  return


contains



  !> This subroutine converts a COO matrix into a CSC
  !! matrix. Optionally the values may be ignored
  !! (this comes handy during the analysis)
  !!
  !! @param[in] in_mat      the input matrix
  !!
  !! @param[in,out] out_mat the output matrix in fmt format
  !!
  !! @param[in]  values     (optional) if values=.true. the output matrix will include
  !!                        numerical values, otherwise only the structure
  !!
  !! @param[in] dups        (optional) remove duplicates. If
  !!                        values=.true. duplicate values will be summed otherwise
  !!                        they will be suppressed.
  !!
  !! @param[in] uplo        (optional) if the matrix is symmetric and uplo.eq.'u' ('l')
  !!                        only the upper (lower) triangular part will be copied in out_mat.
  !!                        If coefficients of in_mat are in the lower (upper) triangular part,
  !!                        they will be transposed to the upper (lower) part.
  !!                        
  subroutine dqrm_coo_to_csc(in_mat, out_mat, values, dups, uplo, info)

    use dqrm_spmat_mod
    use qrm_mem_mod
    use qrm_error_mod
    use qrm_parameters_mod
    implicit none

    type(dqrm_spmat_type), intent(in)  :: in_mat
    type(dqrm_spmat_type)              :: out_mat
    logical, optional                  :: values, dups
    character, optional                :: uplo
    integer, optional                  :: info

    integer, allocatable               :: work(:)
    logical                            :: ivalues, ob, up, lo, ul, idups
    integer                            :: i, j, idx, k, m, n, t, ndups, pnt, savepnt, ii
    
    ! error management
    integer                            :: err
    character(len=*), parameter        :: name='dqrm_coo_to_csc'

    err = 0

    __QRM_IFPRESENT3(ivalues, values, .true.)
    __QRM_IFPRESENT3(idups, dups, .false.)

    __QRM_IFPRESENT4(up, uplo, (in_mat%sym.gt.0).and.(uplo.eq.qrm_upper), .false.)
    __QRM_IFPRESENT4(lo, uplo, (in_mat%sym.gt.0).and.(uplo.eq.qrm_lower), .false.)
    __QRM_IFPRESENT4(ul, uplo, (in_mat%sym.gt.0).and.(uplo.eq.qrm_both), .false.)

    if(err.eq.0) call qrm_alloc(work,           max(in_mat%n,in_mat%m)+1, err)
    __QRM_INFO_CHECK(err, name, 'qrm_(re)alloc', 9999)

    work=0
    ob = .false.

    m = in_mat%m
    n = in_mat%n

    ! first loop to calculate # of nz per column
    do k=1, in_mat%nz
       j = in_mat%jcn(k)
       i = in_mat%irn(k)
       if(ul .and. (i.eq.j)) cycle
       if((up .and. (i.gt.j)) .or. (lo .and. (j.gt.i))) then
          ! transpose coefficient
          t = i
          i = j
          j = t
       end if
       if((j.gt.0) .and. (j.le. n) .and. (i.gt.0) .and. (i.le. m) ) then
          work(j) = work(j)+1
          if(ul) work(i) = work(i)+1
       else
          ! out of bounds coefficients. ignore and print a warning at the end
          ob = .true.
       end if
    end do

    
    if(ob) then
       __QRM_PRNT_DBG('("** Out-of-bounds coefficients present **")')
    end if

    if(err.eq.0) call qrm_realloc(out_mat%jptr, in_mat%n+1, err)
    if(ivalues)  call qrm_realloc(out_mat%val,  sum(work),  err)
    if(err.eq.0) call qrm_realloc(out_mat%irn,  sum(work),  err)
    __QRM_INFO_CHECK(err, name, 'qrm_(re)alloc', 9999)

    ! loop to convert the counts into ptrs
    out_mat%jptr(1) = 1
    do j=2, n+1
       out_mat%jptr(j) = out_mat%jptr(j-1)+work(j-1)
    end do

    
    out_mat%m   = in_mat%m
    out_mat%n   = in_mat%n
    out_mat%nz  = sum(work)
    
    ! last loop to put things into place
    work=0
    ! instead of putting an "if" inside the loop
    ! I put it here to gain some speed
    if(ivalues) then
       do k=1, in_mat%nz
          j = in_mat%jcn(k)
          i = in_mat%irn(k)
          if(ul .and. (i.eq.j)) cycle
          if((j.le.0) .or. (j.gt. n) .or. (i.le.0) .or. (i.gt. m) ) cycle
          if((up .and. (i.gt.j)) .or. (lo .and. (j.gt.i))) then
             ! transpose coefficient
             t = i
             i = j
             j = t
          end if
          idx = out_mat%jptr(j)+work(j)
          out_mat%irn(idx) = i
          out_mat%val(idx) = in_mat%val(k)
          work(j) = work(j)+1

          if(ul) then
             idx = out_mat%jptr(i)+work(i)
             out_mat%irn(idx) = j
             out_mat%val(idx) = in_mat%val(k)
             work(i) = work(i)+1
          end if
       end do
    else
       do k=1, in_mat%nz
          j = in_mat%jcn(k)
          i = in_mat%irn(k)
          if(ul .and. (i.eq.j)) cycle
          if((j.le.0) .or. (j.gt. n) .or. (i.le.0) .or. (i.gt. m) ) cycle
          if((up .and. (i.gt.j)) .or. (lo .and. (j.gt.i))) then
             ! transpose coefficient
             t = i
             i = j
             j = t
          end if
          idx = out_mat%jptr(j)+work(j)
          out_mat%irn(idx) = i
          work(j) = work(j)+1

          if(ul) then
             idx = out_mat%jptr(i)+work(i)
             out_mat%irn(idx) = j
             work(i) = work(i)+1
          end if

       end do
    end if

    if (idups) then
       ! remove duplicates
       work=0
       pnt     = 0
       ndups   = 0
       savepnt = 1
       do j=1, out_mat%n
          t = out_mat%jptr(j)
          do ii=out_mat%jptr(j), out_mat%jptr(j+1)-1
             i = out_mat%irn(ii)
             if(work(i) .ge. t) then
                ! duplicate entry, skip
                ndups = ndups+1
                if(ivalues) out_mat%val(work(i)) = out_mat%val(work(i)) + out_mat%val(ii)
             else
                ! flag the entry as visited
                work(i) = ii
                pnt = pnt+1
                out_mat%irn(pnt) = i
             end if
          end do
          out_mat%jptr(j) = savepnt
          savepnt = pnt+1
       end do
       out_mat%jptr(out_mat%n+1)=savepnt
       out_mat%nz = pnt       
       
    end if
    
    out_mat%nz  = out_mat%nz - ndups
    out_mat%fmt = 'csc'
    out_mat%sym = in_mat%sym

9999 continue
    ! cleanup and return
    call qrm_dealloc(work, info)

    if(err.ne.0) then
       call qrm_dealloc(out_mat%jptr)
       call qrm_dealloc(out_mat%irn)
       call qrm_dealloc(out_mat%val)
    end if

    if(present(info)) info = err
    return

  end subroutine dqrm_coo_to_csc


  !> This subroutine converts a COO matrix into a CSC
  !! matrix. Optionally the values may be ignored
  !! (this comes handy during the analysis)
  !!
  !! @param[in] in_mat      the input matrix
  !!
  !! @param[in,out] out_mat the output matrix in fmt format
  !!
  !! @param[in]  values      (optional) if values=.true. the output matrix will include
  !!                        numerical values, otherwise only the structure
  !!
  !! @param[in] dups        (optional) remove duplicates. If
  !!                        values=.true. duplicate values will be summed otherwise
  !!                        they will be suppressed.
  !!
  !! @param[in] uplo        (optional) if the matrix is symmetric and uplo.eq.'u' ('l')
  !!                        only the upper (lower) triangular part will be copied in out_mat.
  !!                        If coefficients of in_mat are in the lower (upper) triangular part,
  !!                        they will be transposed to the upper (lower) part.
  !!                        
  subroutine dqrm_coo_to_csr(in_mat, out_mat, values, dups, uplo, info)

    use qrm_mem_mod
    use qrm_error_mod
    use qrm_parameters_mod
    implicit none

    type(dqrm_spmat_type), intent(in)  :: in_mat
    type(dqrm_spmat_type)              :: out_mat
    logical, optional                  :: values, dups
    character, optional                :: uplo
    integer, optional                  :: info

    integer, allocatable               :: work(:)
    logical                            :: ivalues, ob, idups
    integer                            :: i, j, idx, k, m, n, pnt, savepnt, ii, ndups, t

    ! error management
    integer                            :: err
    character(len=*), parameter        :: name='dqrm_coo_to_csr'

    err = 0

    __QRM_IFPRESENT3(ivalues, values, .true.)

    if(ivalues) call qrm_realloc(out_mat%val , in_mat%nz, err)
    if(err.eq.0) call qrm_alloc(work, in_mat%m+1, err)
    if(err.eq.0) call qrm_realloc(out_mat%iptr, in_mat%m+1, err)
    if(err.eq.0) call qrm_realloc(out_mat%jcn , in_mat%nz, err)
    __QRM_INFO_CHECK(err, name,'qrm_(re)alloc',9999)

    work=0
    ob = .false.

    m = in_mat%m
    n = in_mat%n

    ! first loop to calculate # of nz per row
    do k=1, in_mat%nz
       j = in_mat%jcn(k)
       i = in_mat%irn(k)
       if((i.gt.0) .and. (i.le. m) .and. (j.gt.0) .and. (j.le.n) ) then
          work(i) = work(i)+1
       else
          ! out of bounds coefficients. ignore and print a warning at the end
          ob = .true.
       end if
    end do

    if(ob) then
       __QRM_PRNT_DBG('("** Out-of-bounds coefficients present **")')
    end if

    ! loop to convert the counts into ptrs
    out_mat%iptr(1) = 1
    do i=2, m+1
       out_mat%iptr(i) = out_mat%iptr(i-1)+work(i-1)
    end do


    ! last loop to put things into place
    work=0
    ! instead of putting an "if" inside the loop
    ! I put it here to gain some speed
    if(ivalues) then
       do k=1, in_mat%nz
          j = in_mat%jcn(k)
          i = in_mat%irn(k)
          if((j.le.0) .or. (j.gt. n) .or. (i.le.0) .or. (i.gt. m) ) cycle
          idx = out_mat%iptr(i)+work(i)
          out_mat%jcn(idx) = j
          out_mat%val(idx) = in_mat%val(k)
          work(i) = work(i)+1
       end do
    else
       do k=1, in_mat%nz
          j = in_mat%jcn(k)
          i = in_mat%irn(k)
          if((j.le.0) .or. (j.gt. n) .or. (i.le.0) .or. (i.gt. m) ) cycle
          idx = out_mat%iptr(i)+work(i)
          out_mat%jcn(idx) = j
          work(i) = work(i)+1
       end do
    end if

    out_mat%m   = in_mat%m
    out_mat%n   = in_mat%n
    out_mat%nz  = in_mat%nz
    out_mat%fmt = 'csr'
    out_mat%sym = in_mat%sym

    ! TODO: add duplicates detection and uplo
    
9999 continue
    ! cleanup and return
    call qrm_dealloc(work)

    if(err.ne.0) then
       call qrm_dealloc(out_mat%iptr)
       call qrm_dealloc(out_mat%jcn)
       call qrm_dealloc(out_mat%val)
    end if

    if(present(info)) info = err
    return

  end subroutine dqrm_coo_to_csr


  !> This subroutine converts a CSC matrix into a CSR
  !! matrix. Optionally the values may be ignored
  !! (this comes handy during the analysis)
  !!
  !! @param[in] in_mat      the input matrix
  !!
  !! @param[in,out] out_mat the output matrix in fmt format
  !!
  !! @param[in]  values      (optional) if values=.true. the output matrix will include
  !!                        numerical values, otherwise only the structure
  !!
  !! @param[in] dups        (optional) remove duplicates. If
  !!                        values=.true. duplicate values will be summed otherwise
  !!                        they will be suppressed.
  !!
  !! @param[in] uplo        (optional) if the matrix is symmetric and uplo.eq.'u' ('l')
  !!                        only the upper (lower) triangular part will be copied in out_mat.
  !!                        If coefficients of in_mat are in the lower (upper) triangular part,
  !!                        they will be transposed to the upper (lower) part.
  !!                        
  subroutine dqrm_csc_to_csr(in_mat, out_mat, values, dups, uplo, info)

    use qrm_mem_mod
    use qrm_error_mod
    use qrm_parameters_mod
    implicit none

    type(dqrm_spmat_type), intent(in)  :: in_mat
    type(dqrm_spmat_type)              :: out_mat
    logical, optional                  :: values, dups
    character, optional                :: uplo
    integer, optional                  :: info

    integer, allocatable               :: work(:)

    logical                            :: ivalues, ob
    integer                            :: i, j, idx, ii, m, n

    ! error management
    integer                            :: err
    character(len=*), parameter        :: name='dqrm_csc_to_csr'

    err = 0

    __QRM_IFPRESENT3(ivalues, values, .true.)

    ob = .false.

    m = in_mat%m
    n = in_mat%n

    if(ivalues)  call qrm_realloc(out_mat%val,  in_mat%nz, err)
    if(err.eq.0) call qrm_alloc(work,           m+1,       err)
    if(err.eq.0) call qrm_realloc(out_mat%iptr, m+1,       err)
    if(err.eq.0) call qrm_realloc(out_mat%jcn,  in_mat%nz, err)
    __QRM_INFO_CHECK(err, name,'qrm_(re)alloc',9999)

    work=0
    ! first loop to calculate # of nz per row
    do j = 1, n
       do ii= in_mat%jptr(j), in_mat%jptr(j+1)-1
          i = in_mat%irn(ii)
          if((i.gt.0) .and. (i.le.m)) then
             work(i) = work(i)+1
          else
             ob = .true.
          end if
       end do
    end do

    if(ob) then
       __QRM_PRNT_DBG('("** Out-of-bounds coefficients present **")')
    end if

    ! loop to convert the counts into ptrs
    out_mat%iptr(1) = 1
    do j=2, m+1
       out_mat%iptr(j) = out_mat%iptr(j-1)+work(j-1)
    end do


    ! last loop to put things into place
    work=0
    ! instead of putting an "if" inside the loop
    ! I put it here to gain some speed
    if(ivalues) then
       do j = 1, n
          do ii= in_mat%jptr(j), in_mat%jptr(j+1)-1
             i = in_mat%irn(ii)
             if((i.le.0) .or. (i.gt.m)) cycle
             idx = out_mat%iptr(i)+work(i)
             out_mat%jcn(idx) = j
             out_mat%val(idx) = in_mat%val(ii)
             work(i) = work(i)+1
          end do
       end do
    else
       do j = 1, n
          do ii= in_mat%jptr(j), in_mat%jptr(j+1)-1
             i = in_mat%irn(ii)
             if((i.le.0) .or. (i.gt.m)) cycle
             idx = out_mat%iptr(i)+work(i)
             out_mat%jcn(idx) = j
             work(i) = work(i)+1
          end do
       end do
    end if

    out_mat%m   = in_mat%m
    out_mat%n   = in_mat%n
    out_mat%nz  = in_mat%nz
    out_mat%fmt = 'csr'
    out_mat%sym = in_mat%sym

    ! TODO: add duplicates detection and uplo
    
9999 continue
    ! cleanup and return
    call qrm_dealloc(work)
    if(err.ne.0) then
       call qrm_dealloc(out_mat%iptr)
       call qrm_dealloc(out_mat%jcn)
       call qrm_dealloc(out_mat%val)
    end if

    if(present(info)) info = err
    return

  end subroutine dqrm_csc_to_csr


end subroutine dqrm_spmat_convert
