!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_spmat_sort.F90
!! This file contains routines to sort qrm_spmat_type matrices
!!
!!
!! ##############################################################################################

#include "qrm_common.h"


!> This subroutine sorts the values of a matrix in order of
!> increasing indexes
!!
!! @param[in] qrm_mat      the input matrix to be sorted
!!
subroutine sqrm_sort_mat(qrm_mat, values, rc, info)

  use sqrm_spmat_mod, savesym => sqrm_sort_mat
  use qrm_error_mod
  implicit none
  type(sqrm_spmat_type)       :: qrm_mat
  logical, optional           :: values
  character(len=3), optional  :: rc
  integer, optional           :: info

  character(len=3)            :: irc
  logical                     :: ivalues

  ! error management
  integer                     :: err
  character(len=*), parameter :: name='sqrm_sort_mat'

  err = 0

  if(present(rc)) then
     irc = rc
  else
     irc = 'row'
  end if

  if(present(values)) then
     ivalues = values
  else
     ivalues = .true.
  end if

  select case(qrm_mat%fmt)
  case('csc')
     call sqrm_sort_csc_mat(qrm_mat, ivalues, err)
     __QRM_INFO_CHECK(err, name,'qrm_sort_csc_mat',9999)
  case('csr')
     call sqrm_sort_csr_mat(qrm_mat, ivalues, err)
     __QRM_INFO_CHECK(err, name,'qrm_sort_csr_mat',9999)
  case('coo')
     call sqrm_sort_coo_mat(qrm_mat, ivalues, irc, err)
     __QRM_INFO_CHECK(err, name,'qrm_sort_coo_mat',9999)
  end select

9999 continue
  if(present(info)) info = err
  return


contains


  !> This subroutine sorts the values of a CSC matrix in order of
  !> increasing indexes
  !!
  !! @param[in] qrm_mat      the input matrix to be sorted
  !!
  subroutine sqrm_sort_csc_mat(qrm_mat, values, info)

    use qrm_error_mod
    use qrm_mem_mod
    use qrm_sort_mod
    implicit none
    type(sqrm_spmat_type)       :: qrm_mat
    logical                     :: values
    integer, optional           :: info

    integer, allocatable        :: aux(:)
    integer                     :: j, m, k1, k2

    integer                     :: err
    character(len=*), parameter :: name='sqrm_sort_csc_mat'

    err = 0

    call qrm_alloc(aux, qrm_mat%m+2, err)
    __QRM_INFO_CHECK(err, name, 'qrm_alloc',9999)

    do j=1,qrm_mat%m
       k1= qrm_mat%jptr(j)
       k2= qrm_mat%jptr(j+1)-1
       m = k2-k1+1
       call qrm_mergesort(m, qrm_mat%irn(k1:k2), aux(1:m+2))
       if(values) then
          call qrm_mergeswap(m, aux(1:m+2), qrm_mat%irn(k1:k2), qrm_mat%val(k1:k2))
       else
          call qrm_mergeswap(m, aux(1:m+2), qrm_mat%irn(k1:k2))
       end if
    end do

    call qrm_dealloc(aux, info)

9999 continue
    if(present(info)) info = err
    return
  end subroutine sqrm_sort_csc_mat


  !> This subroutine sorts the values of a CSR matrix in order of
  !> increasing indexes
  !!
  !! @param[in] qrm_mat      the input matrix to be sorted
  !!
  subroutine sqrm_sort_csr_mat(qrm_mat, values, info)

    use qrm_error_mod
    use qrm_sort_mod
    use qrm_mem_mod
    implicit none

    type(sqrm_spmat_type)       :: qrm_mat
    logical                     :: values
    integer, optional           :: info

    integer, allocatable        :: aux(:)
    integer                     :: i, n, k1, k2

    integer                     :: err
    character(len=*), parameter :: name='sqrm_sort_csr_mat'

    err = 0

    call qrm_alloc(aux, qrm_mat%n+2, err)
    __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)

    do i=1,qrm_mat%m
       k1= qrm_mat%iptr(i)
       k2= qrm_mat%iptr(i+1)-1
       n = k2-k1+1
       call qrm_mergesort(n, qrm_mat%jcn(k1:k2), aux(1:n+2))
       if(values) then
          call qrm_mergeswap(n, aux(1:n+2), qrm_mat%jcn(k1:k2), qrm_mat%val(k1:k2))
       else
          call qrm_mergeswap(n, aux(1:n+2), qrm_mat%jcn(k1:k2))
       end if
    end do

    call qrm_dealloc(aux)

9999 continue

    if(present(info)) info = err
    return

  end subroutine sqrm_sort_csr_mat


  !> This subroutine sorts the values of a COO matrix in order of
  !> increasing indexes
  !!
  !! @param[in] qrm_mat      the input matrix to be sorted
  !!
  subroutine sqrm_sort_coo_mat(qrm_mat, values, rc, info)

    use qrm_error_mod
    use qrm_sort_mod
    use qrm_mem_mod
    implicit none
    type(sqrm_spmat_type)       :: qrm_mat
    logical                     :: values
    character(len=3)            :: rc
    integer, optional           :: info

    integer, allocatable        :: aux(:)
    integer                     :: j, m, k1, k2, nz

    integer                     :: err
    character(len=*), parameter :: name="qrm_sort_coo_mat"

    err = 0

    nz = qrm_mat%nz

    call qrm_alloc(aux, nz+2, err)
    __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)

    if(rc.eq.'row') then
       call qrm_mergesort(nz, qrm_mat%irn(1:nz), aux(1:nz+2))
    else if(rc.eq.'col') then
       call qrm_mergesort(nz, qrm_mat%jcn(1:nz), aux(1:nz+2))
    end if

    if(values) then
       call qrm_mergeswap(nz, aux(1:nz+2), qrm_mat%irn(1:nz), qrm_mat%jcn(1:nz), qrm_mat%val(1:nz))
    else
       call qrm_mergeswap(nz, aux(1:nz+2), qrm_mat%irn(1:nz), qrm_mat%jcn(1:nz))
    end if

    k1 = 1
    if(rc.eq.'row') then
       outerr: do
          if (k1.ge.qrm_mat%nz) exit outerr
          k2 = k1
          innerr: do
             if(k2.eq.qrm_mat%nz) exit innerr
             if(qrm_mat%irn(k2+1).ne.qrm_mat%irn(k1)) exit
             k2 = k2+1
          end do innerr
          m = k2-k1+1
          call qrm_mergesort(m, qrm_mat%jcn(k1:k2), aux(1:m+2))
          if(values) then
             call qrm_mergeswap(m, aux(1:m+2), qrm_mat%jcn(k1:k2), qrm_mat%val(k1:k2))
          else
             call qrm_mergeswap(m, aux(1:m+2), qrm_mat%jcn(k1:k2))
          end if
          k1 = k2+1
       end do outerr
    else if(rc.eq.'col') then
       outerc: do
          if (k1.ge.qrm_mat%nz) exit outerc
          k2 = k1
          innerc: do
             if(k2.eq.qrm_mat%nz) exit innerc
             if(qrm_mat%jcn(k2+1).ne.qrm_mat%jcn(k1)) exit
             k2 = k2+1
          end do innerc
          m = k2-k1+1
          call qrm_mergesort(m, qrm_mat%irn(k1:k2), aux(1:m+2))
          if(values) then
             call qrm_mergeswap(m, aux(1:m+2), qrm_mat%irn(k1:k2), qrm_mat%val(k1:k2))
          else
             call qrm_mergeswap(m, aux(1:m+2), qrm_mat%irn(k1:k2))
          end if
          k1 = k2+1
       end do outerc
    end if

    call qrm_dealloc(aux)

9999 continue

    if(present(info)) info = err
    return
  end subroutine sqrm_sort_coo_mat

end subroutine sqrm_sort_mat
