!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!!
!! Contains a routine which can writes a matrix to a Matrix-Market file
!!
!! ##############################################################################################


#include "qrm_common.h"

!> @brief This subroutine reads a Matrix Market matrix from a file
!! and stores it on the host processor
!!
!! @param[in] matfile   a string containing the name of the matrix file
!!
!! @param[in] qrm_mat   a cqrm_spmat_type data to be written on disk
!!
subroutine cqrm_writemat(matfile, qrm_mat, info)

  use cqrm_spmat_mod
  use qrm_error_mod
  use qrm_mem_mod
  use qrm_parameters_mod
  implicit none

  character(len=*), intent(in)              :: matfile
  type(cqrm_spmat_type), intent(inout)      :: qrm_mat
  integer, optional                         :: info

  character(len=:), allocatable             :: header, sym, typ
  integer                                   :: i
  
  ! error management
  integer                                   :: err
  character(len=*), parameter               :: name='qrm_writemat'

  err = 0
  
  open(4,file=matfile, action='WRITE', iostat=err)
  if(err .gt. 0) then
     err = 25
     call qrm_error_print(err, name, aed=matfile)
     goto 9999
  end if

  if(qrm_allocated(qrm_mat%val)) then
#if defined(dprec) || defined(sprec)
     typ = 'real'
#elif defined(zprec) || defined(cprec)
     typ = 'complex'
#endif
  else
     typ = 'pattern'
  end if

  if(qrm_mat%sym.gt.0) then
     sym = 'symmetric'
  else
     sym = 'general'
  end if
  
  header = '%%MatrixMarket matrix coordinate '//typ//' '//sym

  write(4,'(a)') header

  write(4,'(i0,2x,i0,2x,i0)')qrm_mat%m, qrm_mat%n, qrm_mat%nz
  if(qrm_mat%fmt.eq.'coo') then
     do i = 1, qrm_mat%nz
#if defined(dprec) || defined(sprec)
        write(4,'(i0,2x,i0,2x,f0.40)')qrm_mat%irn(i), &
             qrm_mat%jcn(i), qrm_mat%val(i)
#elif defined(zprec) || defined(cprec)
        write(4,'(i0,2x,i0,2x,f0.20,2x,f0.20)')qrm_mat%irn(i), &
             qrm_mat%jcn(i), qrm_mat%val(i)
#endif
     end do
  end if
  
  
  close(4)

9999 continue
  if(present(info)) info=err
  return

end subroutine cqrm_writemat


