!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!!
!! Contains a routine which can read a matrix from a Matrix-Market file
!!
!! ##############################################################################################


#include "qrm_common.h"


module qrm_ssget_mod

  use qrm_error_mod
  use qrm_parameters_mod
  
  character(len=:), allocatable :: qrm_ssget_repo, qrm_ssget_file

  type qrm_ssget_type
     character(len=:), allocatable :: group, name, dscr
     integer :: m, n, isreal, isbinary, isnd, posdef
     integer(kind=8) :: nz
     real(kind(1.d0)) :: psym, nsym
  end type qrm_ssget_type
  
  type(qrm_ssget_type), allocatable :: qrm_ssget_mats(:)
  
  interface qrm_ssget
     module procedure qrm_ssget_name, qrm_ssget_id
  end interface qrm_ssget
  
contains

  subroutine qrm_ssget_init(repo, info)

    character(len=*)              :: repo
    integer, optional             :: info
    
    logical                       :: ex
    character(len=:), allocatable :: dwnl
    integer                       :: es, nm
    character(len=200)            :: line,a,b,c

    character(len=:), allocatable :: group, dscr
    integer                       :: m, n, nz, isreal, isbinary, isnd, posdef
    real(kind(1.d0))              :: psym, nsym

    ! error management
    integer                       :: err
    character(len=*), parameter   :: name='qrm_ssget_init'

    err = 0
    
    qrm_ssget_repo = repo//"/"

    qrm_ssget_file = qrm_ssget_repo//'/ssstats.csv'

    inquire(file=qrm_ssget_file, exist=ex)

    if(.not.ex) then
       call execute_command_line('wget -q -P '//qrm_ssget_repo//' https://sparse.tamu.edu/files/ssstats.csv', &
            wait=.true., exitstat=err)
       if(err.eq.0) then
          __QRM_PRNT_MSG('("SSstats file read")')
       else
          __QRM_INFO_SET(err, 33, name, 10)
       end if
    end if

    ! parse ssstats.csv file
    open(4, file=qrm_ssget_file)

    read(4,*)nm
    read(4,*)line

    allocate(qrm_ssget_mats(nm))

    do i=1, nm
       read(4,*)a, b,&
            qrm_ssget_mats(i)%m, qrm_ssget_mats(i)%n, qrm_ssget_mats(i)%nz, &
            qrm_ssget_mats(i)%isreal, qrm_ssget_mats(i)%isbinary, &
            qrm_ssget_mats(i)%isnd, qrm_ssget_mats(i)%posdef, &
            qrm_ssget_mats(i)%psym, qrm_ssget_mats(i)%nsym, &
            c
       qrm_ssget_mats(i)%group = trim(a)
       qrm_ssget_mats(i)%name  = trim(b)
       qrm_ssget_mats(i)%dscr  = trim(c)

       ! write(*,'(a,"/",a,", ",i7,", ",i7,", ",i10,4(", ",i1),", ",f4.2,", ",f4.2,", ",a)')&
            ! qrm_ssget_mats(i)%group,qrm_ssget_mats(i)%name, &
            ! qrm_ssget_mats(i)%m, qrm_ssget_mats(i)%n, qrm_ssget_mats(i)%nz, &
            ! qrm_ssget_mats(i)%isreal, qrm_ssget_mats(i)%isbinary, &
            ! qrm_ssget_mats(i)%isnd, qrm_ssget_mats(i)%posdef, &
            ! qrm_ssget_mats(i)%psym, qrm_ssget_mats(i)%nsym, qrm_ssget_mats(i)%dscr
    end do
    close(4)

10  continue
    if(present(info)) info=err
    return
    
  end subroutine qrm_ssget_init


  ! get matrix by name
  function qrm_ssget_name(mname, info)
    implicit none
    character(len=*)              :: mname
    integer, optional             :: info
    
    character(len=:), allocatable :: qrm_ssget_name
    logical                       :: ex
    integer                       :: i, es
    character(len=:), allocatable :: fname

    ! error management
    integer                       :: err
    character(len=*), parameter   :: name='qrm_ssget_name'

    err = 0
    
    ! search for the matrix name
    do i=1, size(qrm_ssget_mats)
       if(qrm_ssget_mats(i)%name.eq.mname) exit
    end do

    if(i.le.size(qrm_ssget_mats)) then
       __QRM_PRNT_MSG('("Matrix found: ",a,"/",a)')qrm_ssget_mats(i)%group,qrm_ssget_mats(i)%name
    else
       __QRM_INFO_SET_A(err, 34, name, mname, 20)
    end if
    
    fname = qrm_ssget_repo//qrm_ssget_mats(i)%name//"/"//qrm_ssget_mats(i)%name//".mtx"
    
    inquire(file=fname, exist=ex)

    if(.not.ex) then
       call execute_command_line('wget -q -P '//qrm_ssget_repo//' https://sparse.tamu.edu/MM/'//&
            qrm_ssget_mats(i)%group//"/"//qrm_ssget_mats(i)%name//".tar.gz", &
            wait=.true., exitstat=err)
       if(err.eq.0) then
          write(*,'(a," file downloaded")')mname//".tar.gz"
          call execute_command_line('tar zxf '//qrm_ssget_repo//mname//'.tar.gz -C '//qrm_ssget_repo, &
            wait=.true., exitstat=err)
          if(es.eq.0) then
             write(*,'(a," file extracted")')mname//".tar.gz"
          end if
          call execute_command_line('rm '//qrm_ssget_repo//mname//'.tar.gz', &
            wait=.true., exitstat=err)
       end if
       if(err.ne.0) then
          __QRM_INFO_SET(err, 33, name, 20)
       end if
    end if

    qrm_ssget_name = fname

20  continue
    if(present(info)) info = err
    
    return
  end function qrm_ssget_name


  ! get matrix by id
  function qrm_ssget_id(id)
    implicit none
    character(len=:), allocatable :: qrm_ssget_id
    integer :: id

    if((id.gt.0) .and. (id.le.size(qrm_ssget_mats))) then
       qrm_ssget_id = qrm_ssget_name(qrm_ssget_mats(id)%name)
    end if
    
    return
  end function qrm_ssget_id
  
end module qrm_ssget_mod
