!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_test_facto.F90
!! This file contains coverage tests for the factorization
!!
!! ##############################################################################################

function dqrm_test_facto(c, m) result(ok)
  use dqrm_testing_mod, protect => dqrm_test_facto
  implicit none

  integer :: c, m
  logical :: ok
  
  integer, parameter :: ncases=15
  logical :: cases(ncases)

  if(c .eq. -1) then
     cases = .true.
  else if (c .le. ncases) then
     cases = .false.
     cases(c) = .true.
  end if

  ok = .true.
  
  if(cases(1 )) ok = case1 (m) .and. ok 
  if(cases(2 )) ok = case2 (m) .and. ok 
  if(cases(3 )) ok = case3 (m) .and. ok 

  return


contains

  !> @brief Test the factorization 
  function case1(m) result(ok)
    use dqrm_mod
    implicit none
    
    integer :: m
    logical :: ok
    
    type(dqrm_spmat_type), pointer :: qrm_spmat
    type(dqrm_spfct_type)          :: qrm_spfct
    integer                        :: info
    real(r64), allocatable      :: x(:), b(:), r(:)
    real(r64)                      :: err, anrm, rnrm, bnrm, xnrm, onrm
    integer                        :: i, j
    character                      :: transp 

    
    if((m.ne.-1) .and. &
         & (m.lt.11) .and. &
         & (m.gt.size(matrices))) then
       write(*,'("Matrix ",i2," is not available for this test")')m
       return
    end if

    ok = .true.
    
    ! loop over all the file-matrices
    do i=11, size(matrices)
       if((m.ne.-1) .and. &
            & (m.ne.i)) cycle
       
       info = 0

       ! get the data
       qrm_spmat => dqrm_get_test_mat(i)
       if(info.eq.0) call qrm_alloc(b, qrm_spmat%m, info)
       if(info.eq.0) call qrm_alloc(r, qrm_spmat%m, info)
       if(info.eq.0) call qrm_alloc(x, qrm_spmat%n, info)
       if(info.ne.0) goto 9999
       call dlarnv(2, iseed, size(b), b(1))
       r = b

       if(qrm_spmat%m .ge. qrm_spmat%n) then
          transp=qrm_no_transp
       else
          transp=qrm_transp
       end if
       
       call dqrm_spfct_init(qrm_spfct, qrm_spmat, info)
       if(info.ne.0) goto 9999
          
       call qrm_analyse(qrm_spmat, qrm_spfct, transp, info)
       if(info.ne.0) goto 9999
       call qrm_factorize(qrm_spmat, qrm_spfct, transp, info)
       if(info.ne.0) goto 9999

       if(qrm_spmat%sym.eq.0) then
          if(transp .eq. qrm_no_transp) then
             r = b ! this copy is to avoid overwriting b
             call qrm_apply(qrm_spfct, qrm_transp, r, info)
             if(info.ne.0) goto 9999
             call qrm_solve(qrm_spfct, qrm_no_transp, r, x, info)
             if(info.ne.0) goto 9999
          else if(transp .eq. qrm_transp) then
             call qrm_solve(qrm_spfct, qrm_transp, b, x, info)
             if(info.ne.0) goto 9999
             call qrm_apply(qrm_spfct, qrm_no_transp, x, info)
             if(info.ne.0) goto 9999
          end if
       else
          call qrm_solve(qrm_spfct, qrm_transp, b, r, info)
          if(info.ne.0) goto 9999
          call qrm_solve(qrm_spfct, qrm_no_transp  , r, x, info)
          if(info.ne.0) goto 9999
       end if
       r = b
       call qrm_residual_norm(qrm_spmat, r, x, rnrm)
       call qrm_vecnrm(x, size(x,1), '2', xnrm)
       call qrm_vecnrm(b, size(b,1), '2', bnrm)
       call qrm_spmat_nrm(qrm_spmat, 'f', anrm)
       call qrm_residual_orth(qrm_spmat, r, onrm)   

       if(transp .eq. qrm_no_transp) then
          if(rnrm.lt.eps) then
             err = rnrm
          else
             err = onrm
          end if
       else if(transp .eq. qrm_transp) then
          err = rnrm
       end if
       
       ! print message
9999   continue
       call dqrm_prnt_testmesg(5, "facto", 1, 1, i, (info .eq. 0) .and. (err .lt. eps))
       ok = ok .and. (info .eq. 0) .and. (err .lt. eps)
       
       ! destroy the sparse factorization object
       call dqrm_spfct_destroy(qrm_spfct)

       call qrm_dealloc(b)
       call qrm_dealloc(r)
       call qrm_dealloc(x)

    end do
    
    return

  end function case1
  

  !> @brief Test the factorization with different
  !! mb, nb, ib and bh parameters
  function case2(m) result(ok)
    use dqrm_mod
    implicit none
    
    integer :: m
    logical :: ok

    type(dqrm_spmat_type), pointer :: qrm_spmat
    type(dqrm_spfct_type)          :: qrm_spfct
    integer                        :: info
    real(r64), allocatable      :: x(:), b(:), r(:)
    real(r64)                      :: err, anrm, rnrm, bnrm, xnrm, onrm
    integer                        :: i, j, p
    character                      :: transp
    integer, parameter             :: np=5
    integer                        :: params(4,np)

    params = reshape((/  64, 64, 32, 0, &
                      & 128, 64, 32, 0, &
                      &  64, 64, 32, 2, &
                      & 160, 80, 40, 1, &
                      & 512, 64, 32, 0 /), shape(params))
    
    if((m.ne.-1) .and. &
         & (m.lt.11) .and. &
         & (m.gt.size(matrices))) then
       write(*,'("Matrix ",i2," is not available for this test")')m
       return
    end if

    ok = .true.
    
    ! loop over all the file-matrices
    do i=11, size(matrices)
       if((m.ne.-1) .and. &
            & (m.ne.i)) cycle
       
       ! get the data
       qrm_spmat => dqrm_get_test_mat(i)
       if(qrm_spmat%m .ge. qrm_spmat%n) then
          transp=qrm_no_transp
       else
          transp=qrm_transp
       end if

       do p=1, np
          info = 0

          if(info.eq.0) call qrm_alloc(b, qrm_spmat%m, info)
          if(info.eq.0) call qrm_alloc(r, qrm_spmat%m, info)
          if(info.eq.0) call qrm_alloc(x, qrm_spmat%n, info)
          if(info.ne.0) goto 9999
          call dlarnv(2, iseed, size(b), b(1))
          r = b

          call qrm_spfct_init(qrm_spfct, qrm_spmat, info)
          if(info.ne.0) goto 9999

          call qrm_set(qrm_spfct, 'qrm_mb_', params(1,p))
          call qrm_set(qrm_spfct, 'qrm_nb_', params(2,p))
          call qrm_set(qrm_spfct, 'qrm_ib_', params(3,p))
          call qrm_set(qrm_spfct, 'qrm_bh_', params(4,p))
          
          call qrm_analyse(qrm_spmat, qrm_spfct, transp, info)
          if(info.ne.0) goto 9999
          call qrm_factorize(qrm_spmat, qrm_spfct, transp, info)
          if(info.ne.0) goto 9999

          if(qrm_spmat%sym.eq.0) then
             if(transp .eq. qrm_no_transp) then
                r = b ! this copy is to avoid overwriting b
                call qrm_apply(qrm_spfct, qrm_transp, r, info)
                if(info.ne.0) goto 9999
                call qrm_solve(qrm_spfct, qrm_no_transp, r, x, info)
                if(info.ne.0) goto 9999
             else if(transp .eq. qrm_transp) then
                call qrm_solve(qrm_spfct, qrm_transp, b, x, info)
                if(info.ne.0) goto 9999
                call qrm_apply(qrm_spfct, qrm_no_transp, x, info)
                if(info.ne.0) goto 9999
             end if
          else
             call qrm_solve(qrm_spfct, qrm_transp, b, r, info)
             if(info.ne.0) goto 9999
             call qrm_solve(qrm_spfct, qrm_no_transp  , r, x, info)
             if(info.ne.0) goto 9999
          end if

          r = b
          call qrm_residual_norm(qrm_spmat, r, x, rnrm)
          call qrm_vecnrm(x, size(x,1), '2', xnrm)
          call qrm_vecnrm(b, size(b,1), '2', bnrm)
          call qrm_spmat_nrm(qrm_spmat, 'f', anrm)
          call qrm_residual_orth(qrm_spmat, r, onrm)   

          if(transp .eq. qrm_no_transp) then
             if(rnrm.lt.eps) then
                err = rnrm
             else
                err = onrm
             end if
          else if(transp .eq. qrm_transp) then
             err = rnrm
          end if
          
9999      continue
          ! print message
          call dqrm_prnt_testmesg(5, "facto", 2, p, i, (info .eq. 0) .and. (err .lt. eps))
          ok = ok .and. (info .eq. 0) .and. (err .lt. eps)
          
          ! destroy the sparse factorization object
          call dqrm_spfct_destroy(qrm_spfct)

          call qrm_dealloc(b)
          call qrm_dealloc(r)
          call qrm_dealloc(x)

       end do
    end do
    return

  end function case2


  !> @brief Test the factorization with different
  !! memory bounds
  function case3(m) result(ok)
    use dqrm_mod
    implicit none
    
    integer :: m
    logical :: ok

    type(dqrm_spmat_type), pointer :: qrm_spmat
    type(dqrm_spfct_type)          :: qrm_spfct
    integer                        :: info
    real(r64), allocatable         :: x(:), b(:), r(:)
    real(r64)                      :: err, anrm, rnrm, bnrm, xnrm, onrm
    integer                        :: i, j, p
    character                      :: transp
    real(r32)                      :: memrelax


    if((m.ne.-1) .and. &
         & (m.lt.11) .and. &
         & (m.gt.size(matrices))) then
       write(*,'("Matrix ",i2," is not available for this test")')m
       return
    end if

    ok = .true.
    
    ! loop over all the file-matrices
    do i=11, size(matrices)
       if((m.ne.-1) .and. &
            & (m.ne.i)) cycle
       
       ! get the data
       qrm_spmat => dqrm_get_test_mat(i)
       if(qrm_spmat%m .ge. qrm_spmat%n) then
          transp=qrm_no_transp
       else
          transp=qrm_transp
       end if

       do p=1, 3
          memrelax = 1.e0+(p-1)*0.25e0
          info = 0

          if(info.eq.0) call qrm_alloc(b, qrm_spmat%m, info)
          if(info.eq.0) call qrm_alloc(r, qrm_spmat%m, info)
          if(info.eq.0) call qrm_alloc(x, qrm_spmat%n, info)
          if(info.ne.0) goto 9999
          call dlarnv(2, iseed, size(b), b(1))
          r = b
          
          call qrm_spfct_init(qrm_spfct, qrm_spmat, info)
          if(info.ne.0) goto 9999

          call qrm_set(qrm_spfct, 'qrm_mem_relax_', memrelax)
          
          call qrm_analyse(qrm_spmat, qrm_spfct, transp, info)
          if(info.ne.0) goto 9999
          call qrm_factorize(qrm_spmat, qrm_spfct, transp, info)
          if(info.ne.0) goto 9999

          if(qrm_spmat%sym.eq.0) then
             if(transp .eq. qrm_no_transp) then
                r = b ! this copy is to avoid overwriting b
                call qrm_apply(qrm_spfct, qrm_transp, r, info)
                if(info.ne.0) goto 9999
                call qrm_solve(qrm_spfct, qrm_no_transp, r, x, info)
                if(info.ne.0) goto 9999
             else if(transp .eq. qrm_transp) then
                call qrm_solve(qrm_spfct, qrm_transp, b, x, info)
                if(info.ne.0) goto 9999
                call qrm_apply(qrm_spfct, qrm_no_transp, x, info)
                if(info.ne.0) goto 9999
             end if
          else
             call qrm_solve(qrm_spfct, qrm_transp, b, r, info)
             if(info.ne.0) goto 9999
             call qrm_solve(qrm_spfct, qrm_no_transp  , r, x, info)
             if(info.ne.0) goto 9999
          end if

          r = b
          call qrm_residual_norm(qrm_spmat, r, x, rnrm)
          call qrm_vecnrm(x, size(x,1), '2', xnrm)
          call qrm_vecnrm(b, size(b,1), '2', bnrm)
          call qrm_spmat_nrm(qrm_spmat, 'f', anrm)
          call qrm_residual_orth(qrm_spmat, r, onrm)   
          
          if(transp .eq. qrm_no_transp) then
             if(rnrm.lt.eps) then
                err = rnrm
             else
                err = onrm
             end if
          else if(transp .eq. qrm_transp) then
             err = rnrm
          end if
          
9999      continue
          ! print message
          call dqrm_prnt_testmesg(5, "facto", 3, p, i, (info .eq. 0) .and. (err .lt. eps))
          ok = ok .and. (info .eq. 0) .and. (err .lt. eps)
          
          ! destroy the sparse factorization object
          call dqrm_spfct_destroy(qrm_spfct)

          call qrm_dealloc(b)
          call qrm_dealloc(r)
          call qrm_dealloc(x)

       end do
    end do
    return

  end function case3

  

end function dqrm_test_facto
