!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!!##############################################################################################
!> @file qrm_sppotr.F90
!!
!! An example program which reads parameters from a file, a matrix from a MM file and
!! solves the problem. Run this program as
!!
!! ./cqrm_sppotr -mat matrix -nrhs
!!
!! (where _ is d, s, c or z) using the provided input.txt file.
!!
!!
!!##############################################################################################

#include "qrm_common.h"

program cqrm_sppotr

  use cqrm_mod
  implicit none

  type(cqrm_spmat_type)             :: qrm_spmat, rfact
  type(cqrm_spfct_type)             :: qrm_spfct
                                     
  integer                           :: info, i, nrhs, ounit, verb, mb
  character                         :: transp 
  complex(r32), allocatable, target :: b(:,:), x(:,:), r(:,:)
  real(kind(1.d0))                  :: tanl, tfct, tslv
  integer(i64)                      :: ts, te, tr
  real(r32)                         :: anrm, mxrnrm, mxxnrm
  real(r32), allocatable            :: rnrm(:), bnrm(:), xnrm(:)
  integer                           :: iseed(4)=(/1,1,1,1/)
  character(len=:), allocatable     :: matname
  integer                           :: ncpu, ngpu
  integer :: mm
  
  call qrm_init()

  ! initialize the matrix data structure. 
  call qrm_spmat_init(qrm_spmat)
  call qrm_read_parms(matname, qrm_spmat, nrhs, verb, info)
  __QRM_INFO_CHECK(info,'qrm_sppotr','qrm_read_parms',10)

  call qrm_spfct_init(qrm_spfct, qrm_spmat)

  ounit = 0
  ! read the input file
  if(verb.eq.3) call qrm_get('qrm_ounit', ounit)

  if(ounit.gt.0) write(ounit,'(30("="))')

  if(qrm_spmat%m .ge. qrm_spmat%n) then
     transp=qrm_no_transp
  else
     if(ounit.gt.0) write(ounit,'("Transpose")')
     transp=qrm_conj_transp
  end if

  if(ounit.gt.0) write(ounit,'("Starting Analysis")')
  call system_clock(ts)
  call qrm_analyse(qrm_spmat, qrm_spfct, transp, info)
  call system_clock(te, tr); tanl = real(te-ts)/real(tr)
  __QRM_INFO_CHECK(info,'qrm_qrm_sppotr','qrm_analyse',10)

  
  if(ounit.gt.0) write(ounit,'("  Estimated nonzeroes in R            : ",i20)')qrm_spfct%gstats(qrm_e_nnz_r_)
  if(ounit.gt.0) write(ounit,'("  Estimated total flops at facto      : ",i20)')qrm_spfct%gstats(qrm_e_facto_flops_)
  if(ounit.gt.0) write(ounit,'("  Estimated memory peak at facto (MB) : ",f20.3)')real(qrm_spfct%gstats(qrm_e_facto_mempeak_))/1e6


  if(ounit.gt.0) write(ounit,'("Starting Factorization")')

  call qrm_start_profiling()
  call system_clock(ts)
  call qrm_factorize(qrm_spmat, qrm_spfct, transp, info)
  call system_clock(te, tr); tfct = real(te-ts)/real(tr)
  call qrm_stop_profiling()
  __QRM_INFO_CHECK(info,'qrm_qrm_sppotr','qrm_factorize',10)

  if(nrhs .gt. 0) then
     if(ounit.gt.0) write(ounit,'("Starting Solve")')
     call qrm_alloc(b, qrm_spmat%m, nrhs)
     call qrm_alloc(r, qrm_spmat%m, nrhs)
     call qrm_alloc(x, qrm_spmat%n, nrhs)
     
     call qrm_alloc(bnrm, nrhs)
     call qrm_alloc(rnrm, nrhs)
     call qrm_alloc(xnrm, nrhs)

     x = qrm_czero

     call clarnv(2, iseed, size(b), b(1,1))
     
     call system_clock(ts)
     call qrm_solve(qrm_spfct, qrm_conj_transp, b, r)
     call qrm_solve(qrm_spfct, qrm_no_transp, r, x)
     call system_clock(te, tr); tslv = real(te-ts)/real(tr)
     r = b

     ! compute the residual
     call qrm_residual_norm(qrm_spmat, r, x, rnrm)
     call qrm_vecnrm(x, size(x,1), '2', xnrm)
     call qrm_vecnrm(b, size(b,1), '2', bnrm)
     call qrm_spmat_nrm(qrm_spmat, 'f', anrm)
     if(ounit.gt.0) then
        write(ounit,'(" ")')
        write(ounit,'("||A||          = ",e10.2)')anrm
        write(ounit,'(" ")')
        write(ounit,'("                 ||b||           ||x||              ||r||          ")')
        write(ounit,'("                                              -----------------    ")')
        write(ounit,'("                                              || |A|*|x|+|b| ||    ")')
        write(ounit,'("===================================================================")')
        do i=1, nrhs
           write(ounit,'("RHS ",i3,"  :   ",2(e10.2,6x),2x,2(e10.2,12x))')i,bnrm(i),xnrm(i),rnrm(i)
        end do
     end if

     mxxnrm = maxval(xnrm)
     mxrnrm = maxval(rnrm)
     
     call qrm_dealloc(b)
     call qrm_dealloc(r)
     call qrm_dealloc(x)
     call qrm_dealloc(bnrm)
     call qrm_dealloc(rnrm)
     call qrm_dealloc(xnrm)
  else
     tslv   = 0.0
     mxxnrm = qrm_czero
     mxrnrm = qrm_czero
  end if

10 continue

  if(ounit.gt.0) write(ounit,'(" ")')
  if(ounit.gt.0) write(ounit,'("Done.")')
  if(ounit.gt.0) write(ounit,'(" ")')
  if(ounit.gt.0) write(ounit,'(" ")')
 
  if(ounit.gt.0) write(ounit,'("  Time to do the analysis  : ",es10.3)'    )tanl
  if(ounit.gt.0) write(ounit,'("  Time to do the facto     : ",es10.3)'    )tfct
  if(ounit.gt.0) write(ounit,'("  Gflops/s at facto        : ",f8.2)'      )real(qrm_spfct%gstats(qrm_e_facto_flops_))/tfct/1e9
  if(ounit.gt.0) write(ounit,'("  Time to compute solution : ",es10.3)'    )tslv
  if(ounit.gt.0) write(ounit,'("  Nonzeroes in R           : ",i20)'       )qrm_spfct%gstats(qrm_nnz_r_)
  if(ounit.gt.0) write(ounit,'("  Total unallocated memory : ",i20)'       )qrm_tot_mem
  if(ounit.gt.0) write(ounit,'("  Memory peak              : ",f9.3," MB")') &
       &real(qrm_max_mem,r32)/1e6


  call qrm_get(qrm_spfct, 'qrm_mb', mb)
  ounit = 0
  if(verb.eq.2) call qrm_get('qrm_ounit', ounit)

  if(ounit.gt.0) write(ounit,'("|      matrix      |  mb  | ncpu | ngpu |   tf    | Gflop/s |   ts   |'// &
       ' ||r||/||A|| |")')
  
  call qrm_get('qrm_ncpu', ncpu)
  call qrm_get('qrm_ngpu', ngpu)
  if(verb.eq.2 .or. verb.eq.1) call qrm_get('qrm_ounit', ounit)
  if(ounit.gt.0) write(ounit,'("| ",a16," |",i5," |",i5," |",i5," |",f8.3," |",f8.2," |",f7.3," |",e12.5," |")')&
       matname(1:min(len(matname),16))//repeat(' ',max(0,16-len(matname))), mb, ncpu, ngpu, tfct, &
       real(qrm_spfct%gstats(qrm_e_facto_flops_))/tfct/1e9, tslv, mxrnrm
  
  
  call qrm_spmat_destroy(qrm_spmat, info)
  call qrm_spfct_destroy(qrm_spfct, info)
  call qrm_finalize()
  stop

contains

  subroutine qrm_read_parms(matname, qrm_spmat, nrhs, verb, info)
    ! Function: read_parms
    !
    ! *Input*:
    !
    ! *Output*:
    !
    use qrm_ssget_mod
    use cqrm_utils_mod
    use qrm_cli_mod
    implicit none

    character(len=:), allocatable :: matname
    type(cqrm_spmat_type)         :: qrm_spmat
    integer                       :: nrhs
    integer                       :: verb
    integer, optional             :: info
                                   
    character(len=:), allocatable :: mat
    type(qrm_cli_type)            :: cli
    character(len=:), allocatable :: matfile, ssrepo
    integer                       :: col1, col2, col3, col4
    integer                       :: nx, ny, nz, n, s

    ! error management
    integer                       :: err
    character(len=*), parameter   :: name='qrm_read_parms'
    
    err = 0

    call cli%init(title='cqrm_sppotr', description='A test program for the Cholesky factorization of sparse matrices')
    call cli%add('-mat'  , 'The matrix file/name.\n'//&
         'This can either be a file in MatrixMarket format, \n'//&
         'a string of the form ss:path:matname, in which case \nmatrix '//&
         'matname from SuiteSparse will be downloaded \nin path and used '//&
         'or a string of the type 7PT:NX:NY:NZ:1 \nwhere a 7PT Poisson matrix of '//&
         'size NXxNYxNZ will be generated.', def='', req=.true.)
    call cli%add('-nrhs' , 'The number of right-hand sides.'                            , def=1 )
    call cli%add('-v'    , 'Verbosity: 3=full, 2=batch w/ header, 1=batch w/o header.'  , def=3 )
    call cli%parse(err)

    if(err.ne.0) then
       
    end if
    
    call cli%get('-mat'  , mat)
    call cli%get('-nrhs' , nrhs)
    call cli%get('-v'    , verb)

    if(index(mat, 'ss:').eq.1) then
       ! this is a SuiteSparse matrix
       col1 = index(mat, ':')
       if(col1.eq.0) then
          write(*,'("SS matrix not correctly formatted.")')
          stop
       end if
       col2 = index(mat(col1+1:),':')+col1
       if(col2.eq.col1 .or. col2.eq.col1+1) then
          ssrepo = '.'
       else
          ssrepo = mat(col1+1:col2-1)
       end if
       call qrm_ssget_init(ssrepo)
       matfile = qrm_ssget(mat(col2+1:))
       call qrm_readmat(matfile, qrm_spmat, .true., err)
       __QRM_INFO_CHECK(err,'qrm_sppotr','qrm_readmat',20)
       matname = mat(col2+1:)
    else if(index(mat, '7PT:').eq.1) then
       ! this is a 7PT matrix
       col1 = index(mat, ':')
       col2 = index(mat(col1+1:),':')+col1
       col3 = index(mat(col2+1:),':')+col2
       col4 = index(mat(col3+1:),':')+col3
       if(col1.eq.0 .or. col2.eq.col1 .or. col3.eq.col2 .or. col4.eq.col3) then
          write(*,'("7PT matrix not correctly formatted.")')
          stop
       end if
       read(mat(col1+1:col2-1),*)nx
       read(mat(col2+1:col3-1),*)ny
       read(mat(col3+1:col4-1),*)nz
       read(mat(col4+1:      ),*)s
       call qrm_matgen_7pt(qrm_spmat, nx, ny, nz, s)
       matname = mat
    ! else if(index(mat, 'Mitt3D_27:').eq.1) then
       ! ! this is a Mitt3D matrix
       ! col1 = index(mat, ':')
       ! if(col1.eq.0) then
          ! write(*,'("Mitt3D_27 matrix not correctly formatted.")')
          ! stop
       ! end if
       ! read(mat(col1+1:col2-1),*)n
       ! call qrm_matgen_mitt3d(qrm_spmat, n)
    else
       matfile = mat
       call qrm_readmat(matfile, qrm_spmat, .true., err)
       __QRM_INFO_CHECK(err,'qrm_sppotr','qrm_readmat',20)
       matname = mat
    end if

    ounit = 0
    if(verb.eq.3) call qrm_get('qrm_ounit', ounit)
    if(ounit.ne.0) write(ounit,'("Matrix ready. M:",i7,"  N:",i7,"  NZ:",i10)')qrm_spmat%m, qrm_spmat%n, qrm_spmat%nz
    if(ounit.ne.0) write(ounit,'(" ")')

20  continue
    if(present(info)) info = err
    return
    
  end subroutine qrm_read_parms

end program cqrm_sppotr







